/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org
*/

#ifndef CONFIG_H
#define CONFIG_H

// WINDOWS overrides
#ifdef _MSC_VER
#define strncasecmp _strnicmp
#define strcasecmp _stricmp
#endif

#define POP_MAX_PATH 256
#define POP_MAX_OPTIONS_SIZE 256

#define PRINCE_VERSION "PRINCE OF PERSIA  V1.5"
#define SDLPOP_VERSION "1.20-modified"
#define WINDOW_TITLE "Prince of Persia v1.5 (powered by SDLPoP 1.20)"

// Default SDL_Joystick button values
#define SDL_JOYSTICK_BUTTON_A 0
#define SDL_JOYSTICK_BUTTON_B 1
#define SDL_JOYSTICK_BUTTON_X 2
#define SDL_JOYSTICK_BUTTON_Y 3
#define SDL_JOYSTICK_X_AXIS 0
#define SDL_JOYSTICK_Y_AXIS 1

#define MAX_FORE_SOUND_ID 62

// Predefined game values
#define GUARD_COLORS 9

#define NUMBER_OF_ROOMS 24
#define PLACEHOLDER_ROOM 99
#define MAX_TILES_PER_ROOM 30

#define MAX_PEELS 50
#define MAX_DRECTS 30
#define MAX_TROBS 30
#define MAX_MOBS 30

#define JOY_AXIS_NUM 6 // Max quantity of analogue inputs on a gamepad/joystick

#define RESTART_AFTER_DEATH_DELAY 288
#define RESTART_NOT_ENOUGH_HITPOINTS 86

#define FEATHER_FALL_IN_SEC 18.75

#define DROPDEAD_FRAMES 8
#define DEADGUARD_OFFSET 140
#define DEADGUARD_OFFSET_LEFT 74
#define DEADGUARD_OFFSET_RIGHT 181

#define DEAD_KID_LEVEL_10_TIMER (is_sound_on ? 160 : 50)
#define DEAD_KID_LEVEL_10_TIMER_P12 DEAD_KID_LEVEL_10_TIMER + 12

#define SKELETON_SHAKE_TIME 60

#define MAX_GRAB_FALLING_SPEED 30
#define GRAB_DELAY 6

#define SHADOW_SEPARATION_COUNT 4

// Guard is 39 pixels tall when standing. Based on the following formula from "can_bump_into_gate()"
// the max modifier is "x / 4 + 6 = 39" or "x = 132".
#define GUARD_HEIGHT_GATE_MODIFIER 132

// level door modifier (in pixels) at which kid can go through.
#define KID_HEIGHT_DOOR_MODIFIER 33

#define ONSCREEN_COLUMN_OFFSET 5
#define ONSCREEN_ROW_OFFSET 1

#define CHAR_OFFSET 14
#define CHAR_10_PX_OFFSET 10
#define CHAR_7_PX_OFFSET 7

// Character fell into the room below
#define OUT_OF_BOUNDS_Y 211

#define DUNGEON_LEVEL 0
#define PALACE_LEVEL 1

#define DUNGEON_OR_PALACE_1_LEVEL 0
#define DUNGEON_OR_PALACE_2_LEVEL 1
#define TEMPLE_LEVEL 2

#define MOUSE_HIT_SPEED 14

#define ALL_HP 100
#define MAX_GUARD_HP 10

#define GATE_GUARDING_SKILL 8

#endif
