/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#ifndef DATA_H
#define DATA_H

#ifdef BODY
// If included from data.c: definitions (without extern and with initialization).
#define INIT(...) __VA_ARGS__
#define extern
#else
// Otherwise: just declarations (with extern and without initialization).
#define INIT(...)
#undef extern
#endif

// executable directory
extern char exe_dir[POP_MAX_PATH] INIT ( = ".");

// renderer
extern char renderer_name[50] INIT( = "UNKNOWN");

// saved folders
extern const char* save_folder INIT( = "SAVED");
extern const char* auto_save_folder INIT( = "SAVED/AUTOSAVE");
extern const char* screenshot_folder INIT( = "SAVED/SCRSHOTS");

// data:5F8A
extern word text_time_remaining;
// data:4C56
extern word text_time_total;
// data:431C
extern word is_show_time;
// data:4C6E
extern word checkpoint;
// data:4E92
extern word upside_down_timer;
// data:3D36
extern short resurrect_time;
// data:42B4
extern word dont_reset_time;
// data:4F7E
extern short rem_min;
// data:4F82
extern word rem_tick;
// data:4608
extern word hitp_beg_lev;
// data:4CAA
extern word need_level1_music;
// data:4380
extern surface_type* offscreen_surface;

// data:295E
extern const rect_type screen_rect INIT( = {0, 0, 200, 320});
// data:42B8
extern short start_level INIT( = -1);
// data:4CE6
extern byte* guard_palettes;
// data:4338
extern chtab_type* chtab_addrs[16];

// data:46C6
extern dialog_type* copyprot_dialog;
// data:46C6
extern dialog_type* about_dialog;
// data:2944
extern dialog_settings_type dialog_settings
INIT( = {
    (void(*)(dialog_type *))(add_dialog_rect),
    (void(*)(dialog_type *))(dialog_method_2_frame),
    4, 4, 4, 4, 3, 4, 1
});
// data:2B76
extern rect_type dialog_rect_1 INIT( = {60, 56, 124, 264});
// data:2B7E
extern rect_type dialog_rect_2 INIT( = {28, 56, 124, 264});

// data:409E
extern word drawn_room;

// custom logic - an ID for placeholder tiles
extern word placeholder_id INIT( = 0);

// data:4CCD
extern byte curr_tile;
// data:4328
extern byte curr_modifier;

// data:4CD8
extern tile_and_mod leftroom_[3];
// data:5950
extern tile_and_mod row_below_left_[10];
// data:2274
extern const word tbl_line[] INIT( = {0, 10, 20});

// data:658A
extern byte* curr_room_tiles;
// data:5F88
extern byte* curr_room_modif;
// data:24C6
extern word col_xh[] INIT( = {0, 4, 8, 12, 16, 20, 24, 28, 32, 36});
// data:5968
extern word draw_xh;
// data:0F9E
extern word current_level INIT( = -1);
// data:2BA6
#define VGA_PALETTE_DEFAULT { \
        {0x00, 0x00, 0x00},\
        {0x00, 0x00, 0x2A},\
        {0x00, 0x2A, 0x00},\
        {0x00, 0x2A, 0x2A},\
        {0x2A, 0x00, 0x00},\
        {0x2A, 0x00, 0x2A},\
        {0x2A, 0x15, 0x00},\
        {0x2A, 0x2A, 0x2A},\
        {0x15, 0x15, 0x15},\
        {0x15, 0x15, 0x3F},\
        {0x15, 0x3F, 0x15},\
        {0x15, 0x3F, 0x3F},\
        {0x3F, 0x15, 0x15},\
        {0x3F, 0x15, 0x3F},\
        {0x3F, 0x3F, 0x15},\
        {0x3F, 0x3F, 0x3F},}
// data:4CC0
extern word room_L;
// data:4CCE
extern word room_R;
// data:4C90
extern word room_A;
// data:4C96
extern word room_B;
// data:461A
extern word room_BR;
// data:43FE
extern word room_BL;
// data:4614
extern word room_AR;
// data:43DE
extern word room_AL;

// data:4F84
extern level_type level;

// data:42AA
extern short table_counts[5];
#define backtable_count table_counts[0]
#define foretable_count table_counts[1]
#define wipetable_count table_counts[2]
#define  midtable_count table_counts[3]
#define  objtable_count table_counts[4]
// data:4D7E
extern short drects_count;
// data:434E
extern short peels_count;

// data:5FF4
extern back_table_type foretable[200];
// data:463C
extern back_table_type backtable[200];
// data:3D38
extern midtable_type midtable[50];
// data:5F1E
extern peel_type* peels_table[MAX_PEELS];
// data:4D9A
extern rect_type drects[MAX_DRECTS];

// data:4CB8
extern sbyte obj_direction;
// data:2588
extern const byte chtab_flip_clip[16] INIT( = {1, 0, 1, 1, 1, 1, 0, 0, 0, 0, 1, 1, 1, 1, 1, 0});
// data:42A6
extern short obj_clip_left;
// data:42C6
extern short obj_clip_top;
// data:42C0
extern short obj_clip_right;
// data:4082
extern short obj_clip_bottom;
// data:34D2
extern wipetable_type wipetable[300];
// data:2592
extern const byte chtab_shift[16] INIT( = {0, 1, 0, 0, 0, 0, 1, 1, 1, 0, 0, 0, 0, 0, 0, 1});
// data:4354
extern word need_drects;
// data:4CC2
extern word is_blind_mode;

// custom logic - should the x coordinate get converted to a screen coordinate
extern const byte chtab_screen_x[16] INIT( = {0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0});

// data:0F86
extern const rect_type rect_top INIT( = {0, 0, 192, 320});
// data:0F96
extern const rect_type rect_bottom_text INIT( = {193, 71, 202, 251});

// data:4058
extern word leveldoor_ybottom;

// data:4CFA
extern byte palace_wall_colors[44 * 3];

// data:2942
extern word seed_was_init INIT( = 0);
// data:4084
extern dword random_seed;


// data:3010
extern surface_type* current_target_surface INIT( = NULL);

// data:4C5C
extern byte* doorlink_map;
// data:4C5A
extern byte* doorlink_loc;

// data:4CC6
extern sbyte control_shift;
// data:461C
extern sbyte control_y;
// data:4612
extern sbyte control_x;

// data:4CCA
extern word is_global_fading;
// data:4400
extern palette_fade_type* fade_palette_buffer;
// data:4358
extern char_type Kid;
// data:295C
extern word is_keyboard_mode INIT( = 0);
// data:4E8A
extern word is_paused;
// data:42D0
extern word is_restart_level;
// data:42C8
extern word is_joyst_mode;
// data:31E7
extern word is_sound_on INIT( = 0x0F);
// data:3D18
extern word next_level;
// data:4C4A
extern short guardhp_delta;
// data:596A
extern word guardhp_curr;
// data:4CC8
extern word next_room;
// data:4C98
extern word hitp_curr;
// data:5FF2
extern word hitp_max;
// data:5FF0
extern short hitp_delta;
// data:4D94
extern word flash_color;
// data:4350
extern word flash_time;
// data:42DC
extern char_type Guard;

// data:4CF8
extern short roomleave_result;
// data:4D96
extern word different_room;
// data:4E94
extern sound_buffer_type* sound_pointers[MAX_FORE_SOUND_ID + 1];
// data:4C58
extern word guardhp_max;
// data:405C
extern word feather_fall_timer;
// data:4CBA
extern chtab_type* chtab_title40;
// data:4CD0
extern chtab_type* chtab_title50;
// data:405E
extern short hof_count;


// data:009A
extern word demo_mode INIT( = 0);

// data:42CA
extern word is_cutscene;
extern bool is_ending_sequence; // added

// data:0FA0
extern cutscene_ptr_type tbl_cutscenes[16] INIT( = {
    NULL,
    NULL,
    cutscene_2_6,
    NULL,
    cutscene_4,
    NULL,
    cutscene_2_6,
    NULL,
    cutscene_8,
    cutscene_9,
    NULL,
    NULL,
    cutscene_12,
    NULL,
    NULL,
    cutscene_15,
});

// data:408C
extern short mobs_count;
// data:4F7A
extern short trobs_count;
// data:4062
extern short next_sound;
// data:34AA
extern word grab_timer;
// data:594C
extern short can_guard_see_kid;
// data:594E
extern word holding_sword;
// data:4E90
extern short united_with_shadow;
// data:409C
extern word leveldoor_open;
// data:4610
extern word demo_index;
// data:4CD4
extern short demo_time;
// data:34A2
extern word have_sword;

// data:3D22
extern char_type Char;
// data:4D80
extern char_type Opp;

// data:42A2
extern short knock;
// data:4370
extern word is_guard_notice;

// data:656C
extern byte wipe_frames[30];
// data:4C72
extern sbyte wipe_heights[30];
// data:34AC
extern byte redraw_frames_anim[30];
// data:43E0
extern byte redraw_frames2[30];
// data:4C1A
extern byte redraw_frames_floor_overlay[30];
// data:4064
extern byte redraw_frames_full[30];
// data:5EFE
extern byte redraw_frames_fore[30];
// data:3484
extern byte tile_object_redraw[30];
// data:4C64
extern byte redraw_frames_above[10];
// data:4CE2
extern word need_full_redraw;
// data:588E
extern short n_curr_objs;
// data:5BAC
extern objtable_type objtable[50];
// data:5F8C
extern short curr_objs[50];

// data:4607
extern byte obj_xh;
// data:4616
extern byte obj_xl;
// data:4613
extern byte obj_y;
// data:4C9A
extern byte obj_chtab;
// data:42A4
extern byte obj_id;
// data:431E
extern byte obj_tilepos;
// data:4604
extern short obj_x;

// data:658C
extern frame_type cur_frame;
// data:5886
extern word seamless;
// data:4CBC
extern trob_type trob;
// data:4382
extern trob_type trobs[MAX_TROBS];
// data:431A
extern short redraw_height;

// data:24DA
extern /*const*/ byte sound_interruptible[] INIT( = {
    0, // sound_0_fell_to_death
    1, // sound_1_falling
    1, // sound_2_tile_crashing
    1, // sound_3_button_pressed
    1, // sound_4_gate_closing
    1, // sound_5_gate_opening
    0, // sound_6_gate_closing_fast
    1, // sound_7_gate_stop
    1, // sound_8_bumped
    1, // sound_9_grab
    1, // sound_10_sword_vs_sword
    1, // sound_11_sword_moving
    1, // sound_12_guard_hurt
    1, // sound_13_kid_hurt
    0, // sound_14_leveldoor_closing
    0, // sound_15_leveldoor_sliding
    1, // sound_16_fat_fart
    1, // sound_17_soft_land
    0, // sound_18_drink
    1, // sound_19_draw_sword
    1, // sound_20_loose_shake_1
    1, // sound_21_loose_shake_2
    1, // sound_22_loose_shake_3
    1, // sound_23_footstep
    0, // sound_24_death_regular
    0, // sound_25_presentation
    0, // sound_26_embrace
    0, // sound_27_cutscene_2_4_6_12
    0, // sound_28_death_in_fight
    1, // sound_29_meet_Jaffar
    0, // sound_30_big_potion
    1, // sound_31_flame_hurts_Jaffar
    0, // sound_32_mouse_music
    0, // sound_33_small_potion
    1, // sound_34_skeleton_hurt
    0, // sound_35_cutscene_8_9
    0, // sound_36_out_of_time
    0, // sound_37_victory
    0, // sound_38_blink
    0, // sound_39_low_weight
    0, // sound_40_cutscene_12_short_time
    0, // sound_41_end_level_music
    1, // sound_42_guard_falling
    0, // sound_43_victory_Jaffar
    0, // sound_44_skel_alive
    0, // sound_45_jump_through_mirror
    0, // sound_46_chomped
    1, // sound_47_chomper
    0, // sound_48_spiked
    1, // sound_49_spikes_out
    0, // sound_50_story_2_princess
    1, // sound_51_princess_door_opening
    0, // sound_52_story_4_Jaffar_leaves
    0, // sound_53_story_3_Jaffar_comes
    0, // sound_54_intro_music
    0, // sound_55_story_1_absence
    0, // sound_56_ending_music,
    0, // sound_57_sword_striked
    0, // sound_58_say_whew
    0, // sound_59_guard_snoring
    1, // sound_60_spikes_in,
    1, // sound_61_shadow_separation,
    0
});

// data:42ED
extern byte curr_tilepos;
// data:432A
extern short curr_room;
// data:4CAC
extern mob_type curmob;
// data:4BB4
extern mob_type mobs[MAX_MOBS];
// data:4332
extern short tile_col;
// data:229C
extern const short y_land[] INIT( = {-8, 55, 118, 181, 244});
// data:288C
extern byte key_states[SDL_NUM_SCANCODES];
// data:24A6
extern const byte x_bump[] INIT( = {0, 2, 16, 30, 44, 58, 72, 86, 100, 114, 128, 142, 156, 170, 184, 198, 212, 226, 240, 254});
// data:42F4
extern word is_screaming;
// data:42EE
extern word offguard; // name from Apple II source
// data:3D32
extern word droppedout; // name from Apple II source

// data:5BAA
extern word exit_room_timer;
// data:4372
extern short char_col_right;
// data:5F86
extern short char_col_left;
// data:599C
extern short char_top_row;
// data:434C
extern short prev_char_top_row;
// data:432C
extern short prev_char_col_right;
// data:42CE
extern short prev_char_col_left;
// data:34A4
extern short char_bottom_row;
// data:3D34
extern short guard_notice_timer;
// data:42A0
extern short jumped_through_mirror;
// data:2292
extern const short y_clip[] INIT( = {-60, 3, 66, 129, 192});
// data:42F9
extern byte curr_tile2;
// data:4336
extern short tile_row;

// data:5F1C
extern word char_width_half;
// data:4618
extern word char_height;
// data:3D1C
extern short char_x_left;
// data:3D20
extern short char_x_left_coll;
// data:42F6
extern short char_x_right_coll;
// data:3D10
extern short char_x_right;
// data:4D98
extern short char_top_y;
// data:4C0E
extern byte through_tile;
// data:5F82
extern sbyte infrontx; // name from Apple II source
// data:228E
extern const sbyte dir_front[] INIT( = {-1, 1});
// data:2290
extern const sbyte dir_behind[] INIT( = {1, -1});
// data:4320
extern word current_sound;
// data:4606
extern sbyte control_shift2;
// data:42A8
extern sbyte control_forward;
// data:4368
extern word guard_skill;
// data:4088
extern sbyte control_backward;
// data:4322
extern sbyte control_up;
// data:409A
extern sbyte control_down;
// data:4CE0
extern sbyte ctrl1_forward;
// data:4CD2
extern sbyte ctrl1_backward;
// data:4D92
extern sbyte ctrl1_up;
// data:4CD6
extern sbyte ctrl1_down;
// data:42F8
extern sbyte ctrl1_shift2;

// data:42F0
extern word shadow_initialized;

// data:4330
extern word guard_refrac;
// data:4098
extern word kid_sword_strike;

// data:6591
extern byte edge_type;

// custom logic - tracks an active chomper
extern byte edge_chomper;

// data:596C
extern SDL_Surface* onscreen_surface_;
extern SDL_Surface* overlay_surface;
extern SDL_Surface* merged_surface;
extern SDL_Renderer* renderer_;
extern SDL_Window* window_;
extern bool is_overlay_displayed;
extern SDL_Texture* texture_sharp;
extern SDL_Texture* target_texture;

extern SDL_GameController* sdl_controller_ INIT( = 0 );
extern SDL_Joystick* sdl_joystick_; // in case our joystick is not compatible with SDL_GameController
extern byte using_sdl_joystick_interface;
extern int joy_axis[JOY_AXIS_NUM]; // hor/ver axes for left/right sticks + left and right triggers (in total 6 axes)
extern int joy_axis_max[JOY_AXIS_NUM]; // Same as above, but stores the highest value reached between game updates
extern int joy_button_states[JOYINPUT_NUM];
extern int joy_left_stick_states[2]; // horizontal, vertical
extern int joy_right_stick_states[2];
extern SDL_Haptic* sdl_haptic;

extern Uint64 perf_counters_per_tick;
extern Uint64 perf_frequency;
extern float milliseconds_per_counter;

extern char** sound_names;

extern int g_argc;
extern char** g_argv;


// data:405A
extern sbyte collision_row;
// data:42C2
extern sbyte prev_collision_row;

// data:4C10
extern sbyte prev_coll_room[10];
// data:4374
extern sbyte curr_row_coll_room[10];
// data:3D06
extern sbyte below_row_coll_room[10];
// data:42D2
extern sbyte above_row_coll_room[10];
// data:5890
extern byte curr_row_coll_flags[10];
// data:4CEA
extern byte above_row_coll_flags[10];
// data:4C4C
extern byte below_row_coll_flags[10];
// data:5BA0
extern byte prev_coll_flags[10];

// data:4F80
extern short pickup_obj_type;

// data:34CA
extern word justblocked; // name from Apple II source

// data:5F84
extern word last_loose_sound;

extern full_image_type full_image[MAX_FULL_IMAGES] INIT( = {
    [TITLE_PRESENTS] = {
        .id = 0,
        .chtab = &chtab_title50,
        .blitter = blitters_10h_transp,
        .xpos = 94,
        .ypos = 142
    },
    [TITLE_GAME] =     {
        .id = 1,
        .chtab = &chtab_title50,
        .blitter = blitters_10h_transp,
        .xpos = 96,
        .ypos = 145
    },
    [TITLE_POP] =      {
        .id = 2,
        .chtab = &chtab_title50,
        .blitter = blitters_10h_transp,
        .xpos = 24,
        .ypos = 17
    },
    [TITLE_HACKED_BY] =  {
        .id = 3,
        .chtab = &chtab_title50,
        .blitter = blitters_10h_transp,
        .xpos = 92,
        .ypos = 160
    },
    [TITLE_TUTORIAL] =  {
        .id = 4,
        .chtab = &chtab_title50,
        .blitter = blitters_10h_transp,
        .xpos = 64,
        .ypos = 182
    },
    [HOF_POP] =        {
        .id = 2,
        .chtab = &chtab_title50,
        .blitter = blitters_10h_transp,
        .xpos = 24,
        .ypos = 24
    },
    [STORY_FRAME] =    {
        .id = 0,
        .chtab = &chtab_title40,
        .blitter = blitters_0_no_transp,
        .xpos = 0,
        .ypos = 0
    },
    [STORY_ABSENCE] =  {
        .id = 1,
        .chtab = &chtab_title40,
        .blitter = blitters_Fh_white,
        .xpos = 24,
        .ypos = 25
    },
    [STORY_MARRY] =    {
        .id = 2,
        .chtab = &chtab_title40,
        .blitter = blitters_Fh_white,
        .xpos = 24,
        .ypos = 25
    },
    [STORY_HAIL] =     {
        .id = 3,
        .chtab = &chtab_title40,
        .blitter = blitters_Fh_white,
        .xpos = 24,
        .ypos = 25
    },
    [STORY_CREDITS] =  {
        .id = 4,
        .chtab = &chtab_title40,
        .blitter = blitters_Fh_white,
        .xpos = 24,
        .ypos = 26
    },
});

// custom logic - custom randomizer seed for debugging
extern dword custom_seed INIT( = 0);

extern bool want_positional_sound;
extern float want_sound_dir_x;
extern float want_sound_distance_loss;
extern bool next_sound_is_positional;
extern float next_sound_dir_x;
extern float next_sound_distance_loss;
extern bool current_sound_is_positional;
extern float current_sound_dir_x;
extern float current_sound_distance_loss;
extern room_offset_type room_offsets[24];

extern int last_key_scancode;
extern font_type hc_font INIT( = {0x01, 0xFF, 7, 2, 1, 1, NULL});
extern textstate_type textstate INIT( = {0, 0, 0, 15, &hc_font});
extern int need_quick_save INIT( = 0);
extern int need_quick_load INIT( = 0);
// custom logic - a flag to load an auto-saved level
extern char* need_auto_load INIT( = "");
// custom logic - a flag to save a loaded level
extern byte need_auto_save INIT( = 0);
// custom logic - do not auto-save level (on Shift-L)
extern byte skip_auto_save INIT( = 0);

extern word pop_window_width INIT( = 640);
extern word pop_window_height INIT( = 400);
extern char levelset_name[POP_MAX_PATH];
extern char mod_data_path[POP_MAX_PATH];

extern byte enable_positional_audio INIT( = 0);
#define is_positional_audio (enable_positional_audio && is_sound_on)
extern byte enable_lighting INIT( = 0);
extern image_type* lighting_mask;
extern game_settings_type* settings;

// data:009C
extern word cheats_enabled INIT( = 0);
extern byte debug_cheats_enabled INIT( = 0);
extern byte is_timer_displayed INIT( = 0);
// custom logic - feather fall timer
extern byte is_feather_timer_displayed INIT( = 0);
// custom logic - upside down timer
extern byte is_upside_down_timer_displayed INIT( = 0);

extern font_type hc_small_font INIT( = {32, 126, 5, 2, 1, 1, NULL});
extern bool have_mouse_input;
extern bool have_keyboard_or_controller_input;
extern int mouse_x, mouse_y;
extern bool mouse_moved;
extern bool mouse_clicked;
extern bool mouse_button_clicked_right;
extern bool pressed_enter;
extern bool escape_key_suppressed;
extern int menu_control_scroll_y;
extern sbyte is_menu_shown;
extern byte enable_pause_menu INIT( = 1);

extern int play_demo_level;
extern bool quotes_needed INIT( = 0);

// custom logic - max saves per level (default difficulty = easy)
extern word max_saves_per_level INIT( = 3);

// custom logic - additional palettes
extern byte* flame_palette;
extern byte* level_6_env_palette;
extern byte* level_6_wall_palette;

// custom logic - chtab for custom graphics
extern chtab_type* custom_graphics_tab;

// custom logic - background music, inspired by SDLPoP JP
extern Mix_Music *track;
extern int track_number;
extern word is_music_on INIT( = 1);
extern int music_mode;
extern int music_suspended_timer;
extern byte music_suspended;
extern byte delayed_music_start INIT( = 0);
extern byte music_paused;
extern byte show_music_menu;

// custom logic - ogg uninterrupted foreground sounds
extern bool feather_fall_track_exists;
extern short curr_ogg_sound_id INIT( = -1);
extern byte ogg_fore_sound_playing;

// custom logic - can inactive guard see kid?
extern byte can_inactive_guard_see_kid;
extern sbyte is_inactive_guard_notice;

// custom logic - loose tile shake counter
extern byte loose_tile_counter[24][30];

// custom logic - tracking whether the kid met Jaffar
extern byte jaffar_presented;

// custom logic - shadow turned into flame
extern bool is_shadow_flame;

// custom logic - tracking whether the kid met Jaffar's vision
extern bool jaffar_vision_presented;

// custom logic - tracking antidote on level 5
extern word took_antidote;

// custom logic - tracking shadow crashing the floor
extern bool level_7_tile_crashed;

// custom logic - tracking if the shadow drank the potions on levels 7 and 9 in the beginning of the level
extern word initial_shadow_potion_counter;

// custom logic - tracking if the shadow drank the potions on levels 7 and 9
extern word shadow_potion_counter;

// custom logic - potion room
extern word potion_room;

// custom logic - potion position
extern word potion_pos;

// custom logic - timer when to put a potion down
extern word potion_timer;

// custom logic - potion size
extern word potion_size;

// custom logic - falling checkpoint on level 6
extern word falling_exit_checkpoint;

// custom logic - kid's direction at the end of level 6
extern word falling_exit_final_direction;

// custom logic - kid exited from the right door on level 2
extern word level_2_right_door;

// custom logic - timer that kills kid on level 12B when there is nowhere to run
extern int level_13_nowhere_to_run_timer;

// custom logic - an ability to restart the level on a timer
extern word level_restart_timer;

// custom logic - tracks whether sword was drawn by kid automatically
extern bool keep_sword_drawn;

// custom logic - timer for the last skeleton on level 3
extern short skeleton_timer;

// custom logic - make hitpoints blink on all levels
extern byte blink_timer;

// custom logic - has kid been turned into shadow
extern byte is_kid_shadow;

// custom logic - tracking when the level ends
extern byte end_of_level_music;

// custom logic - added a timer for delayed sounds
extern int delayed_sound_timer;
extern delayed_sound_type delayed_sound;

// custom logic - dead kid after shadow separation
extern char_type2 Dead;

// custom logic - track dynamic guard
extern char_type2 dynamic_guard;
extern byte dynamic_guard_count[24];

// custom logic - track inactive guard
extern char_type2 inactive_guard[24];
extern word inactive_guard_skill;
extern sbyte inactive_guard_timer;
extern short inactive_obj_x;
extern bool guard_overlap INIT( = false);

// custom logic - extra slots for dead guards
extern char_type2 dead_guard1[24];
extern char_type2 dead_guard2[24];
extern char_type2 dead_guard3[24];
extern char_type2 outbound_dead_guard1[24];
extern char_type2 outbound_dead_guard2[24];
extern char_type2 outbound_dead_guard3[24];

// custom logic - outbound falling guards
extern falling_guard_type falling_guard;

// custom logic - dead kid timer
extern sbyte dead_kid_timer;

// custom logic - dead kid tile, allows restoring buttons and spikes when uniting with shadow or resurrecting
extern stuck_button_type dead_kid_tile INIT( = {0, 0, 0, 0});

// custom logic - should shadow separation text be displayed
extern byte is_shadow_text_shown;

// custom logic - count the number of times kid turned around
extern byte separate_from_shadow_counter;

// custom logic - the previous value of the kid separation counter
extern byte separate_from_shadow_counter_prev;

// custom logic - timer that resets the number of times kid turned around
// if keys have not been pressed quickly enough
extern byte separate_from_shadow_reset_timer;

// custom logic - monitor Alt for manual sword control
extern sbyte control_alt;

// custom logic - dynamic guard timer
extern byte dynamic_guard_timer;

// custom logic - dynamic guard fight indicator
extern byte is_dynamic_guard_fight;

// custom logic - save start time in the beginning of the level
extern short level_rem_min;
extern word level_rem_tick;

// custom logic - guard is following kid into the room below
extern byte guard_dropping_down;

// custom logic - alternate flash in feather mode
extern bool alternate_flash;

// custom logic - Jaffar's timer is a separate variable
extern byte jaffar_delay_timer;

// custom logic - an indicator that a guard knocked down a tile
extern byte guard_knocked_tile;

// custom logic - is guard guarding a gate?
extern byte is_guarding_gate;

// custom logic - blink when flame is hit with sword
extern byte lighting_timer;

// custom logic - what char has been hurt by sword
extern byte charid_hurt_by_sword INIT( = 0xFF);

// custom logic - tracks whether a skeleton has appeared on the level
extern byte skeleton_appeared;

// custom logic - times skeletons that shake
extern byte skeleton_shake_timer;

// custom logic - quick load flag
extern byte quick_loading;

// custom logic - hard jump fall
extern byte hard_jump_fall;

// custom logic - inactive guard fall timer
extern byte inactive_guard_fall_timer;

// custom logic - level 6 initial sequence timer
extern short level_6_sequence_timer INIT( = 0);

// custom logic - super high jump information
extern byte super_jump_timer;
extern byte super_jump_fall;
extern byte super_jump_room;
extern sbyte super_jump_col;
extern sbyte super_jump_row;

// custom logic - a counter that prevents kid and guard moonwalking
extern byte should_turn_around INIT( = 0);

// custom logic - a counter that prevents inactive guard from turning around
// every frame
extern byte should_inactive_turn_around INIT( = 0);

// custom logic - Jaffar's complex parry count
extern word jaffar_parry_count INIT( = 0);

// custom logic - total saves per game
extern short total_saves INIT( = 0);

// custom logic - saves per level count
extern byte saves_per_level INIT( = 0);

// custom logic - keeps the initial state of numlock
extern byte was_numlock_active INIT( = 0);

// custom logic - keeps track of kid palette in cheat mode
extern sbyte kid_cheat_pal_offset;

// custom logic - inactive guard sequence pointer
extern byte inactive_seq_ptr INIT( = 0xFF);

// custom logic - moving wall timer
extern byte moving_wall_timer INIT( = 0);

// custom logic - moving wall direction
extern sbyte moving_wall_direction INIT( = 0);

// custom logic - falling bridge timer
extern word falling_bridge_timer INIT( = 0);

// custom logic - mouse potion animation timer
extern byte mouse_potion_timer INIT( = 0);

// custom logic - just turned with sword
extern byte justturned_engarde INIT( = 0);

// custom logic - the last color of a regular guard that was loaded into the guard chtab
extern byte last_guard_color INIT( = 0);

// custom logic - guard's color in object table
extern byte obj_color INIT( = 0);

// custom logic - skeleton death timer
extern byte skeleton_death_timer INIT( = 0);

// custom logic - 3 punches take away guard's hitpoint
extern byte guard_punch_count INIT( = 0);

// custom logic - joystick shoulder state
extern sbyte control_shoulder;

// custom logic - stores whether prince had sword on the previous level
extern word had_sword INIT( = 0);

#undef INIT
#undef extern

#endif

