/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"

// *INDENT-OFF*
// data:0FE0
//
// { image #, sword bytes, x-offset, y-offset, flags }
//
// image #: references the frame sprite # in the character DAT file (KID.DAT/GUARD.DAT/...)
//
// x-offset, y-offset - on-screen image x/y offsets in screen pixels (x-offset is multiplied by 2)
//
// sword bytes = chtab_value | sword_table_type_id
// ---------------------------------------------------
// chtab_value:
//   0x00 = chtab 2 (=   0 / 64 + 2) - kid
//   0x40 = chtab 3 (=  64 / 64 + 2) - princess in cutscenes
//   0x80 = chtab 4 (= 128 / 64 + 2) - Jaffar in cutscenes
//   0xC0 = chtab 5 (= 192 / 64 + 2) - guard
// sword_table_type_id: sword_table_type array index
//
// flags = frame_value | frame_weight_distance
// -------------------------------------------
// frame_value:
//   0x00 = no floor check, even frame             (even = shift 1 pixel when facing right)
//   0x20 = no floor check, even frame, thin frame (unused)
//   0x40 = floor check, even frame                (floor check = character falls down when a tile is not a floor; even inside a wall)
//   0x60 = floor check, even frame, thin frame    (thin frame = can get closer to a wall type tile before colliding with it)
//   0x80 = no floor check, odd frame              (odd = shift 1 pixel when facing left)
//   0xA0 = no floor check, odd frame, thin frame
//   0xC0 = floor check, odd frame
//   0xE0 = floor check, odd frame, thin frame
// frame_weight_distance: distance in screen pixels from the left of the frame sprite to the base x-coordinate,
//                        typically the center of the foot bearing the character's weight
const frame_type frame_table_kid[] = {
    { 255, 0x00| 0,   0,   0, 0x00| 0}, // frame_0 (do not draw)
    {   0, 0x00| 0,   1,   0, 0xC0| 4}, // frame_1_start_run
    {   1, 0x00| 0,   1,   0, 0x40| 4}, // frame_2_start_run
    {   2, 0x00| 0,   3,   0, 0x40| 7}, // frame_3_start_run
    {   3, 0x00| 0,   4,   0, 0x40| 8}, // frame_4_start_run
    {   4, 0x00| 0,   0,   0, 0xE0| 6}, // frame_5_start_run
    {   5, 0x00| 0,   0,   0, 0x40| 9}, // frame_6_start_run
    {   6, 0x00| 0,   0,   0, 0x40|10}, // frame_7_run
    {   7, 0x00| 0,   0,   0, 0xC0| 5}, // frame_8_run
    {   8, 0x00| 0,   0,   0, 0x40| 4}, // frame_9_run
    {   9, 0x00| 0,   0,   0, 0x40| 7}, // frame_10_run
    {  10, 0x00| 0,   0,   0, 0x40|11}, // frame_11_run
    {  11, 0x00| 0,   0,   0, 0x40| 3}, // frame_12_run
    {  12, 0x00| 0,   0,   0, 0xC0| 3}, // frame_13_run
    {  13, 0x00| 0,   0,   0, 0x40| 7}, // frame_14_run
    {  14, 0x00| 0,   0,   0, 0x40| 3}, // frame_15_stand
    {  15, 0x00| 0,   0,   0, 0xC0| 3}, // frame_16_standing_jump_1
    {  16, 0x00| 0,   0,   0, 0x40| 4}, // frame_17_standing_jump_2
    {  17, 0x00| 0,   0,   0, 0x40| 6}, // frame_18_standing_jump_3
    {  18, 0x00| 0,   0,   0, 0x40| 8}, // frame_19_standing_jump_4
    {  19, 0x00| 0,   0,   0, 0x80| 9}, // frame_20_standing_jump_5
    {  20, 0x00| 0,   0,   0, 0x00|11}, // frame_21_standing_jump_6
    {  21, 0x00| 0,   0,   0, 0x80|11}, // frame_22_standing_jump_7
    {  22, 0x00| 0,   0,   0, 0x00|16}, // frame_23_standing_jump_8
    {  23, 0x00| 0,   0,   0, 0x00| 7}, // frame_24_standing_jump_9
    {  24, 0x00| 0,   0,   0, 0x00| 5}, // frame_25_standing_jump_10
    {  25, 0x00| 0,   0,   0, 0xC0| 1}, // frame_26_standing_jump_11
    {  26, 0x00| 0,   0,   0, 0xC0| 6}, // frame_27_standing_jump_12
    {  27, 0x00| 0,   0,   0, 0x40| 3}, // frame_28_standing_jump_13
    {  28, 0x00| 0,   0,   0, 0x40| 8}, // frame_29_standing_jump_14
    {  29, 0x00| 0,   0,   0, 0x40| 2}, // frame_30_standing_jump_15
    {  30, 0x00| 0,   0,   0, 0x40| 2}, // frame_31_standing_jump_16
    {  31, 0x00| 0,   0,   0, 0xC0| 2}, // frame_32_standing_jump_17
    {  32, 0x00| 0,   0,   0, 0xC0| 2}, // frame_33_standing_jump_18
    {  33, 0x00| 0,   0,   0, 0x40| 3}, // frame_34_start_run_jump_1
    {  34, 0x00| 0,   0,   0, 0x40| 8}, // frame_35_start_run_jump_2
    {  35, 0x00| 0,   0,   0, 0x80|14}, // frame_36_start_run_jump_3
    {  36, 0x00| 0,   0,   0, 0xC0| 1}, // frame_37_start_run_jump_4
    {  37, 0x00| 0,   0,   0, 0x40| 5}, // frame_38_start_run_jump_5
    {  38, 0x00| 0,   0,   0, 0x80|14}, // frame_39_start_run_jump_6
    {  39, 0x00| 0,   0,   0, 0x00|11}, // frame_40_running_jump_1
    {  40, 0x00| 0,   0,   0, 0x80|11}, // frame_41_running_jump_2
    {  41, 0x00| 0,   0,   0, 0x80|10}, // frame_42_running_jump_3
    {  42, 0x00| 0,   0,   0, 0x00| 1}, // frame_43_running_jump_4
    {  43, 0x00| 0,   0,   0, 0xC0| 4}, // frame_44_running_jump_5
    {  44, 0x00| 0,   0,   0, 0xC0| 3}, // frame_45_turn
    {  45, 0x00| 0,   0,   0, 0xC0| 3}, // frame_46_turn
    {  46, 0x00| 0,   0,   0, 0xA0| 5}, // frame_47_turn
    {  47, 0x00| 0,   0,   0, 0xA0| 4}, // frame_48_turn
    {  48, 0x00| 0,   0,   0, 0x60| 6}, // frame_49_turn
    {  49, 0x00| 0,   4,   0, 0x60| 7}, // frame_50_turn
    {  50, 0x00| 0,   3,   0, 0x60| 6}, // frame_51_turn
    {  51, 0x00| 0,   1,   0, 0x40| 4}, // frame_52_turn
    {  64, 0x00| 0,   0,   0, 0xC0| 2}, // frame_53_runturn
    {  65, 0x00| 0,   0,   0, 0x40| 1}, // frame_54_runturn
    {  66, 0x00| 0,   0,   0, 0x40| 2}, // frame_55_runturn
    {  67, 0x00| 0,   0,   0, 0x00| 0}, // frame_56_runturn
    {  68, 0x00| 0,   0,   0, 0x00| 0}, // frame_57_runturn
    {  69, 0x00| 0,   0,   0, 0x80| 0}, // frame_58_runturn
    {  70, 0x00| 0,   0,   0, 0x00| 5}, // frame_59_runturn
    {  71, 0x00| 0,   0,   0, 0x80| 0}, // frame_60_runturn
    {  72, 0x00| 0,   0,   0, 0x00| 0}, // frame_61_runturn
    {  73, 0x00| 0,   0,   0, 0x80| 0}, // frame_62_runturn
    {  74, 0x00| 0,   0,   0, 0x00| 0}, // frame_63_runturn
    {  75, 0x00| 0,   0,   0, 0x00| 0}, // frame_64_runturn
    {  76, 0x00| 0,   0,   0, 0x80| 0}, // frame_65_runturn
    { 255, 0x00| 0,   0,   0, 0x00| 0}, // unused
    {  80, 0x00| 0,  -2,   0, 0x40| 1}, // frame_67_start_jump_up_1
    {  81, 0x00| 0,  -2,   0, 0x40| 1}, // frame_68_start_jump_up_2
    {  82, 0x00| 0,  -1,   0, 0xC0| 2}, // frame_69_start_jump_up_3
    {  83, 0x00| 0,  -2,   0, 0x40| 2}, // frame_70_jumphang
    {  84, 0x00| 0,  -2,   0, 0x40| 1}, // frame_71_jumphang
    {  85, 0x00| 0,  -2,   0, 0x40| 1}, // frame_72_jumphang
    {  86, 0x00| 0,  -2,   0, 0x40| 1}, // frame_73_jumphang
    {  87, 0x00| 0,  -1,   0, 0x00| 7}, // frame_74_jumphang
    {  88, 0x00| 0,  -1,   0, 0x00| 5}, // frame_75_jumphang
    {  89, 0x00| 0,   2,   0, 0x00| 7}, // frame_76_jumphang
    {  90, 0x00| 0,   2,   0, 0x00| 7}, // frame_77_jumphang
    {  91, 0x00| 0,   2,  -3, 0x00| 0}, // frame_78_jumphang
    {  92, 0x00| 0,   2, -10, 0x00| 0}, // frame_79_jumphang
    {  93, 0x00| 0,   2, -11, 0x80| 0}, // frame_80_jumphang
    {  94, 0x00| 0,   3,  -2, 0x40| 3}, // frame_81_hangdrop_1
    {  95, 0x00| 0,   3,   0, 0xC0| 3}, // frame_82_hangdrop_2
    {  96, 0x00| 0,   3,   0, 0xC0| 3}, // frame_83_hangdrop_3
    {  97, 0x00| 0,   3,   0, 0x60| 3}, // frame_84_hangdrop_4
    {  98, 0x00| 0,   4,   0, 0xE0| 3}, // frame_85_hangdrop_5
    {  28, 0x00| 0,   0,   0, 0x00| 0}, // frame_86_test_foot
    {  99, 0x00| 0,   7, -14, 0x80| 0}, // frame_87_hanging_1
    { 100, 0x00| 0,   7, -12, 0x80| 0}, // frame_88_hanging_2
    { 101, 0x00| 0,   4, -12, 0x00| 0}, // frame_89_hanging_3
    { 102, 0x00| 0,   3, -10, 0x80| 0}, // frame_90_hanging_4
    { 103, 0x00| 0,   2, -10, 0x80| 0}, // frame_91_hanging_5
    { 104, 0x00| 0,   1, -10, 0x80| 0}, // frame_92_hanging_6
    { 105, 0x00| 0,   0, -11, 0x00| 0}, // frame_93_hanging_7
    { 106, 0x00| 0,  -1, -12, 0x00| 0}, // frame_94_hanging_8
    { 107, 0x00| 0,  -1, -14, 0x00| 0}, // frame_95_hanging_9
    { 108, 0x00| 0,  -1, -14, 0x00| 0}, // frame_96_hanging_10
    { 109, 0x00| 0,  -1, -15, 0x80| 0}, // frame_97_hanging_11
    { 110, 0x00| 0,  -1, -15, 0x80| 0}, // frame_98_hanging_12
    { 111, 0x00| 0,   0, -15, 0x00| 0}, // frame_99_hanging_13
    { 225, 0x00| 0,  -8,   0, 0x00| 0}, // frame_100_spiking_hanging (custom)
    { 226, 0x00| 0,   0,   0, 0x00| 0}, // frame_101_spiking_hanging (custom)
    { 112, 0x00| 0,   0,   0, 0xC0| 6}, // frame_102_start_fall_1
    { 113, 0x00| 0,   0,   0, 0x40| 6}, // frame_103_start_fall_2
    { 114, 0x00| 0,   0,   0, 0xC0| 5}, // frame_104_start_fall_3
    { 115, 0x00| 0,   0,   0, 0x40| 5}, // frame_105_start_fall_4
    { 116, 0x00| 0,   0,   0, 0xC0| 2}, // frame_106_fall
    { 117, 0x00| 0,   0,   0, 0xC0| 4}, // frame_107_fall_land_1
    { 118, 0x00| 0,   0,   0, 0xC0| 5}, // frame_108_fall_land_2
    { 119, 0x00| 0,   0,   0, 0x40| 6}, // frame_109_crouch
    { 120, 0x00| 0,   0,   0, 0x40| 7}, // frame_110_stand_up_from_crouch_1
    { 121, 0x00| 0,   0,   0, 0x40| 7}, // frame_111_stand_up_from_crouch_2
    { 122, 0x00| 0,   0,   0, 0x40| 9}, // frame_112_stand_up_from_crouch_3
    { 123, 0x00| 0,   0,   0, 0xC0| 8}, // frame_113_stand_up_from_crouch_4
    { 124, 0x00| 0,   0,   0, 0xC0| 9}, // frame_114_stand_up_from_crouch_5
    { 125, 0x00| 0,   0,   0, 0x40| 9}, // frame_115_stand_up_from_crouch_6
    { 126, 0x00| 0,   0,   0, 0x40| 5}, // frame_116_stand_up_from_crouch_7
    { 127, 0x00| 0,   2,   0, 0x40| 5}, // frame_117_stand_up_from_crouch_8
    { 128, 0x00| 0,   2,   0, 0xC0| 5}, // frame_118_stand_up_from_crouch_9
    { 129, 0x00| 0,   0,   0, 0xC0| 3}, // frame_119_stand_up_from_crouch_10
    { 255, 0x00| 0,   0,   0, 0x00| 0}, // unused
    { 133, 0x00| 0,   0,   0, 0x40| 3}, // frame_121_stepping_1
    { 134, 0x00| 0,   0,   0, 0xC0| 4}, // frame_122_stepping_2
    { 135, 0x00| 0,   0,   0, 0xC0| 5}, // frame_123_stepping_3
    { 136, 0x00| 0,   0,   0, 0x40| 8}, // frame_124_stepping_4
    { 137, 0x00| 0,   0,   0, 0x60|12}, // frame_125_stepping_5
    { 138, 0x00| 0,   0,   0, 0xE0|15}, // frame_126_stepping_6
    { 139, 0x00| 0,   0,   0, 0x60| 3}, // frame_127_stepping_7
    { 140, 0x00| 0,   0,   0, 0xC0| 3}, // frame_128_stepping_8
    { 141, 0x00| 0,   0,   0, 0x40| 3}, // frame_129_stepping_9
    { 142, 0x00| 0,   0,   0, 0x40| 3}, // frame_130_stepping_10
    { 143, 0x00| 0,   0,   0, 0x40| 4}, // frame_131_stepping_11
    { 144, 0x00| 0,   0,   0, 0x40| 4}, // frame_132_stepping_12
    { 172, 0x00| 0,   0,   1, 0xC0| 1}, // frame_133_sheathe
    { 173, 0x00| 0,   0,   1, 0xC0| 7}, // frame_134_sheathe
    { 145, 0x00| 0,   0, -12, 0x00| 1}, // frame_135_climbing_1
    { 146, 0x00| 0,   0, -21, 0x00| 0}, // frame_136_climbing_2
    { 147, 0x00| 0,   1, -26, 0x80| 0}, // frame_137_climbing_3
    { 148, 0x00| 0,   3, -32, 0x80| 0}, // frame_138_climbing_4
    { 149, 0x00| 0,   5, -36, 0x80| 0}, // frame_139_climbing_5
    { 150, 0x00| 0,   6, -41, 0x80| 1}, // frame_140_climbing_6
    { 151, 0x00| 0,   1,  17, 0x40| 1}, // frame_141_climbing_7
    { 152, 0x00| 0,   3,   9, 0xC0| 3}, // frame_142_climbing_8
    { 153, 0x00| 0,   3,   4, 0xC0| 8}, // frame_143_climbing_9
    { 154, 0x00| 0,   4,   4, 0xC0| 8}, // frame_144_climbing_10
    { 155, 0x00| 0,   5,   0, 0x60| 9}, // frame_145_climbing_11
    { 156, 0x00| 0,   5,   0, 0xE0| 9}, // frame_146_climbing_12
    { 157, 0x00| 0,   5,   0, 0xE0| 8}, // frame_147_climbing_13
    { 158, 0x00| 0,   5,   0, 0x60| 9}, // frame_148_climbing_14
    { 159, 0x00| 0,   5,   0, 0x60| 9}, // frame_149_climbing_15
    { 184, 0x00|16,   0,   2, 0x80| 0}, // frame_150_parry
    { 174, 0x00|26,   0,   2, 0x80| 0}, // frame_151_strike_1
    { 175, 0x00|18,   3,   2, 0x00| 0}, // frame_152_strike_2
    { 176, 0x00|22,   7,   2, 0xC0| 4}, // frame_153_strike_3
    { 177, 0x00|21,  10,   2, 0x00| 0}, // frame_154_poking
    { 178, 0x00|23,   7,   2, 0x80| 0}, // frame_155_post_strike_1
    { 179, 0x00|25,   4,   2, 0x80| 0}, // frame_156_post_strike_2
    { 180, 0x00|24,   0,   2, 0xC0|14}, // frame_157_walk_with_sword
    { 181, 0x00|15,   0,   2, 0xC0|13}, // frame_158_stand_with_sword
    { 182, 0x00|20,   3,   2, 0x00| 0}, // frame_159_fighting
    { 183, 0x00|31,   3,   2, 0x00| 0}, // frame_160_fighting
    { 184, 0x00|16,   0,   2, 0x80| 0}, // frame_161_parry
    { 185, 0x00|17,   0,   2, 0x80| 0}, // frame_162_block_to_strike
    { 186, 0x00|32,   0,   2, 0x00| 0}, // frame_163_fighting
    { 187, 0x00|33,   0,   2, 0x80| 0}, // frame_164_fighting
    { 188, 0x00|34,   2,   2, 0xC0| 3}, // frame_165_walk_with_sword
    {  14, 0x00| 0,   0,   0, 0x40| 3}, // frame_166_stand_inactive
    { 189, 0x00|19,   7,   2, 0x80| 0}, // frame_167_blocked
    { 190, 0x00|14,   1,   2, 0x80| 0}, // frame_168_back
    { 191, 0x00|27,   0,   2, 0x80| 0}, // frame_169_begin_block
    { 181, 0x00|15,   0,   2, 0xC0|13}, // frame_170_stand_with_sword
    { 181, 0x00|15,   0,   2, 0xC0|13}, // frame_171_stand_with_sword
    { 112, 0x00|43,   0,   0, 0xC0| 6}, // frame_172_jumpfall_2
    { 113, 0x00|44,   0,   0, 0x40| 6}, // frame_173_jumpfall_3
    { 114, 0x00|45,   0,   0, 0xC0| 5}, // frame_174_jumpfall_4
    { 115, 0x00|46,   0,   0, 0x40| 5}, // frame_175_jumpfall_5
    { 114, 0x00| 0,   0,   0, 0xC0| 5}, // unused
    {  78, 0x00| 0,   0,   3, 0x80|10}, // frame_177_spiked
    {  77, 0x00| 0,   4,   3, 0x80| 7}, // frame_178_chomped
    { 211, 0x00| 0,   0,   1, 0x40| 4}, // frame_179_collapse_1
    { 212, 0x00| 0,   0,   1, 0x40| 4}, // frame_180_collapse_2
    { 213, 0x00| 0,   0,   1, 0x40| 4}, // frame_181_collapse_3
    { 214, 0x00| 0,   0,   1, 0x40| 7}, // frame_182_collapse_4
    { 215, 0x00| 0,   0,   7, 0x40| 7}, // frame_183_collapse_5
    { 224, 0x00| 0,  -8,   0, 0x00| 0}, // frame_184_spiked_hanging (custom)
    {  79, 0x00| 0,   4,   7, 0x40| 7}, // frame_185_dead
    { 130, 0x00| 0,   0,   0, 0x40| 4}, // frame_186_mouse_1
    { 131, 0x00| 0,   0,   0, 0x40| 4}, // frame_187_mouse_2
    { 132, 0x00| 0,   0,   2, 0x40| 4}, // frame_188_mouse_stand
    { 255, 0x00| 0,   0,   0, 0x00| 0}, // unused
    { 255, 0x00| 0,   0,   0, 0x00| 0}, // unused
    { 192, 0x00| 0,   0,   0, 0x00| 0}, // frame_191_drink
    { 193, 0x00| 0,   0,   1, 0x00| 0}, // frame_192_drink
    { 194, 0x00| 0,   0,   0, 0x80| 0}, // frame_193_drink
    { 195, 0x00| 0,   0,   0, 0x00| 0}, // frame_194_drink
    { 196, 0x00| 0,  -1,   0, 0x00| 0}, // frame_195_drink
    { 197, 0x00| 0,  -1,   0, 0x00| 0}, // frame_196_drink
    { 198, 0x00| 0,  -1,   0, 0x00| 0}, // frame_197_drink
    { 199, 0x00| 0,  -4,   0, 0x00| 0}, // frame_198_drink
    { 200, 0x00| 0,  -4,   0, 0x80| 0}, // frame_199_drink
    { 201, 0x00| 0,  -4,   0, 0x00| 0}, // frame_200_drink
    { 202, 0x00| 0,  -4,   0, 0x00| 0}, // frame_201_drink
    { 203, 0x00| 0,  -4,   0, 0x00| 0}, // frame_202_drink
    { 204, 0x00| 0,  -4,   0, 0x00| 0}, // frame_203_drink
    { 205, 0x00| 0,  -5,   0, 0x00| 0}, // frame_204_drink
    { 206, 0x00| 0,  -5,   0, 0x00| 0}, // frame_205_drink
    { 255, 0x00| 0,   0,   0, 0x00| 0}, // unused
    { 207, 0x00| 0,   0,   1, 0x40| 6}, // frame_207_draw_1
    { 208, 0x00| 0,   0,   1, 0xC0| 6}, // frame_208_draw_2
    { 209, 0x00| 0,   0,   1, 0xC0| 8}, // frame_209_draw_3
    { 210, 0x00| 0,   0,   1, 0x40|10}, // frame_210_draw_4
    { 220, 0x00| 0,   5,   3, 0x40| 5}, // frame_211_sit_in_cell (custom)
    { 219, 0x00| 0,   0,   0, 0x00| 4}, // frame_212_flame_silhouette (custom)
    { 227, 0x00| 0,   3,   0, 0x40| 7}, // frame_213_kick,
    { 221, 0x00| 0,   0,   0, 0x00| 0}, // frame_214_turn_engarde_1 (custom)
    { 222, 0x00| 0,   0,   0, 0x00| 0}, // frame_215_turn_engarde_2 (custom)
    { 223, 0x00|14,   0,   0, 0x00| 0}, // frame_216_turn_engarde_3 (custom)
    {  52, 0x00| 0,   0,   0, 0x80| 0}, // frame_217_exit_stairs_1
    {  53, 0x00| 0,   0,   0, 0x00| 0}, // frame_218_exit_stairs_2
    {  54, 0x00| 0,   0,   0, 0x00| 0}, // frame_219_exit_stairs_3
    {  55, 0x00| 0,   0,   0, 0x00| 0}, // frame_220_exit_stairs_4
    {  56, 0x00| 0,   0,   0, 0x80| 0}, // frame_221_exit_stairs_5
    {  57, 0x00| 0,   0,   0, 0x00| 0}, // frame_222_exit_stairs_6
    {  58, 0x00| 0,   0,   0, 0x00| 0}, // frame_223_exit_stairs_7
    {  59, 0x00| 0,   0,   0, 0x00| 0}, // frame_224_exit_stairs_8
    {  60, 0x00| 0,   0,   0, 0x80| 0}, // frame_225_exit_stairs_9
    {  61, 0x00| 0,   0,   0, 0x00| 0}, // frame_226_exit_stairs_10
    {  62, 0x00| 0,   0,   0, 0x80| 0}, // frame_227_exit_stairs_11
    {  63, 0x00| 0,   0,   0, 0x00| 0}, // frame_228_exit_stairs_12
    { 160, 0x00|35,   1,   1, 0xC0| 3}, // frame_229_found_sword
    { 161, 0x00|36,   1,   1, 0x40| 9}, // frame_230_sheathe
    { 162, 0x00|37,   2,   1, 0xC0| 3}, // frame_231_sheathe
    { 163, 0x00|38,   2,   1, 0x40| 9}, // frame_232_sheathe
    { 164, 0x00|39,   0,   1, 0xC0| 3}, // frame_233_sheathe
    { 165, 0x00|40,   1,   1, 0x40| 9}, // frame_234_sheathe
    { 166, 0x00|41,   1,   1, 0x40| 3}, // frame_235_sheathe
    { 167, 0x00|42,   1,   1, 0xC0| 9}, // frame_236_sheathe
    { 168, 0x00| 0,   4,   1, 0xC0| 6}, // frame_237_sheathe
    { 169, 0x00| 0,   3,   1, 0xC0|10}, // frame_238_sheathe
    { 170, 0x00| 0,   1,   1, 0x40| 3}, // frame_239_sheathe
    { 171, 0x00| 0,   1,   1, 0xC0| 8}, // frame_240_sheathe
};

// data:1496
// { image #, sword bytes, x-offset, y-offset, flags }
const frame_type frame_tbl_guard[] = {
    { 255, 0x00| 0,   0,   0, 0x00| 0}, // frame_0 (do not draw)
    {  12, 0xC0|13,   2,   1, 0x00| 0}, // frame_150_parry
    {   2, 0xC0| 1,   3,   1, 0x00| 0}, // frame_151_strike_1
    {   3, 0xC0| 2,   4,   1, 0x00| 0}, // frame_152_strike_2
    {   4, 0xC0| 3,   7,   1, 0x40| 4}, // frame_153_strike_3
    {   5, 0xC0| 4,  10,   1, 0x00| 0}, // frame_154_poking
    {   6, 0xC0| 5,   7,   1, 0x80| 0}, // frame_155_post_strike_1
    {   7, 0xC0| 6,   4,   1, 0x80| 0}, // frame_156_post_strike_2
    {   8, 0xC0| 7,   0,   1, 0x80| 0}, // frame_157_walk_with_sword
    {   9, 0xC0| 8,   0,   1, 0xC0|13}, // frame_158_stand_with_sword
    {  10, 0xC0|11,   7,   1, 0x80| 0}, // frame_159_fighting
    {  11, 0xC0|12,   3,   1, 0x00| 0}, // frame_160_fighting
    {  12, 0xC0|13,   2,   1, 0x00| 0}, // frame_161_parry
    {  13, 0xC0| 0,   2,   1, 0x00| 0}, // frame_162_block_to_strike
    {  14, 0xC0|28,   0,   1, 0x00| 0}, // frame_163_fighting
    {  15, 0xC0|29,   0,   1, 0x80| 0}, // frame_164_fighting
    {  16, 0xC0|30,   2,   1, 0xC0| 3}, // frame_165_walk_with_sword
    {  17, 0xC0| 9,  -1,   1, 0x40| 8}, // frame_166_stand_inactive
    {  18, 0xC0|10,   7,   1, 0x80| 0}, // frame_167_blocked
    {  19, 0xC0|14,   3,   1, 0x80| 0}, // frame_168_back
    {   9, 0xC0| 8,   0,   1, 0x80| 0}, // frame_169_begin_block
    {  20, 0xC0| 8,   0,   1, 0xC0|13}, // frame_170_stand_with_sword
    {  21, 0xC0| 8,   0,   1, 0xC0|13}, // frame_171_stand_with_sword
    {  22, 0xC0|47,   0,   0, 0xC0| 6}, // frame_102_start_fall_1
    {  23, 0xC0|48,   0,   0, 0x40| 6}, // frame_103_start_fall_2
    {  24, 0xC0|49,   0,   0, 0xC0| 5}, // frame_104_start_fall_3
    {  24, 0xC0|49,   0,   0, 0xC0| 5}, // frame_105_start_fall_4
    {  24, 0xC0|49,   0,   0, 0xC0| 5}, // frame_106_fall
    {  26, 0xC0| 0,   0,   3, 0x80|10}, // frame_177_spiked
    {  27, 0xC0| 0,   4,   4, 0x80| 7}, // frame_178_chomped
    {  28, 0xC0| 0,  -2,   1, 0x40| 4}, // frame_179_collapse_1
    {  29, 0xC0| 0,  -2,   1, 0x40| 4}, // frame_180_collapse_2
    {  30, 0xC0| 0,  -2,   1, 0x40| 4}, // frame_181_collapse_3
    {  31, 0xC0| 0,  -2,   2, 0x40| 7}, // frame_182_collapse_4
    {  32, 0xC0| 0,  -2,   2, 0x40|10}, // frame_183_collapse_5
    {  48, 0xC0| 0,   4,   0, 0x00| 0}, // frame_184_spiked_hanging (custom - birdman only)
    {  33, 0xC0| 0,   3,   4, 0xC0| 9}, // frame_185_dead
    { 255, 0x00| 0,   0,   0, 0x00| 0}, // unused
    { 255, 0x00| 0,   0,   0, 0x00| 0}, // unused
    { 255, 0x00| 0,   0,   0, 0x00| 0}, // unused
    { 255, 0x00| 0,   0,   0, 0x00| 0}, // unused
    {  35, 0xC0| 0,   0,   3, 0x00| 0}, // frame_190_start_bowing_1 (custom - birdman only)
    {  36, 0xC0| 0,   0,   3, 0x40| 0}, // frame_191_start_bowing_2 (custom - birdman only)
    {  37, 0xC0| 0,   0,   3, 0x40| 0}, // frame_192_start_bowing_3 (custom - birdman only)
    {  38, 0xC0| 0,   0,   3, 0x00| 0}, // frame_193_start_bowing_4 (custom - birdman only)
    {  39, 0xC0| 0,   0,   3, 0x00| 0}, // frame_194_start_bowing_5 (custom - birdman only)
    {  40, 0xC0| 0,   0,   3, 0x00| 0}, // frame_195_start_bowing_6 (custom - birdman only)
    {  41, 0xC0| 0,   2,   3, 0x00| 2}, // frame_196_start_bowing_7 (custom - birdman only)
    {  42, 0xC0| 0,   3,   3, 0x00| 3}, // frame_197_start_bowing_8 (custom - birdman only)
    {  43, 0xC0| 0,   5,   3, 0x00| 5}, // frame_198_start_bowing_9 (custom - birdman only)
    {  44, 0xC0| 0,   6,   3, 0x00| 6}, // frame_199_start_bowing_10 (custom - birdman only)
    {  45, 0xC0| 0,  10,   3, 0x00|10}, // frame_200_start_bowing_11 (custom - birdman only)
    {  46, 0xC0| 0,  10,   4, 0x00| 4}, // frame_201_start_bowing_12 (custom - birdman only)
    {  47, 0xC0| 0,  10,   3, 0x40| 4}, // frame_202_bowing (custom - birdman only)
};

// data:1564
// { image #, sword bytes, x-offset, y-offset, flags }
const frame_type frame_tbl_cuts[] = { // numbers refer to frames in the "alternate" frame sets
    { 255, 0x00| 0,   0,   0, 0x00| 0}, // frame_0 (do not draw)
    {  15, 0x40| 0,   0,   0, 0x00| 0}, // frame_1_princess_slump_1
    {   1, 0x40| 0,   0,   0, 0x80| 0}, // frame_2_princess_turn_1
    {   2, 0x40| 0,   0,   0, 0x80| 0}, // frame_3_princess_turn_2
    {   3, 0x40| 0,   0,   0, 0x80| 0}, // frame_4_princess_turn_3
    {   4, 0x40| 0,  -1,   0, 0x00| 0}, // frame_5_princess_turn_4
    {   5, 0x40| 0,   2,   0, 0x80| 0}, // frame_6_princess_turn_5
    {   6, 0x40| 0,   2,   0, 0x00| 0}, // frame_7_princess_turn_6
    {   7, 0x40| 0,   0,   0, 0x80| 0}, // frame_8_princess_turn_7
    {   8, 0x40| 0,   1,   0, 0x80| 0}, // frame_9_princess_turn_8
    { 255, 0x00| 0,   0,   0, 0x00| 0}, // unused
    {   0, 0x40| 0,   0,   0, 0x80| 0}, // frame_11_princess_stand
    {   9, 0x40| 0,   0,   0, 0x80| 0}, // frame_12_princess_step_back
    {  10, 0x40| 0,   0,   0, 0x00| 0}, // frame_13_princess_step_back
    {  11, 0x40| 0,   0,   0, 0x80| 0}, // frame_14_princess_step_back
    {  12, 0x40| 0,   0,   0, 0x80| 0}, // frame_15_princess_step_back
    {  13, 0x40| 0,   0,   0, 0x80| 0}, // frame_16_princess_step_back
    {  14, 0x40| 0,   0,   0, 0x00| 0}, // frame_17_princess_step_back_stand
    {  16, 0x40| 0,   0,   0, 0x00| 0}, // frame_18_princess_slump_2
    {   0, 0x80| 0,   0,   0, 0x00| 0}, // frame_19_princess_lie
    {   2, 0x80| 0,   0,   0, 0x00| 0}, // frame_20_embrace_1
    {   3, 0x80| 0,   0,   0, 0x00| 0}, // frame_21_embrace_2
    {   4, 0x80| 0,   0,   0, 0x80| 0}, // frame_22_embrace_3
    {   5, 0x80| 0,   0,   0, 0x00| 0}, // frame_23_embrace_4
    {   6, 0x80| 0,   0,   0, 0x80| 0}, // frame_24_embrace_5
    {   7, 0x80| 0,   0,   0, 0x80| 0}, // frame_25_embrace_6
    {   8, 0x80| 0,   0,   0, 0x00| 0}, // frame_26_embrace_7
    {   9, 0x80| 0,   0,   0, 0x00| 0}, // frame_27_embrace_8
    {  10, 0x80| 0,   0,   0, 0x00| 0}, // frame_28_embrace_9
    {  11, 0x80| 0,   0,   0, 0x00| 0}, // frame_29_embrace_10
    {  12, 0x80| 0,   0,   0, 0x00| 0}, // frame_30_embrace_11
    {  13, 0x80| 0,   0,   0, 0x00| 0}, // frame_31_embrace_12
    {  14, 0x80| 0,   0,   0, 0x00| 0}, // frame_32_embrace_13
    {  15, 0x80| 0,   0,   0, 0x00| 0}, // frame_33_embrace_14
    {  16, 0x80| 0,   0,   0, 0x00| 0}, // frame_34_princess_rise_1
    {  17, 0x80| 0,   0,   0, 0x00| 0}, // frame_35_princess_rise_2
    {  18, 0x80| 0,   0,   0, 0x00| 0}, // frame_36_princess_rise_3
    {  19, 0x80| 0,   0,   0, 0x00| 0}, // frame_37_princess_rise_4
    {  20, 0x80| 0,   0,   0, 0x80| 0}, // frame_38_princess_rise_5
    {  21, 0x80| 0,   0,   0, 0x80| 0}, // frame_39_princess_rise_6
    {  22, 0x80| 0,   1,   0, 0x00| 0}, // frame_40_princess_rise_7
    {  23, 0x80| 0,  -1,   0, 0x00| 0}, // frame_41_princess_rise_8
    {  24, 0x80| 0,   2,   0, 0x00| 0}, // frame_42_princess_rise_9
    {  25, 0x80| 0,   1,   0, 0x80| 0}, // frame_43_princess_rise_10
    {  26, 0x80| 0,   0,   0, 0x80| 0}, // frame_44_princess_rise_11
    {  27, 0x80| 0,   0,   0, 0x80| 0}, // frame_45_princess_rise_12
    {  28, 0x80| 0,   0,   0, 0x80| 0}, // frame_46_princess_rise_13
    {  29, 0x80| 0,  -1,   0, 0x00| 0}, // frame_47_princess_rise_14
    {   0, 0x80| 0,   0,   0, 0x80| 0}, // frame_48_vizier_walk_1
    {   1, 0x80| 0,   0,   0, 0x80| 0}, // frame_49_vizier_walk_2
    {   2, 0x80| 0,   0,   0, 0x80| 0}, // frame_50_vizier_walk_3
    {   3, 0x80| 0,   0,   0, 0x00| 0}, // frame_51_vizier_walk_4a
    {   4, 0x80| 0,   0,   0, 0x00| 0}, // frame_52_vizier_walk_5
    {   5, 0x80| 0,   0,   0, 0x80| 0}, // frame_53_vizier_walk_6
    {   6, 0x80| 0,   0,   0, 0x80| 0}, // frame_54_vizier_stand_1
    {   7, 0x80| 0,   0,   0, 0x80| 0}, // frame_55_vizier_stand_2
    {   8, 0x80| 0,   0,   0, 0x80| 0}, // frame_56_vizier_stand_3
    {   9, 0x80| 0,   0,   0, 0x80| 0}, // frame_57_vizier_turn_1
    {  10, 0x80| 0,   0,   0, 0x80| 0}, // frame_58_vizier_turn_2
    {  11, 0x80| 0,   0,   0, 0x80| 0}, // frame_59_vizier_turn_3
    {  12, 0x80| 0,   0,   0, 0x80| 0}, // frame_60_vizier_turn_4
    {  13, 0x80| 0,   0,   0, 0x00| 0}, // frame_61_vizier_turn_5
    {  14, 0x80| 0,   0,   0, 0x80| 0}, // frame_62_vizier_turn_6
    {  15, 0x80| 0,   0,   0, 0x00| 0}, // frame_63_vizier_turn_7
    {  16, 0x80| 0,   0,   0, 0x00| 0}, // frame_64_vizier_turn_8
    {  17, 0x80| 0,   0,   0, 0x80| 0}, // frame_65_vizier_turn_9
    {  18, 0x80| 0,   0,   0, 0x00| 0}, // frame_66_vizier_turn_10
    {  19, 0x80| 0,   3,   0, 0x00| 0}, // frame_67_vizier_cast_1
    {  20, 0x80| 0,   3,   0, 0x00| 0}, // frame_68_vizier_cast_2
    {  21, 0x80| 0,   3,   0, 0x00| 0}, // frame_69_vizier_cast_3
    {  22, 0x80| 0,   2,   0, 0x00| 0}, // frame_70_vizier_cast_4
    {  23, 0x80| 0,   3,   0, 0x80| 0}, // frame_71_vizier_cast_5
    {  24, 0x80| 0,   5,   0, 0x00| 0}, // frame_72_vizier_cast_6
    {  25, 0x80| 0,   5,   0, 0x00| 0}, // frame_73_vizier_cast_7
    {  26, 0x80| 0,   1,   0, 0x80| 0}, // frame_74_vizier_cast_8
    {  27, 0x80| 0,   2,   0, 0x80| 0}, // frame_75_vizier_cast_9
    {  28, 0x80| 0,   2,   0, 0x80| 0}, // frame_76_vizier_cast_10
    {  29, 0x80| 0,   1,   0, 0x80| 0}, // frame_77_vizier_cast_11
    {  30, 0x80| 0,   1,   0, 0x00| 0}, // frame_78_vizier_cast_12
    {  31, 0x80| 0,   2,   0, 0x00| 0}, // frame_79_vizier_cast_13
    {  32, 0x80| 0,   3,   0, 0x00| 0}, // frame_80_vizier_cast_14
    {  33, 0x80| 0,   3,   0, 0x00| 0}, // frame_81_vizier_cast_15
    {  34, 0x80| 0,   0,   0, 0x80| 0}, // frame_82_vizier_cast_16
    {  35, 0x80| 0,   2,   0, 0x80| 0}, // frame_83_vizier_cast_17
    {  36, 0x80| 0,   2,   0, 0x80| 0}, // frame_84_vizier_cast_18
    {  37, 0x80| 0,   1,   0, 0x00| 0}, // frame_85_vizier_cast_19
    {  30, 0x80| 0,   0,   0, 0x00| 0}, // frame_86_dead_princess (custom)
    {  38, 0x80| 0,   0,   0, 0x00| 0}, // frame_87_vizier_walk_4b (custom)
};

// data:1712
// { image #, x-offset, y-offset }
//
// image #: references the sword sprite # in the PRINCE.DAT file
//
// x-offset, y-offset - on-screen image x/y offsets in screen pixels
const sword_table_type sword_tbl[] = {
    { 255,   0,   0}, // no sword
    {   0,   0,  -9}, // guard: frame_151_strike_1
    {   5,   6, -27}, // guard: frame_152_strike_2
    {   1,   7, -25}, // guard: frame_153_strike_3
    {   2,  15, -26}, // guard: frame_154_poking
    {   6,   7, -14}, // guard: frame_155_post_strike_1
    {   7,   0,  -5}, // guard: frame_156_post_strike_2
    {   3,  17, -16}, // guard: frame_157_walk_with_sword
    {   4,  15, -19}, // guard: frame_158_stand_with_sword, frame_169_begin_block, frame_170_stand_with_sword, frame_171_stand_with_sword
    {  30,  12,  -9}, // guard: frame_166_stand_inactive
    {   8,  13, -34}, // guard: frame_167_blocked
    {   9,   7, -25}, // guard: frame_159_fighting
    {  10,  10, -16}, // guard: frame_160_fighting
    {  11,  10, -11}, // guard: frame_161_parry
    {  12,  22, -21}, // kid and guard: frame_168_back, kid: frame_216_turn_engarde_3
    {  13,  27, -21}, // kid: frame_158_stand_with_sword, frame_170_stand_with_sword, frame_171_stand_with_sword
    {  14,  13, -35}, // kid: frame_150_parry
    {  15,   0, -38}, // kid: frame_162_block_to_strike
    {  16,   3, -28}, // kid: frame_152_strike_2
    {  17,  21, -19}, // kid: frame_167_blocked
    {  18,  14, -23}, // kid: frame_159_fighting
    {  19,  21, -20}, // kid: frame_154_poking
    {  19,  20, -23}, // kid: frame_153_strike_3
    {  17,  15, -13}, // kid: frame_155_post_strike_1
    {  12,  21, -20}, // kid: frame_157_walk_with_sword
    {   7,   3,  -5}, // kid: frame_156_post_strike_2
    {   1,  11, -23}, // kid: frame_151_strike_1
    {  28,  14, -28}, // kid: frame_169_begin_block
    {   8,   7, -27}, // guard: frame_163_fighting
    {   4,   5, -23}, // guard: frame_164_fighting
    {   4,   9, -21}, // guard: frame_165_walk_with_sword
    {  12,  23, -20}, // kid: frame_160_fighting
    {  13,  27, -21}, // kid: frame_163_fighting
    {  13,  27, -22}, // kid: frame_164_fighting
    {  13,  27, -21}, // kid: frame_165_walk_with_sword
    {  20,   7, -32}, // kid: frame_229_found_sword
    {  21,   9, -30}, // kid: frame_230_sheathe
    {  22,  11, -31}, // kid: frame_231_sheathe
    {  23,  10, -28}, // kid: frame_232_sheathe
    {  24,  20, -28}, // kid: frame_233_sheathe
    {  25,  25, -27}, // kid: frame_234_sheathe
    {  26,  16, -22}, // kid: frame_235_sheathe
    {  27,   8, -19}, // kid: frame_236_sheathe
    { 255,  14, -25}, // kid: frame_172_jumpfall_2
    {  17,  18, -19}, // kid: frame_173_jumpfall_3
    {  17,  17, -20}, // kid: frame_174_jumpfall_4
    {   8,   8, -37}, // kid: frame_175_jumpfall_5
    {  31,  14, -24}, // guard: frame_102_start_fall_1
    {  32,  13, -24}, // guard: frame_103_start_fall_2
    {  33,   7, -14}, // guard: frame_104_start_fall_3, frame_105_start_fall_4, frame_106_fall
    {   8,   8, -37}, // unused? seems to be a duplicate of the frame_175_jumpfall_5 sword offset
};

// custom logic - skeletons use different swords and the sword table
// { image #, x-offset, y-offset }
const sword_table_type sword_skel_tbl[] = {
    { 255,   0,   0}, // no sword
    {   0,   0,  -9}, // frame_151_strike_1
    {   5,  -9, -29}, // frame_152_strike_2
    {   1,   7, -25}, // frame_153_strike_3
    {   2,  17, -26}, // frame_154_poking
    {   6,   7, -14}, // frame_155_post_strike_1
    {   7,   0,  -5}, // frame_156_post_strike_2
    {   3,  17, -16}, // frame_157_walk_with_sword
    {   4,  16, -19}, // frame_158_stand_with_sword, frame_169_begin_block, frame_170_stand_with_sword, frame_171_stand_with_sword
    {  30,  12,  -9}, // frame_166_stand_inactive
    {   8,  13, -34}, // frame_167_blocked
    {   9,   7, -25}, // frame_159_fighting
    {  10,  10, -16}, // frame_160_fighting
    {  11,  10, -11}, // frame_161_parry
    {  12,  22, -21}, // frame_168_back
    {  13,  27, -21}, // unused
    {  14,  13, -35}, // unused
    {  15,   0, -38}, // unused
    {  16,   0, -29}, // unused
    {  17,  21, -19}, // unused
    {  18,  14, -23}, // unused
    {  19,  21, -22}, // unused
    {  19,  22, -23}, // unused
    {  17,   7, -13}, // unused
    {  17,  15, -18}, // unused
    {   7,   0,  -8}, // unused
    {   1,   7, -27}, // unused
    {  28,  14, -28}, // unused
    {   8,   7, -27}, // frame_163_fighting
    {   4,   6, -23}, // frame_164_fighting
    {   4,   9, -21}, // frame_165_walk_with_sword
    {  10,  11, -18}, // unused
    {  13,  27, -21}, // unused
    {  13,  28, -21}, // unused
    {  13,  27, -21}, // unused
    {  20,   7, -32}, // unused
    {  21,  14, -32}, // unused
    {  22,  14, -31}, // unused
    {  23,  14, -29}, // unused
    {  24,  28, -28}, // unused
    {  25,  28, -28}, // unused
    {  26,  21, -25}, // unused
    {  27,  14, -22}, // unused
    { 255,  14, -25}, // unused
    { 255,  21, -25}, // unused
    {  29,   0, -16}, // unused
    {   8,   8, -37}, // unused
    {  31,  14, -24}, // frame_102_start_fall_1
    {  32,  14, -24}, // frame_103_start_fall_2
    {  33,   7, -14}, // frame_104_start_fall_3, frame_105_start_fall_4, frame_106_fall
    {   8,   8, -37}, // unused
};
// *INDENT-ON*
