/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#ifndef FRAME_TABLES_H
#define FRAME_TABLES_H

enum frameids {
    frame_0 = 0,
    frame_1_start_run = 1,
    frame_2_start_run = 2,
    frame_3_start_run = 3,
    frame_4_start_run = 4,
    frame_5_start_run = 5,
    frame_6_start_run = 6,
    frame_7_run = 7,
    frame_8_run = 8,
    frame_9_run = 9,
    frame_10_run = 10,
    frame_11_run = 11,
    frame_12_run = 12,
    frame_13_run = 13,
    frame_14_run = 14,
    frame_15_stand = 15,
    frame_16_standing_jump_1 = 16,
    frame_17_standing_jump_2 = 17,
    frame_18_standing_jump_3 = 18,
    frame_19_standing_jump_4 = 19,
    frame_20_standing_jump_5 = 20,
    frame_21_standing_jump_6 = 21,
    frame_22_standing_jump_7 = 22,
    frame_23_standing_jump_8 = 23,
    frame_24_standing_jump_9 = 24,
    frame_25_standing_jump_10 = 25,
    frame_26_standing_jump_11 = 26,
    frame_27_standing_jump_12 = 27,
    frame_28_standing_jump_13 = 28,
    frame_29_standing_jump_14 = 29,
    frame_30_standing_jump_15 = 30,
    frame_31_standing_jump_16 = 31,
    frame_32_standing_jump_17 = 32,
    frame_33_standing_jump_18 = 33,
    frame_34_start_run_jump_1 = 34,
    frame_35_start_run_jump_2 = 35,
    frame_36_start_run_jump_3 = 36,
    frame_37_start_run_jump_4 = 37,
    frame_38_start_run_jump_5 = 38,
    frame_39_start_run_jump_6 = 39,
    frame_40_running_jump_1 = 40,
    frame_41_running_jump_2 = 41,
    frame_42_running_jump_3 = 42,
    frame_43_running_jump_4 = 43,
    frame_44_running_jump_5 = 44,
    frame_45_turn = 45,
    frame_46_turn = 46,
    frame_47_turn = 47,
    frame_48_turn = 48,
    frame_49_turn = 49,
    frame_50_turn = 50,
    frame_51_turn = 51,
    frame_52_turn = 52,
    frame_53_runturn = 53,
    frame_54_runturn = 54,
    frame_55_runturn = 55,
    frame_56_runturn = 56,
    frame_57_runturn = 57,
    frame_58_runturn = 58,
    frame_59_runturn = 59,
    frame_60_runturn = 60,
    frame_61_runturn = 61,
    frame_62_runturn = 62,
    frame_63_runturn = 63,
    frame_64_runturn = 64,
    frame_65_runturn = 65,
    frame_67_start_jump_up_1 = 67,
    frame_68_start_jump_up_2 = 68,
    frame_69_start_jump_up_3 = 69,
    frame_70_jumphang = 70,
    frame_71_jumphang = 71,
    frame_72_jumphang = 72,
    frame_73_jumphang = 73,
    frame_74_jumphang = 74,
    frame_75_jumphang = 75,
    frame_76_jumphang = 76,
    frame_77_jumphang = 77,
    frame_78_jumphang = 78,
    frame_79_jumphang = 79,
    frame_80_jumphang = 80,
    frame_81_hangdrop_1 = 81,
    frame_82_hangdrop_2 = 82,
    frame_83_hangdrop_3 = 83,
    frame_84_hangdrop_4 = 84,
    frame_85_hangdrop_5 = 85,
    frame_86_test_foot = 86,
    frame_87_hanging_1 = 87,
    frame_88_hanging_2 = 88,
    frame_89_hanging_3 = 89,
    frame_90_hanging_4 = 90,
    frame_91_hanging_5 = 91,
    frame_92_hanging_6 = 92,
    frame_93_hanging_7 = 93,
    frame_94_hanging_8 = 94,
    frame_95_hanging_9 = 95,
    frame_96_hanging_10 = 96,
    frame_97_hanging_11 = 97,
    frame_98_hanging_12 = 98,
    frame_99_hanging_13 = 99,
    // custom logic - first frame of vertical spike animation
    frame_100_spiking_hanging = 100,
    frame_102_start_fall_1 = 102,
    frame_103_start_fall_2 = 103,
    frame_104_start_fall_3 = 104,
    frame_105_start_fall_4 = 105,
    frame_106_fall = 106,
    frame_107_fall_land_1 = 107,
    frame_108_fall_land_2 = 108,
    frame_109_crouch = 109,
    frame_110_stand_up_from_crouch_1 = 110,
    frame_111_stand_up_from_crouch_2 = 111,
    frame_112_stand_up_from_crouch_3 = 112,
    frame_113_stand_up_from_crouch_4 = 113,
    frame_114_stand_up_from_crouch_5 = 114,
    frame_115_stand_up_from_crouch_6 = 115,
    frame_116_stand_up_from_crouch_7 = 116,
    frame_117_stand_up_from_crouch_8 = 117,
    frame_118_stand_up_from_crouch_9 = 118,
    frame_119_stand_up_from_crouch_10 = 119,
    frame_121_stepping_1 = 121,
    frame_122_stepping_2 = 122,
    frame_123_stepping_3 = 123,
    frame_124_stepping_4 = 124,
    frame_125_stepping_5 = 125,
    frame_126_stepping_6 = 126,
    frame_127_stepping_7 = 127,
    frame_128_stepping_8 = 128,
    frame_129_stepping_9 = 129,
    frame_130_stepping_10 = 130,
    frame_131_stepping_11 = 131,
    frame_132_stepping_12 = 132,
    frame_133_sheathe = 133,
    frame_134_sheathe = 134,
    frame_135_climbing_1 = 135,
    frame_136_climbing_2 = 136,
    frame_137_climbing_3 = 137,
    frame_138_climbing_4 = 138,
    frame_139_climbing_5 = 139,
    frame_140_climbing_6 = 140,
    frame_141_climbing_7 = 141,
    frame_142_climbing_8 = 142,
    frame_143_climbing_9 = 143,
    frame_144_climbing_10 = 144,
    frame_145_climbing_11 = 145,
    frame_146_climbing_12 = 146,
    frame_147_climbing_13 = 147,
    frame_148_climbing_14 = 148,
    frame_149_climbing_15 = 149,
    frame_150_parry = 150,
    frame_151_strike_1 = 151,
    frame_152_strike_2 = 152,
    frame_153_strike_3 = 153,
    frame_154_poking = 154,
    frame_155_post_strike_1 = 155,
    frame_156_post_strike_2 = 156,
    frame_157_walk_with_sword = 157,
    frame_158_stand_with_sword = 158,
    frame_159_fighting = 159,
    frame_160_fighting = 160,
    frame_161_parry = 161,
    frame_162_block_to_strike = 162,
    frame_163_fighting = 163,
    frame_164_fighting = 164,
    frame_165_walk_with_sword = 165,
    frame_166_stand_inactive = 166,
    frame_167_blocked = 167,
    frame_168_back = 168,
    frame_169_begin_block = 169,
    frame_170_stand_with_sword = 170,
    frame_171_stand_with_sword = 171,
    frame_172_jumpfall_2 = 172,
    frame_173_jumpfall_3 = 173,
    frame_174_jumpfall_4 = 174,
    frame_175_jumpfall_5 = 175,
    frame_177_spiked = 177,
    frame_178_chomped = 178,
    frame_179_collapse_1 = 179,
    frame_180_collapse_2 = 180,
    frame_181_collapse_3 = 181,
    frame_182_collapse_4 = 182,
    frame_183_collapse_5 = 183,
    // custom logic - death frame of hanging on a vertical spike
    frame_184_spiked_hanging = 184,
    frame_185_dead = 185,
    frame_186_mouse_1 = 186,
    frame_187_mouse_2 = 187,
    frame_188_mouse_stand = 188,
    frame_191_drink = 191,
    frame_192_drink = 192,
    frame_193_drink = 193,
    frame_194_drink = 194,
    frame_195_drink = 195,
    frame_196_drink = 196,
    frame_197_drink = 197,
    frame_198_drink = 198,
    frame_199_drink = 199,
    frame_200_drink = 200,
    frame_201_drink = 201,
    frame_202_drink = 202,
    frame_203_drink = 203,
    frame_204_drink = 204,
    frame_205_drink = 205,
    frame_207_draw_1 = 207,
    frame_208_draw_2 = 208,
    frame_209_draw_3 = 209,
    frame_210_draw_4 = 210,
    // custom logic - kid sitting in a cell
    frame_211_sit_in_cell = 211,
    // custom logic - adds flame background behind shadow when picking up flaming sword
    frame_212_flame_silhouette = 212,
    // custom logic - kick in a fight
    frame_213_kick = 213,
    // custom logic - kid turns with sword
    frame_214_turn_engarde_1 = 214,
    frame_215_turn_engarde_2 = 215,
    frame_216_turn_engarde_3 = 216,
    frame_217_exit_stairs_1 = 217,
    frame_218_exit_stairs_2 = 218,
    frame_219_exit_stairs_3 = 219,
    frame_220_exit_stairs_4 = 220,
    frame_221_exit_stairs_5 = 221,
    frame_222_exit_stairs_6 = 222,
    frame_223_exit_stairs_7 = 223,
    frame_224_exit_stairs_8 = 224,
    frame_225_exit_stairs_9 = 225,
    frame_226_exit_stairs_10 = 226,
    frame_227_exit_stairs_11 = 227,
    frame_228_exit_stairs_12 = 228,
    frame_229_found_sword = 229,
    frame_230_sheathe = 230,
    frame_231_sheathe = 231,
    frame_232_sheathe = 232,
    frame_233_sheathe = 233,
    frame_234_sheathe = 234,
    frame_235_sheathe = 235,
    frame_236_sheathe = 236,
    frame_237_sheathe = 237,
    frame_238_sheathe = 238,
    frame_239_sheathe = 239,
    frame_240_sheathe = 240,
};

// cutscene frame ids
enum alt_frame_ids {
    frame_1_princess_slump_1 = 1,
    frame_2_princess_turn_1 = 2,
    frame_3_princess_turn_2 = 3,
    frame_4_princess_turn_3 = 4,
    frame_5_princess_turn_4 = 5,
    frame_6_princess_turn_5 = 6,
    frame_7_princess_turn_6 = 7,
    frame_8_princess_turn_7 = 8,
    frame_9_princess_turn_8 = 9,
    frame_11_princess_stand = 11,
    frame_12_princess_step_back = 12,
    frame_13_princess_step_back = 13,
    frame_14_princess_step_back = 14,
    frame_15_princess_step_back = 15,
    frame_16_princess_step_back = 16,
    frame_17_princess_step_back_stand = 17,
    frame_18_princess_slump_2 = 18,
    frame_19_princess_lie = 19,
    frame_20_embrace_1 = 20,
    frame_21_embrace_2 = 21,
    frame_22_embrace_3 = 22,
    frame_23_embrace_4 = 23,
    frame_24_embrace_5 = 24,
    frame_25_embrace_6 = 25,
    frame_26_embrace_7 = 26,
    frame_27_embrace_8 = 27,
    frame_28_embrace_9 = 28,
    frame_29_embrace_10 = 29,
    frame_30_embrace_11 = 30,
    frame_31_embrace_12 = 31,
    frame_32_embrace_13 = 32,
    frame_33_embrace_14 = 33,
    frame_34_princess_rise_1 = 34,
    frame_35_princess_rise_2 = 35,
    frame_36_princess_rise_3 = 36,
    frame_37_princess_rise_4 = 37,
    frame_38_princess_rise_5 = 38,
    frame_39_princess_rise_6 = 39,
    frame_40_princess_rise_7 = 40,
    frame_41_princess_rise_8 = 41,
    frame_42_princess_rise_9 = 42,
    frame_43_princess_rise_10 = 43,
    frame_44_princess_rise_11 = 44,
    frame_45_princess_rise_12 = 45,
    frame_46_princess_rise_13 = 46,
    frame_47_princess_rise_14 = 47,
    frame_48_vizier_walk_1 = 48,
    frame_49_vizier_walk_2 = 49,
    frame_50_vizier_walk_3 = 50,
    frame_51_vizier_walk_4a = 51,
    frame_52_vizier_walk_5 = 52,
    frame_53_vizier_walk_6 = 53,
    frame_54_vizier_stand_1 = 54,
    frame_55_vizier_stand_2 = 55,
    frame_56_vizier_stand_3 = 56,
    frame_57_vizier_turn_1 = 57,
    frame_58_vizier_turn_2 = 58,
    frame_59_vizier_turn_3 = 59,
    frame_60_vizier_turn_4 = 60,
    frame_61_vizier_turn_5 = 61,
    frame_62_vizier_turn_6 = 62,
    frame_63_vizier_turn_7 = 63,
    frame_64_vizier_turn_8 = 64,
    frame_65_vizier_turn_9 = 65,
    frame_66_vizier_turn_10 = 66,
    frame_67_vizier_cast_1 = 67,
    frame_68_vizier_cast_2 = 68,
    frame_69_vizier_cast_3 = 69,
    frame_70_vizier_cast_4 = 70,
    frame_71_vizier_cast_5 = 71,
    frame_72_vizier_cast_6 = 72,
    frame_73_vizier_cast_7 = 73,
    frame_74_vizier_cast_8 = 74,
    frame_75_vizier_cast_9 = 75,
    frame_76_vizier_cast_10 = 76,
    frame_77_vizier_cast_11 = 77,
    frame_78_vizier_cast_12 = 78,
    frame_79_vizier_cast_13 = 79,
    frame_80_vizier_cast_14 = 80,
    frame_81_vizier_cast_15 = 81,
    frame_82_vizier_cast_16 = 82,
    frame_83_vizier_cast_17 = 83,
    frame_84_vizier_cast_18 = 84,
    frame_85_vizier_cast_19 = 85,
    // custom logic - additional cutscene frames
    frame_86_dead_princess = 86,
    frame_87_vizier_walk_4b = 87,
};

// custom logic - bird bowing frames
enum bowingframeids {
    frame_190_start_bowing_1 = 190,
    frame_191_start_bowing_2 = 191,
    frame_192_start_bowing_3 = 192,
    frame_193_start_bowing_4 = 193,
    frame_194_start_bowing_5 = 194,
    frame_195_start_bowing_6 = 195,
    frame_196_start_bowing_7 = 196,
    frame_197_start_bowing_8 = 197,
    frame_198_start_bowing_9 = 198,
    frame_199_start_bowing_10 = 199,
    frame_200_start_bowing_11 = 200,
    frame_201_start_bowing_12 = 201,
    frame_202_bowing = 202,
};

extern const frame_type frame_table_kid[241];
extern const frame_type frame_tbl_guard[54];
extern const frame_type frame_tbl_cuts[88];
extern const sword_table_type sword_tbl[51];
extern const sword_table_type sword_skel_tbl[51];

#endif
