/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

// SEG000.C
void pop_main(void);
void init_game_main(void);
void start_game(void);
// custom logic - is kid going up the stairs?
bool going_up_stairs(void);
int process_key(bool from_paused);
// custom logic - kid says "whew" after climbing up an almost closed gate
// or when the gate is closed on level 1 in front of a guard
void say_whew(void);
void play_frame(void);
void draw_game_frame(void);
void anim_tile_modif(void);
void load_sound_names(void);
// custom logic - load additional palettes
void load_more_palettes(void);
// custom logic - returns if the level uses dungeon 2 graphics
bool is_dungeon_2(void);
// custom logic - returns if the level uses temple graphics
bool is_temple(void);
// custom logic - level 6 has a different color palette
void load_lev6_palette(void);
void load_lev_spr(int level);
void load_level(void);
void reset_level_unused_fields(bool loading_clean_level);
int play_kid_frame(void);
void play_guard_frame(void);
void find_all_rooms_from(byte starting_room);
void check_the_end(void);
void check_fall_flo(void);
void read_joyst_control(void);
void draw_kid_hp(short curr_hp, short max_hp);
// custom logic - clears kid hp
void clear_kid_hp(short max_hp);
void draw_guard_hp(short curr_hp, short max_hp);
void add_life(void);
void set_health_life(void);
void draw_hp(void);
void do_delta_hp(void);
void play_sound(int sound_id);
// custom logic - sets positional sound distance from kid with a specified loss multiplier
void set_sound_pos_with_dist_loss(int room, int col, int row, float distance_loss_multiplier);
// custom logic - sets positional sound distance from kid
void set_sound_pos(int room, int col, int row);
void play_next_sound(void);
void check_sword_vs_sword(void);
void load_chtab_from_file(int chtab_id, int resource, const char *filename, int palette_bits);
void free_all_chtabs_from(int first);
void load_more_opt_graf(const char *filename);
int do_paused(void);
void read_keyb_control(void);
void copy_screen_rect(const rect_type *source_rect_ptr);
void toggle_upside(void);
void feather_fall(void);
void gen_palace_wall_colors(void);
void show_title(void);
void transition_ltr(void);
void release_title_images(void);
void draw_full_image(enum full_image_id id);
void load_kid_sprite(void);
// custom logic - turns shadow into flame
void load_flame_sprite(void);
void clear_screen_and_sounds(void);
void load_all_sounds(void);
void free_optsnd_chtab(void);
void load_title_images(int bgcolor);
// custom logic - check for prison level, set timer and play out-of-time music
void check_prison_level(void);
void show_loading(void);
// custom logic - shadow gets united with kid
bool unite_shadow_with_kid(void);
// custom logic - can kid separate from shadow
sbyte can_kid_separate_from_shadow(void);
// custom logic - turn kid into shadow, if separation is possible
void separate_kid_from_shadow(void);
void check_quick_op(void);
void auto_save_level(int level);
void restore_room_after_quick_load(word prev_level);
// custom logic - load sound and music settings
void config_load(void);
// custom logic - save sound and music settings
void config_save(void);

// SEG001.C
int proc_cutscene_frame(int wait_frames);
void play_both_seq(void);
void draw_proom_drects(void);
void play_kid_seq(void);
void play_opp_seq(void);
void draw_princess_room_bg(void);
void seqtbl_offset_shad_char(int seq_index);
void seqtbl_offset_kid_char(int seq_index);
void init_mouse_cu8(void);
void init_mouse_go(void);
void princess_crouching(void);
void princess_stand(void);
void init_princess_x156(void);
void princess_lying(void);
void init_princess_right(void);
void init_ending_princess(void);
void init_mouse_1(void);
void init_princess(void);
void init_vizier(void);
void init_ending_kid(void);
void cutscene_8(void);
void cutscene_9(void);
// custom logic - cutscene when there is less than 5 min left
void cutscene_short_time(void);
void end_sequence_anim(void);
void time_expired(void);
void cutscene_12(void);
void cutscene_4(void);
void cutscene_2_6(void);
// custom logic - prince captured cut scene
void cutscene_15(void);
void pv_scene(void);
void set_hourglass_state(int state);
int hourglass_frame(void);
void princess_room_torch(void);
void draw_hourglass(void);
void reset_cutscene(void);
void do_flash(short color);
void delay_ticks(Uint32 ticks);
void remove_flash(void);
// custom logic - draws the palace image
void draw_palace_image(void);
void end_sequence(void);
void expired(void);
void load_intro(int which_imgs, void (*func)(void));
// custom logic - draw star in the provided rectangle
void draw_star_rect(rect_type* rect, int which_star, int mark_dirty);
void draw_star(int which_star, int mark_dirty);
void show_hof(void);
void hof_write(void);
void hof_read(void);
void show_hof_text(rect_type *rect, int x_align, int y_align, const char *text);
int fade_in_1(void);
int fade_out_1(void);

// SEG002.C
void do_init_shad(const byte *source, int seq_index);
void get_guard_hp(void);
void check_shadow(void);
// custom logic - enter shadow on level 10
void enter_shadow_level_10(void);
void enter_guard(void);
// custom logic - do guard fallout
void do_guard_fallout(byte start_fall);
void check_guard_fallout(void);
// custom logic - is the guard is a disappearing guard
bool is_disappearing_guard(void);
void leave_guard(void);
void follow_guard(void);
// custom logic - is there is an active guard in the room?
bool has_active_guard(word room_minus_1);
// custom logic - save guard in the specified room into the inactive guard slot
void inactivate_guard(word room_minus_1);
// custom logic - move the active guard in the current room into the inactive guard slot
void inactivate_active_guard(void);
// custom logic - finds the first available dead guard slot
char_type2* get_dead_guard_slot(int room_minus_1);
// custom logic - moves dead guard from the guard slot to a dead slot
void move_dead_guard_to_dead_slot(int room_minus_1);
void exit_room(void);
int goto_other_room(short direction);
short leave_room(void);
// custom logic - leave outbound dead guard
void leave_outbound_guard(word room_minus_1, byte guard_frame, sbyte guard_direction, byte guard_x, sbyte guard_row, byte alt_frame, short color);
// custom logic - move guard in room 18 on level 4 into the next room once the gate is open
void position_level_4_guard(void);
// custom logic - playing presentation music for the spilled poison potion
void play_poison_music(void);
void meet_Jaffar(bool right);
// custom logic - add the mirror and keep entry gate open
void enter_mirror_room(void);
// custom logic - keep the entry gate closed when entering the other side of room 1
void close_mirror_room_gate(void);
// custom logic - play music on level 4 when there is a wall in front of the gate
void play_wall_mus(void);
// custom logic - horse statue makes a neighing sound when prince appears in the same room
void horse_neigh(bool hurt);
// custom logic - drunk guard snores when kid enters the room without a sword on level 1
void guard_snores(void);
void move_0_nothing(void);
void move_1_forward(void);
void move_2_backward(void);
void move_3_up(void);
void move_4_down(void);
void move_up_back(void);
void move_down_back(void);
void move_down_forw(void);
void move_6_shift(void);
void move_7(void);
void autocontrol_opponent(void);
void autocontrol_mouse(void);
void autocontrol_shadow(void);
// custom logic - replace skeleton with bones tile
void inactivate_skel(sbyte direction);
void autocontrol_skeleton(void);
void autocontrol_kid(void);
void autocontrol_guard(void);
void autocontrol_guard_inactive(void);
void autocontrol_guard_active(void);
void autocontrol_guard_kid_far(void);
void guard_follows_kid_down(void);
void autocontrol_guard_kid_in_sight(short distance);
void autocontrol_guard_kid_armed(short distance);
void guard_advance(void);
void guard_block(void);
void guard_strike(void);
void hurt_by_sword(void);
void check_sword_hurt(void);
void check_sword_hurting(void);
void check_hurting(void);
void check_skel(void);
void do_auto_moves(const auto_move_type *moves_ptr);
void autocontrol_shadow_level6(void);
void autocontrol_shadow_level6b(void);
void autocontrol_shadow_level7(void);
void autocontrol_shadow_level9(void);
void autocontrol_shadow_level8(void);
void autocontrol_shadow_level10(void);

// SEG003.C
void init_game(int level);
void play_level(int level_number);
void do_startpos(void);
void set_start_pos(void);
void find_start_level_door(void);
void draw_level_first(void);
// custom logic - position dynamic guards
void pos_dynamic_level_guards(void);
// custom logic - position inactive guards
void pos_inactive_level_guards(void);
void redraw_screen(int drawing_different_room);
int play_level_2(void);
void redraw_at_char(void);
void redraw_at_char2(void);
void check_knock(void);
void timers(void);
// custom logic - process custom events
void process_custom_events(void);
void check_mirror(void);
void jump_through_mirror(void);
void check_mirror_image(void);
void bump_into_opponent(void);
void pos_guards(void);
void check_can_guard_see_kid(void);
// custom logic - check if inactive guard can see kid
void check_can_inactive_guard_see_kid(void);
// custom logic - checks whether the guard was following kid in the previous room before he dropped down
bool check_kid_dropped_into_next_room(void);
byte get_tile_at_kid(int xpos);
void do_mouse(int start_x);
// custom logic - has kid turn into mouse?
bool is_kid_mouse(void);
int flash_if_hurt(void);
void remove_flash_if_hurt(void);

// SEG004.C
void check_collisions(void);
void move_coll_to_prev(void);
void get_row_collision_data(short row, sbyte *row_coll_room_ptr, byte *row_coll_flags_ptr);
int get_left_wall_xpos(int room, int column, int row);
int get_right_wall_xpos(int room, int column, int row);
void check_bumped(void);
void check_bumped_look_left(void);
void check_bumped_look_right(void);
// custom logic - return is the tile is fake wall
int is_fake_wall(void);
int is_obstacle_at_col(int tile_col);
int is_obstacle(void);
int xpos_in_drawn_room(int xpos);
void bumped(sbyte delta_x, sbyte direction, int coll_tile);
void bumped_fall(void);
void bumped_floor(sbyte direction, int coll_tile);
void bumped_sound(void);
void clear_coll_rooms(void);
int can_bump_into_gate(void);
int get_edge_distance(void);
void check_chomped_kid(void);
void chomped(void);
void check_gate_push(void);
void check_guard_bumped(void);
void check_chomped_guard(void);
int check_chomped_here(void);
// custom logic - check for chomped dead or inactive guards
void check_dead_or_inactive_guards_chomped(void);
int dist_from_wall_forward(byte tiletype);
int dist_from_wall_behind(byte tiletype);

// SEG005.C
// custom logic - get sequence word value by its index
word seqtbl_offset_get_seq(short seq_index);
void seqtbl_offset_char(short seq_index);
void seqtbl_offset_opp(int seq_index);
void do_fall(void);
// custom logic - outbound guard is falling
void do_guard_fall(void);
void land(void);
void spiked(void);
// custom logic - is starting level 6 sequence playing?
bool is_level_6_sequence(void);
void control(void);
void control_crouched(void);
void control_standing(void);
void up_pressed(void);
void down_pressed(void);
// custom logic - level door can be closed or a balcony that triggers a different sequence
void go_up_leveldoor(int col, bool is_balcony, bool is_open);
void control_turning(void);
void crouch(void);
void back_pressed(void);
void forward_pressed(void);
void control_running(void);
void safe_step(void);
int check_get_item(void);
void get_item(byte align);
void control_startrun(void);
void control_jumpup(void);
void standing_jump(void);
void check_jump_up(void);
void jump_up_or_grab(void);
void grab_up_no_floor_behind(void);
void jump_up(void);
void control_hanging(void);
void can_climb_up(void);
void hang_fall(void);
void grab_up_with_floor_behind(void);
void run_jump(void);
void back_with_sword(void);
void forward_with_sword(void);
void draw_sword(void);
void control_with_sword(void);
void swordfight(void);
void sword_strike(void);
void parry(void);

// SEG006.C
int get_tile(int room, int col, int row);
int find_room_of_tile(void);
int get_tilepos(int tile_col, int tile_row);
int get_tilepos_nominus(int tile_col, int tile_row);
void load_fram_det_col(void);
void determine_col(void);
void load_frame(void);
short dx_weight(void);
int char_dx_forward(int delta_x);
int obj_dx_forward(int delta_x);
// custom logic - make guard notice kid
void make_guard_notice(bool ignore_lightweight);
// custom logic - show black screen for the provided number of ticks
void show_black_screen(int ticks, byte redraw_hp);
void play_seq(void);
// custom logic - is inactive guard in a ready position?
bool is_inactive_guard_ready(char_type2 inactive);
// custom logic - plays inactive sequence table
void play_inactive_seq(void);
int get_tile_div_mod_m7(int xpos);
int get_tile_div_mod(int xpos);
int y_to_row_mod4(int ypos);
void loadkid(void);
void savekid(void);
void loadshad(void);
void saveshad(void);
void loadkid_and_opp(void);
void savekid_and_opp(void);
void loadshad_and_opp(void);
void saveshad_and_opp(void);
void reset_obj_clip(void);
void x_to_xh_and_xl(int xpos, sbyte *xh_addr, sbyte *xl_addr);
void fall_accel(void);
void fall_speed(void);
void check_action(void);
// custom logic - check if there is an outbound guard that needs to start falling
void check_falling_guard(void);
int tile_is_floor(int tiletype);
// custom logic - checks if the tile is an empty tile
bool tile_is_empty(int tiletype);
// custom logic - returns the expected character tile column number and
// returning the correct column when kid is climbing while facing left
short char_col_climbing_left(void);
// custom logic - vertical spikes can harm characters in additional scenarios
void check_spiked_in_temple(void);
void check_spiked(void);
int take_hp(int count);
int get_tile_at_char(void);
void set_char_collision(void);
void check_on_floor(void);
// custom logic - check if guard can follow dropped out kid
void check_can_guard_follow_kid(void);
// custom logic - set "droppedout" flag if guard can follow kid
void set_kid_droppedout(void);
void start_fall(void);
void check_grab(void);
// custom logic - grab during running jumps
bool check_grab_run_jump(void);
int can_grab_front_above(void);
void in_wall(void);
// custom logic - in wall while getting struck with sword logic
void in_wall_engarde(void);
int get_tile_infrontof_char(void);
int get_tile_infrontof2_char(void);
int get_tile_behind_char(void);
// custom logic - get tile column relative to the current room
int get_tile_col_at_char(void);
int distance_to_edge_weight(void);
int distance_to_edge(int xpos);
void fell_out(void);
// custom logic - restores dead kid tile
void restore_dead_kid_tile(void);
void play_kid(void);
void control_kid(void);
void do_demo(void);
// custom logic - enter dynamic guard after the first static guard is killed
void enter_dynamic_guard(void);
// custom logic - bump into inactive opponent
void bump_into_inactive_opponent(void);
// custom logic - control inactive guard
void autocontrol_inactive_guard(void);
// custom logic - inactive guard turns around
void autocontrol_inactive_guard_turn(char_type2 inactive);
// custom logic - advances inactive guard
void autocontrol_inactive_guard_advance(char_type2 inactive);
// custom logic - makes inactive guard step back
void autocontrol_inactive_guard_retreat(char_type2 inactive);
// custom logic - inactive guard takes or blocks kid's strikes
void autocontrol_inactive_guard_hurt_or_block(char_type2 inactive);
// custom logic - inactive guard strikes kid in the back
void autocontrol_inactive_guard_strike(char_type2 inactive);
// custom logic - defines callbacks when kid is hurt by inactive guard
void autocontrol_inactive_guard_hurt_kid(char_type2 inactive);
// custom logic - simulates inactive guard's death
void inactive_guard_dead(void);
// custom logic - restores collapsed guard
void inactive_guard_restore(void);
void play_guard(void);
void user_control(void);
void flip_control_x(void);
int release_arrows(void);
// custom logic - reset ignored arrows when standing still
void reset_ignored_arrows(void);
void save_ctrl_1(void);
void rest_ctrl_1(void);
void clear_saved_ctrl(void);
void read_user_control(void);
int can_grab(void);
int wall_type(byte tiletype);
// custom logic - does not consider an open gate as a wall
int wall_type2(byte tiletype);
int get_tile_above_char(void);
int get_tile_behind_above_char(void);
int get_tile_front_above_char(void);
int back_delta_x(int delta_x);
void do_pickup(int obj_type);
void check_press(void);
// custom logic - checks whether inactive guard is on a button
void check_inactive_press(void);
void check_spike_near(void);
void clip_char(void);
void stuck_lower(void);
void set_objtile_at_char(void);
// custom logic - reset shadow/flame palette on level 13
void reset_shadow_palette(void);
void proc_get_object(void);
int is_dead(void);
void play_death_music(void);
void on_guard_killed(void);
void clear_char(void);
void save_obj(void);
void load_obj(void);
void draw_hurt_splash(void);
void add_sword_to_objtable(void);
void control_guard_inactive(void);
int char_opp_dist(void);
// custom logic - returns fighting distance between kid and guard
int kid_guard_dist(void);
// custom logic - returns absolute distance between guard and kid
int guard_kid_dist(void);
// custom logic - returns absolute distance between kid and separated shadow
int shad_kid_dist(void);
// custom logic - returns distance between kid and inactive guard
int inactive_guard_kid_dist(void);
// custom logic - returns distance between inactive guard and kid
int kid_inactive_guard_dist(void);
void inc_curr_row(void);
// custom logic - returns whether a bit has been set in the secondary modifier
// represented by the high 3 bits of the tile byte.
byte get_tile_ctrl_bits(short room, short tilepos, byte bit_num);
// custom logic - returns whether the room is not linked to any other room
bool is_room_unlinked(short room);

// SEG007.C
void process_trobs(void);
void animate_tile(void);
short is_trob_in_drawn_room(void);
void set_redraw_anim_right(void);
void set_redraw_anim_curr(void);
void redraw_at_trob(void);
void redraw_21h(void);
void redraw_11h(void);
void redraw_20h(void);
void draw_trob(void);
void redraw_tile_height(void);
short get_trob_pos_in_drawn_room(void);
short get_trob_right_pos_in_drawn_room(void);
short get_trob_right_above_pos_in_drawn_room(void);
void animate_torch(void);
void animate_potion(void);
void animate_sword(void);
void animate_chomper(void);
void animate_spike(void);
void animate_gate(void);
void gate_stop(void);
void animate_leveldoor(void);
// custom logic - animate balcony stars
void animate_balcony(void);
short bubble_next_frame(short curr);
short get_torch_frame(short curr);
void set_redraw_anim(short tilepos, byte frames);
void set_redraw2(short tilepos, byte frames);
void set_redraw_floor_overlay(short tilepos, byte frames);
void set_redraw_full(short tilepos, byte frames);
void set_redraw_fore(short tilepos, byte frames);
void set_wipe(short tilepos, byte frames);
void start_anim_torch(short room, short tilepos);
void start_anim_potion(short room, short tilepos);
void start_anim_sword(short room, short tilepos);
void start_anim_chomper(short room, short tilepos, byte modifier);
void start_anim_spike(short room, short tilepos);
// custom logic - start animating balcony stars
void start_anim_balcony(short room, short tilepos);
short trigger_gate(short room, short tilepos, short button_type);
short trigger_tile(short target_type, short room, short tilepos, short button_type);
void do_trigger_events(short first_event, short button_type);
void add_trob(byte room, byte tilepos, sbyte type);
// custom logic - find trob at a specific room location
short find_trob_at_loc(byte room, byte tilepos);
short find_trob(void);
void clear_tile_wipes(void);
short get_doorlink_timer(short index);
short set_doorlink_timer(short index, byte value);
short get_doorlink_tile(short index);
short get_doorlink_next(short index);
short get_doorlink_room(short index);
void trigger_button(int playsound, int button_type, int modifier);
void died_on_button(word charid);
void animate_button(void);
void start_level_door(short room, short tilepos);
void animate_empty(void);
void animate_loose(void);
void loose_shake(int sound);
int remove_loose(int tilepos);
void make_loose_fall(byte modifier);
// custom logic - added support for random chomper order
short next_chomper_timing(byte timing, bool random_chomper);
void start_chompers(void);
void loose_make_shake(void);
void do_knock(int room, int tile_row, int impact_col);
void add_mob(void);
short get_curr_tile(short tilepos);
void do_mobs(void);
void move_mob(void);
void move_loose(void);
void loose_land(void);
void loose_fall(void);
void redraw_at_cur_mob(void);
void mob_down_a_row(void);
void draw_mobs(void);
void draw_mob(void);
void add_mob_to_objtable(int ypos);
int is_spike_harmful(void);
void check_loose_fall_on_kid(void);
void fell_on_your_head(void);
// custom logic - kill guard if a loose fell on him
void fell_on_his_head(void);
void play_door_sound_if_visible(int sound_id);
// custom logic - animate moving wall on level 12B
void animate_moving_wall(void);
// custom logic - animate balcony stars
void animate_balcony_stars(short tilepos, bool left_room);

// SEG008.C
void redraw_room(void);
void load_room_links(void);
void draw_room(void);
void draw_tile(void);
void draw_tile_aboveroom(word saved_room);
void redraw_needed(short tilepos);
void redraw_needed_above(int column);
int get_tile_to_draw(int room, int column, int row, byte *ptr_tile, byte *ptr_modifier, byte tile_room0);
void load_curr_and_left_tile(void);
void load_leftroom(void);
void load_rowbelow(void);
void draw_tile_floorright(void);
int can_see_bottomleft(void);
void draw_tile_topright(void);
void draw_tile_anim_topright(void);
void draw_tile_right(void);
int get_spike_frame(byte modifier);
void draw_tile_anim_right(void);
void draw_tile_bottom(word redraw);
void draw_loose(void);
void draw_tile_base(void);
void draw_tile_anim(void);
void draw_tile_fore(void);
int get_loose_frame(byte modifier);
int add_backtable(short chtab_id, int id, sbyte xh, sbyte xl, int ybottom, int blit, byte peel);
int add_foretable(short chtab_id, int id, sbyte xh, sbyte xl, int ybottom, int blit, byte peel);
int add_midtable(short chtab_id, int id, sbyte xh, sbyte xl, int ybottom, int blit, byte peel);
void add_peel(int left, int right, int top, int height);
void add_wipetable(sbyte layer, short left, short bottom, sbyte height, short width, sbyte color);
void draw_table(int which_table);
void draw_wipes(int which);
void draw_back_fore(int which_table, int index);
void draw_mid(int index);
void draw_image(image_type *image, image_type *mask, int xpos, int ypos, int blit);
void draw_wipe(int index);
void calc_gate_pos(void);
void draw_gate_back(void);
void draw_gate_fore(void);
void alter_mods_allrm(void);
void load_alter_mod(int tile_ix);
void draw_moving(void);
void redraw_needed_tiles(void);
void draw_tile_wipe(byte height);
void draw_tables(void);
void restore_peels(void);
void add_drect(rect_type *source);
void draw_leveldoor(void);
void get_room_address(int room);
void draw_floor_overlay(void);
void draw_other_overlay(void);
void draw_tile2(void);
void draw_objtable_items_at_tile(byte tilepos);
void sort_curr_objs(void);
int compare_curr_objs(int index1, int index2);
void draw_objtable_item(int index);
int load_obj_from_objtable(int index);
void draw_people(void);
void draw_kid(void);
void draw_guard(void);
void add_kid_to_objtable(void);
void add_guard_to_objtable(void);
void add_objtable(byte obj_type);
void mark_obj_tile_redraw(int index);
void load_frame_to_obj(void);
void show_time(void);
void show_level(void);
short calc_screen_x_coord(short logical_x);
void free_peels(void);
void display_text_bottom(const char *text);
void erase_bottom_text(int clear_text);
void wall_pattern(int which_part, int which_table);
void draw_left_mark (word arg3, word arg2, word arg1);
void draw_right_mark (word arg2, word arg1);
image_type* get_image(short chtab_id, int id, int color);
// custom logic - loads custom images from PRINCE15.DAT
image_type* load_custom_graphics_image(int resource);
// custom logic - draws custom graphics in backtable
void draw_custom_graphics_backtable(void);
// custom logic - draws custom graphics in foretable
void draw_custom_graphics_foretable(void);
// custom logic - draws initial level 6 sword animation
void draw_level_6_sword_animation(void);

// SEG009.C
void sdlperror(const char* header);
void find_exe_dir(void);
// custom logic - create the provided folder, if missing
void create_folder_if_missing(const char* folder);
// custom logic - display help information for the help command line parameter
void display_help(void);
bool file_exists(const char* filename);
#define locate_file(filename) locate_file_(filename, alloca(POP_MAX_PATH), POP_MAX_PATH)
const char* locate_file_(const char* filename, char* path_buffer, int buffer_size);
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
FILE* fopen_UTF8(const char* filename, const char* mode);
#define fopen fopen_UTF8
int chdir_UTF8(const char* path);
#define chdir chdir_UTF8
int access_UTF8(const char* filename_UTF8, int mode);
#ifdef access
#undef access
#endif
#define access access_UTF8
int stat_UTF8(const char *filename_UTF8, struct stat *_Stat);
// We define a function-like macro, because `stat` is also the name of the type, and we don't want to redefine that.
#define stat(filename_UTF8, _Stat) stat_UTF8(filename_UTF8, _Stat)
#endif // WIN32...
int read_key(void);
void clear_kbd_buf(void);
word prandom(word max);
int round_xpos_to_byte(int xpos);
void show_dialog(const char *text, byte which);
void quit(int exit_code);
void restore_stuff(void);
int key_test_quit(void);
const char* check_param(const char *param);
int pop_wait(int timer_index, int time);
dat_type* open_dat(const char *file);
void set_loaded_palette(dat_pal_type *palette_ptr);
chtab_type* load_sprites_from_file(int resource, int palette_bits);
void free_chtab(chtab_type *chtab_ptr);
image_type* decode_image(image_data_type* image_data, dat_pal_type* palette);
image_type* load_image(int index, dat_pal_type* palette);
int set_joy_mode(void);
surface_type* make_offscreen_buffer(const rect_type *rect);
void free_surface(surface_type *surface);
void free_peel(peel_type *peel_ptr);
void set_hc_pal(void);
void flip_not_ega(byte *memory, int height, int stride);
void flip_screen(surface_type *surface);
void fade_in_2(int which_rows);
void fade_out_2(int rows);
int get_line_width(const char *text, int length);
int draw_text_character(byte character);
void draw_rect(const rect_type *rect, int color);
void draw_rect_with_alpha(const rect_type* rect, byte color, byte alpha);
void draw_rect_contours(const rect_type* rect, byte color);
rect_type* shrink2_rect(rect_type *target_rect, const rect_type *source_rect, int delta_x, int delta_y);
void set_curr_pos(int xpos, int ypos);
void restore_peel(peel_type *peel_ptr);
peel_type* read_peel_from_screen(const rect_type *rect);
void show_text(const rect_type *rect_ptr, int x_align, int y_align, const char *text);
int intersect_rect(rect_type *output, const rect_type *input1, const rect_type *input2);
rect_type* union_rect(rect_type *output, const rect_type *input1, const rect_type *input2);
void stop_sounds(void);
void stop_all_sounds(void);
void init_digi(void);
void play_sound_from_buffer(sound_buffer_type *buffer);
void turn_sound_on_off(byte new_state);
int check_sound_playing(void);
void apply_aspect_ratio(int width);
// custom logic - initializes SDL Mixer
void set_music(void);
void set_gr_mode(void);
SDL_Surface* get_final_surface(void);
// custom logic - initialize custom graphic resources from PRINCE15.DAT
void init_custom_graphics(void);
void update_screen(void);
void set_pal_arr(int start, int count, const rgb_type *array);
void set_pal(int index, int red, int green, int blue);
void process_palette(void *target, dat_pal_type *source);
int find_first_pal_row(int which_rows_mask);
int get_text_color(int low_half, int high_half_mask);
void close_dat(dat_type *pointer);
void* load_from_opendats_alloc(int resource, const char* extension, data_location* out_result, int* out_size);
int load_from_opendats_to_area(int resource, void *area, int length, const char* extension);
void rect_to_sdlrect(const rect_type *rect, SDL_Rect* sdlrect);
void method_1_blit_rect(surface_type *target_surface, surface_type *source_surface, const rect_type *target_rect, const rect_type *source_rect, int blit);
image_type* method_3_blit_mono(image_type *image, int xpos, int ypos, int blitter, byte color);
const rect_type* method_5_rect(const rect_type *rect, int blit, byte color);
void draw_rect_with_alpha(const rect_type* rect, byte color, byte alpha);
image_type* method_6_blit_img_to_scr(image_type *image, int xpos, int ypos, int blit);
void reset_timer(int timer_index);
// custom logic - gets ticks per second
double get_ticks_per_sec(int timer_index);
void set_timer_length(int timer_index, int length);
void start_timer(int timer_index, int length);
int do_wait(int timer_index);
void init_timer(int frequency);
void set_clip_rect(const rect_type *rect);
void reset_clip_rect(void);
void set_bg_attr(int vga_pal_index, int hc_pal_index);
rect_type* offset4_rect_add(rect_type *dest, const rect_type *source, int d_left, int d_top, int d_right, int d_bottom);
int input_str(const rect_type *rect, char *buffer, int max_length, const char *initial, int has_initial, int x_offset, int color, int bgcolor);
rect_type* offset2_rect(rect_type *dest, const rect_type *source, int delta_x, int delta_y);
void show_text_with_color(const rect_type *rect_ptr, int x_align, int y_align, const char *text, int color);
void do_simple_wait(int timer_index);
void toggle_fullscreen(void);
void process_events(void);
void idle(void);
void init_dialogs(void);
dialog_type* make_dialog_info(dialog_settings_type *settings, rect_type *text_rect, peel_type *dialog_peel);
void calc_dialog_peel_rect(dialog_type*dialog);
void read_dialog_peel(dialog_type *dialog);
void draw_dialog_frame(dialog_type *dialog);
void add_dialog_rect(dialog_type *dialog);
void dialog_method_2_frame(dialog_type *dialog);
palette_fade_type *make_pal_buffer_fadein(int which_rows, int wait_time);
void pal_restore_free_fadein(palette_fade_type *palette_buffer);
int fade_in_frame(palette_fade_type *palette_buffer);
void fade_out_2(int rows);
palette_fade_type *make_pal_buffer_fadeout(int which_rows, int wait_time);
void pal_restore_free_fadeout(palette_fade_type *palette_buffer);
int fade_out_frame(palette_fade_type *palette_buffer);
void read_palette_256(rgb_type *target);
void set_pal_256(rgb_type *source);
void set_chtab_palette(chtab_type* chtab, byte* colors, int n_colors);
int has_timer_stopped(int index);
sound_buffer_type* load_sound(int index);
// custom logic - calculates size of a const char* array terminated by NULL
int str_arr_size(const char** arr);
// custom logic - converts a character array to upper case
void to_ucase(char* str);
// custom logic - displays "press button to continue"
void display_press_button(void);
#if defined WIN32 || _WIN32 || WIN64 || _WIN64
// custom logic - turns off num lock in Windows because Shift is not detected when Num Lock is on
void set_numlock_state(bool active);
#endif

// OPTIONS.C
void load_global_options(void);

// lighting.c
void init_lighting(void);
void redraw_lighting(void);
void update_lighting(const rect_type *source_rect_ptr);

// screenshot.c
void save_screenshot(bool want_extras);
void save_level_screenshot(bool want_extras);

// menu.c
void init_menu(void);
void draw_menu(void);
void clear_menu_controls(void);
void process_additional_menu_input(void);
int key_test_paused_menu(int key);
void menu_was_closed(void);

// midi.c
void stop_midi(void);
void init_midi(void);
void midi_callback(void *userdata, Uint8 *stream, int len);
void play_midi_sound(sound_buffer_type *buffer);

// musplayer.c
void load_enhanced_sounds(void);
void unload_enhanced_sounds(void);
bool enhanced_sound_file_exists(void);
void unload_enhanced_sounds(void);
void* get_enhanced_sound_data(const char* entry, size_t* length);
void load_music(void);
void free_music(void);
bool is_foreground_sound(int sound_id);
bool is_exclusive_sound(int sound_id);
int play_ogg_fore_sound(int sound_id, double sound_pos, bool trigger_callback);
bool is_music_found(void);
bool is_fight_music(void);
bool is_heartbeat_music(void);
void start_ambient_music(bool fadeIn);
void stop_ambient_music(void);
