/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org
*/

#include "common.h"

#ifndef _MSC_VER // unistd.h does not exist in the Windows SDK.
#include <unistd.h>
#endif
#include <fcntl.h>

// data:4CB4
short cutscene_wait_frames;
// data:3D14
short cutscene_frame_time;
// data:588C
short disable_keys;
// data:436A
short hourglass_sandflow;
// data:5964
short hourglass_state;
// data:4CC4
short which_torch;

#pragma pack(push,1)
typedef struct hof_type {
    char name[25];
    short min, tick, score;
} hof_type;
SDL_COMPILE_TIME_ASSERT(hof_size, sizeof(hof_type) == 31);
#pragma pack(pop)

#define MAX_HOF_COUNT 6
// data:589A
hof_type hof[MAX_HOF_COUNT];

#define N_STARS 9

// data:0D92
rect_type hof_rects[MAX_HOF_COUNT] = {
    { 88,   46,  100,  274},
    {102,   46,  114,  274},
    {116,   46,  128,  274},
    {130,   46,  142,  274},
    {144,   46,  156,  274},
    {158,   46,  170,  274},
};

// custom logic - do we need to render princess room torches
bool need_proom_drects;

// seg001:0004
int proc_cutscene_frame(int wait_frames) {
    cutscene_wait_frames = wait_frames;
    reset_timer(timer_0);
    do {
        set_timer_length(timer_0, cutscene_frame_time);
        if (need_proom_drects) {
            play_both_seq();
            draw_proom_drects(); // changed order of drects and flash
        }
        if (flash_time) {
            do_flash(flash_color);
        }
        if (flash_time) {
            --flash_time;
            remove_flash();
        }
        if (!check_sound_playing()) {
            play_next_sound();
        }
        do {
            if (!disable_keys && do_paused()) {
                stop_all_sounds();
                draw_rect(&screen_rect, color_0_black);
                if (is_global_fading) {
                    fade_palette_buffer->proc_restore_free(fade_palette_buffer);
                    is_global_fading = 0;
                }
                return 1;
            }
            if (is_global_fading) {
                if (fade_palette_buffer->proc_fade_frame(fade_palette_buffer)) {
                    fade_palette_buffer->proc_restore_free(fade_palette_buffer);
                    is_global_fading = 0;
                    return 2;
                }
            } else {
                idle();
                delay_ticks(1);
            }
        } while(!has_timer_stopped(timer_0)); // busy waiting?
    } while(--cutscene_wait_frames);
    return 0;
}

// seg001:00DD
void play_both_seq(void) {
    play_kid_seq();
    play_opp_seq();
}

// seg001:00E6
void draw_proom_drects(void) {
    draw_princess_room_bg();
    if (!is_global_fading) {
        while (drects_count--) {
            copy_screen_rect(&drects[drects_count]);
        }
    }
    drects_count = 0;
    if (cutscene_wait_frames & 1) {
        draw_star(prandom(N_STARS - 1), 1);
    }
}

// seg001:0128
void play_kid_seq(void) {
    loadkid();
    if (Char.frame) {
        play_seq();
        savekid();
    }
}

// seg001:013F
void play_opp_seq(void) {
    loadshad_and_opp();
    if (Char.frame) {
        play_seq();
        saveshad();
    }
}

// seg001:0156
void draw_princess_room_bg(void) {
    memset(table_counts, 0, sizeof(table_counts));
    loadkid();
    if (Char.frame) {
        load_frame_to_obj();
        obj_tilepos = 30;
        add_objtable(objtype_0_cutscene_char);
    }
    loadshad();
    if (Char.frame) {
        load_frame_to_obj();
        obj_tilepos = 30;
        add_objtable(objtype_0_cutscene_char);
    }
    redraw_needed_tiles();
    add_foretable(id_chtab_8_princessroom, 2 /*pillar piece*/, 30, 0, 167, blitters_10h_transp, 0);
    princess_room_torch();
    draw_hourglass();
    draw_tables();
}

// seg001:01E0
void seqtbl_offset_shad_char(int seq_index) {
    loadshad();
    seqtbl_offset_char(seq_index);
    saveshad();
}

// seg001:01F9
void seqtbl_offset_kid_char(int seq_index) {
    loadkid();
    seqtbl_offset_char(seq_index);
    savekid();
}

// seg001:0212
void init_mouse_cu8(void) {
    init_mouse_go();
    Char.x = 144;
    seqtbl_offset_char(seq_106_mouse); // mouse
    play_seq();
}

// seg001:022A
void init_mouse_go(void) {
    Char.charid = charid_24_mouse;
    Char.x = 199;
    Char.y = 167;
    Char.direction = dir_FF_left;
    seqtbl_offset_char(seq_105_mouse_forward); // mouse go
    play_seq();
}

// seg001:024D
void princess_crouching(void) {
    init_princess();
    Char.x = 131;
    Char.y = 169;
    seqtbl_offset_char(seq_110_princess_crouching_PV2); // princess crouching [PV2]
    play_seq();
}

// seg001:026A
void princess_stand(void) {
    init_princess_right();
    Char.x = 144;
    Char.y = 169;
    seqtbl_offset_char(seq_94_princess_stand_PV1); // princess stand [PV1]
    play_seq();
}

// seg001:0287
void init_princess_x156(void) {
    init_princess();
    Char.x = 156;
}

// seg001:0291
void princess_lying(void) {
    init_princess();
    Char.x = 92;
    Char.y = 162;
    seqtbl_offset_char(seq_103_princess_lying_PV2); // princess lying [PV2]
    play_seq();
}

// seg001:02AE
void init_princess_right(void) {
    init_princess();
    Char.direction = dir_0_right;
}

// seg001:02B8
void init_ending_princess(void) {
    init_princess();
    Char.x = 136;
    Char.y = 164;
    seqtbl_offset_char(seq_109_princess_stand_PV2); // princess standing [PV2]
    play_seq();
}

// seg001:02D5
void init_mouse_1(void) {
    init_mouse_go();
    Char.x -= 2;
    Char.y = 164;
}

// seg001:02E4
void init_princess(void) {
    Char.charid = charid_5_princess;
    Char.x = 120;
    Char.y = 166;
    Char.direction = dir_FF_left;
    seqtbl_offset_char(seq_94_princess_stand_PV1); // princess stand [PV1]
    play_seq();
}

// seg001:0307
void init_vizier(void) {
    Char.charid = charid_6_vizier;
    Char.x = 198;
    Char.y = 166;
    Char.direction = dir_FF_left;
    seqtbl_offset_char(seq_95_Jaffar_stand_PV1); // Jaffar stand [PV1]
    play_seq();
}

// seg001:032A
void init_ending_kid(void) {
    Char.charid = charid_0_kid;
    Char.x = 198;
    Char.y = 164;
    Char.direction = dir_FF_left;
    seqtbl_offset_char(seq_1_start_run); // start run
    play_seq();
}

// seg001:034D
void cutscene_8(void) {
    play_sound(sound_35_cutscene_8_9); // cutscene 8, 9
    set_hourglass_state(hourglass_frame());
    init_mouse_cu8();
    savekid();
    princess_crouching();
    saveshad();
    if (fade_in_1()) return;
    if (proc_cutscene_frame(20)) return;
    seqtbl_offset_kid_char(seq_107_mouse_stand_up_and_go); // mouse stand up and go
    if (proc_cutscene_frame(20)) return;
    seqtbl_offset_shad_char(seq_111_princess_stand_up_PV2); // princess stand up [PV2]
    if (proc_cutscene_frame(20)) return;
    Kid.frame = frame_0;
    fade_out_1();
}

// seg001:03B7
void cutscene_9(void) {
    play_sound(sound_35_cutscene_8_9); // cutscene 8, 9
    set_hourglass_state(hourglass_frame());
    princess_stand();
    saveshad();
    if (fade_in_1()) return;
    init_mouse_go();
    savekid();
    if (proc_cutscene_frame(5)) return;
    seqtbl_offset_shad_char(seq_112_princess_crouch_down_PV2); // princess crouch down [PV2]
    if (proc_cutscene_frame(9)) return;
    seqtbl_offset_kid_char(seq_114_mouse_stand); // mouse stand
    if (proc_cutscene_frame(58)) return;
    fade_out_1();
}

// custom logic - cutscene when there is less than 5 min left
void cutscene_short_time(void) {
    set_hourglass_state(hourglass_frame());
    init_princess_x156();
    saveshad();
    play_sound(sound_40_cutscene_12_short_time); // cutscene 12 short time
    if (fade_in_1()) return;
    if (proc_cutscene_frame(2)) return;
    seqtbl_offset_shad_char(seq_98_princess_turn_PV1); // princess turn around [PV1]
    if (proc_cutscene_frame(24)) return;
    fade_out_1();
}

// seg001:041C
void end_sequence_anim(void) {
    stop_ambient_music();
    disable_keys = 1;
    copy_screen_rect(&screen_rect);
    play_sound(sound_26_embrace); // arrived to princess
    init_ending_princess();
    saveshad();
    init_ending_kid();
    savekid();
    if (proc_cutscene_frame(8)) return;
    seqtbl_offset_shad_char(seq_108_princess_turn_and_hug); // princess turn and hug [PV2]
    if (proc_cutscene_frame(5)) return;
    seqtbl_offset_kid_char(seq_13_stop_run); // stop run
    if (proc_cutscene_frame(2)) return;
    Kid.frame = frame_0;
    if (proc_cutscene_frame(39)) return;
    init_mouse_1();
    savekid();
    Kid.x += 2;
    if (proc_cutscene_frame(10)) return;
    seqtbl_offset_kid_char(seq_101_mouse_stands_up); // mouse stands up
    if (proc_cutscene_frame(2)) return;
    Kid.x -= 1;
    if (proc_cutscene_frame(2)) return;
    Kid.x += 1;
    if (proc_cutscene_frame(2)) return;
    Kid.x -= 1;
    if (proc_cutscene_frame(2)) return;
    Kid.x += 1;
    if (proc_cutscene_frame(2)) return;
    Kid.x -= 1;
    if (proc_cutscene_frame(2)) return;
    Kid.x += 1;
    if (proc_cutscene_frame(2)) return;
    Kid.x -= 1;
    if (proc_cutscene_frame(2)) return;
    Kid.x += 1;
    if (proc_cutscene_frame(2)) return;
    Kid.x -= 1;
    if (proc_cutscene_frame(2)) return;
    Kid.x += 1;
    if (proc_cutscene_frame(2)) return;
    Kid.x -= 1;
    if (proc_cutscene_frame(2)) return;
    Kid.x += 1;
    if (proc_cutscene_frame(2)) return;
    Kid.x -= 1;
    if (proc_cutscene_frame(2)) return;
    Kid.x += 1;
    if (proc_cutscene_frame(12)) return;
    fade_out_1();
    while (check_sound_playing()) {
        idle();
        delay_ticks(1);
    }
}

// seg001:04D3
void time_expired(void) {
    stop_ambient_music();
    disable_keys = 1;
    is_ending_sequence = true;
    set_hourglass_state(7);
    hourglass_sandflow = -1;
    // custom logic - add dead princess
    init_princess();
    Char.x = 91;
    Char.y = 182;
    seqtbl_offset_char(seq_131_princess_dead_PV2); // princess lying dead [PV2]
    play_seq();
    saveshad();
    play_sound(sound_36_out_of_time); // time over
    if (fade_in_1()) return;
    if (proc_cutscene_frame(2)) return;
    if (proc_cutscene_frame(100)) return;
    fade_out_1();
    while (check_sound_playing()) {
        idle();
        do_paused();
        delay_ticks(1);
    }
    is_ending_sequence = false;
}

// seg001:0525
void cutscene_12(void) {
    cutscene_2_6();
}

// custom logic - prince captured cut scene
void cutscene_15(void) {
    play_sound(sound_28_death_in_fight);
    if (fade_in_1()) return;
    if (proc_cutscene_frame(26)) return;
    fade_out_1();
}

// seg001:0584
void cutscene_4(void) {
    play_sound(sound_27_cutscene_2_4_6_12); // cutscene 2, 4, 6, 12
    set_hourglass_state(hourglass_frame());
    princess_lying();
    saveshad();
    if (fade_in_1()) return;
    if (proc_cutscene_frame(26)) return;
    fade_out_1();
}

// seg001:05B8
void cutscene_2_6(void) {
    play_sound(sound_27_cutscene_2_4_6_12); // cutscene 2, 4, 6, 12
    set_hourglass_state(hourglass_frame());
    init_princess_right();
    saveshad();
    if (fade_in_1()) return;
    if (proc_cutscene_frame(26)) return;
    fade_out_1();
}

// seg001:05EC
void pv_scene(void) {
    init_princess();
    saveshad();
    if (fade_in_1()) return;
    init_vizier();
    savekid();
    if (proc_cutscene_frame(2)) return;
    play_sound(sound_50_story_2_princess); // story 2: princess waiting
    do {
        if (proc_cutscene_frame(1)) return;
    } while(check_sound_playing());
    cutscene_frame_time = 8;
    if (proc_cutscene_frame(5)) return;
    // custom logic - Jaffar opens a gate instead of a door when enhanced audio is enabled
    bool gate_open_sound_found = is_positional_audio && sound_pointers[sound_62_princess_door_opening] != NULL;
    if (!gate_open_sound_found) {
        play_sound(sound_4_gate_closing); // gate closing
        do {
            if (proc_cutscene_frame(1)) return;
        } while(check_sound_playing());
    }
    play_sound(gate_open_sound_found ? sound_62_princess_door_opening : sound_51_wall_sliding); // princess door opening
    if (proc_cutscene_frame(3)) return;
    seqtbl_offset_shad_char(seq_98_princess_turn_PV1); // princess turn around [PV1]
    if (gate_open_sound_found) {
        while (check_sound_playing()) {
            if (proc_cutscene_frame(1)) return;
        }
    }
    if (proc_cutscene_frame(5)) return;
    seqtbl_offset_kid_char(seq_96_Jaffar_walk_PV1); // Jaffar walk [PV1]
    if (proc_cutscene_frame(6)) return;
    play_sound(sound_53_story_3_Jaffar_comes); // story 3: Jaffar comes
    seqtbl_offset_kid_char(seq_97_Jaffar_stop_PV1); // Jaffar stop [PV1]
    if (proc_cutscene_frame(4)) return;
    if (proc_cutscene_frame(18)) return;
    seqtbl_offset_kid_char(seq_96_Jaffar_walk_PV1); // Jaffar walk [PV1]
    if (proc_cutscene_frame(30)) return;
    seqtbl_offset_kid_char(seq_97_Jaffar_stop_PV1); // Jaffar stop [PV1]
    if (proc_cutscene_frame(35)) return;
    seqtbl_offset_kid_char(seq_102_Jaffar_conjuring_PV1); // Jaffar conjuring [PV1]
    cutscene_frame_time = 7;
    if (proc_cutscene_frame(1)) return;
    seqtbl_offset_shad_char(seq_99_princess_back_PV1); // princess step back [PV1]
    if (proc_cutscene_frame(17)) return;
    hourglass_state = 1;
    flash_time = 5;
    flash_color = color_15_brightwhite;
    do {
        if (proc_cutscene_frame(1)) return;
    } while(check_sound_playing());
    seqtbl_offset_kid_char(seq_100_Jaffar_stop_conjuring_PV1); // Jaffar stop conjuring and walk [PV1]
    hourglass_sandflow = 0;
    if (proc_cutscene_frame(6)) return;
    play_sound(sound_52_story_4_Jaffar_leaves); // story 4: Jaffar leaves
    if (proc_cutscene_frame(24)) return;
    hourglass_state = 2;
    if (proc_cutscene_frame(9)) return;
    seqtbl_offset_shad_char(seq_113_princess_look_down_PV1); // princess look down [PV1]
    if (proc_cutscene_frame(28)) return;
    fade_out_1();
}

// seg001:07C7
void set_hourglass_state(int state) {
    hourglass_sandflow = 0;
    hourglass_state = state;
}

// data:0DEC
short time_bound[] = {6, 17, 33, 65};

// seg001:07DA
int hourglass_frame(void) {
    short bound_index;
    for (bound_index = 0; bound_index < 4; ++bound_index) {
        if (time_bound[bound_index] > rem_min) {
            break;
        }
    }
    return 6 - bound_index;
}

// data:0DF4
short princess_torch_pos_xh[] = {11, 26};
// data:0DF8
short princess_torch_pos_xl[] = {5, 3};
// data:0DFC
short princess_torch_frame[] = {1, 6};

// seg001:0808
void princess_room_torch(void) {
    short which;
    for (which = 2; which--; ) {
        which_torch = !which_torch;
        princess_torch_frame[which_torch] = get_torch_frame(princess_torch_frame[which_torch]);
        int blit = blitters_colored_flame + (11 /* argon blue */ & 0x3F);
        add_backtable(id_chtab_1_flameswordpotion, princess_torch_frame[which_torch] + 1, princess_torch_pos_xh[which_torch], princess_torch_pos_xl[which_torch], 116, blit, 0);
    }
}

// seg001:0863
void draw_hourglass(void) {
    if (hourglass_sandflow >= 0) {
        hourglass_sandflow = (hourglass_sandflow + 1) % 3;
        if (hourglass_state >= 7) return;
        add_foretable(id_chtab_8_princessroom, hourglass_sandflow + 10, 20, 0, 164, blitters_10h_transp, 0);
    }
    if (hourglass_state) {
        add_midtable(id_chtab_8_princessroom, hourglass_state + 2, 19, 0, 168, blitters_10h_transp, 1);
    }
}

// seg001:08CA
void reset_cutscene(void) {
    Guard.frame = frame_0;
    Kid.frame = frame_0;
    which_torch = 0;
    disable_keys = 0;
    hourglass_state = 0;
    hourglass_sandflow = -1;
    cutscene_frame_time = 6;
    clear_tile_wipes();
    next_sound = -1;
}

// seg001:0908
void do_flash(short color) {
    if (color) {
        // custom logic - in feather mode only green color can flash
        if (feather_fall_timer && color != color_2_green) {
            return;
        }
        reset_timer(timer_2);
        set_timer_length(timer_2, 2);
        set_bg_attr(0, color);
        if (color != 0) do_simple_wait(timer_2); // give some time to show the flash
    }
}

void delay_ticks(Uint32 ticks) {
    SDL_Delay(ticks * (1000 / 60));
}

// seg001:0981
void remove_flash(void) {
    set_bg_attr(0, 0);
}

// custom logic - draws the palace image
void draw_palace_image(void) {
    draw_full_image(STORY_FRAME);
    method_6_blit_img_to_scr(custom_graphics_tab->images[image_0_palace], 16, 16, blitters_0_no_transp);
}

// seg001:09D7
void end_sequence(void) {
    peel_type* peel;
    short bgcolor;
    short color;
    rect_type rect;
    short hof_index;
    short score;
    short i;
    color = 0;
    bgcolor = 15;
    load_intro(1, &end_sequence_anim);
    clear_screen_and_sounds();
    is_ending_sequence = true; // added (fix being able to pause the game during the end sequence)
    current_sound_is_positional = false;
    play_sound_from_buffer(sound_pointers[sound_56_ending_music]); // winning theme
    if (offscreen_surface) free_surface(offscreen_surface); // missing in original
    offscreen_surface = make_offscreen_buffer(&screen_rect);
    load_title_images(0);
    current_target_surface = offscreen_surface;
    draw_full_image(STORY_FRAME);
    draw_full_image(STORY_HAIL);
    fade_in_2(0x800);
    pop_wait(timer_0, 0x384);
    start_timer(timer_0, 240);
    draw_palace_image();
    draw_full_image(TITLE_POP);
    draw_full_image(TITLE_HACKED_BY);
    transition_ltr();
    do_wait(timer_0);
    // custom logic - game score is determined from total time and the number of saves
    // you do not get a score when playing in cheat mode or if the time ran out on the last level
    score = (kid_cheat_pal_offset == 0 && rem_min > 0) ? MAX(((rem_min * 72) + (rem_tick / 10)) - (total_saves * 36), 100) : 0;
    if (score > 9999) score = 9999;
    for (hof_index = 0; hof_index < hof_count; ++hof_index) {
        if (hof[hof_index].score < score) {
            break;
        }
    }
    if (score > 0 && hof_index < MAX_HOF_COUNT && hof_index <= hof_count) {
        fade_out_2(0x1000);
        for (i = 5; hof_index + 1 <= i; --i) {
            hof[i] = hof[i - 1];
        }
        hof[i].name[0] = 0;
        hof[i].min = rem_min;
        hof[i].tick = rem_tick;
        hof[i].score = score;
        if (hof_count < MAX_HOF_COUNT) {
            ++hof_count;
        }
        draw_full_image(STORY_FRAME);
        draw_full_image(HOF_POP);
        show_hof();
        offset4_rect_add(&rect, &hof_rects[hof_index], -4, -1, -80, -1);
        peel = read_peel_from_screen(&rect);
        color = 0xBE;
        bgcolor = 0xB7;
        draw_rect(&rect, bgcolor);
        fade_in_2(0x1800);
        current_target_surface = onscreen_surface_;
        while(input_str(&rect, hof[hof_index].name, 18, "", 0, 4, color, bgcolor) <= 0);
        restore_peel(peel);
        show_hof_text(&hof_rects[hof_index], -1, 0, hof[hof_index].name);
        hof_write();
        pop_wait(timer_0, 0x78);
        current_target_surface = offscreen_surface;
    } else if (hof_count) {
        release_title_images();
        load_title_images(1);
        draw_full_image(STORY_FRAME);
        draw_full_image(HOF_POP);
        show_hof();
        transition_ltr();
        pop_wait(timer_0, 0x168);
    }
    draw_full_image(STORY_FRAME);
    method_6_blit_img_to_scr(custom_graphics_tab->images[image_1_the_end], 16, 16, blitters_0_no_transp);
    transition_ltr();
    while (check_sound_playing() &&
            !(key_states[SDL_SCANCODE_LSHIFT] & keystate_held) && // custom logic - shift skips the ending sequence
            !(key_states[SDL_SCANCODE_RSHIFT] & keystate_held) &&
            !(joy_button_states[joyinput_x] & keystate_held) && // custom logic - joystick X button skips the ending sequence
            !key_test_quit()) {
        idle();
        delay_ticks(1);
    }
    fade_out_2(0x1000);
    start_level = -1;
    is_ending_sequence = false;
    start_game();
}

// seg001:0C94
void expired(void) {
    if (offscreen_surface) free_surface(offscreen_surface); // missing in original
    offscreen_surface = NULL;
    clear_screen_and_sounds();
    offscreen_surface = make_offscreen_buffer(&screen_rect);
    load_intro(1, &time_expired);
    start_level = -1;
    start_game();
}

// seg001:0CCD
void load_intro(int which_imgs, cutscene_ptr_type func) {
    short current_star;
    bool is_cutscene_15;
    // custom logic - suspend upside down timer
    word prev_ud_timer = upside_down_timer;
    if (prev_ud_timer > 0) {
        upside_down_timer = 0;
    }
    draw_rect(&screen_rect, color_0_black);
    // custom logic - cut scene 15 uses a custom image instead of the princess room
    is_cutscene_15 = func == cutscene_15;
    if (!is_cutscene_15) {
        need_proom_drects = true;
        free_all_chtabs_from(id_chtab_3_princessinstory);
        load_chtab_from_file(id_chtab_8_princessroom, 950, "PV.DAT", 1 << 13);
        load_chtab_from_file(id_chtab_9_princessbed, 980, "PV.DAT", 1 << 14);
        current_target_surface = offscreen_surface;
        method_6_blit_img_to_scr(get_image(id_chtab_8_princessroom, 0, 0), 0, 0, 0);
        method_6_blit_img_to_scr(get_image(id_chtab_9_princessbed, 0, 0), 0, 142, blitters_10h_transp);

        // Free the images that are not needed anymore.
        free_all_chtabs_from(id_chtab_9_princessbed);
        SDL_FreeSurface(get_image(id_chtab_8_princessroom, 0, 0));
        if (NULL != chtab_addrs[id_chtab_8_princessroom]) chtab_addrs[id_chtab_8_princessroom]->images[0] = NULL;

        load_chtab_from_file(id_chtab_3_princessinstory, 800, "PV.DAT", 1 << 9);
        load_chtab_from_file(id_chtab_4_jaffarinstory_princessincutscenes,
                             50 * which_imgs + 850, "PV.DAT", 1 << 10);
        for (current_star = 0; current_star < N_STARS; ++current_star) {
            draw_star(current_star, 0);
        }
    } else {
        need_proom_drects = false;
        free_surface(offscreen_surface);
        offscreen_surface = make_offscreen_buffer(&screen_rect);
        current_target_surface = offscreen_surface;
        load_title_images(1);
        draw_full_image(STORY_FRAME);
        release_title_images();
        method_6_blit_img_to_scr(custom_graphics_tab->images[image_10_prince_captured], 16, 16, blitters_0_no_transp);
    }
    current_target_surface = onscreen_surface_;
    while (check_sound_playing()) {
        idle();
        do_paused();
        delay_ticks(1);
    }
    need_drects = 1;
    reset_cutscene();
    is_cutscene = 1;
    func();
    is_cutscene = 0;
    // custom logic - clean up off screen surface for cut scene 15
    if (is_cutscene_15) {
        free_surface(offscreen_surface);
        offscreen_surface = make_offscreen_buffer(&rect_top);
    } else {
        free_all_chtabs_from(3);
    }
    draw_rect(&screen_rect, color_0_black);
    // custom logic - reset upside down timer
    if (prev_ud_timer > 0) {
        upside_down_timer = prev_ud_timer;
    }
}

typedef struct star_type {
    short x, y, color;
} star_type;

// data:0DC2
star_type stars[N_STARS] = {
    { 20,  97, 0},
    { 16, 104, 1},
    { 23, 110, 2},
    { 17, 116, 3},
    { 24, 120, 4},
    { 18, 128, 0},
    {306,  92, 1},
    {308,  98, 2},
    {309, 111, 3},
};

#define N_STAR_COLORS 5
// data:0DE6
const byte star_colors[N_STAR_COLORS] = {8, 7, 15, 15, 7};

// custom logic - draw star in the provided rectangle
void draw_star_rect(rect_type* rect, int which_star, int mark_dirty) {
    short star_color = 15;
    stars[which_star].color = (stars[which_star].color + 1) % N_STAR_COLORS;
    star_color = star_colors[stars[which_star].color];
    // stars have green/cyan colors on level 6 with a purple palette
    if (!is_cutscene &&
            current_level == 6 &&
            level_6_sequence_timer == 0 &&
            (falling_exit_checkpoint < 4 || leveldoor_open != 1)
       ) {
        star_color -= 5;
    }
    draw_rect(rect, star_color);
    if (mark_dirty) {
        add_drect(rect);
    }
}

// seg001:0E1C
void draw_star(int which_star, int mark_dirty) {
    // The stars in the window of the princess's room.
    rect_type rect;
    rect.right = rect.left = stars[which_star].x;
    ++rect.right;
    rect.bottom = rect.top = stars[which_star].y;
    ++rect.bottom;
    draw_star_rect(&rect, which_star, mark_dirty);
}

// seg001:0E94
void show_hof(void) {
    // Hall of Fame
    short index;
    char time_text[24];
    for (index = 0; index < hof_count; ++index) {
        snprintf(time_text, sizeof(time_text), "%d:%02d (%04d)", hof[index].min - 1, hof[index].tick / 12, hof[index].score);
        show_hof_text(&hof_rects[index], -1, 0, hof[index].name);
        show_hof_text(&hof_rects[index], 1, 0, time_text);
    }
}

static const char* hof_file = "PRINCE.HOF";

const char* get_hof_path(char* custom_path_buffer, size_t max_len) {
    snprintf(custom_path_buffer, max_len, "%s/%s", save_folder, hof_file);
    return custom_path_buffer;
}

// seg001:0F17
void hof_write(void) {
    int handle;
    char custom_hof_path[POP_MAX_PATH];
    create_folder_if_missing(save_folder);
    const char* hof_path = get_hof_path(custom_hof_path, sizeof(custom_hof_path));
    // no O_TRUNC
    handle = open(hof_path, O_WRONLY | O_CREAT | O_BINARY, 0600);
    if (handle < 0 ||
            write(handle, &hof_count, 2) != 2 ||
            write(handle, &hof, sizeof(hof)) != sizeof(hof) ||
            close(handle))
        perror(hof_path);
    if (handle >= 0)
        close(handle);
}

// seg001:0F6C
void hof_read(void) {
    int handle;
    hof_count = 0;
    char custom_hof_path[POP_MAX_PATH];
    const char* hof_path = get_hof_path(custom_hof_path, sizeof(custom_hof_path));
    handle = open(hof_path, O_RDONLY | O_BINARY);
    if (handle < 0)
        return;
    if (read(handle, &hof_count, 2) != 2 ||
            read(handle, &hof, sizeof(hof)) != sizeof(hof)) {
        perror(hof_path);
        hof_count = 0;
    }
    close(handle);
}

// seg001:0FC3
void show_hof_text(rect_type *rect, int x_align, int y_align, const char *text) {
    short shadow_color;
    short text_color;
    rect_type rect2;
    text_color = 15;
    shadow_color = 0;
    text_color = 0xB7;
    offset2_rect(&rect2, rect, 1, 1);
    show_text_with_color(&rect2, x_align, y_align, text, shadow_color);
    show_text_with_color(rect, x_align, y_align, text, text_color);
}

// seg001:1029
int fade_in_1(void) {
//	sbyte index;
    word interrupted;
    fade_palette_buffer = make_pal_buffer_fadein(0x6689, /*0*/ 2);
    is_global_fading = 1;
    do {
        interrupted = proc_cutscene_frame(1);
        if (interrupted == 1) {
            return 1;
        }
    } while (interrupted == 0);
    is_global_fading = 0;
    return 0;
}

// seg001:112D
int fade_out_1(void) {
    word interrupted;
    fade_palette_buffer = make_pal_buffer_fadeout(0x6689, /*0*/ 2);
    is_global_fading = 1;
    do {
        interrupted = proc_cutscene_frame(1);
        if (interrupted == 1) {
            return 1;
        }
    } while (interrupted == 0);
    is_global_fading = 0;
    return 0;
}
