/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"

// seg002:0000
void do_init_shad(const byte *source, int seq_index) {
    // custom logic - inactivate guard in the room
    inactivate_guard(Char.room - 1);
    memcpy(&Char, source, 8);
    seqtbl_offset_char(seq_index);
    Char.alive = -1;
    Char.charid = charid_1_shadow;
    Char.sword = sword_0_sheathed;
    Char.fall_x = 0;
    Char.fall_y = 0;
    demo_time = 0;
    guard_skill = 3;
    // custom logic - set the shadow hp
    word shadhp = 1;
    if (shadow_potion_counter < 3) {
        shadhp += shadow_potion_counter;
    }
    guardhp_delta = guardhp_curr = guardhp_max = shadhp;
    saveshad();
}

// seg002:0044
void get_guard_hp(void) {
    guardhp_delta = guardhp_curr = guardhp_max = MIN(settings->extrastrength[guard_skill] + settings->tbl_guard_hp[current_level], MAX_GUARD_HP);
}

// custom logic - play a special shadow music when positional audio is enabled
void play_shadow_music(void) {
    stop_sounds();
    play_sound(is_positional_audio ? sound_78_separated_from_shadow : sound_25_presentation);
}

// seg002:0064
void check_shadow(void) {
    offguard = 0;
    // custom logic - mirror or princess appears in room 1, level 6 instead of shadow
    if (current_level == 6) {
        // Special event: level 6 princess
        Char.room = drawn_room;
        if (Char.room == 1 && Kid.room == Char.room) {
            if (Kid.x > 86) {
                if (leveldoor_open != 0x4D) {
                    get_tile(1, 9, 1);
                    curr_room_modif[curr_tilepos] = 0;
                    get_tile(5, 9, 1);
                    curr_room_modif[curr_tilepos] = 0xFF;
                    get_tile(16, 4, 0);
                    curr_room_modif[curr_tilepos] = 1;
                    play_sound(is_positional_audio ? sound_67_Jaffar_spell : sound_25_presentation);
                    delayed_sound_timer = 2;
                    delayed_sound = (struct delayed_sound_type) {
                        sound_6_gate_closing_fast, 1, 9, 1, 0, 0, 0
                    };
                    leveldoor_open = 0x4D;
                    dead_guard3[0].room = 0;
                    show_black_screen(6, 1);
                }
            } else if (level.roomlinks[4].right == 1) {
                Kid.x = x_bump[/*0 +*/ ONSCREEN_COLUMN_OFFSET] + 2;
                Kid.y = y_land[Kid.curr_row + ONSCREEN_ROW_OFFSET];
                Kid.direction = dir_0_right;
                Kid.frame = frame_15_stand;
                play_sound(is_positional_audio ? sound_68_come_to_princess : sound_29_meet_Jaffar);
                get_tile(5, 4, 0);
                curr_room_modif[curr_tilepos] = 1;
                level.roomlinks[4].right = 4;
                potion_timer = 32;
                show_black_screen(6, 1);
                text_time_remaining = 80;
                text_time_total = 80;
                display_text_bottom("FIND THE WAY TO OTHER SIDE");
            }
            return;
            // custom logic - open the rightmost gate in room 5 after seeing princess
        } else if (Char.room == 5 && level.roomlinks[4].right == 4) {
            get_tile(5, 9, 1);
            curr_room_modif[curr_tilepos] = 0xFF;
            return;
            // custom logic - shadow in room 22
        } else if (Char.room == 22 &&
                   Kid.room == Char.room &&
                   shadow_initialized == 0 &&
                   have_sword) {
            get_tile(21, 9, 0);
            if (curr_room_modif[curr_tilepos] < 40) {
                play_shadow_music();
                do_init_shad(settings->init_shad_6b, seq_2_stand);
                shadow_initialized = 1;
            }
            return;
        }
        // Special event: level 7 shadow
        // custom logic - shadow appears on level 7, room 6 when the gate is open
        // and extra life potion is in the room
    } else if (current_level == 7) {
        Char.room = drawn_room;
        if (Char.room == 6 &&
                Kid.room == Char.room &&
                shadow_initialized == 0) {
            // there is no full potion
            if (get_tile(6, 7, 2) != tiles_10_potion ||
                    curr_room_modif[curr_tilepos] >= 0x38) {
                enter_guard();
                return;
            }
            if (get_tile(6, 5, 1) != tiles_4_gate) {
                enter_guard();
                return;
            }
            // the gate is mostly closed
            if (curr_room_modif[curr_tilepos] < 80) {
                enter_guard();
                return;
            }
            get_tile(7, 3, 2);
            curr_room_tiles[curr_tilepos] = tiles_0_empty;
            get_tile(6, 3, 1);
            trigger_button(1, tiles_6_closer, -1);
            curr_room_tiles[curr_tilepos] = tiles_14_debris;
            play_shadow_music();
            do_init_shad(settings->init_shad_7, seq_2_stand);
            shadow_initialized = 1;
            return;
        } else if (Char.room == 4 &&
                   Kid.room == Char.room &&
                   !level_7_tile_crashed) {
            level_7_tile_crashed = true;
            // shadow only appears (breaks through floor) when the gate is mostly open
            if (get_tile(6, 5, 1) == tiles_4_gate && curr_room_modif[curr_tilepos] >= 80) {
                stop_sounds();
                set_sound_pos(6, 5, 1);
                play_sound(sound_2_tile_crashing);
            }
        }
        // Special event: level 8 shadow
        // custom logic - shadow appears on level 8, room 16 after level door was opened
    } else if (current_level == 8) {
        Char.room = drawn_room;
        if (Char.room == 16 &&
                Kid.room == Char.room &&
                shadow_initialized == 0 &&
                leveldoor_open < 2) {
            if (get_tile(3, 1, 2) != tiles_16_level_door_left ||
                    curr_room_modif[curr_tilepos] == 0) {
                enter_guard();
                return;
            }
            if (get_tile(16, 3, 0) != tiles_4_gate) {
                enter_guard();
                return;
            }
            if (get_tile(16, 7, 0) != tiles_15_opener) {
                enter_guard();
                return;
            }
            curr_room_tiles[curr_tilepos] = tiles_6_closer;
            get_tile(16, 9, 0);
            curr_room_modif[curr_tilepos] = 0xFF;
            if (leveldoor_open == 0) {
                leveldoor_open = 1;
            }
            play_shadow_music();
            do_init_shad(settings->init_shad_8, seq_50_crouch);
            // ensure kid gets inside the room
            if (Kid.curr_col >= 9) {
                Kid.x = 73;
            }
            shadow_initialized = 1;
            return;
        }
        // Special event: level 9 shadow
        // custom logic - the shadow appears on level 9, room 1 when the gate is open
        // and room 23 when extra life potion is in the room
    } else if (current_level == 9) {
        Char.room = drawn_room;
        if (Char.room == 1 &&
                Kid.room == Char.room &&
                Kid.alive < 0 &&
                upside_down_timer == 0) {
            if (Kid.curr_row == 0 && Kid.x >= x_bump[7 + ONSCREEN_COLUMN_OFFSET]) {
                enter_guard();
                return;
            }
            if (get_tile(1, 1, 1) != tiles_4_gate || curr_room_modif[curr_tilepos] < 80) {
                enter_guard();
                return;
            }
            // activate spikes
            if (get_tile(1, 7, 2) == tiles_2_spike) {
                curr_room_modif[curr_tilepos] = 0x8F; // fully out before retracting
                add_trob(1, curr_tilepos, 1);
            }
            play_shadow_music();
            do_init_shad(settings->init_shad_9, seq_2_stand);
            return;
        } else if (Char.room == 23 &&
                   Kid.room == Char.room &&
                   Kid.curr_row == 2 &&
                   shadow_initialized == 0) {
            if (upside_down_timer > 0) {
                // close gate in room 1 that causes shadow to appear when it is open
                if (get_tile(1, 1, 1) == tiles_4_gate) {
                    curr_room_modif[curr_tilepos] = 0;
                }
                enter_guard();
                return;
            }
            if (get_tile(23, 6, 2) != tiles_4_gate || curr_room_modif[curr_tilepos] != 0) {
                enter_guard();
                return;
            }
            if (get_tile(23, 4, 2) != tiles_10_potion || curr_room_modif[curr_tilepos] >= 0x38) {
                enter_guard();
                return;
            }
            play_shadow_music();
            do_init_shad(settings->init_shad_9_2, seq_2_stand);
            shadow_initialized = 1;
            return;
        }
    } else if (current_level == 10 && is_shadow_text_shown == 1) {
        Char.room = drawn_room;
        if (Char.room == 19 &&
                Kid.room == Char.room &&
                Kid.alive < 0) {
            ++is_shadow_text_shown;
            if (shadow_initialized == 0) {
                play_shadow_music();
            }
        }
    } else if (current_level == 11 && is_shadow_text_shown == 0) {
        Char.room = drawn_room;
        if (Char.room == 20 &&
                Kid.room == Char.room &&
                Kid.alive < 0 &&
                !is_kid_shadow) {
            ++is_shadow_text_shown;
            if (hitp_max <= 6) {
                rem_min = 1;
                rem_tick = 90;
                stop_ambient_music();
                music_suspended_timer = 90;
                display_text_bottom("NOT ENOUGH LIVES TO CONTINUE");
                text_time_total = RESTART_NOT_ENOUGH_HITPOINTS;
                text_time_remaining = text_time_total;
                is_show_time = 0;
                delayed_sound_timer = 6;
                delayed_sound = (struct delayed_sound_type) {
                    sound_38_blink, 0, 0, 0, 0, 0, 1
                };
            } else {
                stop_sounds();
                play_sound(is_positional_audio ? sound_79_united_with_shadow : sound_25_presentation);
            }
        }
    }
    enter_guard();
}

// custom logic - enter shadow on level 10
void enter_shadow_level_10(void) {
    word room_minus_1 = Guard.room - 1;
    shadow_initialized = 1;
    if (Guard.room == Kid.room && Guard.alive < 0) {
        // settings for "inactivate_guard" that is called from the "do_init_shad" method
        level.guards_color[room_minus_1] = Guard.color;
        level.guards_color[room_minus_1] |= (guardhp_curr << 4);
        level.guards_x[room_minus_1] = Guard.x;
        level.guards_dir[room_minus_1] = Guard.direction;
    }
    loadkid();
    do_init_shad(settings->init_shad_10, 2);
    delayed_sound_timer = 12;
    delayed_sound = (struct delayed_sound_type) {
        sound_78_separated_from_shadow, 0, 0, 0, 0, 0, 0
    };
}

// custom logic - enter dynamic or inactive guard, if one exists
void enter_dynamic_or_inactive_guard(word room_minus_1, bool outbound_guard) {
    char_type2 inactive;
    short guard_color;
    inactive = inactive_guard[room_minus_1];
    if (inactive.room == drawn_room &&
            (is_inactive_guard_ready(inactive) ||
             (inactive.frame2 >= frame_179_collapse_1 && inactive.frame2 <= frame_185_dead)) &&
            (Guard.alive >= 0 ||
             (!outbound_guard && Guard.direction == dir_56_none) ||
             Guard.charid == charid_1_shadow ||
             Guard.charid == charid_24_mouse)) {
        level.guards_seq_hi[room_minus_1] = 0;
        level.guards_dir[room_minus_1] = inactive.direction;
        level.guards_x[room_minus_1] = inactive.x;
        level.guards_tile[room_minus_1] = inactive.row * 10;
        level.guards_color[room_minus_1] = inactive.color;
        if (inactive_guard_skill < NUM_GUARD_SKILLS) {
            level.guards_skill[room_minus_1] = inactive_guard_skill;
            inactive_guard_skill = 0xFF;
        }
        inactive_guard[room_minus_1].room = 0;
        is_inactive_guard_notice = 0;
        draw_guard_hp(0, MAX_GUARD_HP);
        enter_guard();
    } else if (dynamic_guard.room == drawn_room &&
               Kid.room == drawn_room &&
               Kid.curr_row >= dynamic_guard.row &&
               Kid.curr_row - dynamic_guard.row < 2 &&
               !is_kid_shadow &&
               (Guard.alive >= 0 || Guard.curr_row != dynamic_guard.row) &&
               (dynamic_guard_count[room_minus_1] == 0 ||
                (dynamic_guard_count[room_minus_1] == 1 && dynamic_guard.frame > 0))) {
        if (Guard.alive < 0 && can_guard_see_kid == 2) {
            level.guards_seq_hi[room_minus_1] = 0;
            level.guards_dir[room_minus_1] = dynamic_guard.direction;
            level.guards_tile[room_minus_1] = dynamic_guard.row * 10;
            level.guards_x[room_minus_1] = dynamic_guard.x;
            if (settings->tbl_guard_type[current_level] == guard_type_guard) {
                if (dynamic_guard.color > 0 && dynamic_guard.color <= GUARD_COLORS) {
                    guard_color = dynamic_guard.color;
                } else {
                    guard_color = level.guards_color[room_minus_1] & 0x0F;
                }
            } else {
                guard_color = 0;
            }
            level.guards_color[room_minus_1] = guard_color;
            level.guards_color[room_minus_1] |= ((settings->tbl_guard_hp[current_level] - 1) << 4);
            return;
        }
        if ((dynamic_guard.direction == dir_FF_left && Kid.x > 128) ||
                (dynamic_guard.direction == dir_0_right && Kid.x < 128)) {
            return;
        }
        is_dynamic_guard_fight = 2;
        enter_dynamic_guard();
    }
}

// seg002:0112
void enter_guard(void) {
    word room_minus_1;
    word guard_tile;
    word frame;
    byte seq_hi;

    // arrays are indexed 0..23 instead of 1..24
    room_minus_1 = drawn_room - 1;
    guard_tile = level.guards_tile[room_minus_1];

    // custom logic - inactivate guard when shadow is active
    if ((Guard.charid == charid_1_shadow || Guard.charid == charid_24_mouse) &&
            guardhp_curr > 0 &&
            Guard.alive < 0 &&
            Guard.direction != dir_56_none) {
        if (guard_tile >= 30) {
            draw_guard_hp(0, MAX_GUARD_HP);
            return;
        }
        link_type room_links = level.roomlinks[Guard.room - 1];
        if (Kid.room == Guard.room ||
                Kid.room == room_links.left ||
                Kid.room == room_links.right ||
                Kid.room == room_links.up ||
                Kid.room == room_links.down) {
            inactivate_guard(room_minus_1);
            draw_guard_hp((level.guards_color[room_minus_1] & 0xF0) >> 4, MAX_GUARD_HP);
            return;
        }
    }

    // custom logic - set dynamic guard in the room
    dynamic_guard = settings->dynamic_guards[current_level][room_minus_1];

    if (guard_tile >= 30) {
        // try to see if there are offscreen guards in the left and right rooms that might be visible from this room
        word left_guard_tile = 31;
        word right_guard_tile = 31;
        if (room_L > 0) left_guard_tile = level.guards_tile[room_L - 1];
        if (room_R > 0) right_guard_tile = level.guards_tile[room_R - 1];

        int other_guard_x;
        sbyte other_guard_dir;
        int delta_x;
        int other_room_minus_1;
        if (right_guard_tile < MAX_TILES_PER_ROOM &&
                // only retrieve non-dead, non-dynamic guards
                level.guards_seq_hi[room_R - 1] == 0 &&
                is_dynamic_guard_fight == 0) {
            other_room_minus_1 = room_R - 1;
            other_guard_x = level.guards_x[other_room_minus_1];
            other_guard_dir = level.guards_dir[other_room_minus_1];
            // custom logic - do not restore guards that are standing next to walls in other rooms
            if (get_tile(room_R, get_tile_div_mod_m7(other_guard_x) - 1, right_guard_tile / 10) == tiles_20_wall ||
                    wall_type2(curr_tile2) == wall_type_2_left) {
                // check for guards in the left room
                if (left_guard_tile < MAX_TILES_PER_ROOM) {
                    goto loc_left_guard;
                }
                // enter dynamic or inactive guard, if one exists
                enter_dynamic_or_inactive_guard(room_minus_1, false);
                return;
            }
            // left edge of the guard matters
            if (other_guard_dir == dir_0_right) other_guard_x -= 9; // only retrieve a guard if they will be visible
            if (other_guard_dir == dir_FF_left) other_guard_x += 1; // getting these right was mostly trial and error
            // only retrieve offscreen guards
            if (!(other_guard_x < 58 + 4)) {
                // check for guards in the left room
                if (left_guard_tile < MAX_TILES_PER_ROOM) {
                    goto loc_left_guard;
                }
                // enter dynamic or inactive guard, if one exists
                enter_dynamic_or_inactive_guard(room_minus_1, false);
                return;
            }
            delta_x = 140; // guard leaves to the left
            guard_tile = right_guard_tile;
        } else if (left_guard_tile < MAX_TILES_PER_ROOM &&
                   // only retrieve non-dead, non-dynamic guards
                   level.guards_seq_hi[room_L - 1] == 0 &&
                   is_dynamic_guard_fight == 0) {
loc_left_guard:
            other_room_minus_1 = room_L - 1;
            // only retrieve non-dead, non-dynamic guards
            other_guard_x = level.guards_x[other_room_minus_1];
            other_guard_dir = level.guards_dir[other_room_minus_1];
            // custom logic - do not restore guards that are standing next to walls in other rooms
            if (get_tile(room_L, get_tile_div_mod_m7(other_guard_x) + 1, left_guard_tile / 10) == tiles_20_wall ||
                    wall_type2(curr_tile2) == wall_type_1_right) {
                // enter dynamic or inactive guard, if one exists
                enter_dynamic_or_inactive_guard(room_minus_1, false);
                return;
            }
            // right edge of the guard matters
            if (other_guard_dir == dir_0_right) other_guard_x -= 9;
            if (other_guard_dir == dir_FF_left) other_guard_x += 1;
            // only retrieve offscreen guards
            if (!(other_guard_x > 190 - 4)) {
                // enter dynamic or inactive guard, if one exists
                enter_dynamic_or_inactive_guard(room_minus_1, false);
                return;
            }
            delta_x = -140; // guard leaves to the right
            guard_tile = left_guard_tile;
        } else {
            // enter dynamic or inactive guard, if one exists
            enter_dynamic_or_inactive_guard(room_minus_1, false);
            return;
        }

        // retrieve guard from adjacent room
        level.guards_x[room_minus_1] = level.guards_x[other_room_minus_1] + delta_x;
        level.guards_color[room_minus_1] = level.guards_color[other_room_minus_1];
        level.guards_dir[room_minus_1] = level.guards_dir[other_room_minus_1];
        level.guards_seq_hi[room_minus_1] = level.guards_seq_hi[other_room_minus_1];
        level.guards_seq_lo[room_minus_1] = level.guards_seq_lo[other_room_minus_1];
        level.guards_skill[room_minus_1] = level.guards_skill[other_room_minus_1];

        level.guards_tile[other_room_minus_1] = 0xFF;
        level.guards_seq_hi[other_room_minus_1] = 0;

        // enter dynamic or inactive guard, if one exists
        enter_dynamic_or_inactive_guard(room_minus_1, true);
    }

    // custom logic - clear inactive guard when shadow/mouse enters
    if (inactive_guard[room_minus_1].room == drawn_room &&
            (Guard.charid == charid_1_shadow || Guard.charid == charid_24_mouse)) {
        inactive_guard[room_minus_1].room = 0;
        is_inactive_guard_notice = 0;
    }

    Char.room = drawn_room;
    Char.curr_row = guard_tile / 10;
    Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
    Char.x = level.guards_x[room_minus_1];
    Char.curr_col = get_tile_div_mod_m7(Char.x);
    Char.direction = level.guards_dir[room_minus_1];
    Char.repeat = 0;

    guard_skill = level.guards_skill[room_minus_1];
    if (guard_skill >= NUM_GUARD_SKILLS) {
        guard_skill = 3;
    }

    // custom logic - only regular guards have different colors
    if (settings->tbl_guard_type[current_level] == guard_type_guard) {
        // fat guard has the default color
        if (current_level == settings->fat_guard_level && guard_skill == settings->fat_guard_skill) {
            Char.color = 0;
        } else {
            Char.color = level.guards_color[room_minus_1] & 0x0F;
        }
    } else {
        Char.color = 0;
    }

    // custom logic - reset guard punch count
    guard_punch_count = 0;

    int remembered_hp = (level.guards_color[room_minus_1] & 0xF0) >> 4;

    if (settings->tbl_guard_type[current_level] == guard_type_skeleton) {
        Char.charid = charid_4_skeleton;
    } else {
        Char.charid = charid_2_guard;
    }
    seq_hi = level.guards_seq_hi[room_minus_1];
    if (seq_hi == 0) {
        if (Char.charid == charid_4_skeleton || can_inactive_guard_see_kid == 2) {
            Char.sword = sword_2_drawn;
            seqtbl_offset_char(seq_90_guard_stand_active); // stand active (when entering room)
        } else {
            Char.sword = sword_0_sheathed;
            seqtbl_offset_char(seq_77_guard_stand_inactive); // stand inactive (when entering room)
        }
    } else {
        Char.curr_seq = level.guards_seq_lo[room_minus_1] + (seq_hi << 8);
    }
    play_seq();
    frame = Char.frame;
    // custom logic - Jaffar's vision reappears on level 13 when the level door is "open"
    if (current_level == 13 &&
            Char.charid != charid_0_kid &&
            Kid.room == 1 &&
            Kid.x < 100 && // kid is entering from left the gate into room 1
            leveldoor_open == 0) {
        frame = frame_166_stand_inactive;
    }
    if (frame == frame_185_dead || frame == frame_177_spiked || frame == frame_178_chomped) {
        Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
        Char.alive = 1;
        draw_guard_hp(0, guardhp_curr);
        guardhp_curr = 0;
        // custom logic - adds the fallout logic when guard is dead on a non-floor tile
        if (frame == frame_185_dead &&
                Char.curr_row == 2 &&
                !tile_is_floor(get_tile(Char.room, Char.curr_col, Char.curr_row))) {
            Char.y = OUT_OF_BOUNDS_Y;
            saveshad();
            return;
        }
    } else {
        Char.alive = -1;
        justblocked = 0;
        guard_refrac = 0;
        is_guard_notice = 0;
        should_turn_around = 0;
        get_guard_hp();
        // custom logic - do not remember hitpoints for Jaffar's vision
        if (remembered_hp > 0 && settings->tbl_guard_type[current_level] != guard_type_Jaffar_vision) {
            guardhp_delta = guardhp_max = guardhp_curr = (word) remembered_hp;
        }
        // custom logic - keep the opener button pressed
        determine_col();
        if (get_tile_at_char() == tiles_15_opener) {
            trigger_button(0, 0, -1);
        }
    }
    Char.fall_y = 0;
    Char.fall_x = 0;
    Char.action = actions_1_run_jump;
    saveshad();

    // custom logic - enter inactive guard if there is no alive guard
    if (inactive_guard[room_minus_1].room == drawn_room && Guard.alive >= 0) {
        char_type2* slot = get_dead_guard_slot(room_minus_1);
        if (slot != NULL) {
            slot->room = room_minus_1 + 1;
            slot->direction = Guard.direction;
            slot->row = Guard.curr_row;
            slot->x = Guard.x;
            slot->color = Guard.color;
            slot->frame2 = Guard.frame;
        }
        enter_dynamic_or_inactive_guard(room_minus_1, false);
        inactive_guard[room_minus_1].room = 0;
        is_inactive_guard_notice = 0;
        return;
    }

    // custom logic - dynamic guard enters when the current guard is dead
    if (dynamic_guard.room == drawn_room &&
            dynamic_guard_timer == 0 &&
            Kid.room == drawn_room &&
            (dynamic_guard_count[room_minus_1] == 0 ||
             (dynamic_guard_count[room_minus_1] == 1 && dynamic_guard.frame > 0))) {
        if ((dynamic_guard.direction == dir_FF_left && Kid.x > 128) ||
                (dynamic_guard.direction == dir_0_right && Kid.x < 128)) {
            return;
        }
        if (Char.alive == 1) {
            Char.alive = DROPDEAD_FRAMES - 1;
            saveshad();
        } else if (Char.alive < 0 &&
                   Char.curr_row != Kid.curr_row &&
                   dynamic_guard.row == Kid.curr_row &&
                   (inactive_guard[room_minus_1].room != drawn_room || inactive_guard[room_minus_1].direction == dir_56_none)) {
            inactivate_active_guard();
            enter_dynamic_guard();
        }
        return;
    }

    if (Guard.alive < 0 && inactive_guard[room_minus_1].room == 0) {
        // custom logic - guards in rooms 5 or 8 of level 10 should not follow kid
        if (current_level == 10 && (room_minus_1 + 1 == 5 || room_minus_1 + 1 == 8)) {
            return;
        }

        word left_guard_tile = 31;
        word right_guard_tile = 31;
        if (room_L > 0) left_guard_tile = level.guards_tile[room_L - 1];
        if (room_R > 0) right_guard_tile = level.guards_tile[room_R - 1];

        int other_guard_x;
        sbyte other_guard_dir;
        int other_room_minus_1;
        int delta_x;
        if (right_guard_tile < MAX_TILES_PER_ROOM &&
                // only retrieve non-dead, non-dynamic guards
                level.guards_seq_hi[room_R - 1] == 0) {
            other_room_minus_1 = room_R - 1;
            other_guard_x = level.guards_x[other_room_minus_1];
            other_guard_dir = level.guards_dir[other_room_minus_1];
            // custom logic - do not restore guards that are standing next to walls in other rooms
            if (get_tile(room_R, get_tile_div_mod_m7(other_guard_x) - 1, right_guard_tile / 10) == tiles_20_wall ||
                    wall_type2(curr_tile2) == wall_type_2_left) {
                // check for guards in the left room
                if (left_guard_tile < MAX_TILES_PER_ROOM) {
                    goto loc_left_inactive_guard;
                }
                return;
            }
            // left edge of the guard matters
            if (other_guard_dir == dir_0_right) other_guard_x -= 9; // only retrieve a guard if they will be visible
            if (other_guard_dir == dir_FF_left) other_guard_x += 1; // getting these right was mostly trial and error
            // only retrieve offscreen guards
            if (!(other_guard_x < 58 + 4)) {
                // check for guards in the left room
                if (left_guard_tile < MAX_TILES_PER_ROOM) {
                    goto loc_left_inactive_guard;
                }
                return;
            }
            delta_x = 140; // guard leaves to the left
        } else if (left_guard_tile < MAX_TILES_PER_ROOM &&
                   // only retrieve non-dead, non-dynamic guards
                   level.guards_seq_hi[room_L - 1] == 0) {
loc_left_inactive_guard:
            other_room_minus_1 = room_L - 1;
            // only retrieve non-dead, non-dynamic guards
            other_guard_x = level.guards_x[other_room_minus_1];
            other_guard_dir = level.guards_dir[other_room_minus_1];
            // custom logic - do not restore guards that are standing next to walls in other rooms
            if (get_tile(room_L, get_tile_div_mod_m7(other_guard_x) + 1, left_guard_tile / 10) == tiles_20_wall ||
                    wall_type2(curr_tile2) == wall_type_1_right) {
                return;
            }
            // right edge of the guard matters
            if (other_guard_dir == dir_0_right) other_guard_x -= 9;
            if (other_guard_dir == dir_FF_left) other_guard_x += 1;
            // only retrieve offscreen guards
            if (!(other_guard_x > 190 - 4)) {
                return;
            }
            delta_x = -140; // guard leaves to the right
        } else {
            return;
        }
        inactive_guard[room_minus_1].room = room_minus_1 + 1;
        inactive_guard[room_minus_1].x = level.guards_x[other_room_minus_1] + delta_x;
        inactive_guard[room_minus_1].row = level.guards_tile[other_room_minus_1] / 10;
        inactive_guard[room_minus_1].color = level.guards_color[other_room_minus_1];
        inactive_guard[room_minus_1].direction = level.guards_dir[other_room_minus_1];
        inactive_guard_skill = level.guards_skill[other_room_minus_1];

        level.guards_tile[other_room_minus_1] = 0xFF;
        level.guards_seq_hi[other_room_minus_1] = 0;
    }
}

// custom logic - do guard fallout
void do_guard_fallout(byte start_fall) {
    word room_minus_1;
    room_minus_1 = Guard.room - 1;
    loadshad();
    load_fram_det_col();
    // falling from above
    if (guard_dropping_down && Char.curr_row <= 0 && Char.y >= OUT_OF_BOUNDS_Y) {
        Char.y = 0;
        Char.curr_row = 0;
        Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_10_PX_OFFSET;
    } else if (Kid.room == level.roomlinks[room_minus_1].down) {
        Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
    }
    // guard keeps falling in a different room
    if (Char.curr_col < 0) {
        Char.room = level.roomlinks[Char.room - 1].left;
        Char.curr_col += 10;
        Char.x += 140;
    } else if (Char.curr_col >= 10) {
        Char.room = level.roomlinks[Char.room - 1].right;
        Char.curr_col -= 10;
        Char.x -= 140;
    }
    saveshad();
    falling_guard.room = Char.room;
    falling_guard.direction = Char.direction;
    falling_guard.x = Char.x;
    falling_guard.y = Char.y;
    falling_guard.fall_y = Char.fall_y; // fall y
    falling_guard.skill = guard_skill; // skill
    falling_guard.color = Char.color;
    falling_guard.start_fall = start_fall; // start falling
    falling_guard.hitp = guardhp_curr; // hitpoints
    level.guards_tile[room_minus_1] = 0xFF;
    Guard.direction = dir_56_none;
    draw_guard_hp(0, guardhp_curr);
    guardhp_curr = 0;
    guard_knocked_tile = 0;
    if (can_inactive_guard_see_kid) {
        enter_dynamic_or_inactive_guard(room_minus_1, Char.room != room_minus_1 + 1);
        if (guard_overlap) {
            Guard.y -= 2;
        }
    }
}

// seg002:0269
void check_guard_fallout(void) {
    if (Guard.direction == dir_56_none) {
        return;
    }
    // custom logic - Jaffar's vision should not fall out
    if (settings->tbl_guard_type[current_level] == guard_type_Jaffar_vision) {
        return;
    }
    // custom logic - proceed with fallout once guard is or below kid's row in the same room
    if (guard_dropping_down) {
        if (guard_dropping_down == 1 && Guard.y <= Kid.y + 3) {
            guard_dropping_down = 0;
        }
        return;
    }
    if (Guard.y < OUT_OF_BOUNDS_Y) {
        return;
    }
    if (Guard.charid == charid_1_shadow) {
        if (Guard.action != actions_4_in_freefall) {
            return;
        }
        loadshad();
        clear_char();
        saveshad();
    } else if (Guard.charid == charid_4_skeleton &&
               current_level == 3 &&
               level.roomlinks[Guard.room - 1].down == 3) {
        Guard.room = 3;
        // custom logic - suspend ambient music
        music_suspended_timer = 32;
        stop_ambient_music();
        // if skeleton falls down into room 3
        if (get_tile(Guard.room, 2, 1) == tiles_11_loose) {
            Guard.x = 133;
            Guard.curr_row = 1;
            Guard.direction = dir_0_right;
            Guard.alive = -1;
            leave_guard();
            // custom logic - skeleton pressed the opener tile
            delayed_sound_timer = 6;
            delayed_sound = (struct delayed_sound_type) {
                sound_17_soft_land, Guard.room, 0, 2, tiles_15_opener, 2, 0
            };
            // custom logic - if skeleton falls into an abyss
        } else if (get_tile(Guard.room, 2, 1) == tiles_0_empty) {
            leave_guard();
            if (get_tile(Guard.room, 0, 2) == tiles_15_opener) {
                trigger_button(1, tiles_15_opener, -1);
            }
            falling_bridge_timer = 18;
            falling_guard = (struct falling_guard_type) {
                Guard.room,
                dir_0_right,
                x_bump[2 + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET,
                y_land[3 + ONSCREEN_ROW_OFFSET],
                0, // fall y
                guard_skill,
                guardhp_curr,
                Guard.color,
                0 // start fall
            };
        }
    } else {
        do_guard_fallout(false);
    }
}

// custom logic - is the guard is a disappearing guard
bool is_disappearing_guard(void) {
    return (Guard.charid == charid_4_skeleton && Guard.frame == frame_185_dead) ||
           settings->tbl_guard_type[current_level] == guard_type_Jaffar_vision ||
           settings->tbl_guard_type[current_level] == guard_type_purple;
}

// custom logic - is the guard a dead guard
bool is_dead_guard(void) {
    return Guard.frame == frame_177_spiked ||
           Guard.frame == frame_178_chomped ||
           Guard.frame == frame_185_dead;
}

// seg002:02F5
void leave_guard(void) {
    word room_minus_1;
    byte guard_frame;
    word guard_seq;
    char_type2* dead_slot;
    // arrays are indexed 0..23 instead of 1..24
    room_minus_1 = Guard.room - 1;

    if (Guard.charid == charid_1_shadow || Guard.charid == charid_24_mouse) {
        // custom logic - clear inactive guard slot
        if (inactive_guard[drawn_room - 1].room == drawn_room) {
            inactive_guard[drawn_room - 1].room = 0;
        }
        draw_guard_hp(0, guardhp_curr);
        return;
    }

    if (Guard.direction == dir_56_none) {
        // custom logic - store fallen out guard
        if (Guard.y >= OUT_OF_BOUNDS_Y && Guard.frame >= frame_102_start_fall_1 && Guard.frame <= frame_106_fall) {
            outbound_dead_guard3[room_minus_1].room = Guard.room;
            outbound_dead_guard3[room_minus_1].direction = Guard.direction;
            outbound_dead_guard3[room_minus_1].x = Guard.x;
            outbound_dead_guard3[room_minus_1].row = Guard.curr_row;
            outbound_dead_guard3[room_minus_1].color = Guard.color;
            outbound_dead_guard3[room_minus_1].frame2 = Guard.frame;
        }
        return;
    }

    // custom logic - do not change the guard's position on level 6
    if (current_level == 6 && falling_exit_checkpoint < 4 && Guard.alive < 0) {
        goto loc_leave_guard;
        // custom logic - do not leave skeleton falling from the bridge on level 12
    } else if (current_level == 12 && falling_bridge_timer != 0) {
        return;
    }

    // custom logic - some guard types disappear when they die
    bool disappearing_guard = is_disappearing_guard();

    level.guards_tile[room_minus_1] = get_tilepos(0, Guard.curr_row);
    level.guards_color[room_minus_1] = Guard.color;
    if (guardhp_curr < 16) // can remember 1..15 hp
        level.guards_color[room_minus_1] |= (guardhp_curr << 4);
    level.guards_x[room_minus_1] = Guard.x;
    level.guards_dir[room_minus_1] = Guard.direction;
    level.guards_skill[room_minus_1] = guard_skill;
    guard_frame = Guard.frame;
    // custom logic - reset guard standing on an opener button
    level.opener_col[room_minus_1] = 0;
    level.opener_row[room_minus_1] = 0;
    if ((Guard.alive < 0 && guardhp_curr > 0) ||
            // custom logic - bowing bird guards are still alive
            (guard_frame >= frame_190_start_bowing_1 && guard_frame <= frame_202_bowing)) {
        // custom logic - store guard standing on a button
        if (guard_frame == frame_166_stand_inactive && get_tile(Guard.room, Guard.curr_col, Guard.curr_row) == tiles_15_opener) {
            // 0 = default. Stored value is incremented by 1.
            level.opener_col[room_minus_1] = Guard.curr_col + 1;
            level.opener_row[room_minus_1] = Guard.curr_row + 1;
        }
        level.guards_seq_hi[room_minus_1] = 0;
        // custom logic - guard on level 5, room 15 needs to wake up
    } else if (current_level == 5 && Guard.room == 15 && took_antidote != 2) {
        level.guards_seq_lo[room_minus_1] = Guard.curr_seq;
        level.guards_seq_hi[room_minus_1] = Guard.curr_seq >> 8;
        // custom logic - move dead guards to dead guard slots for non-dynamic fights
    } else if (!disappearing_guard && (dead_slot = get_dead_guard_slot(room_minus_1)) != NULL) {
        level.guards_tile[room_minus_1] = 0xFF;
        dead_slot->room = Guard.room;
        dead_slot->direction = Guard.direction;
        dead_slot->x = Guard.x;
        dead_slot->row = Guard.curr_row;
        dead_slot->color = Guard.color;
        dead_slot->alt_frame = Guard.repeat;
        if (guard_frame != frame_185_dead &&
                guard_frame != frame_177_spiked &&
                guard_frame != frame_178_chomped) {
            guard_frame = frame_185_dead;
        }
        dead_slot->frame2 = guard_frame;
        leave_outbound_guard(room_minus_1, guard_frame, Guard.direction, Guard.x, Guard.curr_row, Guard.repeat, Guard.color);
    } else {
        guard_seq = Guard.curr_seq;
        if (guard_frame != frame_185_dead &&
                guard_frame != frame_177_spiked &&
                guard_frame != frame_178_chomped) {
            guard_frame = frame_185_dead;
            guard_seq = seqtbl_offset_get_seq(seq_22_crushed);
        }
        level.guards_seq_lo[room_minus_1] = guard_seq;
        level.guards_seq_hi[room_minus_1] = guard_seq >> 8;
        if (disappearing_guard) {
            level.guards_tile[room_minus_1] = 0xFF;
        } else {
            leave_outbound_guard(room_minus_1, guard_frame, Guard.direction, Guard.x, Guard.curr_row, Guard.repeat, Guard.color);
        }
    }
loc_leave_guard:
    // custom logic - do not hide guard if kid fell out
    if (Kid.room != 0) {
        Guard.direction = dir_56_none;
        draw_guard_hp(0, guardhp_curr);
        guardhp_curr = 0;
    }
}

// custom logic - leave outbound dead guard
void leave_outbound_guard(word room_minus_1,
                          byte guard_frame,
                          sbyte guard_direction,
                          byte guard_x,
                          sbyte guard_row,
                          byte alt_frame,
                          short color) {
    word room_left, room_right, other_room;
    int offset;
    room_left = level.roomlinks[room_minus_1].left;
    room_right = level.roomlinks[room_minus_1].right;
    if (guard_x < DEADGUARD_OFFSET_LEFT && room_left && room_left != room_minus_1 + 1) {
        other_room = room_left - 1;
        offset = DEADGUARD_OFFSET;
    } else if (guard_x > DEADGUARD_OFFSET_RIGHT && room_right && room_right != room_minus_1 + 1) {
        other_room = room_right - 1;
        offset = -DEADGUARD_OFFSET;
    } else {
        other_room = 24;
        offset = 0;
    }
    if (other_room < 24) {
        if (outbound_dead_guard1[other_room].room == 0) {
            outbound_dead_guard1[other_room].room = other_room + 1;
            outbound_dead_guard1[other_room].direction = guard_direction;
            outbound_dead_guard1[other_room].x = guard_x + offset;
            outbound_dead_guard1[other_room].row = guard_row;
            outbound_dead_guard1[other_room].color = color;
            outbound_dead_guard1[other_room].frame2 = guard_frame;
            outbound_dead_guard1[other_room].alt_frame = alt_frame;
        } else if (outbound_dead_guard2[other_room].room == 0) {
            outbound_dead_guard2[other_room].room = other_room + 1;
            outbound_dead_guard2[other_room].direction = guard_direction;
            outbound_dead_guard2[other_room].x = guard_x + offset;
            outbound_dead_guard2[other_room].row = guard_row;
            outbound_dead_guard2[other_room].color = color;
            outbound_dead_guard2[other_room].frame2 = guard_frame;
            outbound_dead_guard2[other_room].alt_frame = alt_frame;
        } else if (outbound_dead_guard3[other_room].room == 0) {
            outbound_dead_guard3[other_room].room = other_room + 1;
            outbound_dead_guard3[other_room].direction = guard_direction;
            outbound_dead_guard3[other_room].x = guard_x + offset;
            outbound_dead_guard3[other_room].row = guard_row;
            outbound_dead_guard3[other_room].color = color;
            outbound_dead_guard3[other_room].frame2 = guard_frame;
            outbound_dead_guard3[other_room].alt_frame = alt_frame;
        }
    }
}

// seg002:039E
void follow_guard(void) {
    word kid_room_m1;
    kid_room_m1 = Kid.room - 1;

    // custom logic - move a dead guard into a dead slot
    if (level.guards_seq_hi[kid_room_m1] != 0 &&
            level.guards_tile[kid_room_m1] < MAX_TILES_PER_ROOM &&
            level.guards_dir[kid_room_m1] != dir_56_none) {
        move_dead_guard_to_dead_slot(kid_room_m1);
    }

    level.guards_tile[kid_room_m1] = 0xFF;
    level.guards_tile[Guard.room - 1] = 0xFF;

    loadshad();
    goto_other_room(roomleave_result);
    saveshad();
}

// custom logic - is there is an active guard in the room?
bool has_active_guard(word room_minus_1) {
    return level.guards_seq_hi[room_minus_1] == 0 &&
           level.guards_dir[room_minus_1] != dir_56_none &&
           level.guards_tile[room_minus_1] < MAX_TILES_PER_ROOM;
}

// custom logic - save guard in the specified room into the inactive guard slot
void inactivate_guard(word room_minus_1) {
    if (!has_active_guard(room_minus_1)) {
        return;
    }
    int remembered_skill = level.guards_skill[room_minus_1];
    if (remembered_skill >= NUM_GUARD_SKILLS) {
        remembered_skill = 3;
    }
    int hitp = (level.guards_color[room_minus_1] & 0xF0) >> 4;
    if (hitp == 0) {
        hitp = settings->extrastrength[remembered_skill] + settings->tbl_guard_hp[current_level];
        level.guards_color[room_minus_1] |= (hitp << 4);
    }
    inactive_guard_skill = remembered_skill;
    inactive_guard[room_minus_1] = (struct char_type2) {
        room_minus_1 + 1,
                     level.guards_dir[room_minus_1],
                     level.guards_x[room_minus_1],
                     level.guards_tile[room_minus_1] / 10,
                     0,
                     level.guards_color[room_minus_1],
                     frame_166_stand_inactive,
                     0,
                     0,
                     0
    };
    inactive_seq_ptr = 0xFF;
    is_inactive_guard_notice = 0;
}

// custom logic - move the active guard in the current room into the inactive guard slot
void inactivate_active_guard(void) {
    if (guardhp_curr > 0 &&
            Guard.alive < 0 &&
            Guard.direction != dir_56_none &&
            Guard.curr_row >= 0 &&
            Guard.curr_row < 3 &&
            Guard.curr_col >= 0 &&
            Guard.curr_col < 10) {
        int guard_color = Guard.color;
        guard_color |= guardhp_curr << 4;
        inactive_guard_skill = guard_skill;
        inactive_guard[Guard.room - 1] = (struct char_type2) {
            Guard.room,
            Guard.direction,
            Guard.x,
            Guard.curr_row,
            0,
            guard_color,
            frame_166_stand_inactive,
            0,
            0,
            0
        };
        inactive_seq_ptr = 0xFF;
        is_inactive_guard_notice = 0;
        Guard.alive = 0;
        Guard.direction = dir_56_none;
        check_can_inactive_guard_see_kid();
    }
}

// custom logic - finds the first available dead guard slot
char_type2* get_dead_guard_slot(int room_minus_1) {
    if (dead_guard1[room_minus_1].room == 0 || dead_guard1[room_minus_1].direction == dir_56_none) {
        return &dead_guard1[room_minus_1];
    }
    if (dead_guard2[room_minus_1].room == 0 || dead_guard2[room_minus_1].direction == dir_56_none) {
        return &dead_guard2[room_minus_1];
    }
    if (dead_guard3[room_minus_1].room == 0 || dead_guard3[room_minus_1].direction == dir_56_none) {
        return &dead_guard3[room_minus_1];
    }
    if (outbound_dead_guard1[room_minus_1].room == 0 || outbound_dead_guard1[room_minus_1].direction == dir_56_none) {
        return &outbound_dead_guard1[room_minus_1];
    }
    if (outbound_dead_guard2[room_minus_1].room == 0 || outbound_dead_guard2[room_minus_1].direction == dir_56_none) {
        return &outbound_dead_guard2[room_minus_1];
    }
    if (outbound_dead_guard3[room_minus_1].room == 0 || outbound_dead_guard3[room_minus_1].direction == dir_56_none) {
        return &outbound_dead_guard3[room_minus_1];
    }
    return NULL;
}

// custom logic - moves dead guard from the guard slot to a dead slot
void move_dead_guard_to_dead_slot(int room_minus_1) {
    char_type2* slot;
    byte guard_frame;
    word guard_seq;
    slot = get_dead_guard_slot(room_minus_1);
    if (slot != NULL) {
        guard_seq = level.guards_seq_lo[room_minus_1] + (level.guards_seq_hi[room_minus_1] << 8);
        if (guard_seq - seqtbl_offset_get_seq(seq_51_spiked) >= 0 && guard_seq - seqtbl_offset_get_seq(seq_51_spiked) < 10) {
            guard_frame = frame_177_spiked;
        } else if (guard_seq - seqtbl_offset_get_seq(seq_54_chomped) >= 0 && guard_seq - seqtbl_offset_get_seq(seq_54_chomped) < 10) {
            guard_frame = frame_178_chomped;
        } else {
            guard_frame = frame_185_dead;
        }
        slot->room = room_minus_1 + 1;
        slot->direction = level.guards_dir[room_minus_1];
        slot->x = level.guards_x[room_minus_1];
        slot->row = level.guards_tile[room_minus_1] / 10;
        slot->color = level.guards_color[room_minus_1];
        slot->frame2 = guard_frame;
    }
}

// custom logic - visible guard should continue falling if kid exited the room
bool guard_fell_out(void) {
    if (Guard.charid != charid_1_shadow &&
            !is_dead_guard() &&
            !is_disappearing_guard() &&
            Guard.room != next_room) {
        if (Guard.fall_y > 0 || Guard.action == actions_3_in_midair || Guard.action == actions_4_in_freefall) {
            do_guard_fallout(false);
            return true;
        }
        if (!tile_is_floor(get_tile(Guard.room, Guard.curr_col, Guard.curr_row)) ||
                (curr_tile2 == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0)) {
            do_guard_fallout(true);
            return true;
        }
    }
    return false;
}

// custom logic - guard follows kid down
void follow_kid_down(void) {
    sbyte char_col;
    short distance = ABS(get_tile_div_mod_m7(Kid.x) - get_tile_div_mod_m7(Guard.x));
    for (byte col_delta = 1; col_delta <= distance + 1; ++col_delta) {
        // find the empty tile where kid dropped down
        if (tile_is_empty(get_tile(Guard.room, get_tile_div_mod_m7(Guard.x) + dir_front[Guard.direction + 1] * col_delta, Guard.curr_row))) {
            char_col = tile_col;
            // if the tile below is a safe floor, drop down
            if (tile_is_floor(get_tile(Kid.room, char_col, Kid.curr_row)) &&
                    curr_tile2 != tiles_2_spike &&
                    curr_tile2 != tiles_11_loose &&
                    curr_tile2 != tiles_18_chomper) {
                // Set a timer for the number of ticks based on the distance between kid and guard.
                // We cannot control the guard above in a case when there is already a guard in the room below,
                // so the delay of guard walking to the abyss is simulated by a timer.
                guard_dropping_down = (distance + 1) * 4;
                guard_refrac = guard_dropping_down + 6;
                if (char_col > 0 && char_col < 9) {
                    // position the guard in right on the abyss tile before dropping down
                    Guard.x = x_bump[char_col + ONSCREEN_COLUMN_OFFSET] + (Guard.direction == dir_0_right ? 0 : 14);
                    Guard.curr_col = char_col;
                }
                break;
            }
        }
    }
}

// seg002:03C7
void exit_room(void) {
    word leave;
    word kid_room_m1;
    word room_below;
    sbyte char_col;
    leave = 0;
    if (exit_room_timer != 0) {
        --exit_room_timer;
        // We cannot skip 2 frames when kid is out-of-bounds below the room threshold.
        // Otherwise, the "Char.y" value will end up above 240 (-16 when cast as sbyte)
        // in some scenarios (https://www.popot.org/documentation.php?doc=OverflowBug)
        // causing the room to switch up instead of down (see the "leave_room" function).
        if (!(Char.y >= OUT_OF_BOUNDS_Y && Char.curr_row >= 2)) {
            return;
        }
    }
    loadkid();
    load_frame_to_obj();
    set_char_collision();
    roomleave_result = leave_room();
    if (roomleave_result < 0) {
        return;
    }
    savekid();
    // custom logic - do not switch rooms when falling/hanging down in the flame room on level 10
    if (current_level == 10 && drawn_room == 8 && (Char.room == 8 || Char.room == 24) && roomleave_result >= 2) {
        return;
    }
    next_room = Char.room;
    // custom logic - guards can disappear during super high jumps from floor 1 or during broken room links
    // if the room does not actually switch
    if (next_room == drawn_room) {
        return;
    }
    kid_room_m1 = Kid.room - 1;
    if (Guard.direction == dir_56_none) {
        leave_guard();
        return;
    }
    // custom logic - guard fell out in another room
    if (guard_fell_out()) {
        return;
    }
    if (Guard.alive < 0 &&
            (Guard.sword == sword_2_drawn || can_guard_see_kid == 2 || (is_kid_shadow && have_sword))) {
        if (kid_room_m1 <= 23 &&
                // custom logic - Jaffar's vision does not follow kid
                current_level != 13 &&
                // custom logic - follow if there is no inactive guard
                (level.guards_tile[kid_room_m1] >= 30 ||
                 level.guards_seq_hi[kid_room_m1] != 0 ||
                 inactive_guard[kid_room_m1].room == 0)) {
            determine_col();
            if (roomleave_result == room_dir_left) {
                if (droppedout && Kid.action == actions_3_in_midair) {
                    room_below = level.roomlinks[kid_room_m1].down;
                    if (!(room_below > 0 && (level.guards_tile[room_below - 1] >= 30 || level.guards_seq_hi[room_below - 1] != 0))) {
                        leave = 1;
                    }
                } else if (can_guard_see_kid == 0 &&
                           !check_kid_dropped_into_next_room() &&
                           (!is_kid_shadow || !have_sword || Kid.curr_row != Guard.curr_row)) {
                    leave = 1;
                } else if (Guard.curr_col >= Kid.curr_col && Guard.curr_row == Kid.curr_row) {
                    for (char_col = Guard.curr_col; char_col >= Kid.curr_col; char_col--) {
                        if (!tile_is_floor(get_tile(Guard.room, char_col, Guard.curr_row)) ||
                                curr_tile2 == tiles_18_chomper ||
                                (char_col < Guard.curr_col && curr_tile2 == tiles_4_gate && curr_room_modif[curr_tilepos] < GUARD_HEIGHT_GATE_MODIFIER)) {
                            leave = 1;
                            break;
                        }
                    }
                    char_col = get_tile_div_mod_m7(Kid.x);
                    if (char_col >= 9 &&
                            (!tile_is_floor(get_tile(Kid.room, char_col, Kid.curr_row)) ||
                             curr_tile2 == tiles_18_chomper ||
                             (curr_tile2 == tiles_4_gate && curr_room_modif[curr_tilepos] < GUARD_HEIGHT_GATE_MODIFIER))) {
                        leave = 1;
                    }
                }
                // custom logic - guards follow kid more aggressively
                if (Guard.x > (114 + CHAR_OFFSET)) {
                    leave = 1;
                    // custom logic - guard in room 5 of level 10 should not follow kid to the level door room
                } else if (current_level == 10 && Guard.room == 5) {
                    leave = 1;
                }
            } else if (roomleave_result == room_dir_right) {
                if (droppedout && Kid.action == actions_3_in_midair) {
                    room_below = level.roomlinks[kid_room_m1].down;
                    if (!(room_below > 0 && (level.guards_tile[room_below - 1] >= 30 || level.guards_seq_hi[room_below - 1] != 0))) {
                        leave = 1;
                    }
                } else if (can_guard_see_kid == 0 &&
                           !check_kid_dropped_into_next_room() &&
                           (!is_kid_shadow || !have_sword || Kid.curr_row != Guard.curr_row)) {
                    leave = 1;
                } else if (Guard.curr_col <= Kid.curr_col && Guard.curr_row == Kid.curr_row) {
                    for (char_col = Guard.curr_col; char_col <= Kid.curr_col; char_col++) {
                        if (!tile_is_floor(get_tile(Guard.room, char_col, Guard.curr_row)) ||
                                (char_col > Guard.curr_col && curr_tile2 == tiles_18_chomper) ||
                                (curr_tile2 == tiles_4_gate && curr_room_modif[curr_tilepos] < GUARD_HEIGHT_GATE_MODIFIER && char_col != Kid.curr_col)) {
                            leave = 1;
                            break;
                        }
                    }
                    char_col = get_tile_div_mod_m7(Kid.x);
                    if (char_col <= 0 && (!tile_is_floor(get_tile(Kid.room, char_col, Kid.curr_row)) || curr_tile2 == tiles_18_chomper)) {
                        leave = 1;
                    }
                }
                // custom logic - guards follow kid more aggressively
                if (Guard.x < (141 - CHAR_OFFSET)) {
                    leave = 1;
                    // custom logic - skeleton that is backing down should not follow kid
                } else if (current_level == 3 && Guard.room == 6) {
                    leave = 1;
                    // custom logic - guard in room 8 of level 10 should not follow kid out of the level door room
                } else if (current_level == 10 && Guard.room == 8) {
                    leave = 1;
                }
            } else if (roomleave_result == room_dir_up) {
                if (Guard.curr_row >= 0) leave = 1;
            } else { // down
                if (Guard.curr_row < 2) {
                    leave = 1;
                    // custom logic - guard tries to follow the kid down from row 2
                } else if (droppedout &&
                           Kid.alive < 0 &&
                           Kid.curr_row == 0 &&
                           Kid.room == level.roomlinks[Guard.room - 1].down &&
                           Guard.curr_row == 2 &&
                           ABS(Kid.curr_col - MAX(MIN(Guard.curr_col, 10), -1)) <= 4) {
                    follow_kid_down();
                    leave = guard_dropping_down ? 0 : 1;
                } else {
                    leave = 1;
                }
            }
        } else {
            leave = 1;
        }
        // custom logic - guard falling into the room below should be followed
    } else if (Guard.curr_row == 3 && (Guard.fall_y > 0 || Guard.action == actions_3_in_midair || Guard.action == actions_4_in_freefall)) {
        leave = 0;
    } else {
        leave = 1;
    }
    if (leave) {
        droppedout = 0;
        leave_guard();
    } else {
        // custom logic - inactivate guard
        inactivate_guard(kid_room_m1);
        // custom logic - ensure the guard does not get stuck behind the gate
        if (roomleave_result == room_dir_right &&
                get_tile(Guard.room, Guard.curr_col, Guard.curr_row) == tiles_4_gate &&
                get_tile(Guard.room, Guard.curr_col + 1, Guard.curr_row) == tiles_6_closer) {
            Guard.x = x_bump[Guard.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
        }
        follow_guard();
    }
}

void follow_inactive_guard_left(word char_room_minus_1, word other_room_minus_1) {
    int left_bound = can_inactive_guard_see_kid == 2 ? 114 : 71;
    // inactive guard should not follow active guard into a room with yet another guard
    if (can_guard_see_kid == 2 && level.guards_tile[other_room_minus_1] < 30 && level.guards_seq_hi[other_room_minus_1] == 0) {
        inactive_guard[char_room_minus_1].frame2 = frame_166_stand_inactive;
        return;
    }
    if (inactive_guard[char_room_minus_1].room == char_room_minus_1 + 1 &&
            inactive_guard[char_room_minus_1].direction != dir_56_none &&
            inactive_guard[char_room_minus_1].x <= left_bound &&
            (inactive_guard[other_room_minus_1].room != other_room_minus_1 + 1 ||
             inactive_guard[other_room_minus_1].direction == dir_56_none)) {
        if (wall_type2(get_tile(inactive_guard[char_room_minus_1].room,
                                MAX(get_tile_div_mod_m7(inactive_guard[char_room_minus_1].x) - 1, -1),
                                inactive_guard[char_room_minus_1].row)) != wall_type_0_none) {
            inactive_guard[char_room_minus_1].frame2 = frame_166_stand_inactive;
            return;
        }
        inactive_guard[other_room_minus_1].room = other_room_minus_1 + 1;
        inactive_guard[other_room_minus_1].direction = inactive_guard[char_room_minus_1].direction;
        inactive_guard[other_room_minus_1].x = inactive_guard[char_room_minus_1].x + 140;
        inactive_guard[other_room_minus_1].row = inactive_guard[char_room_minus_1].row;
        inactive_guard[other_room_minus_1].frame2 = inactive_guard[char_room_minus_1].frame2;
        inactive_guard[other_room_minus_1].color = inactive_guard[char_room_minus_1].color;
        inactive_guard[char_room_minus_1].room = 0;
        level.guards_seq_hi[other_room_minus_1] = 0;
    } else {
        inactive_guard[char_room_minus_1].frame2 = frame_166_stand_inactive;
    }
}

void follow_inactive_guard_right(word char_room_minus_1, word other_room_minus_1) {
    int right_bound = can_inactive_guard_see_kid == 2 ? 141 : 184;
    // inactive guard should not follow active guard into a room with yet another guard
    if (can_guard_see_kid == 2 && level.guards_tile[other_room_minus_1] < 30 && level.guards_seq_hi[other_room_minus_1] == 0) {
        inactive_guard[char_room_minus_1].frame2 = frame_166_stand_inactive;
        return;
    }
    if (inactive_guard[char_room_minus_1].room == char_room_minus_1 + 1 &&
            inactive_guard[char_room_minus_1].direction != dir_56_none &&
            inactive_guard[char_room_minus_1].x >= right_bound &&
            (inactive_guard[other_room_minus_1].room != other_room_minus_1 + 1 ||
             inactive_guard[other_room_minus_1].direction == dir_56_none)) {
        if (wall_type2(get_tile(inactive_guard[char_room_minus_1].room,
                                MIN(get_tile_div_mod_m7(inactive_guard[char_room_minus_1].x) + 1, 9),
                                inactive_guard[char_room_minus_1].row)) != wall_type_0_none) {
            inactive_guard[char_room_minus_1].frame2 = frame_166_stand_inactive;
            return;
        }
        inactive_guard[other_room_minus_1].room = other_room_minus_1 + 1;
        inactive_guard[other_room_minus_1].direction = inactive_guard[char_room_minus_1].direction;
        inactive_guard[other_room_minus_1].x = inactive_guard[char_room_minus_1].x - 140;
        inactive_guard[other_room_minus_1].row = inactive_guard[char_room_minus_1].row;
        inactive_guard[other_room_minus_1].frame2 = inactive_guard[char_room_minus_1].frame2;
        inactive_guard[other_room_minus_1].color = inactive_guard[char_room_minus_1].color;
        inactive_guard[char_room_minus_1].room = 0;
        level.guards_seq_hi[other_room_minus_1] = 0;
    } else {
        inactive_guard[char_room_minus_1].frame2 = frame_166_stand_inactive;
    }
}

// seg002:0486
int goto_other_room(short direction) {
    short opposite_dir;
    word char_room_minus_1;
    // custom logic - do not go into non-existing rooms
    byte other_room = ((byte*)&level.roomlinks[Char.room - 1])[direction];
    if (Char.room == 0 && other_room != 0) {
        other_room = 0;
        redraw_screen(0);
    }
    // custom logic - stuck button becomes unstuck on level 9, room 23 in the upside down mode
    if (current_level == 9 && other_room == 23) {
        get_tile(other_room, 8, 2);
        curr_room_tiles[curr_tilepos] = upside_down_timer == 0 ? tiles_5_stuck : tiles_15_opener;
    }
    char_room_minus_1 = Char.room - 1;
    Char.room = other_room;
    if (direction == room_dir_left) {
        // custom logic - the cell room on level 1 is dim
        if (current_level == 1 && Char.room == 24 && !enable_lighting && get_tile(24, 9, 0) == tiles_4_gate) {
            enable_lighting = 1;
            // custom logic - the first room on level 6 is well lit
        } else if (current_level == 6 && Char.room == 24 && enable_lighting) {
            enable_lighting = 0;
            // custom logic - rooms in the second part of level 13 are dim
        } else if (current_level == 14 && Char.room == 12 && !enable_lighting) {
            enable_lighting = 1;
            // custom logic - last room of level 13 is well lit
        } else if (current_level == 14 && Char.room == 23 && enable_lighting) {
            enable_lighting = 0;
        }
        // custom logic - guards follow kid more aggressively
        if (Char.x > 114) {
            Char.x = 254;
        } else {
            Char.x += 140;
        }
        if (other_room != 0) {
            follow_inactive_guard_left(char_room_minus_1, other_room - 1);
        }
        opposite_dir = 1;
    } else if (direction == room_dir_right) {
        // custom logic - the room next to the cell on level 1 is well lit
        if (current_level == 1 && Char.room == 18 && enable_lighting) {
            enable_lighting = 0;
            // custom logic - the right-most room on level 6 is dim
        } else if (current_level == 6 && Char.room == 10 && !enable_lighting) {
            enable_lighting = 1;
            // custom logic - rooms in the second part of level 13 are well lit
        } else if (current_level == 14 && Char.room == 11 && enable_lighting) {
            enable_lighting = 0;
            // custom logic - room 19 of level 13 is dim
        } else if (current_level == 14 && Char.room == 19 && !enable_lighting) {
            enable_lighting = 1;
        }
        // custom logic - guards follow kid more aggressively
        if (Char.x < 141) {
            Char.x = 1;
        } else {
            Char.x -= 140;
        }
        if (other_room != 0) {
            follow_inactive_guard_right(char_room_minus_1, other_room - 1);
        }
        opposite_dir = 0;
    } else if (direction == room_dir_up) {
        Char.y += 189;
        Char.curr_row = y_to_row_mod4(Char.y);
        opposite_dir = 3;
    } else { // down
        // custom logic - the right-most room on level 6 is dim
        if (current_level == 6 && Char.room == 10 && !enable_lighting) {
            enable_lighting = 1;
        }
        Char.y -= 189;
        Char.curr_row = y_to_row_mod4(Char.y);
        opposite_dir = 2;
    }
    // custom logic - dynamic fight ended
    is_dynamic_guard_fight = 0;
    // custom logic - disable dynamic guards in the room if kid exited it in their direction
    if (char_room_minus_1 + 1 == dynamic_guard.room &&
            ((dynamic_guard.direction == dir_0_right && direction == room_dir_left) ||
             (dynamic_guard.direction == dir_FF_left && direction == room_dir_right))
       ) {
        dynamic_guard_count[char_room_minus_1] = 0xFF;
    }
    return opposite_dir;
}

// seg002:0504
short leave_room(void) {
    short frame;
    word action;
    short chary;
    short room;
    short leave_dir;
    chary = Char.y;
    action = Char.action;
    frame = Char.frame;
    room = Char.room;
    if (action != actions_5_bumped &&
            action != actions_4_in_freefall &&
            action != actions_3_in_midair &&
            (sbyte)chary < 10 && (sbyte)chary > -16
       ) {
        leave_dir = room_dir_up;
    } else if (chary >= OUT_OF_BOUNDS_Y) {
        // custom logic - do not leave room while a loose tile is dropping down to avoid falling down 2 rooms (Trick 156)
        short col = get_tile_div_mod_m7(Char.x);
        if (col < 0) { // adjust room/column if the tile is in the room on the left
            col += 10;
            room = level.roomlinks[room - 1].left;
        }
        // do not switch rooms yet while falling with the tile
        if (((curmob.room == room && curmob.row == 0) ||
                (curmob.room == level.roomlinks[room - 1].up && curmob.row == 2)) &&
                (curmob.xh >> 2) == col &&
                curmob.speed <= 15 &&
                frame >= frame_102_start_fall_1 && frame < frame_106_fall) {
            return -1;
        }
        leave_dir = room_dir_down;
    } else if (
        // frames 135..149: climb up
        (frame >= frame_135_climbing_1 && frame < frame_150_parry) ||
        // custom logic - jumping above the guard into the next room
        (Opp.alive < 0 && frame >= frame_16_standing_jump_1 && frame <= frame_44_running_jump_5 && !tile_is_floor(get_tile_at_char())) ||
        // frames 110..119: standing up from crouch
        (frame >= frame_110_stand_up_from_crouch_1 && frame <= frame_119_stand_up_from_crouch_10) ||
        // frames 150..162: with sword
        (frame >= frame_150_parry && frame < frame_163_fighting
         // By repeatedly pressing 'back' in a swordfight, you can retreat out of a room without the room changing. (Trick 35)

         // The game waits for a 'legal frame' (e.g. frame_170_stand_with_sword) until leaving is possible;
         // However, this frame is never reached if you press 'back' in the correct pattern!

         // Solution: also allow the room to be changed on frame_157_walk_with_sword
         // Note that this means that the delay for leaving rooms in a swordfight becomes noticably shorter.

         && frame != frame_157_walk_with_sword) ||
        // frames 166..168: with sword
        (frame >= frame_166_stand_inactive && frame < frame_169_begin_block) ||
        action == actions_7_turn // turn
    ) {
        return -1;
    } else if (Char.direction != dir_0_right) {
        // looking left
        if (char_x_left <= 54) {
            leave_dir = room_dir_left;
        } else if (char_x_left >= 198) {
            // custom logic - do not exit the room if there is a wall in the first col with no floor on top
            if (get_tile(room, 10, Char.curr_row) == tiles_20_wall && curr_room != 0 && !tile_is_floor(get_tile(room, 10, Char.curr_row - 1))) {
                return -1;
            }
            leave_dir = room_dir_right;
        } else {
            return -1;
        }
    } else {
        // looking right
        get_tile(room, 9, Char.curr_row);
        if (curr_tile2 != tiles_7_doortop_with_floor &&
                curr_tile2 != tiles_12_doortop &&
                char_x_right >= 201
           ) {
            // custom logic - do not exit the room if there is a wall in the first col with no floor on top
            if (get_tile(room, 10, Char.curr_row) == tiles_20_wall && curr_room != 0 && !tile_is_floor(get_tile(room, 10, Char.curr_row - 1))) {
                return -1;
            }
            leave_dir = room_dir_right;
        } else if (char_x_right <= 57) {
            leave_dir = room_dir_left;
        } else {
            return -1;
        }
    }
    switch (leave_dir) {
        case room_dir_left:
            play_wall_mus();
            // custom logic - presentation music for the spilled poison potion
            play_poison_music();
            // custom logic - add the mirror and keep entry gate open
            enter_mirror_room();
            meet_Jaffar(false);
            break;
        case room_dir_right:
            // custom logic - keep the entry gate closed when entering the other side of room 1
            close_mirror_room_gate();
            meet_Jaffar(true);
            // custom logic - drunk guard snores when kid enters the room without a sword on level 1
            guard_snores();
            break;
        case room_dir_up:
            break;
        case room_dir_down:
            // custom logic - position guard on level 4, room 18
            position_level_4_guard();
            // Special event: falling exit
            if (current_level == settings->falling_exit_level) {
                if (Char.room == 7) { // custom logic - level 6 exit
                    if (++falling_exit_checkpoint >= 4) {
                        stop_all_sounds();
                        saves_per_level = 0;
                        is_restart_level = 1;
                        return -1;
                    } else if (falling_exit_checkpoint == 2) {
                        if (Char.alive < 0) {
                            if (hitp_max > 3) --hitp_max;
                            if (hitp_curr > hitp_max) hitp_curr = hitp_max;
                            hitp_beg_lev = hitp_max;
                        }
                    } else if (falling_exit_checkpoint == 3) {
                        auto_save_level(current_level);
                    }
                } else if (Char.room == settings->falling_exit_room) { // custom logic - level 6B exit
                    falling_exit_final_direction = Kid.direction;
                    return -2;
                }
            }
            break;
    }
    // custom logic - reset skeleton shake timer on skeleton level
    if (current_level == 3) {
        skeleton_shake_timer = 0;
    }
    goto_other_room(leave_dir);
    return leave_dir;
}

// custom logic - move guard in room 18 on level 4 into the next room once the gate is open
void position_level_4_guard(void) {
    if (current_level == 4 && Char.room == 8 &&
            get_tile(17, 1, 0) == tiles_14_debris &&
            level.guards_tile[17] < MAX_TILES_PER_ROOM &&
            level.guards_seq_hi[17] == 0 &&
            level.guards_dir[17] != dir_56_none &&
            level.guards_color[17] > 0) {
        level.guards_tile[17] = 0xFF;
        level.guards_tile[9] = 0;
        level.guards_x[9] = x_bump[1 + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
        level.guards_seq_hi[9] = 0;
        level.guards_dir[9] = dir_0_right;
        level.guards_color[9] = level.guards_color[17];
        level.guards_skill[9] = level.guards_skill[17];
    }
}

// custom logic - playing presentation music for the spilled poison potion
void play_poison_music(void) {
    if (current_level == 5 && Char.room == 16 && took_antidote == 0) {
        ++took_antidote;
        stop_sounds();
        play_sound(sound_25_presentation);
    }
}

// seg002:06AE
void meet_Jaffar(bool right) { // custom logic - added direction tracking parameter to use for level 13
    // Special event: play music
    if (right) {
        // custom logic - level 12B Jaffar's vision tricks
        if (current_level == 13 && Char.room == 12) {
            if (Char.curr_row == 0)  {
                // remove Jaffar's vision when kid is at the top or level door is "open"
                level.guards_dir[0] = dir_56_none;
                level.guards_tile[0] = 0xFF;
                if (leveldoor_open == 0) {
                    get_tile(1, 0, 0);
                    curr_room_modif[curr_tilepos] = 0; // close the gate
                    level_13_nowhere_to_run_timer = 0;
                }
                // Jaffar's vision reappears in the original position if level door is still "closed"
                // and kid is entering from the left room
            } else if (Char.curr_row == 2 && leveldoor_open == 0) {
                level.guards_dir[0] = dir_FF_left;
                level.guards_tile[0] = 27;
                level.guards_x[0] = 160;
                level.guards_skill[0] = 7;
                level.guards_seq_hi[0] = 0;
                if (!jaffar_vision_presented) {
                    jaffar_vision_presented = true;
                    stop_sounds();
                    play_sound(sound_25_presentation);
                }
            }
        }
    } else {
        // custom logic - meet Jaffar
        if (jaffar_presented != 7 && current_level == settings->win_level && Char.room == 3) {
            jaffar_presented = 7;
            jaffar_parry_count = 0;
            stop_sounds();
            play_sound(sound_29_meet_Jaffar);
            // Special event: Jaffar waits a bit (28/12=2.33 seconds)
            jaffar_delay_timer = 28;
            // replace floor tiles in room 3 on level 13
            for (int col = 5; col <= 7; col++) {
                get_tile(3, col, 1);
                curr_room_tiles[curr_tilepos] = tiles_1_floor;
                curr_room_modif[curr_tilepos] = 0;
                curr_room_tiles[curr_tilepos + 10] = tiles_20_wall;
                curr_room_modif[curr_tilepos + 10] = 0;
            }
            // move dead kid into the room 3 from any of the rooms below
            if (is_kid_shadow &&
                    Dead.direction != dir_56_none &&
                    (Dead.room == 7 || Dead.room == 9 || Dead.room == 10)) {
                Dead.room = 3;
                Dead.x = x_bump[3 + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
                Dead.row = 1;
            }
        }
    }
}

// custom logic - add the mirror and keep entry gate open
void enter_mirror_room(void) {
    if (current_level == 6 && leveldoor_open != 0x4D) {
        if (Char.room == 16) {
            get_tile(1, 5, 1);
            curr_room_tiles[curr_tilepos] = tiles_13_mirror;
            curr_room_modif[curr_tilepos] = 1;
            get_tile(1, 6, 1);
            curr_room_tiles[curr_tilepos] = tiles_5_stuck;
        } else if (Char.room == 13) {
            if (get_tile(1, 9, 1) == tiles_4_gate && curr_room_modif[curr_tilepos] < 0xFF) {
                curr_room_modif[curr_tilepos] = 0xFF;
            }
        }
    }
}

// custom logic - keep the entry gate closed when entering the other side of room 1
void close_mirror_room_gate(void) {
    if (current_level == 6 && Char.room == 4) {
        if (get_tile(1, 9, 1) == tiles_4_gate && curr_room_modif[curr_tilepos] > 0) {
            curr_room_modif[curr_tilepos] = 0;
        }
    }
}

// custom logic - play music on level 4 when there is a wall in front of the gate
void play_wall_mus(void) {
    // Special event: wall appears
    if (leveldoor_open != 0 &&
            leveldoor_open != 0x4D && // has the music played already?
            current_level == 4 &&
            Kid.room == 11 &&
            Kid.curr_row == 1
       ) {
        play_sound(sound_25_presentation); // presentation
        leveldoor_open = 0x4D;
    }
}

// custom logic - horse statue makes a neighing sound when prince appears in the same room
void horse_neigh(bool hurt) {
    if (!is_positional_audio) {
        return;
    }
    if (current_level == 10 &&
            Kid.room == 4 &&
            Kid.curr_row == 2 &&
            leveldoor_open < DEAD_KID_LEVEL_10_TIMER &&
            !skeleton_appeared) {
        ++skeleton_appeared;
        delayed_sound_timer = 2;
        if (hurt) {
            play_sound(sound_13_kid_hurt);
            delayed_sound_timer += 10;
        }
        delayed_sound = (struct delayed_sound_type) {
            sound_71_horse_neigh, 4, 5, 2, 0, 0, 0
        };
    }
}

// custom logic - drunk guard snores when kid enters the room without a sword on level 1
void guard_snores(void) {
    if (!have_sword &&
            (!check_sound_playing() || current_sound != sound_59_guard_snoring) &&
            current_level == 1 &&
            Char.room == 22) {
        delayed_sound_timer = 6;
        delayed_sound = (struct delayed_sound_type) {
            sound_59_guard_snoring, 16, 4, 0, 0, 0, 0
        };
    }
}

// seg002:0706
void move_0_nothing(void) {
    control_shift = control_0_released;
    control_alt = control_0_released;
    control_y = control_0_released;
    control_x = control_0_released;
    control_shift2 = control_0_released;
    control_down = control_0_released;
    control_up = control_0_released;
    control_backward = control_0_released;
    control_forward = control_0_released;
}

// seg002:0721
void move_1_forward(void) {
    // custom logic - do not advance kid in the demo past column 5
    if (current_level == 0 &&
            !play_demo_level &&
            Char.room == 4 &&
            Char.charid == charid_0_kid &&
            Char.sword >= sword_2_drawn &&
            Char.curr_col <= 5) {
        return;
    }
    // custom logic - guard on level 10, room 8 should not move beyond column 7
    if (current_level == 10 &&
            Char.room == 8 &&
            Char.charid == charid_2_guard &&
            Char.sword >= sword_2_drawn &&
            Char.curr_row == 2 &&
            Char.direction == dir_0_right &&
            (Kid.room != Char.room || Kid.curr_col >= 8)) {
        offguard = 1; // do not try to fight level door guard out of range
        return;
    }
    control_x = control_m1_held_left;
    control_forward = control_m1_held;
}

// seg002:072A
void move_2_backward(void) {
    control_x = control_1_held_right;
    control_backward = control_m1_held;
}

// seg002:0735
void move_3_up(void) {
    control_y = control_m1_held_up;
    control_up = control_m1_held;
}

// seg002:073E
void move_4_down(void) {
    control_y = control_1_held_down;
    control_down = control_m1_held;
}

// seg002:0749
void move_up_back(void) {
    control_up = control_m1_held;
    move_2_backward();
}

// seg002:0753
void move_down_back(void) {
    control_down = control_m1_held;
    move_2_backward();
}

// seg002:075D
void move_down_forw(void) {
    control_down = control_m1_held;
    move_1_forward();
}

// seg002:0767
void move_6_shift(void) {
    control_shift = control_m1_held;
    control_shift2 = control_m1_held;
}

// seg002:0770
void move_7(void) {
    control_shift = control_0_released;
}

// custom logic - jump grab move
void move_8_jump_grab(void) {
    control_shift = control_m1_held;
    control_shift2 = control_m1_held;
    control_y = control_m1_held_up;
    control_up = control_m1_held;
    control_x = control_0_released;
    control_forward = control_0_released;
}

// seg002:0776
void autocontrol_opponent(void) {
    word charid;
    move_0_nothing();
    charid = Char.charid;
    if (charid == charid_4_skeleton && skeleton_death_timer > 0) {
        return;
    }
    if (charid == charid_0_kid) {
        autocontrol_kid();
    } else {
        if (justblocked) --justblocked;
        if (kid_sword_strike) --kid_sword_strike;
        if (guard_refrac) --guard_refrac;
        if (charid == charid_24_mouse) {
            autocontrol_mouse();
        } else if (charid == charid_4_skeleton) {
            autocontrol_skeleton();
        } else if (charid == charid_1_shadow) {
            autocontrol_shadow();
        } else {
            autocontrol_guard();
        }
    }
}

// seg002:07EB
void autocontrol_mouse(void) {
    if (Char.direction == dir_56_none) {
        return;
    }
    if (current_level == settings->mouse_level) {
        if (Char.action != actions_0_stand && Char.x < 166) {
            if (is_positional_audio) {
                set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
                play_sound(sound_70_mouse_squeak);
            }
            seqtbl_offset_char(seq_107_mouse_stand_up_and_go); // mouse
            play_seq();
        }
        // custom logic - mouse appears and helps kid to complete level 12
    } else if (current_level == 12) {
        if (Char.curr_col < 0) {
            clear_char();
        } else if (Char.curr_col == 0 && Char.curr_row == 0) {
            // create missing floor tiles in the room on the left
            get_tile(level.roomlinks[settings->mouse_2_room - 1].left, settings->mouse_2_room_left_first_empty_col, 0);
            curr_room_tiles[curr_tilepos] = tiles_1_floor;
            while (++curr_tilepos < 9) {
                curr_room_tiles[curr_tilepos] = tiles_1_floor;
            }
            // move the guard closer to kid
            level.guards_x[12] += 5;
        }
    }
}

// seg002:081D
void autocontrol_shadow(void) {
    switch (current_level) {
        case 6:
            autocontrol_shadow_level6();
            autocontrol_shadow_level6b();
            break;
        case 7:
            autocontrol_shadow_level7();
            break;
        case 8:
            autocontrol_shadow_level8();
            break;
        case 9:
            autocontrol_shadow_level9();
            break;
        case 10:
            autocontrol_shadow_level10();
            break;
    }
}

// custom logic - replace skeleton with bones tile
void inactivate_skel(sbyte direction) {
    if (skeleton_death_timer > 0) {
        return;
    }
    if (Char.direction != direction) {
        Char.direction = direction;
        Char.x = char_dx_forward(7);
    }
    seqtbl_offset_char(seq_141_skeleton_collapse);
    skeleton_death_timer = 3;
}

// seg002:0850
void autocontrol_skeleton(void) {
    // custom logic - skeleton on level 3 room 6 does not fight but backs up
    // if kid stays at a reasonable distance
    if (current_level == 3 &&
            (skeleton_timer == 0 || Opp.alive >= 0 || blink_timer == 16) &&
            drawn_room == 6 &&
            (char_opp_dist() >= 29 || Opp.alive >= 0 || blink_timer == 16) &&
            (Char.frame < frame_151_strike_1 || Char.frame > frame_154_poking)) {
        if (Char.direction == dir_0_right && get_tile_behind_char() != tiles_20_wall && Char.curr_col != 0) {
            move_2_backward();
        } else if (Char.direction == dir_FF_left && get_tile_infrontof_char() != tiles_20_wall && Char.curr_col != 0) {
            move_1_forward();
        } else {
            inactivate_skel(dir_0_right);
            skeleton_timer = 1;
        };
        return;
    }
    // custom logic - skeleton on level 12 returns to its place after killing kid
    if (current_level == 12 &&
            drawn_room == 24 &&
            Char.curr_row == 1 &&
            Char.action <= actions_1_run_jump &&
            can_guard_see_kid == 0) {
        if (Char.direction == dir_FF_left && get_tile_behind_char() != tiles_20_wall && Char.curr_col != 7) {
            move_2_backward();
        } else if (Char.direction == dir_0_right && get_tile_infrontof_char() != tiles_20_wall && Char.curr_col != 7) {
            move_1_forward();
        } else {
            inactivate_skel(dir_FF_left);
        };
        return;
    }
    // custom logic - fallen down skeleton on level 12 returns to its place after killing kid
    if (current_level == 12 &&
            drawn_room == 4 &&
            Char.curr_row == 2 &&
            Char.action <= actions_1_run_jump &&
            can_guard_see_kid == 0) {
        if (Char.direction == dir_FF_left && get_tile_behind_char() != tiles_20_wall && Char.curr_col != 8) {
            move_2_backward();
        } else if (Char.direction == dir_0_right && get_tile_infrontof_char() != tiles_20_wall && Char.curr_col != 8) {
            move_1_forward();
        } else {
            inactivate_skel(dir_FF_left);
        };
        return;
    }
    Char.sword = sword_2_drawn;
    autocontrol_guard();
}

// seg002:085F
void autocontrol_kid(void) {
    // custom logic - do not put away sword during a dynamic guard fight when a guard is killed
    if (is_dynamic_guard_fight && Opp.alive >= 0) {
        return;
    }
    autocontrol_guard();
}

// custom logic - check if guard should turn around
bool check_turn_around(sbyte direction) {
    int tile_behind;
    if (can_guard_see_kid < 2 && droppedout) {
        return false;
    }
    tile_behind = get_tile_behind_char();
    // guard should not turn around on a single tile between a wall and an abiss
    if (wall_type2(tile_behind) == wall_type_4_wall && !tile_is_floor(get_tile_infrontof_char())) {
        return false;
    }
    switch (direction) {
        case dir_FF_left:
            return wall_type2(tile_behind) != wall_type_2_left &&
                   wall_type2(tile_behind) != wall_type_3_chomper &&
                   (wall_type2(get_tile_at_char()) != wall_type_1_right || curr_tile2 == tiles_4_gate);
        case dir_0_right:
            return wall_type2(tile_behind) != wall_type_1_right;
        default:
            return false;
    }
}

// custom logic - returns if guard is near a closing/closed gate
// -1 = the gate is on the right, 1 = gate is on the left
int is_near_gate(void) {
    if (Char.direction == dir_0_right &&
            (get_tile_at_char() == tiles_4_gate || get_tile_infrontof_char() == tiles_4_gate) &&
            curr_room_modif[curr_tilepos] < GUARD_HEIGHT_GATE_MODIFIER) {
        return -1;
    }
    if (Char.direction == dir_FF_left &&
            get_tile_infrontof_char() == tiles_4_gate &&
            curr_room_modif[curr_tilepos] < GUARD_HEIGHT_GATE_MODIFIER) {
        return 1;
    }
    return 0;
}

// seg002:0864
void autocontrol_guard(void) {
    // custom logic - do not control guard on prison level
    if (current_level == settings->prison_level) {
        return;
    }
    // custom logic - notify inactive guard
    if (is_guard_notice &&
            !is_inactive_guard_notice &&
            Kid.alive < 0 &&
            inactive_guard[drawn_room - 1].room == drawn_room &&
            (can_guard_see_kid != 2 || can_inactive_guard_see_kid)) {
        is_inactive_guard_notice = 1;
    }
    // custom logic - dynamic fight status
    byte dynamic_count = dynamic_guard_count[drawn_room - 1];
    if (is_dynamic_guard_fight == 1 &&
            can_guard_see_kid < 2 &&
            ((dynamic_count == 1 && dynamic_guard.frame == 0) || dynamic_count > 1)) {
        is_dynamic_guard_fight = 2;
    } else if (is_dynamic_guard_fight == 2 && can_guard_see_kid == 2) {
        is_dynamic_guard_fight = 1;
    }
    if (is_dynamic_guard_fight == 1 &&
            can_guard_see_kid < 2 &&
            inactive_guard[drawn_room - 1].room == drawn_room &&
            can_inactive_guard_see_kid < 2 &&
            !droppedout) {
        is_dynamic_guard_fight = 2;
    }
    // custom logic - dynamic guards enter the room
    if (dynamic_guard_timer > 0 &&
            Char.charid == charid_2_guard &&
            Char.alive < 0 &&
            Char.direction != dir_56_none &&
            guard_kid_dist() >= 50 &&
            (tile_is_floor(get_tile_infrontof_char()) ||
             tile_is_floor(get_tile_infrontof2_char())) &&
            curr_tile2 != tiles_18_chomper) {
        if (Char.frame == frame_166_stand_inactive) {
            move_down_forw();
        }
        move_1_forward();
        return;
    }
    // custom logic - guards do not react to shadow except for skeletons, Jaffar and his vision
    // after kid unites with shadow, birdmen bow to him
    int distance = char_opp_dist();
    if (Char.charid != charid_0_kid &&
            Char.charid != charid_4_skeleton &&
            current_level == 10 &&
            leveldoor_open >= DEAD_KID_LEVEL_10_TIMER &&
            (distance >= 0 || is_guard_notice)) {
        if (can_guard_see_kid && Char.frame < frame_190_start_bowing_1 && Char.alive < 0 && guardhp_curr > 0) {
            bool flip = false;
            if (Char.x < Opp.x) {
                flip = Char.direction != dir_0_right;
            } else {
                flip = Char.direction == dir_0_right;
            }
            if (!flip && !tile_is_floor(get_tile_infrontof_char())) {
                Char.x = char_dx_forward(-15 + distance_to_edge_weight());
            }
            draw_guard_hp(0, MAX_GUARD_HP);
            Char.alive = 0;
            Char.action = 0;
            seqtbl_offset_char(flip ? seq_124_bird_turn_bow : seq_123_bird_bow);
            play_seq();
        } else if (is_kid_shadow && Char.frame == frame_177_spiked) {
            enter_guard();
        } else if (Char.frame != frame_166_stand_inactive) {
            move_down_back();
        } else if (distance < 0 &&
                   (word)distance < (word) - 4 &&
                   !is_kid_shadow) {
            move_4_down();
        }
        return;
    }
    if (Char.sword < sword_2_drawn) {
        autocontrol_guard_inactive();
    } else {
        autocontrol_guard_active();
    }
}

// seg002:0876
void autocontrol_guard_inactive(void) {
    short distance, turn_delta_x;
    bool is_right_behind;
    // custom logic - do not turn around when falling from above or when kid is dead
    if (Opp.alive >= 0) return;
    if (Char.curr_row == -1) return;
    // custom logic - the guard on level 7, room 7 should not react to kid
    if (current_level == 7 &&
            Char.alive < 0 &&
            Char.direction == dir_FF_left &&
            Opp.room == 7 &&
            Opp.curr_row > 0 &&
            Opp.curr_col <= 3 &&
            Char.x <= 131) {
        if (Char.x - Opp.x >= 35) {
            Char.x = 124;
        }
        if (can_guard_see_kid > 0) {
            move_down_forw();
        }
        return;
    }
    distance = char_opp_dist();
    turn_delta_x = -7;
    // custom logic - avoid guard being idle at turn_delta_x = -9 when kid is climbing
    if ((Opp.frame >= frame_141_climbing_7 && Opp.frame <= frame_148_climbing_14) || Opp.action == actions_2_hang_climb) {
        turn_delta_x -= 2;
    }
    if (Opp.curr_row != Char.curr_row || (word)distance < (word)turn_delta_x) {
        // custom logic - guard will turn around if kid is right behind him
        is_right_behind = (can_guard_see_kid >= 2 &&
                           distance < turn_delta_x && distance >= -25 &&
                           (Opp.frame < frame_39_start_run_jump_6 || Opp.frame > frame_41_running_jump_2) &&
                           (Opp.frame <= frame_80_jumphang ||
                            (Opp.frame >= frame_91_hanging_5 && Opp.frame <= frame_93_hanging_7) ||
                            (Opp.frame >= frame_109_crouch && Opp.frame <= frame_149_climbing_15)));
        // If Kid made a sound ...
        if (is_guard_notice || is_right_behind) {
            is_guard_notice = 0;
            if (distance < 0) {
                // custom logic - gate guarding guard on level 7 should not change directions
                if (current_level == 7 &&
                        Opp.room == 7 &&
                        Opp.curr_row == 0 &&
                        Opp.curr_col > 3) {
                    return;
                }
                // custom logic - gate guarding guard on level 11 should not change directions
                if (current_level == 11 &&
                        Char.direction == dir_0_right &&
                        Char.room == 13 &&
                        Char.curr_row == 1 &&
                        Char.curr_col <= 2) {
                    return;
                }
                // custom logic - a counter that prevents guard from turning around every frame
                if (should_turn_around > 0) {
                    --should_turn_around;
                }
                // ... and Kid is behind Guard, Guard turns around.
                // custom logic - guard should not turn around if there is a wall or a doortop behind
                // he would not be able to see anything and can fall out through the tile
                if ((word)distance < (word) - 4 &&
                        should_turn_around == 0 &&
                        (check_turn_around(Char.direction) ||
                         (can_guard_see_kid == 2 && Char.direction == Opp.direction))) {
                    should_turn_around = 6;
                    move_4_down();
                }
                // custom logic - try going into engarde mode when kid is climbing down
                if (Opp.frame < frame_136_climbing_2 || Opp.frame > frame_138_climbing_4) {
                    return;
                }
            }
        } else if (distance < 0) {
            // custom logic - guard should be alert if kid landed right behind him
            if (can_guard_see_kid == 2 &&
                    Char.action == actions_1_run_jump &&
                    (Char.direction != Opp.direction || is_right_behind) &&
                    ((Opp.frame >= frame_40_running_jump_1 && Opp.frame <= frame_44_running_jump_5) ||
                     (Opp.frame >= frame_107_fall_land_1 && Opp.frame <= frame_108_fall_land_2))
               )  {
                if (tile_is_floor(get_tile_infrontof_char()) || distance_to_edge_weight() > 8) {
                    Char.x = char_dx_forward(8);
                }
                move_4_down();
            }
            // custom logic - a counter that prevents guard from turning around every frame
            if (should_turn_around > 0) {
                --should_turn_around;
            }
            if (droppedout &&
                    Opp.curr_row == Char.curr_row + 1 &&
                    Opp.action <= actions_1_run_jump &&
                    (Opp.frame < frame_135_climbing_1 || Opp.frame > frame_149_climbing_15)) {
                move_down_forw();
            }
            return;
        }
    }
    should_turn_around = 0;
    check_can_guard_see_kid();
    // custom logic - gate guarding guards should be active
    if (can_guard_see_kid >= (is_guarding_gate ||
                              (distance <= 42 && Opp.direction != Char.direction && Opp.sword == sword_2_drawn && !is_near_gate()) ? 1 : 2)
       ) {
        // If Guard can see Kid, Guard moves to fighting pose.
        if (current_level != settings->win_level || jaffar_delay_timer == 0 || distance <= 42) {
            // custom logic - do not turn around into an abyss
            if (!tile_is_floor(get_tile_infrontof_char()) && distance < 12) {
                Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
            }
            move_down_forw();
        }
        // custom logic - approach kid who is falling/jumping in front of the guard
    } else if (can_guard_see_kid &&
               distance > 0 &&
               !tile_is_floor(get_tile_at_kid(Opp.x)) &&
               Opp.action == actions_4_in_freefall) {
        loadkid_and_opp();
        set_kid_droppedout();
        loadshad_and_opp();
    }
    // custom logic - guard is active if kid dropped out
    if (droppedout && can_guard_see_kid != 1 && distance > 0 && Opp.curr_row >= Char.curr_row) {
        // custom logic - check if kid can land on a tile right below or grab a tile in front of him
        if (Opp.curr_row == Char.curr_row + 1 &&
                Opp.action >= actions_3_in_midair &&
                Opp.action <= actions_4_in_freefall &&
                (!tile_is_floor(get_tile(Opp.room, Opp.curr_col, Opp.curr_row)) ||
                 curr_tile2 == tiles_2_spike ||
                 curr_tile2 == tiles_11_loose)
           ) {
            // climbing down 1 floor
            if (curr_tile2 == tiles_20_wall &&
                    Opp.frame >= frame_136_climbing_2 && Opp.frame <= frame_140_climbing_6 &&
                    tile_is_floor(get_tile(Opp.room, Opp.curr_col + dir_behind[Char.direction + 1], Opp.curr_row)) &&
                    curr_tile2 != tiles_2_spike &&
                    curr_tile2 != tiles_11_loose) {
                move_down_forw();
                return;
            }
            bool can_kid_grab;
            int max_edge_distance;
            loadkid_and_opp();
            can_kid_grab = can_grab_front_above(); // tile right in front
            max_edge_distance = Char.fall_x == 2 ? 8 : 5;
            if (!can_kid_grab && !tile_is_floor(curr_tile2) && curr_tile2 != tiles_20_wall && distance_to_edge_weight() <= max_edge_distance) { // can grab the next tile
                int incr = dir_front[Char.direction + 1];
                Char.curr_col += incr;
                can_kid_grab = can_grab_front_above();
                Char.curr_col -= incr;
            }
            loadshad_and_opp();
            if (!can_kid_grab ||
                    !tile_is_floor(get_tile(Char.room, get_tile_div_mod_m7(Char.x + dir_front[Char.direction + 1] * 7), Char.curr_row))) {
                droppedout = 0;
                return;
            }
        }
        // do not walk/turn around into an abyss
        if (!tile_is_floor(get_tile_infrontof_char()) && distance < 12) {
            Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
        }
        move_down_forw();
        if (!can_guard_see_kid) {
            guard_follows_kid_down();
        }
        // custom logic - guard reacts to spikes on the row below in front of him
    } else if (!droppedout &&
               !can_guard_see_kid &&
               distance > 0 &&
               !is_temple() &&
               check_sound_playing() &&
               current_sound == sound_49_spikes_out &&
               Opp.curr_row == Char.curr_row + 1 &&
               trob.type != -1 &&
               trob.room == Opp.room &&
               trob.tilepos / 10 == Opp.curr_row &&
               Char.x == level.guards_x[Char.room - 1] &&
               Char.direction == level.guards_dir[Char.room - 1] &&
               Char.curr_row == level.guards_tile[Char.room - 1] / 10 &&
               wall_type2(get_tile_at_char()) == 0 &&
               tile_is_floor(get_tile_infrontof_char()) &&
               wall_type2(curr_tile2) == 0 &&
               curr_tile2 != tiles_11_loose &&
               tile_is_floor(get_tile_infrontof2_char()) &&
               wall_type2(curr_tile2) == 0) {
        loadkid_and_opp();
        can_guard_see_kid = 1;
        set_kid_droppedout();
        can_guard_see_kid = 0;
        loadshad_and_opp();
        if (droppedout) {
            move_down_forw();
        }
    }
}

// seg002:08DC
void autocontrol_guard_active(void) {
    short opp_frame;
    short char_frame;
    short distance;
    char_frame = Char.frame;
    is_guard_notice = 0;
    // custom logic - the guard on level 1, room 18 should not react to kid on upper floors
    if (current_level == 1 && guard_skill == GATE_GUARDING_SKILL && Opp.room == 18 && Opp.curr_row < 2) {
        return;
        // custom logic - the guard on level 7, room 7 should not react to kid
    } else if (current_level == 7 &&
               Char.alive < 0 &&
               Opp.room == 7 &&
               Opp.curr_row > 0 &&
               Opp.curr_col <= 3 &&
               curr_room_modif[20] != 0xFF && // neither gate is stuck open
               curr_room_modif[22] != 0xFF &&
               Char.direction == dir_FF_left &&
               Char.x <= 131 &&
               !can_inactive_guard_see_kid) {
        // set guard on the closing tile when not within fighting distance
        if (Char.x - Opp.x >= 35) {
            Char.x = 124;
        }
        if (can_guard_see_kid == 0) {
            move_down_back();
        }
        return;
        // custom logic - guard on level 11, room 13 should not react to kid on row 1
    } else if (current_level == 11 &&
               Char.curr_col <= 2 &&
               Char.direction == dir_0_right &&
               Opp.room == 13 &&
               Opp.curr_row == 1) {
        if (can_guard_see_kid == 0) {
            move_down_back();
        }
        return;
    } else if (current_level == 12 &&
               Char.room == 13 &&
               Char.curr_row == 0 &&
               falling_bridge_timer == 0) {
        falling_bridge_timer = settings->falling_bridge_delay;
    }
    // custom logic - skeletons have sword drawn when they are inactive and should be activated when they can see alive kid
    if (Char.charid == charid_4_skeleton && char_frame == frame_166_stand_inactive && can_guard_see_kid == 2 && Opp.alive < 0) {
        seqtbl_offset_char(seq_90_guard_stand_active);
        play_seq();
        return;
    }
    if (char_frame != frame_166_stand_inactive && char_frame >= frame_150_parry && can_guard_see_kid != 1) {
        distance = char_opp_dist();
        if (can_guard_see_kid == 0) {
            should_turn_around = 0;
            if (droppedout != 0) {
                // custom logic - turn around when kid is falling out behind guard
                if (Char.direction == Opp.direction &&
                        Char.sword == sword_2_drawn &&
                        Char.curr_row == 2 &&
                        Opp.curr_row == 3 &&
                        distance < 0) {
                    seqtbl_offset_char(seq_60_turn_with_sword);
                    play_seq();
                } else {
                    guard_follows_kid_down();
                }
            } else if (Char.charid != charid_4_skeleton || is_dungeon_2()) { // custom logic - dungeon 2 skeletons can put sword down
                move_down_back();
            }
        } else { // can_guard_see_kid == 2
            offguard = 0;
            opp_frame = Opp.frame;
            if (distance >= 12 &&
                    // frames 102..117: falling and landing
                    opp_frame >= frame_102_start_fall_1 && opp_frame < frame_109_crouch &&
                    Opp.action == actions_5_bumped
               ) {
                should_turn_around = 0;
                return;
            }
            if (distance < 35) {
                if ((Char.sword < sword_2_drawn && distance < 8) || distance < 12) {
                    if (Char.direction == Opp.direction) {
                        // custom logic - do not get stuck behind a wall or fall into an abyss
                        int wall_type_behind = wall_type2(get_tile_behind_char());
                        if (!tile_is_floor(curr_tile2) ||
                                (Char.direction == dir_0_right && wall_type_behind == wall_type_1_right) ||
                                (Char.direction == dir_FF_left &&
                                 (wall_type_behind == wall_type_2_left ||
                                  wall_type_behind == wall_type_3_chomper ||
                                  wall_type2(get_tile_at_char()) == wall_type_1_right))) {
                            if (distance_to_edge_weight() >= 8) {
                                Char.x = char_dx_forward(4);
                                return;
                            }
                        }
                        // turn around
                        ++should_turn_around; // custom logic - a counter that prevents kid and guard moonwalking
                        move_2_backward();
                    } else {
                        should_turn_around = 0;
                        move_1_forward();
                    }
                } else {
                    should_turn_around = 0;
                    autocontrol_guard_kid_in_sight(distance);
                }
            } else {
                should_turn_around = 0;
                if (guard_refrac != 0) return;
                if (Char.direction != Opp.direction &&
                        // custom logic - guards that have an advance probability should be moving towards running/jumping
                        // kid if they can hurt him
                        (can_guard_see_kid != 2 || settings->advprob[guard_skill] == 0)) {
                    // frames 7..14: running
                    // frames 34..43: run-jump
                    if (opp_frame >= frame_1_start_run && opp_frame <= frame_14_run) {
                        return;
                    } else if (opp_frame >= frame_34_start_run_jump_1 && opp_frame <= frame_44_running_jump_5) {
                        return;
                    }
                }
                autocontrol_guard_kid_far();
            }
        }
        // custom logic - additional guard logic
    } else if (can_guard_see_kid == 1 &&
               !is_guarding_gate &&
               dynamic_guard_timer == 0 &&
               Char.charid == charid_2_guard &&
               Char.action != actions_5_bumped &&
               Opp.action <= actions_1_run_jump) {
        should_turn_around = 0;
        determine_col();
        int near_gate = is_near_gate();
        // do not walk through closed gates (facing right)
        if (near_gate == -1 && Char.direction == ~Opp.direction && Char.sword == sword_2_drawn) {
            // custom logic - recalculate offscreen gate column
            int gate_col = get_tile_col_at_char();
            // check if guard is standing in the air on the right side of the gate
            if (!tile_is_floor(get_tile_infrontof_char())) {
                Char.x = char_dx_forward(7);
                return;
            }
            // guard on the first level should move closer to the gate when kid does not have a sword
            if (gate_col <= Char.curr_col || (!have_sword && Opp.curr_col > gate_col)) {
                Char.x = x_bump[gate_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
                Char.x = char_dx_forward(7);
                resurrect_time = -4;
            }
            guard_refrac = 6;
            move_down_back();
            // do not walk through closed gates (facing left)
        } else if (near_gate == 1 && Char.direction == ~Opp.direction && Char.sword == sword_2_drawn) {
            // custom logic - recalculate offscreen gate column
            int gate_col = get_tile_col_at_char() + 1;
            Char.x = x_bump[gate_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
            guard_refrac = 6;
            move_down_back();
        } else if ((Char.curr_col < 0 || Char.curr_col > 9) &&
                   !near_gate &&
                   (Opp.frame < frame_151_strike_1 || Opp.frame > frame_157_walk_with_sword) &&
                   tile_is_floor(get_tile_infrontof_char()) &&
                   curr_tile2 != tiles_11_loose &&
                   tile_is_floor(get_tile_infrontof2_char()) &&
                   curr_tile2 != tiles_11_loose) {
            move_1_forward(); // do not stop in between rooms
        } else if (tile_is_floor(get_tile_at_char()) &&
                   curr_tile2 != tiles_11_loose &&
                   curr_tile2 != tiles_18_chomper &&
                   // check_chomped_guard() is looking at the current and the next column
                   get_tile(Char.room, Char.curr_col + 1, Char.curr_row) != tiles_18_chomper) {
            if (char_opp_dist() > 42 || Opp.sword == sword_0_sheathed) {
                move_down_back(); // put sword down when "can_guard_see_kid" is 1
            }
        } else if (curr_tile2 == tiles_18_chomper || (curr_tile2 == tiles_11_loose && curr_room == Char.room)) {
            if (curr_tile2 != tiles_18_chomper ||
                    distance_to_edge_weight() < 8 ||
                    wall_type(get_tile_behind_char()) != wall_type_0_none || // checks for the tile behind
                    !tile_is_floor(curr_tile2) ||
                    curr_tile2 == tiles_18_chomper) {
                move_1_forward();
            } else {
                move_2_backward();
            }
        }
    } else {
        should_turn_around = 0;
    }
}

// seg002:09CB
void autocontrol_guard_kid_far(void) {
    if (tile_is_floor(get_tile_infrontof_char()) ||
            tile_is_floor(get_tile_infrontof2_char())) {
        move_1_forward();
    } else {
        move_2_backward();
    }
}

// seg002:09F8
void guard_follows_kid_down(void) {
    // This is called from autocontrol_guard_active, so char=Guard, Opp=Kid
    int wall_infrontof;
    if (// custom logic - guard might be standing "on air" in the room above
        tile_is_floor(get_tile_at_char()) &&
        // there is wall in front of Guard
        ((wall_infrontof = wall_type2(get_tile_infrontof_char())) != wall_type_0_none ||
         curr_tile2 == tiles_11_loose ||
         (!tile_is_floor(curr_tile2) && (
              (get_tile(curr_room, tile_col, ++tile_row) == tiles_2_spike ||
               // Guard would fall on loose floor
               curr_tile2 == tiles_11_loose ||
               // ... or wall (?)
               wall_type2(curr_tile2) != wall_type_0_none ||
               // ... or into a chasm
               ! tile_is_floor(curr_tile2)) ||
              // ... or Kid is not below
              Char.curr_row + 1 != Opp.curr_row
          )))
    ) {
        // custom logic - come closer to a chomper
        if (Char.direction == dir_FF_left && wall_infrontof == wall_type_3_chomper && distance_to_edge_weight() >= 7) {
            move_1_forward();
            return;
        }
        // don't follow
        droppedout = 0;
        move_2_backward();
        // custom logic - do not follow falling or dead kid
    } else if (Opp.alive < 0 &&
               (tile_is_floor(
                    get_tile(Char.room, get_tile_div_mod_m7(Char.x) + dir_front[Char.direction + 1], Char.curr_row)
                ) ||
                guard_dropping_down ||
                (Opp.action != actions_2_hang_climb &&
                 Opp.action != actions_3_in_midair &&
                 Opp.action != actions_4_in_freefall &&
                 Opp.action != actions_6_hang_straight &&
                 (Opp.frame < frame_135_climbing_1 || Opp.frame > frame_140_climbing_6) &&
                 Opp.curr_row > Char.curr_row))) {
        // follow
        move_1_forward();
    }
}

// seg002:0A93
void autocontrol_guard_kid_in_sight(short distance) {
    int edge_distance;
    if (Opp.sword == sword_2_drawn) {
        autocontrol_guard_kid_armed(distance);
    } else if (guard_refrac == 0 && resurrect_time == 0) {
        if (distance < 29) {
            // custom logic - do not fall off trying to strike kid in the back
            edge_distance = distance_to_edge_weight();
            if (tile_is_empty(get_tile_behind_char()) &&
                    Char.direction == Opp.direction &&
                    edge_distance > 0 &&
                    edge_distance < 6) {
                Char.x = char_dx_forward(edge_distance);
            }
            move_6_shift();
            // custom logic - kid's action can change without registering a hit
            if (Opp.charid == charid_0_kid) {
                check_hurting();
                saveshad_and_opp();
            }
        } else {
            // custom logic - do not forward into an abyss when kid is turning and running over it
            if (Opp.frame >= frame_53_runturn && Opp.frame <= frame_65_runturn &&
                    !tile_is_floor(get_tile_at_kid(Char.x + dir_front[Char.direction + 1] * 7))) {
                return;
            }
            move_1_forward();
        }
    }
}

// seg002:0AC1
void autocontrol_guard_kid_armed(short distance) {
    // custom logic - do not advance if there is an empty tile in front of kid
    if (!tile_is_floor(get_tile(Opp.room, Opp.curr_col, Opp.curr_row))) {
        if (distance >= 10 && distance < 29) {
            guard_block();
        }
        return;
    }
    if (distance < 10 || distance >= 29) {
        guard_advance();
    } else {
        guard_block();
        if (guard_refrac == 0 && resurrect_time == 0) {
            if (distance < 12 || distance >= 29) {
                guard_advance();
            } else {
                guard_strike();
            }
        }
    }
}

// custom logic - returns whether the guard is the skeleton on the falling bridge on level 12
bool is_falling_bridge_guard(void) {
    return current_level == 12 && Char.charid == charid_4_skeleton && guard_skill == 7;
}

// seg002:0AF5
void guard_advance(void) {
    word adv_prob;
    adv_prob = settings->advprob[guard_skill];
    if (guard_skill == 0 || kid_sword_strike == 0) {
        // custom logic - skeleton at the top of the bridge with the skill 7
        // should advance when kid is on the left side
        if (is_falling_bridge_guard() && Char.direction == dir_FF_left) {
            adv_prob = 200;
        }
        if (adv_prob > prandom(255)) {
            move_1_forward();
        }
    }
}

// seg002:0B1D
void guard_block(void) {
    word opp_frame;
    opp_frame = Opp.frame;
    if (opp_frame == frame_152_strike_2 || opp_frame == frame_153_strike_3 || opp_frame == frame_162_block_to_strike) {
        if (justblocked != 0) {
            if (settings->impblockprob[guard_skill] > prandom(255)) {
                move_3_up();
            }
        } else {
            if (settings->blockprob[guard_skill] > prandom(255)) {
                move_3_up();
            }
        }
    }
}

// seg002:0B73
void guard_strike(void) {
    word opp_frame;
    word char_frame;
    opp_frame = Opp.frame;
    if (opp_frame == frame_169_begin_block || opp_frame == frame_151_strike_1) return;
    char_frame = Char.frame;
    if (char_frame == frame_161_parry || char_frame == frame_150_parry) {
        if (settings->restrikeprob[guard_skill] > prandom(255)) {
            move_6_shift();
        }
    } else {
        word strikeprob = settings->strikeprob[guard_skill];
        // custom logic - guards with skills 0 and 1 have a lower percentage of
        // restriking in a fist fight
        if (!have_sword && guard_skill <= 1) {
            strikeprob /= 2;
        }
        if (strikeprob > prandom(255)) {
            move_6_shift();
        }
    }
}

// custom logic - checks if there a guard in a dead slot nearby
bool is_dead_guard_near(char_type2 dead) {
    return dead.room == Char.room &&
           dead.row == Char.curr_row &&
           dead.frame == frame_185_dead &&
           ABS(Char.x - dead.x) < 42;
}

// custom logic - sets an alternative dead frame for a regular guard if there are dead guards nearby
void set_alt_dead_frame(void) {
    bool near_dead1;
    bool near_dead2;
    bool near_dead3;
    word room = Char.room;
    word room_minus_1 = room - 1;
    char_type2 dead1 = dead_guard1[room_minus_1];
    char_type2 dead2 = dead_guard2[room_minus_1];
    char_type2 dead3 = dead_guard3[room_minus_1];
    // only regular and bird guards have an alternate death frame
    if (Char.charid != charid_2_guard ||
            (settings->tbl_guard_type[current_level] != guard_type_guard &&
             settings->tbl_guard_type[current_level] != guard_type_bird)) {
        return;
    }
    // keep active dead guard's frame when no dead guard slots are available
    // as the "level" variable has no means of saving the alternate frame
    if (dead1.room == room && dead2.room == room && dead3.room == room) {
        return;
    }
    near_dead1 = is_dead_guard_near(dead1);
    near_dead2 = is_dead_guard_near(dead2);
    near_dead3 = is_dead_guard_near(dead3);
    // do not put 2 guards with alternate dead frame images next to each other
    if ((near_dead1 && dead1.alt_frame != 0xFF && (!near_dead2 || dead2.alt_frame != 0xFF) && (!near_dead3 || dead3.alt_frame != 0xFF)) ||
            (near_dead2 && dead2.alt_frame != 0xFF && (!near_dead1 || dead1.alt_frame != 0xFF) && (!near_dead3 || dead3.alt_frame != 0xFF)) ||
            (near_dead3 && dead3.alt_frame != 0xFF && (!near_dead1 || dead1.alt_frame != 0xFF) && (!near_dead2 || dead2.alt_frame != 0xFF))) {
        Char.repeat = 0xFF;
    }
}

// seg002:0BCD
void hurt_by_sword(void) {
    short distance;
    int wall_col;
    if (Char.alive >= 0) {
        if (Char.charid == charid_2_guard && Char.frame >= frame_190_start_bowing_1 && Char.frame <= frame_202_bowing) {
            Char.direction = ~Kid.direction;
            take_hp(ALL_HP);
            seqtbl_offset_char(seq_85_stabbed_to_death);
            saveshad();
            play_sound(sound_12_guard_hurt);
            play_seq();
        }
        return;
    }
    // custom logic - You can't hurt mouse
    if (Char.charid == charid_24_mouse) return;
    if (Char.sword != sword_2_drawn) {
        // custom logic - bump kid/guard out of the wall
        if (get_tile_at_char() == tiles_20_wall) {
            in_wall();
        }
        // custom logic - kid should not die if he is facing an armed guard with the sword
        // and the sword is not drawn
        if (Char.charid == charid_0_kid &&
                ((!is_kid_shadow && have_sword && current_level != settings->flame_level) || (is_kid_shadow && is_shadow_flame)) &&
                guard_refrac == 0 &&
                Char.fall_y == 0 &&
                Char.action != actions_2_hang_climb &&
                Char.action != actions_3_in_midair &&
                Char.action != actions_4_in_freefall &&
                Char.action != actions_6_hang_straight &&
                (Char.frame <= frame_56_runturn ||
                 (Char.frame >= frame_121_stepping_1 && Char.frame <= frame_132_stepping_12)) &&
                Char.direction != Opp.direction &&
                char_opp_dist() >= 0) {
            // tall and fat guard takes takes 2 hit points
            if (take_hp(current_level == settings->fat_guard_level && guard_skill == settings->fat_guard_skill ? 2 : 1)) {
                seqtbl_offset_char(seq_85_stabbed_to_death);
            } else {
                if (!tile_is_floor(get_tile_at_char())) {
                    Char.x = char_dx_forward(distance_to_edge_weight() - 1);
                }
                draw_sword();
                goto loc_hit;
            }
        } else {
            // guard cannot hurt kid through a closed gate
            if (Char.charid == charid_0_kid && can_guard_see_kid == 1 && Guard.direction == dir_0_right && get_tile_at_kid(Guard.x) == tiles_4_gate) {
                Guard.x -= 10;
                return;
            }
            // kid cannot kill a jumping guard with a fist
            if (Char.charid != charid_0_kid && !have_sword && guardhp_curr > 1) {
                Char.sword = sword_2_drawn;
                goto loc_hit;
            }
            // custom logic - do not re-animate death if the character has been chomped
            if (Char.frame == frame_178_chomped) {
                seqtbl_offset_char(seq_54_chomped);
                play_seq();
                return;
            }
            // Being hurt when not in fighting pose means death.
            take_hp(ALL_HP);
            seqtbl_offset_char(seq_85_stabbed_to_death); // dying (stabbed unarmed)
            // custom logic - kid is falling close to the edge
            distance = distance_to_edge_weight();
            if (Char.charid == charid_0_kid &&
                    ((Char.frame >= frame_102_start_fall_1 && Char.frame <= frame_105_start_fall_4) ||
                     (Char.frame >= frame_141_climbing_7 && Char.frame <= frame_148_climbing_14))
               ) {
                if (tile_is_floor(get_tile_at_char())) {
                    Char.x = char_dx_forward(-(14 - distance) - 5);
                } else if (distance < 5) {
                    Char.x = char_dx_forward(distance - 5);
                }
            } else if (Char.charid == charid_0_kid &&
                       // The current Char sequence is "seq_85_stabbed_to_death" but the Kid sequence has not been updated yet
                       Kid.curr_seq >= seqtbl_offsets[seq_6_run_turn] &&
                       Kid.curr_seq < seqtbl_offsets[seq_81_kid_pushed_off_ledge] &&
                       Char.frame >= frame_56_runturn &&
                       Char.frame <= frame_65_runturn) {
                // custom logic - kid gets striked in the back while running and turning
                Char.direction = ~Char.direction;
                Char.x = char_dx_forward(7);
            }
            // custom logic - use an alternate dead frame image for a regular or bird guard when needed
            set_alt_dead_frame();
        }
loc_dead_from_hit:
        distance = distance_to_edge_weight();
        // custom logic - guards should not fall through tiles that act as walls
        bool near_wall = false;
        for (wall_col = 0; wall_col <= (Char.direction < dir_0_right ? 2 : 1); ++wall_col) {
            if (get_tile(Char.room, dir_behind[Char.direction + 1] * wall_col + Char.curr_col, Char.curr_row) == tiles_20_wall ||
                    curr_tile2 == tiles_7_doortop_with_floor ||
                    curr_tile2 == tiles_12_doortop ||
                    (curr_tile2 == tiles_4_gate && can_bump_into_gate())) {
                near_wall = true;
                break;
            }
        }
        bool tile_behind_empty = tile_is_empty(get_tile_behind_char());
        // custom logic - makes pushing a guard off a ledge easier with 1 hitpoint
        bool tile_behind_2_empty = tile_is_empty(get_tile(Char.room, get_tile_div_mod_m7(char_dx_forward(distance - 16)) + dir_behind[Char.direction + 1], Char.curr_row)) &&
                                   curr_room == drawn_room;
        if ((!tile_behind_empty && !tile_behind_2_empty) ||
                Char.sword == sword_0_sheathed ||
                near_wall ||
                !have_sword || // custom logic - do not push guards far with a fist
                // custom logic - Jaffar's vision cannot be thrown off a ledge
                (Char.charid != charid_0_kid && settings->tbl_guard_type[current_level] == guard_type_Jaffar_vision)) {
            seqtbl_offset_char(seq_85_stabbed_to_death); // dying (stabbed)
            // custom logic - use an alternate dead frame image for a r=egular or bird guard when needed
            set_alt_dead_frame();
            if (Char.direction < dir_0_right && // looking left
                    (get_tile_at_char() == tiles_4_gate || get_tile_behind_char() == tiles_4_gate) &&
                    can_bump_into_gate()
               ) {
                // custom logic - recalculate offscreen gate column
                int gate_col = get_tile_col_at_char();
                Char.x = x_bump[gate_col - (curr_tile2 != tiles_4_gate) + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
                Char.x = char_dx_forward(10);
            }
            Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
            Char.fall_y = 0;
        } else {
            if (!tile_behind_empty && tile_behind_2_empty) {
                distance += 4;
            }
            Char.x = char_dx_forward(distance);
            Char.frame = frame_158_stand_with_sword; // custom logic - prevents guard falling too far behind due to dx_weight()
            load_fram_det_col();
            inc_curr_row();
            if (Char.charid == charid_2_guard && guardhp_curr + guardhp_delta == 0) {
                set_alt_dead_frame();
            }
            seqtbl_offset_char(seq_158_guard_pushed_far); // Kid/Guard is killed and pushed off the ledge
        }
    } else {
        // custom logic - kid gets killed trying to stab Jaffar
        if (current_level == settings->flame_level &&
                Char.charid == charid_2_guard &&
                !is_kid_shadow) {
            loadkid();
            release_arrows();
            keep_sword_drawn = false;
            holding_sword = 0;
            Char.sword = sword_0_sheathed;
            seqtbl_offset_char(seq_93_put_sword_away_fast);
            play_sound(sound_11_sword_moving);
            savekid();
            loadshad();
            seqtbl_offset_char(seq_90_guard_stand_active);
            play_seq();
            flash_color = color_12_brightred;
            flash_time = 12;
            return;
        }
        // custom logic - reset Jaffar's complex parry count
        if (jaffar_parry_count > 0) {
            jaffar_parry_count = 0;
        }
        // You can't hurt skeletons
        if (Char.charid != charid_4_skeleton) {
            int hp = 1;
            // custom logic - Jaffar kills kid, not turned into shadow, with 1 strike
            if (current_level == settings->flame_level && Char.charid == charid_0_kid && !is_kid_shadow) {
                hp = ALL_HP;
                // custom logic - tall and fat guard on level 8 produces double damage before farting
            } else if (current_level == settings->fat_guard_level &&
                       guard_skill == settings->fat_guard_skill &&
                       Char.charid == charid_0_kid &&
                       Opp.color == 0) {
                ++hp;
            }
            // custom logic - it takes 3 fist punches to hurt a guard
            if (!have_sword && Char.charid != charid_0_kid && ++guard_punch_count < 3) {
                goto loc_hit;
            }
            if (take_hp(hp)) {
                goto loc_dead_from_hit;
            }
        }
loc_hit:
        // custom logic - do not fall through the wall after a strike
        Char.frame = frame_158_stand_with_sword; // custom logic - prevents guard falling too far behind due to dx_weight()
        in_wall_engarde();
        seqtbl_offset_char(seq_74_hit_by_sword); // being hit with sword
        Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
        Char.fall_y = 0;
    }
    // sound 13: Kid hurt (by sword), sound 12: Guard hurt (by sword)
    switch (Char.charid) {
        case charid_0_kid:
            play_sound(sound_34_hit_by_sword);
            if (!is_kid_shadow) {
                delayed_sound_timer = 1;
                delayed_sound = (struct delayed_sound_type) {
                    sound_13_kid_hurt, 0, 0, 0, 0, 0, 0
                };
            }
            // custom logic - blink when flame is hit with sword
            if (is_kid_shadow &&
                    is_shadow_flame &&
                    hitp_curr > 1) {
                lighting_timer = 4;
            }
            break;
        case charid_4_skeleton:
            // custom logic - skeletons have a different hurt sound
            play_sound(sound_34_hit_by_sword);
            break;
        default:
            // custom logic - punch without a sword
            if (!have_sword && guard_punch_count != 3) {
                seqtbl_offset_char(seq_157_fist_fight_hit);
                play_sound(sound_34_hit_by_sword);
                // custom logic - Jaffar's vision does not get hurt
            } else if (settings->tbl_guard_type[current_level] == guard_type_Jaffar_vision) {
                play_sound(sound_57_sword_striked);
                // custom logic - flame hurts Jaffar
            } else if (is_kid_shadow) {
                play_sound(is_shadow_flame ? sound_31_flame_hurts_Jaffar : sound_12_guard_hurt);
                // custom logic - fat guard on level 8 farts and lose his skill
            } else if (current_level == settings->fat_guard_level &&
                       guard_skill == settings->fat_guard_skill &&
                       Char.alive == -1 &&
                       Char.color == 0 &&
                       guardhp_curr == 3) {
                Char.x = char_dx_forward(-5);
                in_wall_engarde();
                guard_skill = 0;
                play_sound(sound_34_hit_by_sword);
                delayed_sound_timer = 1;
                delayed_sound = (struct delayed_sound_type) {
                    sound_16_fat_fart, 0, 0, 0, 0, 0, 0
                };
            } else {
                play_sound(sound_34_hit_by_sword);
                delayed_sound_timer = 1;
                delayed_sound = (struct delayed_sound_type) {
                    sound_12_guard_hurt, 0, 0, 0, 0, 0, 0
                };
            }
            break;
    }
    play_seq();
    // custom logic - holding sword should be true when hit by sword
    if (Char.charid == charid_0_kid && hitp_curr + hitp_delta == 0) {
        holding_sword = 1;
    }
    // custom logic - Jaffar takes kid's lives up until he has 10 hit points
    if (current_level == settings->win_level && Char.charid == charid_0_kid && guardhp_curr < guardhp_max) {
        ++guardhp_delta;
    }
}

// seg002:0CD4
void check_sword_hurt(void) {
    if (Guard.action == actions_99_hurt) {
        // custom logic - guard was hurt by sword
        charid_hurt_by_sword = charid_2_guard;
        if (Kid.action == actions_99_hurt) {
            Kid.action = actions_1_run_jump;
        }
        loadshad();
        hurt_by_sword();
        saveshad();
        // custom logic - punches have a different recovery rate
        guard_refrac = have_sword || guard_punch_count == 3 ?
                       settings->refractimer[guard_skill] :
                       settings->refractimer2[guard_skill];
        // custom logic - skeleton at the top of the bridge with the skill 7 waits longer after getting hit
        if (is_falling_bridge_guard()) {
            guard_refrac *= 2;
        }
    } else if (Kid.action == actions_99_hurt) {
        // custom logic - dead guard should not hurt kid to avoid a bug where kid
        // gets repeatedly hurt by an invisible sword between rooms
        if (Guard.alive >= 0 && hitp_curr > 0) {
            charid_hurt_by_sword = 0xFF;
            Kid.action = Kid.frame == frame_106_fall ? actions_4_in_freefall : actions_1_run_jump;
        } else {
            // custom logic - kid was hurt by sword
            charid_hurt_by_sword = charid_0_kid;
            loadkid();
            hurt_by_sword();
            savekid();
        }
    }
}

// seg002:0D1A
void check_sword_hurting(void) {
    short kid_frame;
    kid_frame = Kid.frame;
    if (kid_frame != 0 && (kid_frame < frame_219_exit_stairs_3 || kid_frame > frame_228_exit_stairs_12)) {
        loadshad_and_opp();
        check_hurting();
        saveshad_and_opp();
        loadkid_and_opp();
        check_hurting();
        savekid_and_opp();
    }
}

// seg002:0D56
void check_hurting(void) {
    short opp_frame, char_frame, distance, min_hurt_range, max_hurt_range;
    if (Char.sword != sword_2_drawn) return;
    opp_frame = Opp.frame;
    char_frame = Char.frame;
    // custom logic - kid can strike with the opponent on another row
    if (Char.curr_row != Opp.curr_row) {
        goto loc_sword_moving;
    }
    // frames 153..154: poking with sword
    if (char_frame != frame_153_strike_3 && char_frame != frame_154_poking && char_frame != frame_213_kick) {
        return;
    }
    // custom logic - striking a bowing bird guard
    if (Opp.charid == charid_2_guard &&
            opp_frame >= frame_199_start_bowing_10 &&
            opp_frame <= frame_202_bowing) {
        distance = guard_kid_dist();
        if (Opp.direction != Char.direction) {
            distance -= 10;
        }
        if (distance > -4 && distance < 18) {
            Opp.x -= (distance - 4) * (Char.direction | 1) /* converts dir_0_right to 1 */;
            Opp.action = actions_99_hurt;
        }
        return;
    }
    distance = char_opp_dist();
    // frames 161 and 150: parrying
    if (distance < 0 || distance >= 29 ||
            (opp_frame != frame_161_parry && opp_frame != frame_150_parry)
       ) {
        // ... and Opp is not parrying
        // frame 154: poking
        if (char_frame == frame_154_poking || char_frame == frame_213_kick) {
            if (Opp.sword < sword_2_drawn) {
                min_hurt_range = 8;
            } else {
                min_hurt_range = 12;
            }
            distance = char_opp_dist();
            max_hurt_range = 29;
            // custom logic - fist fighting happens in a different range
            if (Char.charid == charid_0_kid && !have_sword) {
                if (char_frame == frame_213_kick) {
                    min_hurt_range = -5;
                    max_hurt_range = 15;
                } else {
                    min_hurt_range = 1;
                    max_hurt_range = 19;
                }
            }
            if (distance >= min_hurt_range && distance < max_hurt_range) {
                // custom logic - do not hurt kid if he is jumping/falling significantly above the floor (63 px per floor)
                if ((super_jump_timer > 0 || Opp.action == actions_3_in_midair || Opp.action == actions_4_in_freefall) &&
                        (y_land[Opp.curr_row + ONSCREEN_ROW_OFFSET] - Opp.y) >= 37) {
                    if (Opp.charid == charid_0_kid) {
                        guard_refrac = 5;
                    }
                } else {
                    Opp.action = actions_99_hurt;
                }
            }
        }
    } else {
        Opp.frame = frame_161_parry;
        if (Char.charid != charid_0_kid) {
            justblocked = 4;
        }
        if (Char.charid != charid_0_kid && !have_sword) {
            Opp.action = actions_99_hurt; // custom logic - kid cannot parry with his fists
        } else {
            seqtbl_offset_char(seq_69_attack_was_parried); // attack was parried
        }
        play_seq();
    }
loc_sword_moving:
    if (Char.alive >= 0 || Char.direction == dir_56_none || (Char.charid != charid_0_kid && !can_guard_see_kid)) {
        return; // Fix looping "sword moving" sound.
    }
    // frame 154: poking
    // frame 161: parrying
    if (char_frame == frame_154_poking &&
            opp_frame != frame_161_parry &&
            (Opp.action != actions_99_hurt || Opp.alive >= 0 || Opp.direction == dir_56_none)) {
        if (Char.charid == charid_0_kid && !have_sword) {
            return;
        }
        play_sound(sound_11_sword_moving); // sword moving
    }
}

// custom logic - initialize skeleton
void do_init_skel(byte tilepos, sbyte direction, byte col, byte row, word skill, byte presentation) {
    if (resurrect_time > 0 || is_kid_mouse()) {
        return;
    }
    curr_room_tiles[tilepos] = tiles_1_floor;
    redraw_height = 24;
    set_redraw_full(tilepos, 1);
    set_wipe(tilepos, 1);
    set_redraw_full(tilepos + 1, 1);
    set_wipe(tilepos + 1, 1);
    Char.room = drawn_room;
    Char.curr_row = row;
    Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
    Char.curr_col = col;
    Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
    Char.direction = direction;
    seqtbl_offset_char(seq_88_skel_wake_up);
    play_seq();
    set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
    play_sound(sound_44_skel_alive);
    if (presentation && !skeleton_appeared && is_positional_audio) {
        delayed_sound_timer = 4;
        delayed_sound = (struct delayed_sound_type) {
            sound_76_skel_presentation, Char.room, Char.curr_col, Char.curr_row, 0, 0, 0
        };
        skeleton_appeared = 1;
    }
    guard_skill = skill;
    Char.alive = -1;
    guardhp_max = guardhp_curr = 3;
    Char.fall_x = Char.fall_y = 0;
    is_guard_notice = guard_refrac = 0;
    Char.sword = sword_2_drawn;
    Char.charid = charid_4_skeleton;
    saveshad();
}

// seg002:0E1F
void check_skel(void) {
    // custom logic - the original skeleton
    bool is_skel_tile_1 = current_level == 3 &&
                          Guard.direction == dir_56_none &&
                          drawn_room == 1 &&
                          hitp_curr > 0 &&
                          Kid.alive < 0 &&
                          Kid.room == drawn_room &&
                          leveldoor_open != 0 &&
                          (Kid.curr_col == 2 || Kid.curr_col == 3);
    // custom logic - the potion guarding skeleton
    bool is_skel_tile_2 = current_level == 3 &&
                          Guard.direction == dir_56_none &&
                          drawn_room == 17 &&
                          hitp_curr > 0 &&
                          Kid.alive < 0 &&
                          Kid.fall_y == 0 &&
                          Kid.room == drawn_room &&
                          Kid.curr_row == 2 &&
                          Kid.curr_col >= 0 &&
                          Kid.curr_col <= 3;
    // custom logic - the top level skeleton
    bool is_skel_tile_3 = current_level == 3 &&
                          Guard.direction == dir_56_none &&
                          drawn_room == 14 &&
                          hitp_curr > 0 &&
                          Kid.alive < 0 &&
                          Kid.room == drawn_room &&
                          Kid.curr_row == 0 &&
                          Kid.curr_col >= 0 &&
                          Kid.curr_col <= 2;
    // custom logic - the level door guarding skeleton
    bool is_skel_tile_4 = current_level == 3 &&
                          Guard.direction == dir_56_none &&
                          drawn_room == 6 &&
                          leveldoor_open != 0 &&
                          blink_timer < 16 && // kid did not go up the level door
                          hitp_curr > 0 &&
                          Kid.alive < 0 &&
                          Kid.room == drawn_room &&
                          Kid.curr_row == 2;
    // custom logic - the fallen down skeleton
    bool is_skel_tile_5 = current_level == 3 &&
                          (Guard.direction == dir_56_none || Guard.frame == frame_178_chomped) &&
                          drawn_room == 13 &&
                          hitp_curr > 0 &&
                          Kid.alive < 0 &&
                          Kid.fall_y == 0 &&
                          Kid.room == drawn_room &&
                          Kid.curr_row == 2;
    // custom logic - the mid-way skeleton on level 12
    bool is_skel_tile_6 = current_level == 12 &&
                          Guard.direction == dir_56_none &&
                          drawn_room == 24 &&
                          hitp_curr > 0 &&
                          Kid.alive < 0 &&
                          Kid.room == drawn_room &&
                          Kid.curr_row == 1 &&
                          get_tile_div_mod_m7(Kid.x) >= 1;
    // custom logic - the fallen skeleton on level 12
    bool is_skel_tile_7 = current_level == 12 &&
                          Guard.direction == dir_56_none &&
                          drawn_room == 4 &&
                          hitp_curr > 0 &&
                          Kid.alive < 0 &&
                          Kid.room == drawn_room &&
                          Kid.curr_row == 2 &&
                          get_tile_div_mod_m7(Kid.x) >= 5;
    // Special event: skeleton wakes
    if (is_skel_tile_1) {
        get_tile(drawn_room, 5, 1);
        if (curr_tile2 == tiles_21_skeleton) {
            do_init_skel(curr_tilepos, dir_FF_left, tile_col, tile_row, 4, 1);
        }
    } else if (is_skel_tile_2) {
        if (get_tile(drawn_room, 4, 2) == tiles_21_skeleton) {
            do_init_skel(curr_tilepos, dir_FF_left, tile_col, tile_row, 3, 1);
        }
    } else if (is_skel_tile_3) {
        if (get_tile(drawn_room, 2, 0) == tiles_21_skeleton) {
            do_init_skel(curr_tilepos, dir_FF_left, 3 /* makes skeleton fall */, tile_row, 1, 0);
        }
    } else if (is_skel_tile_4) {
        // skeleton is resurrected after 60 ticks
        int skel_col = skeleton_timer == 60 ? 0 : 6;
        if (get_tile(drawn_room, skel_col, 2) == tiles_21_skeleton) {
            do_init_skel(curr_tilepos, dir_0_right, tile_col, tile_row, 1, 1);
        }
    } else if (is_skel_tile_5) {
        if (get_tile(drawn_room, 3, 2) == tiles_21_skeleton) {
            // move chomped skeleton to a dead slot
            if (Guard.frame == frame_178_chomped) {
                word room_minus_1 = drawn_room - 1;
                char_type2* slot = get_dead_guard_slot(room_minus_1);
                if (slot != NULL) {
                    slot->room = Guard.room;
                    slot->direction = Guard.direction;
                    slot->x = Guard.x;
                    slot->row = Guard.curr_row;
                    slot->frame2 = Guard.frame;
                    slot->frame = 0;
                    slot->color = 0;
                }
            }
            sbyte skel_dir = Kid.x <= x_bump[tile_col + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET ? dir_FF_left : dir_0_right;
            do_init_skel(curr_tilepos, skel_dir, tile_col, tile_row, 1, 1);
        }
    } else if (is_skel_tile_6) {
        if (get_tile(drawn_room, 7, 1) == tiles_21_skeleton) {
            do_init_skel(curr_tilepos, dir_FF_left, tile_col, tile_row, 4, 1);
        }
    } else if (is_skel_tile_7) {
        if (get_tile(drawn_room, 7, 2) == tiles_21_skeleton || get_tile(drawn_room, 8, 2) == tiles_21_skeleton) {
            // do not react to kid falling to death
            if (Kid.action == actions_4_in_freefall && Kid.fall_y >= 33) {
                return;
            }
            do_init_skel(curr_tilepos, dir_FF_left, tile_col, tile_row, 4, 1);
        }
    }
}

void do_instruct_moves(int demo_part) {
    const char* instruction = NULL;
    if (demo_part == 1) {
        switch (demo_index) {
            case 5:
                instruction = "SHIFT = PICK UP";
                break;
            case 7:
                instruction = "FORWARD ARROW = RUN";
                break;
            case 11:
                instruction = "UP ARROW = RUNNING JUMP";
                break;
            case 17:
                instruction = "FORWARD + UP = STANDING JUMP";
                break;
            case 21:
                instruction = "SHIFT = GRAB    UP = CLIMB UP";
                break;
            case 26:
                instruction = "SHIFT = STRIKE    UP = BLOCK";
                break;
        }
    } else if (demo_part == 2) {
        switch (demo_index) {
            case 3:
                instruction = "SHIFT + BACK = TURN AND GRAB";
                break;
            case 8:
                instruction = "SHIFT + FORWARD = STEP";
                break;
            case 12:
                instruction = "SHIFT = GRAB WHILE FALLING";
                break;
            case 18:
                instruction = "SHIFT + UP = JUMP AND GRAB";
                break;
            case 22:
                instruction = "UP = ENTER LEVEL DOOR";
                break;
        }
    }
    if (instruction != NULL) {
        display_text("%s", instruction);
    }
}

// seg002:0F3F
void do_auto_moves(const auto_move_type *moves_ptr) {
    short demoindex;
    short curr_move;
    if (demo_time >= 0x128) return;
    ++demo_time;
    demoindex = demo_index;
    if (moves_ptr[demoindex].time <= demo_time) {
        ++demo_index;
    } else {
        demoindex = demo_index - 1;
    }
    curr_move = moves_ptr[demoindex].move;
    switch (curr_move) {
        case -1:
            break;
        case 0:
            move_0_nothing();
            break;
        case 1:
            move_1_forward();
            break;
        case 2:
            move_2_backward();
            break;
        case 3:
            move_3_up();
            break;
        case 4:
            move_4_down();
            break;
        case 5:
            move_3_up();
            move_1_forward();
            break;
        case 6:
            move_6_shift();
            break;
        case 7:
            move_7();
            break;
        case 8:
            move_8_jump_grab();
            break;
        case 9:
            move_6_shift();
            move_1_forward();
            break;
    }
    if (moves_ptr == settings->demo_moves) {
        if (demo_index == 1) {
            display_press_button();
            text_time_total = text_time_remaining = 10000;
            const char* game_instructions =
                "Prince can run, jump and crouch using left/right/up/down arrow keys.\n\n"
                "Press Shift + forward arrow to walk.\n\n"
                "Read more instructions in the bottom text area.\n";
            show_dialog(game_instructions, instructions_dialog);
            erase_bottom_text(1);
        }
        do_instruct_moves(1);
    } else if (moves_ptr == settings->demo_moves_2) {
        do_instruct_moves(2);
    }
}

// seg002:1000
void autocontrol_shadow_level6(void) {
    if (Char.room == 1 && Char.direction != dir_56_none) {
        if (Char.x > 100) {
            move_1_forward();
        }
    }
}

// custom logic - shadow presses the opener tile and jumps down on level 6, room 22
void autocontrol_shadow_level6b(void) {
    if (Char.room == 22 && Char.direction != dir_56_none) {
        if (demo_time == 0) {
            demo_index = 0;
        }
        do_auto_moves(settings->shad_level_6b);
    }
}

// seg002:101A
void autocontrol_shadow_level7(void) {
    // custom logic - shadow appears on level 7, room 6
    if (Char.room == 6 && Char.direction != dir_56_none) {
        if (demo_time == 0) {
            demo_index = 0;
        }
        do_auto_moves(settings->shad_level_7);
    }
}

void autocontrol_shadow_level8(void) {
    // custom logic - shadow appears on level 8, room 16 and closes the gate
    if (Char.room == 16 && Char.direction != dir_56_none && leveldoor_open == 1) {
        if (demo_time == 0) {
            demo_index = 0;
        }
        do_auto_moves(settings->shad_level_8);
    }
}

void autocontrol_shadow_level9(void) {
    // custom logic - shadow appears on level 9, room 1 and closes the gate
    if (Char.room == 1 && Char.direction != dir_56_none) {
        if (Kid.room == Char.room && Kid.curr_row == 1 && (Kid.curr_col < 0 || Kid.curr_col >= 8)) {
            return;
        }
        if (demo_time == 0) {
            demo_index = 0;
        }
        do_auto_moves(settings->shad_level_9_2);
    } else if (Char.room == 23 && Char.direction != dir_56_none) {
        if (demo_time == 0) {
            demo_index = 0;
        }
        do_auto_moves(settings->shad_level_9);
    }
}

void autocontrol_shadow_level10(void) {
    // custom logic - unite with the shadow on level 10 after dying
    word potion_iter;
    short distance;
    if (shadow_initialized < 1 || shadow_initialized > 18) {
        return;
    }
    distance = guard_kid_dist();
    if (distance < 35) {
        if (shadow_initialized == 1 && Char.frame != frame_15_stand) {
            return;
        }
        move_6_shift();
    }
    if (distance > 0) {
        move_1_forward();
    } else {
        ++shadow_initialized;
        if (shadow_initialized == 2) {
            Char.direction = Kid.direction;
            Char.x = char_dx_forward(3);
            Char.y = Kid.y;
            seqtbl_offset_char(seq_71_dying);
            saveshad();
            play_seq();
            return;
        } else if (shadow_initialized < 18) { // stay dead for 18 frames
            return;
        }
        clear_saved_ctrl();
        erase_bottom_text(1);
        // unite with the shadow shimmer
        united_with_shadow = 42;
        shadow_potion_counter = 0;
        // add shadow's potion hitpoints to kid
        for (potion_iter = 0; potion_iter < initial_shadow_potion_counter; potion_iter++) {
            add_life();
        }
        // revive kid
        loadkid();
        Char.alive = -1;
        Char.sword = sword_0_sheathed;
        holding_sword = 0;
        resurrect_time = 20;
        hitp_delta = hitp_max;
        if (inactive_guard[Char.room - 1].room == Char.room) {
            if (Char.x < inactive_guard[Char.room - 1].x && Char.direction == dir_FF_left) {
                Char.direction = dir_0_right;
                Char.x = char_dx_forward(4);
            } else if (Char.x > inactive_guard[Char.room - 1].x && Char.direction == dir_0_right) {
                Char.direction = dir_FF_left;
                Char.x = char_dx_forward(4);
            }
        }
        Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
        seqtbl_offset_char(seq_142_rise);
        play_seq();
        load_fram_det_col();
        savekid();
        loadshad();
        Char.direction = dir_56_none;
        saveshad();
        droppedout = 0;
        offguard = 1;
        enter_guard();
        // play victory sound
        stop_sounds();
        play_sound(is_positional_audio ? sound_79_united_with_shadow : sound_37_victory);
    }
}
