/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"

// data:432F
sbyte bump_col_left_of_wall;
// data:436E
sbyte bump_col_right_of_wall;
// data:4C0A
sbyte right_checked_col;
// data:408A
sbyte left_checked_col;


// data:4C0C
short coll_tile_left_xpos;
// These two arrays are indexed with the return value of wall_type.
// data:24BA
const sbyte wall_dist_from_left[] = {0, 10, 0, -1, 0, 0};
// data:24C0
const sbyte wall_dist_from_right[] = {0, 0, 10, 13, 0, 0};

// seg004:0004
void check_collisions(void) {
    short column;
    bump_col_left_of_wall = bump_col_right_of_wall = -1;
    if (Char.action == actions_7_turn) return;
    collision_row = Char.curr_row;
    move_coll_to_prev();
    prev_collision_row = collision_row;
    right_checked_col = MIN(get_tile_div_mod_m7(char_x_right_coll) + 2, 11);
    left_checked_col = get_tile_div_mod_m7(char_x_left_coll) - 1;
    get_row_collision_data(collision_row, curr_row_coll_room, curr_row_coll_flags);
    get_row_collision_data(collision_row + 1, below_row_coll_room, below_row_coll_flags);
    get_row_collision_data(collision_row - 1, above_row_coll_room, above_row_coll_flags);
    for (column = 9; column >= 0; --column) {
        if (curr_row_coll_room[column] >= 0 &&
                prev_coll_room[column] == curr_row_coll_room[column]
           ) {
            // char bumps into left of wall
            if (
                (prev_coll_flags[column] & 0x0F) == 0 &&
                (curr_row_coll_flags[column] & 0x0F) != 0
            ) {
                bump_col_left_of_wall = column;
            }
            // char bumps into right of wall
            if (
                (prev_coll_flags[column] & 0xF0) == 0 &&
                (curr_row_coll_flags[column] & 0xF0) != 0
            ) {
                bump_col_right_of_wall = column;
            }
        }
    }
}

// seg004:00DF
void move_coll_to_prev(void) {
    sbyte* row_coll_room_ptr;
    byte* row_coll_flags_ptr;
    short column;
    if (collision_row == prev_collision_row ||
            collision_row + 3 == prev_collision_row ||
            collision_row - 3 == prev_collision_row
       ) {
        row_coll_room_ptr = curr_row_coll_room;
        row_coll_flags_ptr = curr_row_coll_flags;
    } else if (
        collision_row + 1 == prev_collision_row ||
        collision_row - 2 == prev_collision_row
    ) {
        row_coll_room_ptr = above_row_coll_room;
        row_coll_flags_ptr = above_row_coll_flags;
    } else {
        row_coll_room_ptr = below_row_coll_room;
        row_coll_flags_ptr = below_row_coll_flags;
    }
    for (column = 0; column < 10; ++column) {
        prev_coll_room[column] = row_coll_room_ptr[column];
        prev_coll_flags[column] = row_coll_flags_ptr[column];
        below_row_coll_room[column] = -1;
        above_row_coll_room[column] = -1;
        curr_row_coll_room[column] = -1;
        // bugfix:
        curr_row_coll_flags[column] = 0;
        below_row_coll_flags[column] = 0;
        above_row_coll_flags[column] = 0;
    }
}

// seg004:0185
void get_row_collision_data(short row, sbyte *row_coll_room_ptr, byte *row_coll_flags_ptr) {
    short right_wall_xpos;
    byte curr_flags;
    short room;
    short column;
    short left_wall_xpos;
    room = Char.room;
    coll_tile_left_xpos = x_bump[left_checked_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
    for (column = left_checked_col; column <= right_checked_col; ++column) {
        left_wall_xpos = get_left_wall_xpos(room, column, row);
        right_wall_xpos = get_right_wall_xpos(room, column, row);
        // char bumps into left of wall
        curr_flags = (left_wall_xpos < char_x_right_coll) * 0x0F;
        // char bumps into right of wall
        curr_flags |= (right_wall_xpos > char_x_left_coll) * 0xF0;
        row_coll_flags_ptr[tile_col] = curr_flags;
        row_coll_room_ptr[tile_col] = curr_room;
        coll_tile_left_xpos += 14;
    }
}

// seg004:0226
int get_left_wall_xpos(int room, int column, int row) {
    short type;
    type = wall_type(get_tile(room, column, row));
    if (type) {
        return wall_dist_from_left[type] + coll_tile_left_xpos;
    } else {
        return 0xFF;
    }
}

// seg004:025F
int get_right_wall_xpos(int room, int column, int row) {
    short type;
    type = wall_type(get_tile(room, column, row));
    if (type) {
        return coll_tile_left_xpos - wall_dist_from_right[type] + 13;
    } else {
        return 0;
    }
}

// seg004:029D
void check_bumped(void) {
    if (Char.action != actions_2_hang_climb &&
            Char.action != actions_6_hang_straight &&
            // frames 135..149: climb up
            (Char.frame < frame_135_climbing_1 || Char.frame >= frame_149_climbing_15)
       ) {
        // custom logic - do not bump the invisible wall on the right in rooms 2 and 13 on level 13
        if (bump_col_left_of_wall >= 0) {
            if (bump_col_left_of_wall == 0 &&
                    curr_room == 0 &&
                    current_level == 13 &&
                    (drawn_room == 2 || drawn_room == 13)) {
                return;
            }
            check_bumped_look_right();
        }
        if (bump_col_right_of_wall >= 0) {
            check_bumped_look_left();
        }
    }
}

// seg004:02D2
void check_bumped_look_left(void) {
    if ((Char.sword == sword_2_drawn || Char.direction < dir_0_right) && // looking left
            is_obstacle_at_col(bump_col_right_of_wall)
       ) {
        // custom logic - kid can grab a tile above after bumping a wall
        if (control_shift < control_0_released) {
            if (check_grab_run_jump()) {
                return;
            }
            // reset tile coordinates
            is_obstacle_at_col(bump_col_right_of_wall);
        }
        bumped(get_right_wall_xpos(curr_room, tile_col, tile_row) - char_x_left_coll, dir_0_right, curr_tile2);
    }
}

// seg004:030A
void check_bumped_look_right(void) {
    if ((Char.sword == sword_2_drawn || Char.direction == dir_0_right) && // looking right
            is_obstacle_at_col(bump_col_left_of_wall)
       ) {
        // custom logic - kid can grab a tile above after bumping a wall
        if (control_shift < control_0_released) {
            if (check_grab_run_jump()) {
                return;
            }
            // reset tile coordinates
            is_obstacle_at_col(bump_col_left_of_wall);
        }
        bumped(get_left_wall_xpos(curr_room, tile_col, tile_row) - char_x_right_coll, dir_FF_left, curr_tile2);
    }
}

// custom logic - returns whether the tile is a fake wall
int is_fake_wall(void) {
    if (curr_tile2 != tiles_20_wall && curr_tile2 != tiles_7_doortop_with_floor) {
        return 0;
    }
    if (current_level == 2 &&
            curr_room == 24 &&
            curr_tilepos == 9) {
        return 1; // fake wall near life potion on level 2
    }
    if (current_level == 14 &&
            curr_room == 19 &&
            curr_tilepos == 9) {
        return 1; // fake wall near gate opener in the wall on level 13
    }
    if (current_level == 14 &&
            curr_room == 12 &&
            curr_tilepos == 19 &&
            Char.direction == dir_FF_left) {
        if (drawn_room == level.roomlinks[curr_room - 1].right &&
                char_x_right >= x_bump[1 + ONSCREEN_COLUMN_OFFSET]) {
            return 0;
        }
        return 1; // doortop with floor on level 13, room 12 can be walked through right to left
    }
    return 0;
}

// seg004:0343
int is_obstacle_at_col(int tile_col) {
    short tile_row;
    tile_row = Char.curr_row;
    if (tile_row < 0) {
        tile_row += 3;
    }
    if (tile_row >= 3) {
        tile_row -= 3;
    }
    // custom logic - fixed a bump issue with a doortop facing a non-existing room
    // and kid bumping into in the right direction
    if (curr_row_coll_room[tile_col] == 0 && tile_col == 0) {
        get_tile(Char.room, 9, tile_row);
        if (is_obstacle()) {
            return 1;
        } else {
            tile_col = 0;
        }
    }
    get_tile(curr_row_coll_room[tile_col], tile_col, tile_row);
    return is_obstacle();
}

// seg004:037E
int is_obstacle(void) {
    if (curr_tile2 == tiles_10_potion) {
        return 0;
    } else if (curr_tile2 == tiles_4_gate) {
        if (!can_bump_into_gate()) {
            return 0;
        }
    } else if (curr_tile2 == tiles_18_chomper) {
        // custom logic - kid as shadow should not bump into chompers
        if (Char.charid == charid_0_kid && (is_kid_shadow || is_kid_mouse())) {
            return 0;
        }
        // is the chomper closed? (with or without blood)
        if ((curr_room_modif[curr_tilepos] & 0x7F) != 2) return 0;
    } else if (
        curr_tile2 == tiles_13_mirror &&
        Char.charid == charid_0_kid &&
        Char.frame >= frame_39_start_run_jump_6 && Char.frame < frame_44_running_jump_5 && // run-jump
        Char.direction < dir_0_right // right-to-left only
    ) {
        // custom logic - keep track of whether kid has jumped through the mirror already
        jumped_through_mirror = jumped_through_mirror ? -2 : -1;
        return 0;
        // custom logic - fake walls are not walls
    } else if (is_fake_wall() == 1) {
        return 0;
        // custom logic - doortop with floor on level 13, room 12 can be walked through
    } else if (current_level == 14 &&
               Char.room == 11 &&
               Char.curr_col <= 0 &&
               Char.curr_row == 1) {
        return 0;
    } else if (
        // custom logic - wall on level 4 room 20 can be jumped through
        current_level == 4 &&
        curr_tile2 == tiles_20_wall &&
        Char.charid == charid_0_kid &&
        Char.room == 20 &&
        Char.curr_row == 1 &&
        Char.frame >= frame_39_start_run_jump_6 && Char.frame < frame_44_running_jump_5 && // run-jump
        Char.direction == dir_FF_left
    ) {
        if (hitp_curr > 1) {
            hitp_delta = -(hitp_curr - 1);
        } else {
            hitp_delta = -hitp_curr;
        }
        stop_sounds();
        play_sound(sound_13_kid_hurt);
        return 0;
    } else if (
        // custom logic - door top with floor on level 4, room 20 is not an obstacle
        current_level == 4 &&
        curr_tile2 == tiles_7_doortop_with_floor &&
        Char.charid == charid_0_kid &&
        Char.room == 20 &&
        Char.curr_row == 1 &&
        Char.direction == dir_FF_left
    ) {
        // remove the wall
        get_tile(20, 4, 1);
        curr_room_tiles[curr_tilepos] = tiles_4_gate;
        curr_room_modif[curr_tilepos] = 0xFF;
        get_tile(20, 5, 1);
        curr_room_tiles[curr_tilepos] = tiles_1_floor;
        curr_room_modif[curr_tilepos] = 0;
        redraw_height = 63;
        set_redraw_full(curr_tilepos - 1, 1);
        set_wipe(curr_tilepos - 1, 1);
        set_redraw_full(curr_tilepos, 1);
        set_wipe(curr_tilepos, 1);
        set_redraw_full(curr_tilepos + 1, 1);
        set_wipe(curr_tilepos + 1, 1);
        return 0;
    }
    coll_tile_left_xpos = xpos_in_drawn_room(x_bump[tile_col + ONSCREEN_COLUMN_OFFSET]) + CHAR_7_PX_OFFSET;
    return 1;
}

// seg004:0405
int xpos_in_drawn_room(int xpos) {
    if (curr_room != drawn_room) {
        if (curr_room == room_L || curr_room == room_BL) {
            xpos -= 140;
        } else if (curr_room == room_R || curr_room == room_BR) {
            xpos += 140;
        }
    }
    return xpos;
}

// seg004:0448
void bumped(sbyte delta_x, sbyte push_direction, int coll_tile) {
    short distance;
    // custom logic - do not bump character when falling down into room 0. It can cause a crash due to recycled memory.
    if (curr_room == 0 && Char.curr_row > 2) {
        return;
    }
    // custom logic - moving wall should not be bumping kid backwards
    if (current_level == settings->moving_wall_level &&
            curr_room == settings->moving_wall_room &&
            tile_row == settings->moving_wall_row &&
            tile_col >= settings->moving_wall_col_left &&
            tile_col <= settings->moving_wall_col_right &&
            ((Char.direction == dir_FF_left && tile_col - Char.curr_col >= 2 && delta_x > 0) ||
             (Char.direction == dir_0_right && Char.curr_col - tile_col >= 2 && delta_x < 0))
       ) {
        return;
    }
    if ((Char.alive < 0 || Char.action == actions_4_in_freefall) && Char.frame != frame_177_spiked) {
        Char.x += delta_x;
        if (push_direction < dir_0_right) {
            // pushing left
            if (curr_tile2 == tiles_20_wall) {
                get_tile(curr_room, --tile_col, tile_row);
            }
        } else {
            // pushing right
            if (curr_tile2 == tiles_12_doortop ||
                    curr_tile2 == tiles_7_doortop_with_floor ||
                    curr_tile2 == tiles_20_wall ||
                    // custom logic - falling prince should move back when hitting a closed gate to be able to grab
                    // a floor above; standing bumps are handled by the check_gate_push() method.
                    (Char.action == actions_4_in_freefall && curr_tile2 == tiles_4_gate && can_bump_into_gate())
               ) {
                ++tile_col;
                if (curr_room == 0 && tile_col == 10) {
                    curr_room = Char.room;
                    tile_col = 0;
                }
                get_tile(curr_room, tile_col, tile_row);
            }
        }
        if (tile_is_floor(curr_tile2)) {
            bumped_floor(push_direction, coll_tile);
            // custom logic - prevent bumping character through a wall
        } else if (curr_tile2 == tiles_20_wall &&
                   Opp.sword == sword_2_drawn &&
                   (distance = char_opp_dist()) > 0 &&
                   distance < 29 &&
                   tile_is_floor(get_tile(Char.room, get_tile_div_mod_m7(Char.x + (dir_front[Char.direction + 1] * 7)), Char.curr_row))) {
            Char.x = char_dx_forward(7);
            determine_col();
            if (!tile_is_floor(get_tile_at_char())) {
                bumped_fall();
            }
            // custom logic - do not bump fall during a sword strike
        } else if (Char.frame >= frame_151_strike_1 && Char.frame <= frame_154_poking) {
            Char.x = char_dx_forward(-4);
            determine_col();
            bumped_sound();
            // custom logic - do not bump fall if kid can grab and Shift is pressed
        } else if (Char.action == actions_3_in_midair &&
                   can_grab_front_above() &&
                   distance_to_edge_weight() <= 4 &&
                   control_shift < control_0_released) {
            return;
        } else {
            bumped_fall();
        }
    }
}

// seg004:04E4
void bumped_fall(void) {
    short action;
    action = Char.action;
    // custom logic - sheathe sword after the bump fall
    Char.sword = sword_0_sheathed;
    Char.x = char_dx_forward(-4);
    if (action == actions_4_in_freefall) {
        Char.fall_x = 0;
        // custom logic - do not bump/fall when jumping up in feather mode
    } else if (super_jump_fall == 0) {
        // custom logic - set hard jump fall flag
        if (Char.charid == charid_0_kid) {
            if (Char.frame >= frame_42_running_jump_3 && Char.frame <= frame_44_running_jump_5) {
                hard_jump_fall = 2; // loses a hitpoint
            } else {
                hard_jump_fall = 1;
            }
            // does guard need to follow kid?
            if (droppedout == 0) {
                droppedout = (Opp.charid == charid_2_guard || Opp.charid == charid_4_skeleton) &&
                             !is_guarding_gate &&
                             can_guard_see_kid
                             ? 1
                             : 0;
                // guard should be facing in the right direction
                if (droppedout &&
                        ((Char.curr_col < Opp.curr_col && Opp.direction != dir_FF_left) ||
                         (Char.curr_col > Opp.curr_col && Opp.direction != dir_0_right))) {
                    loadshad();
                    Char.direction = ~Char.direction;
                    Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
                    saveshad();
                    loadkid();
                }
            }
        }
        seqtbl_offset_char(seq_45_bumpfall); // fall after bumped
        play_seq();
    }
    bumped_sound();
}

// seg004:0520
void bumped_floor(sbyte push_direction, int coll_tile) {
    short frame;
    short seq_index;
    if (Char.sword != sword_2_drawn && (word)(y_land[Char.curr_row + ONSCREEN_ROW_OFFSET] - Char.y) >= (word)15) {
        bumped_fall();
    } else {
        frame = Char.frame;
        // custom logic - do not bump kid after a medium landing
        if (frame == frame_109_crouch &&
                delayed_sound_timer > 0 &&
                delayed_sound_timer <= 2 &&
                delayed_sound.sound_id == sound_13_kid_hurt) {
            return;
        }
        Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
        if (Char.fall_y >= 22) {
            Char.x = char_dx_forward(-5);
        } else {
            Char.fall_y = 0;
            if (Char.alive) {
                if (Char.sword == sword_2_drawn) {
                    if (push_direction == Char.direction) {
                        seqtbl_offset_char(seq_65_bump_forward_with_sword); // pushed forward with sword (Kid)
                        play_seq();
                        Char.x = char_dx_forward(1);
                        return;
                    } else {
                        // custom logic - improved wall collision logic with sword
                        if (push_direction == dir_0_right && coll_tile == tiles_4_gate) {
                            Char.x = x_bump[get_tile_col_at_char() + ONSCREEN_COLUMN_OFFSET + 1] + 13;
                        } else {
                            Char.x = x_bump[get_tile_col_at_char() + ONSCREEN_COLUMN_OFFSET] + 13;
                        }
                        if (Char.direction == dir_0_right && push_direction == dir_FF_left && coll_tile != tiles_20_wall) {
                            Char.x = char_dx_forward(-4);
                        } else if (Char.direction == dir_FF_left) {
                            Char.x = char_dx_forward(-2);
                        }
                        seq_index = seq_64_pushed_back_with_sword; // pushed back with sword
                        // custom logic - colliding with sword against gate makes metallic sounds
                        if (have_sword && coll_tile != tiles_12_doortop && coll_tile != tiles_7_doortop_with_floor) {
                            play_sound(sound_10_sword_vs_sword);
                        }
                    }
                } else {
                    if (frame == frame_24_standing_jump_9 || frame == frame_25_standing_jump_10 ||
                            (frame >= frame_40_running_jump_1 && frame < frame_43_running_jump_4) ||
                            (frame >= frame_102_start_fall_1 && frame < frame_107_fall_land_1)
                       ) {
                        seq_index = seq_46_hardbump; // bump into wall after run-jump (crouch)
                        bumped_sound();
                    } else {
                        // custom logic - only certain frames should make a bump sound
                        if ((frame >= frame_12_run && frame <= frame_14_run) ||
                                (frame >= frame_43_running_jump_4 && frame <= frame_44_running_jump_5) ||
                                (frame >= frame_57_runturn && frame <= frame_63_runturn) ||
                                (frame >= frame_186_mouse_1 && frame <= frame_187_mouse_2)
                           ) {
                            bumped_sound();
                        }
                        // custom logic - prevent kid from getting bumped through a gate between 2 walls/gates/tapestries
                        if (push_direction == dir_0_right &&
                                (get_tile_infrontof_char() == tiles_20_wall || wall_type2(curr_tile2) == wall_type_1_right) &&
                                wall_type2(get_tile_at_char()) == wall_type_1_right) {
                            Char.x = char_dx_forward(4);
                        } else if (push_direction == dir_FF_left &&
                                   wall_type2(get_tile_at_char()) == wall_type_1_right &&
                                   (get_tile_behind_char() == tiles_20_wall || wall_type2(curr_tile2) == wall_type_1_right)) {
                            Char.x = char_dx_forward(4);
                        }
                        // custom logic - handle kid as mouse bumping into walls
                        if (Char.charid == charid_0_kid && is_kid_mouse()) {
                            Char.x = char_dx_forward(-5);
                            determine_col();
                            if (Char.fall_x >= MOUSE_HIT_SPEED) {
                                Char.fall_x = 0;
                                seq_index = seq_47_bump;
                                if (current_level == 14 && Char.room == 20 && Char.x < 72) {
                                    play_sound(sound_36_out_of_time);
                                    level_restart_timer = 168;
                                    display_text_bottom("YOU HAVE BEEN CAPTURED");
                                    text_time_total = 48;
                                    text_time_remaining = 48;
                                }
                            } else {
                                release_arrows();
                                Char.fall_x = 0;
                                seq_index = seq_106_mouse;
                            }
                        } else {
                            seq_index = seq_47_bump; // bump into wall
                        }
                    }
                    // custom logic - keep kid in front of the gate in room 1, level 6
                    if (current_level == 6 && Char.room == 1 && Char.x <= 85) {
                        seq_index = seq_2_stand;
                    }
                }
                // custom logic - kid falls when jumping and hitting the gate
                if (Char.charid == charid_0_kid &&
                        Char.direction == dir_FF_left &&
                        coll_tile == tiles_4_gate &&
                        distance_to_edge(Char.x) == 13 &&
                        !tile_is_floor(get_tile_at_kid(Char.x + 7)) &&
                        ((frame >= frame_23_standing_jump_8 && frame <= frame_33_standing_jump_18) ||
                         (frame >= frame_40_running_jump_1 && frame <= frame_44_running_jump_5))
                   ) {
                    if (!check_sound_playing()) {
                        bumped_sound();
                    }
                    if (frame >= frame_40_running_jump_1) {
                        Char.x = char_dx_forward(-3);
                    }
                    start_fall();
                    return;
                }
                seqtbl_offset_char(seq_index);
                play_seq();
            }
        }
    }
}

// seg004:05F1
void bumped_sound(void) {
    // custom logic - do not play bumped sound during a feather fall
    if (Char.fall_y > 0 && feather_fall_timer > 0) {
        return;
    }
    // custom logic - in feather fall mode guards do not notice kid
    if (!feather_fall_timer) {
        set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
        play_sound(sound_8_bumped);
    }
    make_guard_notice(true);
}

// seg004:0601
void clear_coll_rooms(void) {
    memset(prev_coll_room, -1, sizeof(prev_coll_room));
    memset(curr_row_coll_room, -1, sizeof(curr_row_coll_room));
    memset(below_row_coll_room, -1, sizeof(below_row_coll_room));
    memset(above_row_coll_room, -1, sizeof(above_row_coll_room));
    // workaround
    memset(prev_coll_flags, 0, sizeof(prev_coll_flags));
    memset(curr_row_coll_flags, 0, sizeof(curr_row_coll_flags));
    memset(below_row_coll_flags, 0, sizeof(below_row_coll_flags));
    memset(above_row_coll_flags, 0, sizeof(above_row_coll_flags));
    prev_collision_row = -1;
}

// seg004:0657
int can_bump_into_gate(void) {
    // custom logic - mouse can walk through closed gates
    if (Char.charid == charid_24_mouse || (Char.charid == charid_0_kid && is_kid_mouse())) {
        return 0;
    }
    byte modifier = curr_room_modif[curr_tilepos];
    // custom logic - gates on level 12B act inversely for kid
    if (current_level == settings->inverse_level && Char.charid == charid_0_kid) {
        return modifier > 0;
    }
    // custom logic - fully open gate should let guards in that can have sprites over 53 pixels
    if (modifier >= 188) {
        return 0;
    }
    // custom logic - do not fly through semi-closed gate
    if ((Char.action == actions_3_in_midair || Char.action == actions_4_in_freefall) &&
            (word)(y_land[Char.curr_row + ONSCREEN_ROW_OFFSET] - Char.y) >= (word)15) {
        return modifier < 180;
    }
    // custom logic - account for scenarios between rooms when char_height is small
    if (modifier == 0) {
        return 1;
    }
    return (modifier >> 2) + 6 < char_height;
}

// seg004:067C
int get_edge_distance(void) {
    short distance;
    byte tiletype;
    determine_col();
    load_frame_to_obj();
    set_char_collision();
    tiletype = get_tile_at_char();
    if (wall_type(tiletype) != wall_type_0_none) {
        tile_col = Char.curr_col;
        distance = dist_from_wall_forward(tiletype);
        if (distance >= 0) {
loc_wall_dist:
            if (distance < 14) {
                edge_type = edge_type_1_edge;
            } else {
                edge_type = edge_type_2_floor;
                distance = 11;
            }
        } else {
            goto loc_no_wall_dist;
        }
    } else {
loc_no_wall_dist:
        tiletype = get_tile_infrontof_char();
        if (tiletype == tiles_12_doortop && Char.direction >= dir_0_right) {
loc_step_dist:
            edge_type = edge_type_0_object_or_loose;
            distance = distance_to_edge_weight();
        } else {
            if (wall_type(tiletype) != wall_type_0_none) {
                tile_col = infrontx;
                distance = dist_from_wall_forward(tiletype);
                if (distance >= 0) goto loc_wall_dist;
            }
            // custom logic - safe step on closer tiles
            if (tiletype == tiles_6_closer) {
                goto loc_step_dist;
            }
            // custom logic - do not consider stuck loose tiles as loose
            if (tiletype == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0) {
                goto loc_step_dist;
            }
            if (tiletype == tiles_22_sword || tiletype == tiles_10_potion) {
                distance = distance_to_edge_weight();
                if (distance != 0) {
                    edge_type = edge_type_0_object_or_loose;
                } else {
                    edge_type = edge_type_2_floor;
                    distance = 11;
                }
            } else {
                if (tile_is_floor(tiletype)) {
                    edge_type = edge_type_2_floor;
                    distance = 11;
                } else {
                    goto loc_step_dist;
                }
            }
        }
    }
    // custom logic - tracks an active chomper
    edge_chomper = tiletype == tiles_18_chomper && find_trob_at_loc(curr_room, curr_tilepos) >= 0 ? 1 : 0;
    curr_tile2 = tiletype;
    return distance;
}

// seg004:076B
void check_chomped_kid(void) {
    short tile_col;
    short tile_row;
    tile_row = Char.curr_row;
    for (tile_col = 0; tile_col < 10; ++tile_col) {
        if (curr_row_coll_flags[tile_col] == 0xFF &&
                get_tile(curr_row_coll_room[tile_col], tile_col, tile_row) == tiles_18_chomper &&
                (curr_room_modif[curr_tilepos] & 0x7F) == 2  && // closed chomper
                // custom logic - closed chomper should not chomp before it is activated
                find_trob_at_loc(curr_room, curr_tilepos) >= 0
           ) {
            chomped();
        }
    }
}

// seg004:07BF
void chomped(void) {
    // custom logic - kid as shadow is immune to chompers
    if (Char.charid == charid_0_kid && is_kid_shadow) {
        return;
    }
    // custom logic - mouse and shadow are immune to chompers
    if (Char.charid == charid_1_shadow || Char.charid == charid_24_mouse) {
        return;
    }
    if (Char.frame != frame_178_chomped && Char.room == curr_room) {
        byte room_minus_1;
        room_minus_1 = curr_room - 1;
        // custom logic - only show blood when the character is actually killed (put the check inside the if statement)
        if (Char.charid != charid_4_skeleton) {
            curr_room_modif[curr_tilepos] |= 0x80; // put blood
        }
        // custom logic - guards' chomped sprites are facing in the wrong direction
        if (Char.charid == charid_2_guard) {
            Char.direction = ~Char.direction;
        }
        // custom logic - reverse direction of the second chomped guard
        bool guard1_chomped = dead_guard1[room_minus_1].room == curr_room &&
                              dead_guard1[room_minus_1].frame == frame_178_chomped &&
                              dead_guard1[room_minus_1].row == Char.curr_row &&
                              ABS(dead_guard1[room_minus_1].x - Char.x) <= 24;
        bool guard2_chomped = dead_guard2[room_minus_1].room == curr_room &&
                              dead_guard2[room_minus_1].frame == frame_178_chomped &&
                              dead_guard2[room_minus_1].row == Char.curr_row &&
                              ABS(dead_guard2[room_minus_1].x - Char.x) <= 24;
        if (Char.charid == charid_2_guard || Char.charid == charid_4_skeleton) {
            if (guard1_chomped && dead_guard1[room_minus_1].direction == Char.direction) {
                Char.direction = ~Char.direction;
            } else if (guard2_chomped &&
                       dead_guard2[room_minus_1].direction == Char.direction &&
                       dead_guard1[room_minus_1].frame != frame_178_chomped) {
                Char.direction = ~Char.direction;
            }
        }
        // custom logic - recalculate offscreen chomper column
        int chomper_col = get_tile_col_at_char();
        Char.x = x_bump[chomper_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
        Char.x = char_dx_forward(7 + !Char.direction);
        // custom logic - align skeleton in the chomper
        if (Char.charid == charid_4_skeleton && Char.direction == dir_0_right) {
            Char.x = char_dx_forward(-4);
        }
        // custom logic - temple has different chompers
        if (is_temple()) {
            Char.x -= 2;
        }
        Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
        Char.fall_y = 0;
        Char.sword = sword_0_sheathed;
        // custom logic - do not try to stand up and draw sword if chomped near guard
        if (Char.charid == charid_0_kid && char_opp_dist() < 35 && Char.curr_row == Opp.curr_row) {
            Char.alive = 0;
        }
        take_hp(ALL_HP);
        set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
        play_sound(sound_46_chomped); // something chomped
        seqtbl_offset_char(seq_54_chomped); // chomped
        play_seq();
        // custom logic - hide older chomped guards to avoid newly chomped
        // guards of different color to disappear
        if (Char.charid == charid_2_guard && guard1_chomped && guard2_chomped) {
            if (Char.color > 0 &&
                    Char.color != (dead_guard1[room_minus_1].color & 0x0F) &&
                    dead_guard1[room_minus_1].direction == Char.direction) {
                dead_guard1[room_minus_1].direction = dir_56_none;
            } else if (Char.color > 0 &&
                       Char.color != (dead_guard2[room_minus_1].color & 0x0F) &&
                       dead_guard2[room_minus_1].direction == Char.direction) {
                dead_guard2[room_minus_1].direction = dir_56_none;
            }
        }
    }
}

// seg004:0833
void check_gate_push(void) {
    // Closing gate pushes Kid
    short frame;
    short orig_col;
    frame = Char.frame;
    if (Char.action == actions_7_turn ||
            frame <= frame_15_stand || // stand/run
            (frame >= frame_45_turn && frame <= frame_65_runturn) || // turn
            (frame >= frame_108_fall_land_2 && frame < frame_111_stand_up_from_crouch_2) || // crouch
            (frame == frame_106_fall && Char.fall_x > 0 && Char.direction == dir_FF_left) || // jumping forward facing left
            (frame == frame_106_fall && distance_to_edge_weight() < 5 && Char.direction == dir_0_right) || // falling with the back towards the gate
            frame == frame_157_walk_with_sword ||
            frame == frame_158_stand_with_sword ||
            frame == frame_165_walk_with_sword ||
            frame == frame_166_stand_inactive ||
            frame == frame_170_stand_with_sword ||
            frame == frame_171_stand_with_sword) { // engarde
        get_tile_at_char();
        orig_col = tile_col;
        int orig_room = curr_room;
        if ((curr_tile2 == tiles_4_gate ||
                get_tile(curr_room, --tile_col, tile_row) == tiles_4_gate) &&
                (curr_row_coll_flags[tile_col] & prev_coll_flags[tile_col]) == 0xFF &&
                can_bump_into_gate()
           ) {
            // custom logic - fixed falling through gate when jumping
            if (frame == frame_106_fall) {
                bumped_sound();
                Char.fall_x = 0; // cancel forward movement
                Char.x += 5; // bump away
                return;
            }
            // custom logic - only certain frames should make a bump sound
            if ((frame >= frame_12_run && frame <= frame_14_run) ||
                    (frame >= frame_43_running_jump_4 && frame <= frame_44_running_jump_5) ||
                    (frame >= frame_57_runturn && frame <= frame_63_runturn)) {
                bumped_sound();
            }
            // If get_tile() changed curr_room from orig_room to the left neighbor of orig_room (because tile_col was outside room orig_room),
            // then change tile_col (and curr_room) so that orig_col and tile_col are meant in the same room.
            if (curr_room == level.roomlinks[orig_room - 1].left) {
                tile_col -= 10;
                curr_room = orig_room;
            }
            // custom logic - adjust push direction when the sword is drawn
            if (Char.sword == sword_2_drawn && distance_to_edge_weight() < 6) {
                if (Char.direction == dir_0_right) {
                    ++orig_col;
                } else {
                    --orig_col;
                }
            }
            // push Kid left if orig_col <= tile_col, gate at char's tile
            // push Kid right if orig_col > tile_col, gate is left from char's tile
            Char.x += 5 - (orig_col <= tile_col) * 10;
        }
    }
}

// seg004:08C3
void check_guard_bumped(void) {
    // custom logic - ensure guards gets bumped while falling
    if (((Char.action == actions_1_run_jump || Char.action == actions_5_bumped || Char.action == actions_7_turn) && Char.alive < 0) ||
            Char.action == actions_4_in_freefall) {
        if (get_tile_behind_char() == tiles_20_wall ||
                get_tile_at_char() == tiles_20_wall ||
                curr_tile2 == tiles_7_doortop_with_floor ||
                curr_tile2 == tiles_12_doortop ||
                (curr_tile2 == tiles_4_gate && can_bump_into_gate()) ||
                (Char.direction >= dir_0_right && (
                     get_tile(curr_room, --tile_col, tile_row) == tiles_7_doortop_with_floor ||
                     curr_tile2 == tiles_12_doortop ||
                     (curr_tile2 == tiles_4_gate && can_bump_into_gate())
                 ))
           ) {
            load_frame_to_obj();
            set_char_collision();
            if (is_obstacle()) {
                short delta_x;
                delta_x = dist_from_wall_behind(curr_tile2);
                if (delta_x < 0 && delta_x >= -13) {
                    if (delta_x > -2) delta_x = -2;
                    Char.x = char_dx_forward(-delta_x);
                    load_fram_det_col();
                    if (Char.sword >= sword_2_drawn && Char.action == actions_1_run_jump) {
                        seqtbl_offset_char(seq_65_bump_forward_with_sword); // pushed to wall with sword (Guard)
                        play_seq();
                    }
                }
            }
        }
    }
}

// seg004:0989
void check_chomped_guard(void) {
    get_tile_at_char();
    if (!check_chomped_here()) {
        get_tile(curr_room, ++tile_col, tile_row);
        check_chomped_here();
    }
}

// seg004:09B0
int check_chomped_here(void) {
    if (curr_tile2 == tiles_18_chomper &&
            (curr_room_modif[curr_tilepos] & 0x7F) == 2
       ) {
        coll_tile_left_xpos = x_bump[tile_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
        if (get_left_wall_xpos(curr_room, tile_col, tile_row) < char_x_right_coll &&
                get_right_wall_xpos(curr_room, tile_col, tile_row) > char_x_left_coll
           ) {
            chomped();
            return 1;
        } else {
            return 0;
        }
    } else {
        return 0;
    }
}

// custom logic - check for chomped inactive/dead guard
void check_dead_or_inactive_guard_chomped(char_type2* dead, sbyte alive) {
    if (dead->room != drawn_room ||
            dead->frame == frame_0 ||
            dead->frame == frame_178_chomped ||
            dead->frame2 == frame_178_chomped) {
        return;
    }
    short dead_col, chomped_col;
    byte chomped;
    dead_col = get_tile_div_mod_m7(dead->x_left_coll);
    chomped = 0;
    for (chomped_col = dead_col; chomped_col <= dead_col + 1; ++chomped_col) {
        get_tile(dead->room, chomped_col, dead->row);
        if (curr_tile2 == tiles_18_chomper && (curr_room_modif[curr_tilepos] & 0x7F) == 2) {
            coll_tile_left_xpos = x_bump[tile_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
            if (get_left_wall_xpos(dead->room, chomped_col, dead->row) < dead->x_right_coll &&
                    get_right_wall_xpos(dead->room, chomped_col, dead->row) > dead->x_left_coll) {
                chomped = 1;
                break;
            }
        }
    }
    if (chomped == 0) {
        return;
    }
    // guards' chomped sprites are in the wrong direction
    dead->direction = ~dead->direction;
    short delta_x = 7 - !(dead->direction);
    if (dead->direction == dir_FF_left) {
        delta_x = -delta_x;
    }
    dead->frame2 = frame_178_chomped;
    dead->x = x_bump[chomped_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
    dead->x += delta_x;
    // temple has different chompers
    if (is_temple()) {
        dead->x -= 2;
    }
    if (alive < 0) {
        char_type2* slot = get_dead_guard_slot(drawn_room - 1);
        if (slot != NULL) {
            slot->room = dead->room;
            slot->direction = dead->direction;
            slot->row = dead->row;
            slot->x = dead->x;
            slot->color = dead->color;
            slot->frame2 = dead->frame2;
            leave_outbound_guard(slot->room - 1, slot->frame, slot->direction, slot->x, slot->row, slot->alt_frame, slot->color);
        }
        inactive_guard[drawn_room - 1].room = 0;
    }
    set_sound_pos(dead->room, chomped_col, dead->row);
    play_sound(sound_46_chomped);
}

// custom logic - check for chomped dead or inactive guards
void check_dead_or_inactive_guards_chomped(void) {
    word room_minus_1;
    room_minus_1 = drawn_room - 1;
    check_dead_or_inactive_guard_chomped(&inactive_guard[room_minus_1], -1);
    check_dead_or_inactive_guard_chomped(&dead_guard1[room_minus_1], 0);
    check_dead_or_inactive_guard_chomped(&dead_guard2[room_minus_1], 0);
    check_dead_or_inactive_guard_chomped(&dead_guard3[room_minus_1], 0);
}

// seg004:0A10
int dist_from_wall_forward(byte tiletype) {
    short type;
    if (tiletype == tiles_4_gate && !can_bump_into_gate()) {
        return -1;
    } else {
        coll_tile_left_xpos = x_bump[tile_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
        type = wall_type(tiletype);
        if (type == wall_type_0_none) return -1;
        if (Char.direction < dir_0_right) {
            // looking left
            return char_x_left_coll - (coll_tile_left_xpos + 13 - wall_dist_from_right[type]);
        } else {
            // looking right
            return wall_dist_from_left[type] + coll_tile_left_xpos - char_x_right_coll;
        }
    }
}

// seg004:0A7B
int dist_from_wall_behind(byte tiletype) {
    short type;
    type = wall_type(tiletype);
    if (type == wall_type_0_none) {
        return 99;
    } else {
        if (Char.direction >= dir_0_right) {
            // looking right
            return char_x_left_coll - (coll_tile_left_xpos + 13 - wall_dist_from_right[type]);
        } else {
            // looking left
            return wall_dist_from_left[type] + coll_tile_left_xpos - char_x_right_coll;
        }
    }
}
