/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"
#include "math.h"

// seg006:0006
int get_tile(int room, int col, int row) {
    curr_room = room;
    tile_col = col;
    tile_row = row;
    curr_room = find_room_of_tile();
    // bugfix: check_chomped_kid may call with room = -1
    if (curr_room > 0) {
        get_room_address(curr_room);
        curr_tilepos = tbl_line[tile_row] + tile_col;
        curr_tile2 = curr_room_tiles[curr_tilepos] & 0x1F;
    } else {
        // wall in room 0
        curr_tile2 = tiles_20_wall;
    }
    return curr_tile2;
}

// seg006:005D
int find_room_of_tile(void) {
again:
    // Check tile_row < 0 first, this way the prince can grab a ledge at the bottom right corner of a room with no room below.
    // Details: https://forum.princed.org/viewtopic.php?p=30410#p30410
    if (tile_row < 0) {
        tile_row += 3;
        if (curr_room) {
            curr_room = level.roomlinks[curr_room - 1].up;
        }
        //find_room_of_tile();
        goto again;
    }
    if (tile_col < 0) {
        tile_col += 10;
        if (curr_room) {
            curr_room = level.roomlinks[curr_room - 1].left;
        }
        //find_room_of_tile();
        goto again;
    }
    if (tile_col >= 10) {
        tile_col -= 10;
        if (curr_room) {
            curr_room = level.roomlinks[curr_room - 1].right;
        }
        //find_room_of_tile();
        goto again;
    }
    // if (tile_row < 0) was here originally
    if (tile_row >= 3) {
        tile_row -= 3;
        if (curr_room) {
            curr_room = level.roomlinks[curr_room - 1].down;
        }
        //find_room_of_tile();
        goto again;
    }
    return curr_room;
}

// seg006:00EC
int get_tilepos(int tile_col, int tile_row) {
    if (tile_row < 0) {
        return -(tile_col + 1);
    } else if (tile_row >= 3 || tile_col >= 10 || tile_col < 0) {
        return 30;
    } else {
        return tbl_line[tile_row] + tile_col;
    }
}

// seg006:0124
int get_tilepos_nominus(int tile_col, int tile_row) {
    short tile_pos;
    tile_pos = get_tilepos(tile_col, tile_row);
    if (tile_pos < 0) return 30;
    else return tile_pos;
}

// seg006:0144
void load_fram_det_col(void) {
    load_frame();
    determine_col();
}

// seg006:014D
void determine_col(void) {
    Char.curr_col = get_tile_div_mod_m7(dx_weight());
}

void get_frame_internal(const frame_type frame_table[], int frame, const char* frame_table_name, int count) {
    if (frame >= 0 && frame < count) {
        cur_frame = frame_table[frame];
    } else {
        printf("Tried to use %s[%d], not in 0..%d\n", frame_table_name, frame, count - 1);
        static const frame_type blank_frame = {255, 0, 0, 0, 0};
        cur_frame = blank_frame;
    }
}
#define get_frame(frame_table, frame) get_frame_internal(frame_table, frame, #frame_table, COUNT(frame_table))

// seg006:015A
void load_frame(void) {
    short frame;
    short add_frame;
    frame = Char.frame;
    add_frame = 0;
    switch (Char.charid) {
        case charid_0_kid:
        case charid_1_shadow: // custom logic - shadow always uses kid table
        case charid_24_mouse:
loc_use_kid_table:
            get_frame(frame_table_kid, frame);
            break;
        case charid_2_guard:
            if (settings->tbl_guard_type[current_level] == guard_type_kid) { // kid as guard
                goto loc_use_kid_table;
            }
        // fallthrough!
        case charid_4_skeleton:
            if (frame >= 102 && frame < 107) add_frame = 70;
            get_frame(frame_tbl_guard, frame + add_frame - 149);
            break;
        case charid_5_princess:
        case charid_6_vizier:
            get_frame(frame_tbl_cuts, frame);
            break;
    }
}
#undef get_frame

// seg006:01F5
short dx_weight(void) {
    sbyte delta_x;
    delta_x = cur_frame.dx - (cur_frame.flags & FRAME_WEIGHT_X);
    return char_dx_forward(delta_x);
}

// seg006:0213
int char_dx_forward(int delta_x) {
    if (Char.direction < dir_0_right) {
        delta_x = -delta_x;
    }
    return delta_x + Char.x;
}

// seg006:0234
int obj_dx_forward(int delta_x) {
    if (obj_direction < dir_0_right) {
        delta_x = -delta_x;
    }
    obj_x += delta_x;
    return obj_x;
}

// custom logic - make guard notice kid
void make_guard_notice(bool ignore_lightweight) {
    if (ignore_lightweight) {
        if (feather_fall_timer > 0) {
            return;
        }
        if (Char.charid == charid_0_kid && is_kid_shadow) {
            return;
        }
    }
    is_guard_notice = 1;
}

// custom logic - jumping off balcony
void jump_off_balcony(void) {
    Char.y = OUT_OF_BOUNDS_Y - 1; // prevent room switching
    Char.direction = dir_56_none;
    Char.curr_seq = seqtbl_offset_get_seq(seq_127_dead);
    feather_fall_timer = 0;
    hitp_delta = -hitp_curr;
}

// custom logic - rooms to teleport on level 6A
int teleport_rooms[] = { 4, 5, 13, 14, 18 };

// custom logic - teleport between rooms on level 6A
void teleport(void) {
    int dest_room, dest_tilepos;
    int rnd_index;
    dest_tilepos = 11;
    dest_room = 13;
    if (custom_seed != 0) {
        random_seed = custom_seed;
        seed_was_init = 1;
    }
    rnd_index = prandom(4);
    dest_room = teleport_rooms[rnd_index];
    if (dest_room == Char.room) { // the destination room cannot be the same
        if (rnd_index == 0) {
            dest_room = teleport_rooms[rnd_index + 1];
        } else {
            dest_room = teleport_rooms[rnd_index - 1];
        }
    }
    if (get_tile(1, 9, 1) == tiles_4_gate && curr_room_modif[curr_tilepos] < 0xFF) {
        curr_room_modif[curr_tilepos] = 0xFF;
    }
    Char.room = dest_room;
    Char.curr_col = dest_tilepos % 10;
    Char.curr_row = dest_tilepos / 10;
    Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET + CHAR_7_PX_OFFSET; // center on the destination teleport.
    Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
    next_room = Char.room;
    clear_coll_rooms(); // without this, the prince will sometimes end up at the wrong place.
    leave_guard();
    seqtbl_offset_char(seq_5_turn);
    play_sound(sound_45_jump_through_mirror);
    show_black_screen(6, 1);
}

// custom logic - show black screen for the provided number of ticks
void show_black_screen(int ticks, byte redraw_hp) {
    draw_rect(&screen_rect, color_0_black);
    update_screen();
    delay_ticks(ticks);
    if (redraw_hp) {
        draw_kid_hp(hitp_curr, hitp_max);
    }
}

// seg006:0254
void play_seq(void) {
    for (;;) {
        byte seq_value = *(seqtbl + Char.curr_seq++);
        switch (seq_value) {
            case SEQ_DX: // dx
                Char.x = char_dx_forward(*(seqtbl + Char.curr_seq++));
                break;
            case SEQ_DY: // dy
                Char.y += *(seqtbl + Char.curr_seq++);
                break;
            case SEQ_FLIP: // flip
                Char.direction = ~Char.direction;
                break;
            case SEQ_JMP_IF_SHAD_FEATHER:
                if (Char.action == actions_2_hang_climb) {
                    // custom logic - shadow or kid in feather mode can hang of an edge forever
                    if (is_kid_shadow || feather_fall_timer > 0) {
                        Char.curr_seq = hang1; // hanging instead of dropping
                        break;
                    }
                    // custom logic - negate dx(3) from "climbdrop" to avoid being able to grab a floor tile 2 rows below
                    Char.x = char_dx_forward(-3);
                } else if (Char.action == actions_3_in_midair) {
                    if (!feather_fall_timer) {
                        ++Char.curr_seq;
                        ++Char.curr_seq;
                        break;
                    }
                } else if (Char.action == actions_6_hang_straight) {
                    // custom logic - shadow or kid in feather mode can hang of an edge forever
                    if (is_kid_shadow || feather_fall_timer > 0) {
                        Char.curr_seq = hangstraight1; // hanging straight instead of dropping
                        break;
                    } else if (tile_is_floor(get_tile_behind_char())) {
                        Char.x = char_dx_forward(-7);
                        Char.curr_seq = hangdrop; // drop at the same level
                        break;
                    }
                }
            // fallthrough!
            case SEQ_JMP: // jump
                Char.curr_seq = *(const word*)(seqtbl + Char.curr_seq);
                break;
            case SEQ_UP: // up
                --Char.curr_row;
                start_chompers();
                break;
            case SEQ_DOWN: // down
                // custom logic - follow kid when he climbs down near the guard
                set_kid_droppedout();
                inc_curr_row();
                start_chompers();
                break;
            case SEQ_ACTION: // action
                Char.action = *(seqtbl + Char.curr_seq++);
                break;
            case SEQ_SET_FALL: // set fall
                Char.fall_x = *(seqtbl + Char.curr_seq++);
                Char.fall_y = *(seqtbl + Char.curr_seq++);
                break;
            case SEQ_KNOCK_UP: // knock up
                // custom logic - store char type and row
                if (tile_is_floor(get_tile_above_char()) ||
                        curr_tile2 == tiles_20_wall ||
                        (distance_to_edge_weight() >= 8 &&
                         (tile_is_floor(get_tile_behind_above_char()) ||
                          curr_tile2 == tiles_20_wall))
                   ) {
                    knock = Char.curr_row + 1;
                    knock |= Char.charid << 4;
                }
                break;
            case SEQ_KNOCK_DOWN: // knock down
                // custom logic - hangdrop can cause tiles loose to fall
                if (tile_is_floor(get_tile_at_char())) {
                    make_guard_notice(true);
                    // custom logic - store char type and row
                    knock = Char.curr_row + 2;
                    knock |= Char.charid << 4;
                }
                break;
            case SEQ_SOUND: // sound
                switch (*(seqtbl + Char.curr_seq++)) {
                    case SND_SILENT: // no sound actually played, but guards still notice the kid (used in a fight)
                        if (can_guard_see_kid >= 2) {
                            make_guard_notice(false);
                        }
                        break;
                    case SND_FOOTSTEP: // feet
                        // custom logic - in feather fall mode guards do not notice kid
                        if (!feather_fall_timer &&
                                tile_is_floor(get_tile_at_char()) &&
                                blink_timer != 20 /* fell off balcony on level 6 */) {
                            set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
                            if (want_positional_sound) { // reduce the surround effect
                                want_sound_dir_x /= 2.0f;
                            }
                            play_sound(sound_23_footstep); // footstep
                        }
                        make_guard_notice(true);
                        break;
                    case SND_BUMP: // bump
                        bumped_sound();
                        break;
                    case SND_DRINK: // drink
                        play_sound(sound_18_drink);
                        break;
                    case SND_DRAW_SWORD: { // draw sword
                        if (Char.charid != charid_0_kid) {
                            set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
                        }
                        play_sound(sound_19_draw_sword);
                        // custom logic - do not notice shadow with a drawn sword
                        if (Char.charid != charid_0_kid || !is_kid_shadow || is_shadow_flame) {
                            make_guard_notice(false);
                        }
                        break;
                    }
                    case SND_LEVEL: // level
                        feather_fall_timer = 0;
                        if (!is_sound_on) break;
                        // custom logic - various end-of-level music
                        if (current_level > 0 && current_level < 14) {
                            if (current_level == 10 && Char.room == 20 && !shadow_initialized) {
                                break;
                            }
                            stop_all_sounds();
                            // allows to skip end of level music with shift/enter
                            end_of_level_music = 1;
                            // play different music on level 10
                            if (is_positional_audio && current_level == 10) {
                                play_sound(sound_73_shadow_level_music); // PoP2 end of level
                            } else {
                                play_sound(sound_41_end_level_music); // end level
                            }
                        } else if (current_level == 0 && play_demo_level) {
                            stop_all_sounds();
                            end_of_level_music = 1;
                            play_sound(sound_41_end_level_music);
                        }
                        break;
                }
                break;
            case SEQ_END_LEVEL: // end level
                // custom logic - teleport/balcony
                if (*(seqtbl + Char.curr_seq++) == tiles_23_balcony_left) {
                    if (current_level == 6 && falling_exit_checkpoint < 4) {
                        teleport();
                    } else {
                        jump_off_balcony();
                    }
                } else {
                    stop_ambient_music();
                    // custom logic - handle demo level
                    if (current_level == 0) {
                        if (play_demo_level) {
                            word key;
                            while (check_sound_playing() && (key = key_test_quit()) == 0) {
                                if (key) break;
                                if (key_states[SDL_SCANCODE_LSHIFT] & keystate_held || key_states[SDL_SCANCODE_RSHIFT] & keystate_held) {
                                    key_states[SDL_SCANCODE_LSHIFT] = key_states[SDL_SCANCODE_RSHIFT] = 0;
                                    break;
                                }
                                idle();
                                process_trobs();
                                draw_game_frame();
                                do_simple_wait(timer_1);
                            }
                            clear_kbd_buf();
                        }
                        draw_rect(&screen_rect, color_0_black);
                        if (!play_demo_level) {
                            start_level = 0;
                            quotes_needed = 0;
                            play_demo_level = 1;
                        } else {
                            start_level = -1;
                            quotes_needed = 1;
                        }
                        start_game();
                        break;
                    }
                    // custom logic - guards kill kid in room 20 of level 10 after entering the level door
                    if (current_level == 10 && Char.room == 20 && !shadow_initialized) {
                        if (upside_down_timer) {
                            upside_down_timer = 0;
                            need_full_redraw = 1;
                        }
                        if (is_positional_audio) {
                            play_sound(sound_72_guards_kill_kid);
                        }
                        level_restart_timer = 32;
                        break;
                    }
                    ++next_level;
                }
                break;
            case SEQ_GET_ITEM: // get item
                if (*(seqtbl + Char.curr_seq++) == 1) {
                    proc_get_object();
                }
                break;
            case SEQ_DIE: // nop
                break;
            default:
                // custom logic - Jaffar walks funny on the way back
                if (seq_value == frame_51_vizier_walk_4a && Char.charid == charid_6_vizier && Char.direction == dir_0_right) {
                    seq_value = frame_87_vizier_walk_4b;
                }
                // custom logic - spiking frame is different when kid is facing to the right
                if (seq_value == frame_100_spiking_hanging && Char.charid == charid_0_kid && Char.direction == dir_0_right) {
                    ++seq_value;
                }
                // custom logic - reset kid's speed as a mouse
                if (Char.charid == charid_0_kid &&
                        Char.frame >= frame_186_mouse_1 &&
                        Char.frame <= frame_188_mouse_stand &&
                        (seq_value < frame_186_mouse_1 || seq_value > frame_188_mouse_stand)) {
                    Char.fall_x = 0;
                    if (Char.action == actions_5_bumped) { // kid bumped into a wall
                        flash_color = color_7_lightgray;
                        flash_time = 6;
                    }
                }
                Char.frame = seq_value;
                return;
        }
    }
}

// custom logic - prevents inactive guard from striking kid too quickly
byte inactive_refrac_timer = 0;

// custom logic - is inactive guard in a ready position?
bool is_inactive_guard_ready(char_type2 inactive) {
    return ((inactive.frame >= frame_150_parry && inactive.frame <= frame_171_stand_with_sword) ||
            (inactive.frame2 >= frame_150_parry && inactive.frame2 <= frame_171_stand_with_sword));
}

// custom logic - sets inactive guard in a ready position
void reset_inactive_guard(char_type2* inactive) {
    if (can_inactive_guard_see_kid == 2 &&
            Kid.curr_row == inactive->row &&
            (!is_kid_shadow || Guard.charid == charid_4_skeleton)) {
        inactive->frame2 = frame_171_stand_with_sword;
    } else {
        inactive->frame2 = frame_166_stand_inactive;
    }
    is_inactive_guard_notice = 0;
}

// custom logic - plays inactive sequence table
void play_inactive_seq(void) {
    byte seq_value;
    char_type2* inactive;
    inactive = &inactive_guard[drawn_room - 1];
    if (inactive->room != drawn_room || inactive->direction == dir_56_none) {
        inactive_seq_ptr = inactive_seq_stand;
        return;
    }
    if (inactive->room != Kid.room &&
            (inactive_seq_ptr < inactive_seq_dropdead || inactive_seq_ptr >= inactive_seq_dropdead + 8)) {
        inactive_seq_ptr = inactive_seq_stand;
        return;
    }
    if (inactive_seq_ptr == inactive_seq_stand) {
        reset_inactive_guard(inactive);
        return;
    }
    for (;;) {
        seq_value = *(inactive_seqtbl + inactive_seq_ptr++);
        switch (seq_value) {
            case inactive_seq_dead: {
                char_type2* slot = get_dead_guard_slot(inactive->room - 1);
                if (slot != NULL) {
                    slot->room = inactive->room;
                    slot->direction = inactive->direction;
                    slot->row = inactive->row;
                    slot->x = inactive->x;
                    slot->color = inactive->color;
                    slot->frame2 = inactive->frame;
                    slot->alt_frame = 0x05;
                    leave_outbound_guard(slot->room - 1, slot->frame, slot->direction, slot->x, slot->row, slot->alt_frame, slot->color);
                }
                inactive_guard[drawn_room - 1].room = 0;
                return;
            }
            case inactive_seq_flip:
                inactive->direction = ~inactive->direction;
                break;
            case inactive_seq_dx_fwd:
                seq_value = *(inactive_seqtbl + inactive_seq_ptr++);
                if (inactive->direction == dir_FF_left) {
                    seq_value = -seq_value;
                }
                inactive->x += seq_value;
                break;
            case inactive_seq_kid_hurt:
                autocontrol_inactive_guard_hurt_kid(*inactive);
                return;
            case inactive_seq_snd:
                seq_value = *(inactive_seqtbl + inactive_seq_ptr++);
                play_sound(seq_value);
                break;
            case inactive_seq_stand:
                reset_inactive_guard(inactive);
                inactive_seq_ptr = inactive_seq_stand;
                return;
            default:
                inactive->frame2 = seq_value;
                return;
        }
    }
}

// custom logic - sets inactive sequence pointer
void seqtbl_inactive_char(int seq_index) {
    inactive_seq_ptr = seq_index;
}

// seg006:03DE
int get_tile_div_mod_m7(int xpos) {
    return get_tile_div_mod(xpos - 7);
}

// seg006:03F0
int get_tile_div_mod(int xpos) {
    int x = xpos - 58; // first 58 pixels are allocated to the invisible 5 columns on the left
    int xl = x % 14;
    int xh = x / 14;
    if (xl < 0) {
        // Integer division rounds towards zero, but we want to round down.
        --xh;
        // Remainder returns a negative number if x is negative, but we want 0 <= xl < 14.
        xl += 14;
    }
    obj_xl = xl;
    return xh;
}

// seg006:0433
int y_to_row_mod4(int ypos) {
    return (ypos + 60) / 63 % 4 - 1;
}

// seg006:044F
void loadkid(void) {
    Char = Kid;
}

// seg006:0464
void savekid(void) {
    Kid = Char;
}

// seg006:0479
void loadshad(void) {
    Char = Guard;
}

// seg006:048E
void saveshad(void) {
    Guard = Char;
}

// seg006:04A3
void loadkid_and_opp(void) {
    loadkid();
    Opp = Guard;
}

// seg006:04BC
void savekid_and_opp(void) {
    savekid();
    Guard = Opp;
}

// seg006:04D5
void loadshad_and_opp(void) {
    loadshad();
    Opp = Kid;
}

// seg006:04EE
void saveshad_and_opp(void) {
    saveshad();
    Kid = Opp;
}

// seg006:0507
void reset_obj_clip(void) {
    obj_clip_left = 0;
    obj_clip_top = 0;
    obj_clip_right = 320;
    obj_clip_bottom = 192;
}

// seg006:051C
void x_to_xh_and_xl(int xpos, sbyte *xh_addr, sbyte *xl_addr) {
    // custom logic - the original logic had some strange arithmetic for negative xpos values
    // that did not produce correct results for values divisible by -8
    *xh_addr = xpos >> 3;
    *xl_addr = xpos & 0x07;
}

// seg006:057C
void fall_accel(void) {
    if (Char.action == actions_4_in_freefall) {
        if (feather_fall_timer) {
            ++Char.fall_y;
            // custom logic - feather mode affects forward acceleration like in PoP2
            if (Char.fall_x > 1) Char.fall_x = 1;
            if (Char.fall_y > 4) Char.fall_y = 4;
            // custom logic - shadow falls softer than kid
        } else if (Char.charid == charid_0_kid && is_kid_shadow) {
            Char.fall_y += 2;
            if (Char.fall_y > 16) Char.fall_y = 16;
        } else {
            Char.fall_y += 3;
            if (Char.fall_y > 33) Char.fall_y = 33;
        }
    }
}

// seg006:05AE
void fall_speed(void) {
    Char.y += Char.fall_y;
    if (Char.action == actions_4_in_freefall && super_jump_fall == 0) {
        Char.x = char_dx_forward(Char.fall_x);
        load_fram_det_col();
    }
}

// seg006:05CD
void check_action(void) {
    short frame;
    short action;
    action = Char.action;
    frame = Char.frame;
    // custom logic - kid can grab a tile while jumping
    if (action == actions_1_run_jump && control_shift < control_0_released && check_grab_run_jump()) {
        return;
    }
    // frame 109: crouching
    if (action == actions_6_hang_straight ||
            action == actions_5_bumped
       ) {
        if (frame == frame_109_crouch ||
                (frame >= frame_110_stand_up_from_crouch_1 && frame <= frame_119_stand_up_from_crouch_10)) {
            check_on_floor();
        }
    } else if (action == actions_4_in_freefall) {
        do_fall();
    } else if (action == actions_3_in_midair) {
        // frame 102..106: start fall + fall
        if (frame >= frame_102_start_fall_1 && frame < frame_106_fall) {
            check_grab();
        }
    } else if (action != actions_2_hang_climb) {
        check_on_floor();
    }
    // custom logic - do not keep sword drawn after falling/hanging
    if (Char.charid == charid_0_kid &&
            keep_sword_drawn &&
            (action == actions_2_hang_climb ||
             action == actions_3_in_midair ||
             action == actions_4_in_freefall ||
             action == actions_6_hang_straight)) {
        keep_sword_drawn = false;
    }
}

// custom logic - check if there is an outbound guard that needs to start falling
void check_falling_guard(void) {
    // main skeleton fall on level 3 is delayed
    if (current_level == 3 && falling_bridge_timer > 0) {
        return;
    }
    if (falling_guard.room != 0 && falling_guard.direction != dir_56_none) {
        do_guard_fall();
    }
}

// seg006:0628
int tile_is_floor(int tiletype) {
    // custom logic - prevents kid from falling through fake walls
    if (is_fake_wall() == 1) {
        return 1;
    }
    // custom logic - prevents kid from falling through the special wall on level 4, room 20
    if (current_level == 4 &&
            Char.charid == charid_0_kid &&
            Char.room == 20 &&
            Char.direction == dir_FF_left &&
            Char.curr_row == 1 &&
            tiletype == tiles_20_wall) {
        return 1;
    }
    switch (tiletype) {
        case tiles_0_empty:
        case tiles_9_bigpillar_top:
        case tiles_12_doortop:
        case tiles_20_wall:
        case tiles_26_lattice_down: // custom logic - also covers dungeon torches without a floor
        case tiles_27_lattice_small:
        case tiles_28_lattice_left:
        case tiles_29_lattice_right:
            return 0;
        // custom logic - spikes in the temple with control bits set to 010 do not have a floor
        case tiles_2_spike:
            if (curr_tile2 == tiles_2_spike &&
                    is_temple() &&
                    get_tile_ctrl_bits(curr_room, get_tilepos(tile_col, tile_row), 3) == 2) {
                return 0;
            }
            return 1;
        default:
            return 1;
    }
}

// custom logic - checks if the tile is an empty tile
bool tile_is_empty(int tiletype) {
    return tiletype == tiles_0_empty || (tiletype == tiles_26_torch_empty && settings->tbl_level_type[current_level] == DUNGEON_LEVEL);
}

// custom logic - returns the expected character tile column number and
// returning the correct column when kid is climbing while facing left
short char_col_climbing_left(void) {
    short char_col = Char.curr_col;
    if (Char.charid == charid_0_kid &&
            Char.direction == dir_FF_left &&
            Char.frame >= frame_90_hanging_4 &&
            Char.frame < frame_99_hanging_13) {
        ++char_col;
    }
    return char_col;
}

// custom logic - vertical spikes can harm characters in additional scenarios
void check_spiked_in_temple(void) {
    short harmful;
    short frame;
    frame = Char.frame;
    if (get_tile(Char.room, char_col_climbing_left(), Char.curr_row) == tiles_2_spike && Char.room == curr_room) {
        harmful = is_spike_harmful();
        // vertical spike is harmful right away
        if (harmful < 2 &&
                (Char.charid != charid_0_kid || !is_kid_shadow) &&
                resurrect_time == 0 &&
                find_trob_at_loc(curr_room, curr_tilepos) >= 0 &&
                ((get_tile_ctrl_bits(curr_room, curr_tilepos, 3) == 2 && curr_room_modif[curr_tilepos] != 0xFF) ||
                 curr_room_modif[curr_tilepos] == 0)
           ) {
            harmful = 2;
        }
        if (harmful >= 2 &&
                (((frame >= frame_7_run && frame < frame_15_stand) ||
                  (frame >= frame_34_start_run_jump_1 && frame < frame_40_running_jump_1) ||
                  (frame >= frame_59_runturn && frame <= frame_65_runturn)) ||
                 ((frame == frame_43_running_jump_4 || frame == frame_26_standing_jump_11) && harmful != 0) ||
                 ((frame > frame_0 && frame < frame_7_run) ||
                  (frame >= frame_53_runturn && frame < frame_59_runturn) ||
                  (frame >= frame_109_crouch && frame <= frame_119_stand_up_from_crouch_10) ||
                  (Char.direction == dir_FF_left && frame == frame_79_jumphang) ||
                  (Char.direction == dir_FF_left && frame >= frame_91_hanging_5 && frame < frame_99_hanging_13) ||
                  (Char.direction == dir_0_right && frame >= frame_156_post_strike_2 && frame <= frame_158_stand_with_sword) ||
                  (Char.direction == dir_0_right && frame >= frame_170_stand_with_sword && frame <= frame_171_stand_with_sword) ||
                  (Char.action == actions_5_bumped && Char.sword == sword_0_sheathed && frame != frame_83_hangdrop_3) ||
                  (frame >= frame_102_start_fall_1 && frame <= frame_106_fall)))
           ) {
            if (frame >= frame_58_runturn && frame <= frame_65_runturn) {
                Char.direction = ~Char.direction;
            }
            spiked();
        }
    }
}

// seg006:0658
void check_spiked(void) {
    short harmful;
    short frame;
    bool run_turn = false;
    // custom logic - vertical spikes can harm characters in additional scenarios
    if (is_temple()) {
        check_spiked_in_temple();
        return;
    }
    frame = Char.frame;
    if (get_tile(Char.room, Char.curr_col, Char.curr_row) == tiles_2_spike) {
        harmful = is_spike_harmful();
        if ((harmful >= 2 &&
                ((frame >= frame_7_run && frame < frame_15_stand) ||
                 (frame >= frame_34_start_run_jump_1 && frame < frame_40_running_jump_1) ||
                 // custom logic - spiked when running and turning
                 (run_turn = (frame >= frame_59_runturn && frame <= frame_65_runturn)) ||
                 // custom logic - fell on spikes
                 (frame >= frame_107_fall_land_1 && frame <= frame_108_fall_land_2) ||
                 // custom logic - bumped the wall next to spikes
                 (Char.action == actions_5_bumped && frame >= frame_50_turn && frame <= frame_52_turn))) ||
                ((frame == frame_43_running_jump_4 || frame == frame_26_standing_jump_11) && harmful != 0)
           ) {
            if (run_turn) {
                if (distance_to_edge_weight() <= 7) {
                    Char.direction = ~Char.direction;
                } else {
                    return;
                }
            }
            spiked();
            return;
        }
    }
    // custom logic - if kid falls dead on a spike tile facing right, activate spikes
    if (Char.charid == charid_0_kid &&
            Char.action == actions_5_bumped &&
            Char.frame == frame_185_dead &&
            Char.direction == dir_0_right &&
            get_tile(Char.room, MIN(char_col_left + 1, char_col_right), Char.curr_row) == tiles_2_spike &&
            is_spike_harmful() != 0) {
        stop_sounds();
        spiked();
    }
}

// seg006:06BD
int take_hp(int count) {
    word dead;
    dead = 0;
    if (Char.charid == charid_0_kid) {
        // custom logic - kid playing as shadow does not die except for levels 12/12B and 13
        if (is_kid_shadow &&
                current_level != 12 &&
                current_level != 13 &&
                current_level != 14) {
            return dead;
        }
        if (count >= hitp_curr) {
            hitp_delta = -hitp_curr;
            dead = 1;
        } else {
            hitp_delta = -count;
        }
    } else {
        if (count >= guardhp_curr) {
            guardhp_delta = -guardhp_curr;
            dead = 1;
        } else {
            guardhp_delta = -count;
        }
    }
    return dead;
}

// seg006:070D
int get_tile_at_char(void) {
    return get_tile(Char.room, Char.curr_col, Char.curr_row);
}

// seg006:0723
void set_char_collision(void) {
    image_type* image = get_image(obj_chtab, obj_id, 0);
    if (image == NULL) {
        char_width_half = 0;
        char_height = 0;
    } else {
        char_width_half = (image->/*width*/w + 1) / 2;
        char_height = image->/*height*/h;
    }
    char_x_left = obj_x / 2 + 58;
    if (Char.direction >= dir_0_right) {
        char_x_left -= char_width_half;
    }
    char_x_left_coll = char_x_left;
    char_x_right_coll = char_x_right = char_x_left + char_width_half;
    char_top_y = obj_y - char_height + 1;
    if (char_top_y >= 192) {
        char_top_y = 0;
    }
    char_top_row = y_to_row_mod4(char_top_y);
    char_bottom_row = y_to_row_mod4(obj_y);
    if (char_bottom_row == -1) {
        char_bottom_row = 3;
    }
    char_col_left = MAX(get_tile_div_mod(char_x_left), 0);
    char_col_right = MIN(get_tile_div_mod(char_x_right), 9);
    if (cur_frame.flags & FRAME_THIN) {
        // "thin" this frame for collision detection
        char_x_left_coll += 4;
        char_x_right_coll -= 4;
    }
}

short dx_weight_engarde(void) {
    sbyte delta_x;
    sbyte max_delta_x = -13;
    delta_x = cur_frame.dx - (cur_frame.flags & FRAME_WEIGHT_X);
    if (delta_x < max_delta_x) {
        delta_x = max_delta_x;
    }
    return char_dx_forward(delta_x);
}

// seg006:0815
void check_on_floor(void) {
    if (cur_frame.flags & FRAME_NEEDS_FLOOR) {
        // custom logic - frame 153 should not need a floor like frames 151, 152 and 154
        // It can cause guard/kid to fall through a floor during a strike.
        // But we do enable it on the falling bridge as it can look unnatural.
        if (Char.frame == frame_153_strike_3 &&
                (falling_bridge_timer == 0 || tile_is_floor(get_tile(Char.room, get_tile_div_mod_m7(Char.x), Char.curr_row)))
           ) {
            return;
        }
        // custom logic - Jaffar's vision should not fall through the floor
        if (Char.charid != charid_0_kid && settings->tbl_guard_type[current_level] == guard_type_Jaffar_vision) {
            return;
        }
        // custom logic - shadow on level 9 should not fall walking backwards
        if (current_level == 9 &&
                Char.charid == charid_1_shadow &&
                Char.room == 1 &&
                Char.curr_row == 1) {
            return;
        }
        // custom logic - bowing guard column can be miscalculated due to a large dx weight
        if (Char.charid == charid_2_guard &&
                Char.frame == frame_202_bowing &&
                tile_is_floor(get_tile(Char.room, get_tile_div_mod_m7(Char.x), Char.curr_row))) {
            return;
        }
        if (get_tile_at_char() == tiles_20_wall) {
            in_wall();
        }
        if (!tile_is_floor(curr_tile2)) {
            int distance = distance_to_edge_weight();
            // custom logic - bump kid back at the end of a jump when landing on an edge
            if (((Char.frame == frame_7_run && distance == 0) ||
                    Char.frame == frame_27_standing_jump_12 ||
                    Char.frame == frame_44_running_jump_5) &&
                    feather_fall_timer == 0 &&
                    !is_kid_shadow &&
                    tile_is_floor(get_tile(Char.room, get_tile_div_mod_m7(Char.x), Char.curr_row))) {
                if (curr_tile2 == tiles_11_loose) {
                    make_loose_fall(1);
                }
                if (Char.frame == frame_7_run) {
                    Char.x = char_dx_forward(1);
                    determine_col();
                    return;
                }
                bumped_sound();
                Char.x = char_dx_forward(distance_to_edge(Char.x) - 14);
                determine_col();
                return;
                // bump kid forward when he stops running at the end of an edge
            } else if (Char.frame >= frame_53_runturn &&
                       Char.frame <= frame_55_runturn &&
                       distance >= 10 &&
                       control_forward != control_m1_held &&
                       control_backward == control_0_released &&
                       tile_is_floor(get_tile(Char.room, get_tile_div_mod_m7(Char.x) + dir_behind[Char.direction + 1], Char.curr_row))) {
                Char.x = char_dx_forward(distance - 11);
                determine_col();
            }
            // custom logic - start falling during a run/turn sooner
            if (Char.frame == frame_59_runturn) {
                // change the direction
                Char.direction = ~Char.direction;
                Char.x = char_dx_forward(distance_to_edge(Char.x) - 1);
            }
            // Special event: floors appear
            // custom logic - mouse can create tiles on air
            if (current_level == settings->mouse_2_level &&
                    Char.charid == charid_24_mouse &&
                    Char.room == settings->mouse_2_room) {
                curr_room_tiles[curr_tilepos] = tiles_1_floor;
                if (Char.room == curr_room) {
                    set_wipe(curr_tilepos, 1);
                    set_redraw_full(curr_tilepos, 1);
                }
                ++curr_tilepos;
                if (curr_tilepos == 10) { // redraw tile 0 in the drawn room instead of tile 10 in the room on the left
                    curr_tilepos = 0;
                }
                set_wipe(curr_tilepos, 1);
                set_redraw_full(curr_tilepos, 1);
            } else {
                if (Char.frame >= frame_110_stand_up_from_crouch_1 && Char.frame <= frame_119_stand_up_from_crouch_10) {
                    // We need to prevent the Kid from stepping off a ledge accidentally while standing up.
                    // (This can happen because the "standing up" frames now require a floor.)
                    // --> Cancel the fall, if the tile at dx=2 behind the kid is a valid floor.
                    int col = get_tile_div_mod_m7(dx_weight() + back_delta_x(2));
                    if (tile_is_floor(get_tile(Char.room, col, Char.curr_row))) {
                        return;
                    }
                }
                start_fall();
            }
        }
    }
}

// seg006:08B9
void start_fall(void) {
    short frame;
    short char_dist;
    short distance;
    short tile_infront, tile_behind, wall_infront;
    bool floor_infront;
    word seq_id;
    frame = Char.frame;
    // custom logic - guard should not fall down until the timer is equal to 1
    if ((Char.charid == charid_2_guard || Char.charid == charid_4_skeleton) && guard_dropping_down > 1) {
        return;
    }
    Char.sword = sword_0_sheathed;
    distance = distance_to_edge_weight();
    char_dist = char_opp_dist();
    // custom logic - kid is about to be striked with the sword, do not fall yet
    if (can_guard_see_kid == 2 &&
            Char.charid == charid_0_kid &&
            Char.sword == sword_0_sheathed &&
            Opp.frame == frame_153_strike_3 &&
            char_dist >= 12 &&
            char_dist < 29) {
        if (distance < 5) {
            Char.x = char_dx_forward(distance - 5);
        }
        return;
    }
    // custom logic - check for walls at the end of a running jump
    tile_infront = get_tile_infrontof_char();
    wall_infront = wall_type2(tile_infront);
    wall_infront = (wall_infront == wall_type_4_wall || (Char.direction == dir_FF_left && wall_infront == wall_type_1_right));
    floor_infront = tile_is_floor(tile_infront);
    tile_behind = get_tile(Char.room, (Char.direction == dir_FF_left ? char_col_right : char_col_left), Char.curr_row);
    if (frame == frame_44_running_jump_5 && wall_infront && distance < 8) {
        Char.x = char_dx_forward(distance - 2);
        Char.y += ceil(distance / 2.0);
        Char.curr_row = y_to_row_mod4(Char.y);
        bumped_fall();
        return;
    }
    // custom logic - makes run-turn at the edge easier
    if (Char.charid == charid_0_kid &&
            frame >= frame_54_runturn &&
            frame <= frame_55_runturn &&
            ((Char.direction == dir_0_right && control_x == control_m1_held_left) ||
             (Char.direction == dir_FF_left && control_x == control_1_held_right)) &&
            distance_to_edge(Char.x) >= 8 + wall_infront) {
        Char.x = char_dx_forward(-3);
        return;
    }
    // custom logic - do not fall on the first pixel of the empty tile on frame 14
    if (frame == frame_14_run && tile_is_floor(tile_infront) && distance == 0) {
        Char.x = char_dx_forward(-3);
    }
    inc_curr_row();
    start_chompers();
    if (frame == frame_13_run || frame == frame_54_runturn) {
        seq_id = seq_19_fall; // put kid one step ahead
    } else if (frame == frame_26_standing_jump_11) {
        seq_id = seq_18_fall_after_standing_jump; // fall after standing jump
    } else if (frame == frame_44_running_jump_5) {
        ++Char.fall_x;
        seq_id = seq_21_fall_after_running_jump; // fall after running jump
    } else if (frame >= frame_81_hangdrop_1 && frame < frame_86_test_foot) {
        seq_id = seq_19_fall;
        Char.x = char_dx_forward(5);
        load_fram_det_col();
    } else if (frame == frame_15_stand || (frame >= frame_107_fall_land_1 && frame <= frame_108_fall_land_2)) {
        // custom - falling after standing, crouching or jumping up very close from the edge
        seq_id = seq_19_fall;
        if (distance > 11) {
            Char.x = char_dx_forward(distance - 11);
        }
    } else if (frame == frame_109_crouch && distance <= 2 && tile_is_floor(tile_infront)) {
        // frame 109: crouch (guards cannot crouch)
        seq_id = seq_81_kid_pushed_off_ledge; // fall after landing at the edge of a floor tile
    } else if ((frame >= frame_150_parry && frame < frame_180_collapse_2) ||
               (frame >= frame_207_draw_1 && frame <= frame_210_draw_4)) {
        // frame 150..179: with sword + fall + dead
        if (Char.charid == charid_2_guard || Char.charid == charid_4_skeleton) {
            // custom logic - prevent bridge skeleton from falling through a wall
            if (current_level == 12 &&
                    Char.charid == charid_4_skeleton &&
                    Char.direction == dir_0_right &&
                    Char.room == 13 &&
                    Char.curr_col == 6 &&
                    falling_bridge_timer != 0) {
                Char.fall_x = 0;
            }
            if (Char.fall_x < 0 ||
                    (can_guard_see_kid && Char.direction == ~Opp.direction && Char.charid != charid_4_skeleton)) {
                seq_id = seq_82_guard_pushed_off_ledge; // Guard is pushed off the ledge
                if (Char.direction < dir_0_right && distance <= 7) {
                    Char.x = char_dx_forward(-5);
                }
                // custom logic - guard fell back while drawing sword
            } else if (can_guard_see_kid == 2 && Opp.sword == sword_0_sheathed && char_dist > -7) {
                seq_id = seq_82_guard_pushed_off_ledge;
                // custom logic - guard should fall backwards when engarde near a wall
            } else if (Char.fall_x > 0 &&
                       (frame == frame_158_stand_with_sword || frame == frame_170_stand_with_sword) &&
                       distance <= 2 &&
                       get_tile_infrontof_char() == tiles_20_wall) {
                seq_id = seq_82_guard_pushed_off_ledge;
            } else {
                droppedout = 0;
                // custom logic - start falling not too close to the floor
                if (distance >= 11 && tile_is_floor(get_tile_behind_above_char())) {
                    Char.x = char_dx_forward(distance - 10);
                }
                seq_id = seq_83_guard_fall; // fall after forwarding with sword
            }
        } else {
            // custom logic - fall backwards in a fight or when holding sword
            if (Char.fall_x < 0 ||
                    can_guard_see_kid == 2 ||
                    (Opp.alive < 0 && Char.curr_row - 1 == Opp.curr_row && Opp.sword == sword_2_drawn)) {
                if ((Char.direction < dir_0_right || Char.fall_x < 0) && distance <= 7) {
                    Char.x = char_dx_forward(-5);
                }
                seq_id = seq_81_kid_pushed_off_ledge; // fall after backing with sword / Kid is pushed off the ledge
            } else {
                // custom logic - start falling not too close to the floor
                if (distance >= 11 && tile_is_floor(get_tile_behind_above_char())) {
                    Char.x = char_dx_forward(14 - distance);
                }
                seq_id = seq_19_fall; // fall after forwarding with sword
            }
        }
        // custom logic - kid is falling after guard pushes him off without the sword drawn
    } else if (Char.charid == charid_0_kid &&
               frame >= frame_179_collapse_1 &&
               frame <= frame_183_collapse_5 &&
               Char.direction == ~Opp.direction &&
               Opp.frame >= frame_151_strike_1 &&
               Opp.frame <= frame_157_walk_with_sword &&
               char_dist < 35) {
        seq_id = seq_81_kid_pushed_off_ledge;
        // custom logic - dying guard pushed off the edge
    } else if (Char.charid == charid_2_guard && Char.alive >= 0 && Char.direction == ~Opp.direction && Char.action == actions_99_hurt) {
        seq_id = seq_82_guard_pushed_off_ledge; // Guard is pushed off the ledge
    } else {
        seq_id = seq_7_fall; // fall after step, run, crouch or some post jump frames
    }
    // custom logic - make a loose tile behind fall when falling backwards
    if ((seq_id == seq_81_kid_pushed_off_ledge || seq_id == seq_82_guard_pushed_off_ledge) && tile_behind == tiles_11_loose) {
        make_loose_fall(3);
    }
    tile_infront = get_tile_infrontof_char();
    // custom logic - kid can fall through tapestry when standing/stepping/running/crouching near the edge
    // this logic needs to happen before the animation has started
    if (distance <= 2 &&
            (tile_infront == tiles_20_wall ||
             (Char.direction == dir_FF_left && (tile_infront == tiles_12_doortop || tile_infront == tiles_7_doortop_with_floor))) &&
            (frame <= frame_15_stand || (frame >= frame_109_crouch && frame <= frame_132_stepping_12))) {
        Char.x = char_dx_forward(distance - 2);
        seq_id = seq_45_bumpfall;
    }
    // custom logic - makes grabbing while running easier
    if (Char.charid == charid_0_kid && control_shift < control_0_released && floor_infront) {
        if ((seq_id == seq_7_fall || seq_id == seq_19_fall) &&
                frame >= frame_4_start_run && frame <= frame_14_run &&
                distance >= 7) {
            Char.x = char_dx_forward(distance - 7);
            if (frame >= frame_12_run) {
                Char.x = char_dx_forward(1);
            }
        }
    }
    // custom logic - guards should follow kid as often as possible except for level 6
    if (current_level == 6) {
        droppedout = 0;
    } else if (Char.charid == charid_0_kid) {
        set_kid_droppedout();
    } else if (inactive_guard[Char.room - 1].room != Char.room || inactive_guard[Char.room - 1].direction == dir_56_none) {
        droppedout = 0;
    }
    // custom logic - reset victory sound if guard is falling
    if (delayed_sound.sound_id == sound_37_victory) {
        delayed_sound_timer = 0;
    }
    seqtbl_offset_char(seq_id);
    play_seq();
    load_fram_det_col();
    tile_infront = get_tile_infrontof_char();
    if (tile_infront == tiles_20_wall ||
            // Also treat tapestries (when approached to the left) like a wall here.
            (Char.direction == dir_FF_left &&
             (tile_infront == tiles_12_doortop || tile_infront == tiles_7_doortop_with_floor))
       ) {
        distance = distance_to_edge_weight();
        if (frame != frame_44_running_jump_5 || distance >= 6) {
            Char.x = char_dx_forward(-1);
            // custom logic - falling with no forward movement
            if ((seq_id == seq_7_fall || seq_id == seq_19_fall) && distance <= 8) {
                seqtbl_offset_char(seq_45_bumpfall);
                play_seq();
            }
        } else {
            // falling close to the wall at the end of a running jump
            seqtbl_offset_char(seq_104_start_fall_in_front_of_wall);
            play_seq();
        }
        load_fram_det_col();
    }
    if (get_tile_at_char() == tiles_20_wall &&
            seq_id != seq_81_kid_pushed_off_ledge &&
            seq_id != seq_82_guard_pushed_off_ledge) {
        in_wall();
    }
}

// custom logic - check if guard can follow dropped out kid
void check_can_guard_follow_kid(void) {
    short left_pos;
    short right_pos;
    short right_pos_row;
    short walltype;
    bool right_to_left;
    bool abyss;
    left_pos = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
    right_pos = x_bump[Opp.curr_col + dir_front[Opp.direction + 1] + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
    if (Char.frame >= frame_135_climbing_1 && Char.frame <= frame_148_climbing_14) {
        left_pos += Char.direction == dir_FF_left ? 14 : -14;
    }
    right_to_left = right_pos >= left_pos;
    abyss = 0;
    while (right_to_left ? right_pos >= left_pos : right_pos <= left_pos) {
        right_pos_row = get_tile_div_mod_m7(right_pos);
        walltype = wall_type2(get_tile(Opp.room, right_pos_row, Opp.curr_row));
        if (walltype > 0 && walltype != wall_type_3_chomper) {
            droppedout = 0;
            break;
        } else if (!tile_is_floor(curr_tile2) || curr_tile2 == tiles_11_loose) {
            abyss = true;
        }
        if (abyss &&
                (!tile_is_floor(get_tile(Opp.room, right_pos_row, Opp.curr_row + 1)) ||
                 wall_type2(curr_tile2) != wall_type_0_none) &&
                tile_col != Char.curr_col) {
            droppedout = 0;
            break;
        }
        if (right_to_left) {
            right_pos -= 14;
        } else {
            right_pos += 14;
        }
    }
}

// custom logic - set "droppedout" flag if guard can follow kid
void set_kid_droppedout(void) {
    if (!can_guard_see_kid || is_guarding_gate) {
        droppedout = 0;
        return;
    }
    droppedout = 1;
    // custom logic - there should only be 1 abyss (a non-floor or loose tile) between kid and guard
    // when "can_guard_see_kid" is 1. Char = kid, Opp = guard
    if (can_guard_see_kid == 1) {
        if ((Char.curr_col < Opp.curr_col && Opp.direction != dir_FF_left) ||
                (Char.curr_col > Opp.curr_col && Opp.direction != dir_0_right)) {
            droppedout = 0;
            return;
        }
        check_can_guard_follow_kid();
    }
}

// custom logic - plays the grab sound
void play_grab_sound(void) {
    if (feather_fall_timer > 0) {
        return;
    }
    set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
    if (want_positional_sound) { // reduce the surround effect
        want_sound_dir_x /= 2.0f;
    }
    play_sound(sound_9_grab);
}

// seg006:0A19
void check_grab(void) {
    word old_x;
    word delta_x;
    if ((control_shift < control_0_released || (super_jump_fall && control_y < control_0_released)) && // press shift or up arrow to grab
            Char.fall_y < MAX_GRAB_FALLING_SPEED && // you can't grab if you're falling too fast ...
            Char.alive < 0 && // ... or dead
            (word)y_land[Char.curr_row + ONSCREEN_ROW_OFFSET] <= (word)(Char.y + 25)
       ) {
        old_x = Char.x;
        // custom logic - delta_x makes it easier for kid to grab tiles during super high and slow
        // hard jump falls
        delta_x = super_jump_fall ? 3 - Char.direction : 0;
        Char.x = char_dx_forward(-8 + delta_x);
        // custom logic - makes it possible to catch the tile below when swinging forward and dropping down
        if (!super_jump_fall &&
                Char.fall_y == 18 &&
                Char.fall_x == 1 &&
                tile_is_floor(get_tile(Char.room, Char.curr_col, Char.curr_row - 2))) {
            get_tile_front_above_char();
            if (can_grab()) {
                Char.x = char_dx_forward(3);
            }
        }
        load_fram_det_col();
        // custom logic - do not grab during regular hard jumps
        if (hard_jump_fall && Char.fall_y == 0) {
            Char.x = old_x;
        } else if (!can_grab_front_above()) {
            Char.x = old_x;
        } else {
            Char.x = char_dx_forward(distance_to_edge_weight() - delta_x);
            Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
            seqtbl_offset_char(seq_15_grab_ledge_midair); // grab a ledge (after falling)
            play_seq();
            // custom logic - hang twice as long after a long jump unless you are a shadow or have 1 hitpoint left
            grab_timer = (Char.fall_x >= 3 || hitp_curr == 1) && !is_kid_shadow ? GRAB_DELAY * 2 : GRAB_DELAY;
            Char.fall_x = 0;
            Char.fall_y = 0;
            play_grab_sound();
            make_guard_notice(true);
            is_screaming = 0;
            // custom logic - reset hard jump fall flag
            hard_jump_fall = 0;
            start_chompers();
        }
    }
}

// custom logic - check if kid can grab a tile infront while jumping
bool check_grab_run_jump(void) {
    word frame;
    bool is_jump, is_running_jump;
    int grabbed_tile;
    frame = Char.frame;
    is_jump = frame >= frame_22_standing_jump_7 && frame <= frame_23_standing_jump_8;
    is_running_jump = frame >= frame_39_start_run_jump_6 && frame <= frame_41_running_jump_2;
    if (Char.action == actions_1_run_jump &&
            !hard_jump_fall &&
            (is_jump || is_running_jump) &&
            control_x == control_0_released &&
            control_y < control_0_released) {
        if (can_grab_front_above()) { // can grab a ledge at a specific point of during a jump
            grabbed_tile = curr_tile2;
            Char.x = x_bump[get_tile_col_at_char() + ONSCREEN_COLUMN_OFFSET] + CHAR_7_PX_OFFSET;
            Char.x = char_dx_forward(Char.direction == dir_FF_left ? -12 : 2);
            Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
            // kid cannot grab tiles on top of a wall while jumping on level 6
            if (current_level == 6 && falling_exit_checkpoint < 4 && get_tile_infrontof_char() == tiles_20_wall) {
                seqtbl_offset_char(seq_47_bump);
                play_seq();
                bumped_sound();
                return 0;
            }
            seqtbl_offset_char(frame == frame_40_running_jump_1 || frame == frame_41_running_jump_2
                               ? seq_140_grab_while_run_jump
                               : seq_9_grab_while_jumping); // grab a ledge
            play_seq();
            grab_timer = hitp_curr == 1 && !is_kid_shadow ? GRAB_DELAY * 2 : GRAB_DELAY;
            play_grab_sound();
            make_guard_notice(true);
            if (grabbed_tile == tiles_15_opener || grabbed_tile == tiles_6_closer) {
                trigger_button(1, 0, -1);
            } else if (grabbed_tile == tiles_11_loose) {
                make_loose_fall(1);
            }
            if (!droppedout &&
                    Opp.alive < 0 &&
                    Opp.direction != dir_56_none &&
                    Char.curr_row == Opp.curr_row + 1 &&
                    char_opp_dist() > 0) {
                can_guard_see_kid = 1;
                set_kid_droppedout();
                can_guard_see_kid = 0;
            }
            return 1;
        }
    }
    return 0;
}

// seg006:0ABD
int can_grab_front_above(void) {
    int old_x = 0;
    int grab;
    // custom logic - ensure kid does not jump too far into the second floor tile
    if (Char.frame == frame_106_fall &&
            Char.fall_x == 3 &&
            tile_is_floor(get_tile_at_char()) &&
            distance_to_edge_weight() >= 9 &&
            !tile_is_floor(get_tile_behind_above_char()) &&
            curr_tile2 != tiles_20_wall) {
        old_x = Char.x;
        Char.x = char_dx_forward(-14);
        load_fram_det_col();
    }
    through_tile = get_tile_above_char();
    get_tile_front_above_char();
    grab = can_grab();
    // custom logic - reset the original position if kid still cannot grab
    if (old_x != 0 && grab == 0) {
        Char.x = old_x;
        load_fram_det_col();
        through_tile = get_tile_above_char();
        get_tile_front_above_char();
        grab = can_grab();
    }
    return grab;
}

// seg006:0ACD
void in_wall(void) {
    short delta_x;
    // custom logic - do not bump in walls with the lowest control byte of 1
    if (curr_tile2 == tiles_20_wall && get_tile_ctrl_bits(curr_room, curr_tilepos, 1) == 1) {
        return;
    }
    delta_x = distance_to_edge_weight();
    // custom logic - fake walls should not push kid back out
    // fall at distance to edge of 0 to prevent safe step or run jumping off the edge
    if (is_fake_wall() && (delta_x > 0 || Char.sword == sword_2_drawn)) {
        if (Char.frame == frame_143_climbing_9) {
            goto loc_push_forward;
        }
        return;
    }
    if (delta_x >= 8 || get_tile_infrontof_char() == tiles_20_wall) {
        delta_x = 6 - MAX(delta_x, 8);
    } else {
loc_push_forward:
        delta_x += 4;
    }
    Char.x = char_dx_forward(delta_x);
    load_fram_det_col();
    get_tile_at_char();
}

// custom logic - in wall while getting struck with sword logic
void in_wall_engarde(void) {
    int wall_col;
    load_fram_det_col();
    if (get_tile_at_char() == tiles_20_wall) {
        in_wall();
        if (get_tile_at_char() != tiles_20_wall) {
            return;
        }
    }
    if (Char.direction == dir_0_right &&
            get_tile_behind_char() == tiles_20_wall &&
            tile_is_floor(get_tile_at_char()) &&
            distance_to_edge_weight() >= 8) {
        wall_col = get_tile_col_at_char();
        Char.x = x_bump[wall_col + ONSCREEN_COLUMN_OFFSET + dir_front[Char.direction + 1]] + CHAR_7_PX_OFFSET;
        Char.curr_col = wall_col;
    } else if (get_tile_at_char() == tiles_20_wall &&
               tile_is_floor(get_tile_infrontof_char()) &&
               distance_to_edge_weight() >= 8) {
        wall_col = get_tile_col_at_char();
        Char.x = x_bump[wall_col + ONSCREEN_COLUMN_OFFSET + dir_front[Char.direction + 1]] + CHAR_7_PX_OFFSET;
        Char.curr_col = wall_col;
    }
}

// seg006:0B0C
int get_tile_infrontof_char(void) {
    return get_tile(Char.room, infrontx = dir_front[Char.direction + 1] + Char.curr_col, Char.curr_row);
}

// seg006:0B30
int get_tile_infrontof2_char(void) {
    short direction_delta;
    direction_delta = dir_front[Char.direction + 1];
    return get_tile(Char.room, infrontx = (direction_delta << 1) + Char.curr_col, Char.curr_row);
}

// seg006:0B66
int get_tile_behind_char(void) {
    return get_tile(Char.room, dir_behind[Char.direction + 1] + Char.curr_col, Char.curr_row);
}

// custom logic - get tile column relative to the current room
int get_tile_col_at_char(void) {
    int left, right, up;
    int char_col = tile_col;
    int room_minus_1 = Char.room - 1;
    if (curr_room != Char.room)	{
        left = level.roomlinks[room_minus_1].left;
        right = level.roomlinks[room_minus_1].right;
        up = level.roomlinks[room_minus_1].up;
        if (curr_room == right) {
            char_col += 10;
        } else if (curr_room == left) {
            char_col -= 10;
        } else if (right && curr_room == level.roomlinks[right - 1].up) {
            if (right == curr_room) return char_col; // broken room link fix
            char_col += 10;
        } else if (left && curr_room == level.roomlinks[left - 1].up) {
            if (left == curr_room) return char_col;
            char_col -= 10;
        } else if (up && curr_room == level.roomlinks[up - 1].right) {
            if (up == curr_room) return char_col;
        } else if (up && curr_room == level.roomlinks[up - 1].left) {
            if (up == curr_room) return char_col;
            char_col -= 10;
        }
    }
    return char_col;
}

// seg006:0B8A
int distance_to_edge_weight(void) {
    return distance_to_edge(dx_weight());
}

// seg006:0B94
int distance_to_edge(int xpos) {
    short distance;
    get_tile_div_mod_m7(xpos);
    distance = obj_xl;
    if (Char.direction == dir_0_right) {
        distance = 13 - distance;
    }
    return distance;
}

// seg006:0BC4
void fell_out(void) {
    if (Char.room != 0) {
        return;
    }
    // custom logic - keep level 13 connected to level 12
    if (!is_kid_shadow &&
            current_level == 13 &&
            (drawn_room == 2 || drawn_room == 13) &&
            Char.curr_col < 0) {
        checkpoint = drawn_room == 13 ? 1 : 2;
        next_level = 12;
        seamless = 1;
        return;
    }
    if (Char.alive < 0) {
        take_hp(ALL_HP);
        Char.alive = 0;
        erase_bottom_text(1);
    }
    // custom logic - ensure kid's frame is actually dead
    if (!is_dead()) {
        Char.frame = frame_185_dead;
    }
}

// custom logic - restores dead kid tile
void restore_dead_kid_tile(void) {
    if (dead_kid_tile.room == 0 || dead_kid_tile.room > level.used_rooms) {
        return;
    }
    get_room_address(dead_kid_tile.room);
    // check for a loose tile falling onto the button/spikes
    if ((curr_room_tiles[dead_kid_tile.tilepos] & 0x1F) != tiles_14_debris) {
        // reset button
        curr_room_tiles[dead_kid_tile.tilepos] = dead_kid_tile.type;
        curr_room_modif[dead_kid_tile.tilepos] = dead_kid_tile.modifier;
        add_trob(dead_kid_tile.room, dead_kid_tile.tilepos, 1);
        // reset gates
        if ((dead_kid_tile.type & 0x1F) == tiles_15_opener) {
            do_trigger_events(dead_kid_tile.modifier, tiles_5_stuck);
        }
    }
    dead_kid_tile = (struct stuck_button_type) {
        0, 0, 0, 0
    };
}

// seg006:0BEE
void play_kid(void) {
    fell_out();
    control_kid();
    // custom logic - put potion back when kid is dead or fighting
    if ((hitp_curr == 0 ||
            Char.alive >= 0 ||
            (Char.sword == sword_2_drawn && Opp.charid != charid_1_shadow)) &&
            potion_timer > 0 &&
            (current_level != settings->mouse_2_level || Char.room != settings->mouse_2_room)) {
        if (current_sound == sound_18_drink && check_sound_playing()) {
            stop_sounds();
        }
        potion_timer = (potion_timer < 100 ? 100 : 0);
    }
    // custom logic - if kid dies when not in combat in room 8 of level 10, redraw the screen to hide flame
    if (Char.alive >= 0 && current_level == 10 && drawn_room == 8 && (!holding_sword || Char.action == actions_4_in_freefall)) {
        holding_sword = 0;
        need_full_redraw = 1;
    }
    // custom logic - kid should not hang indefinitely if he dies hanging straight
    if (Char.alive >= 0 && Char.action == actions_6_hang_straight) {
        Char.action = actions_3_in_midair;
        hang_fall();
        return;
    }
    if (Char.alive >= 0 && is_dead()) {
        // custom logic - when kid dies, shadow dies
        if (Guard.charid == charid_1_shadow && Guard.alive < 0) {
            if (current_level != 10 || shadow_initialized == 0) {
                guardhp_curr = 0;
            }
        }
        // custom logic - spiked kid should be crushed by a loose tile
        if (Char.frame == frame_177_spiked && get_tile_at_char() == tiles_14_debris) {
            seqtbl_offset_char(seq_22_crushed); // dead (because of loose floor)
            Char.x = char_dx_forward(-12);
        }
        // custom logic - dead shadow turns back into kid
        if (is_kid_shadow && Dead.room > 0 && Dead.direction != dir_56_none && Char.alive < 6) {
            if (is_shadow_flame) {
                is_shadow_flame = false;
                reset_shadow_palette();
                hitp_curr = hitp_max = MIN(settings->start_hitp, settings->max_hitp_allowed);
            }
            Char.room = Dead.room;
            Char.direction = Dead.direction;
            Char.x = Dead.x;
            Char.x = char_dx_forward(3);
            Char.y = y_land[Dead.row + ONSCREEN_ROW_OFFSET];
            Char.curr_row = Dead.row;
            have_sword = 1;
            draw_guard_hp(0, MAX_GUARD_HP);
            next_room = Char.room;
            seqtbl_offset_char(seq_71_dying);
            Char.fall_y = 0;
            Char.fall_x = 0;
            Char.alive = 6; // play music and show press button message
            play_seq();
            flash_color = color_12_brightred;
            flash_time = 12;
        }
        if (is_kid_shadow &&
                (Char.frame == frame_177_spiked ||
                 Char.frame == frame_178_chomped ||
                 Char.frame == frame_185_dead)) {
            Dead.room = 0;
            Dead.direction = dir_56_none;
            Dead.color = 0;
            if (flash_time == 0) {
                is_kid_shadow = 0;
                if (is_shadow_flame) {
                    is_shadow_flame = false;
                    reset_shadow_palette();
                }
                draw_kid_hp(0, hitp_max);
            } else if (is_shadow_flame && dead_kid_timer == 0) {
                dead_kid_timer = 7;
                load_flame_sprite();
                dead_kid_timer = 0;
            }
        }
        // custom logic - turn kid into a skeleton if killed in room 1, level 12B, row 2
        if (current_level == 13 &&
                level_13_nowhere_to_run_timer < 0 &&
                Char.room == 1 &&
                Char.curr_row == 2 &&
                get_tile(1, 4, 0) != tiles_21_skeleton) {
            level_13_nowhere_to_run_timer = 0;
        }
        if (resurrect_time) {
            is_kid_shadow = 0;
            reset_shadow_palette();
            stop_all_sounds();
            hitp_delta = hitp_max;
            // custom logic - restore dead kid tile
            restore_dead_kid_tile();
            seqtbl_offset_char(seq_2_stand); // stand
            // custom logic - do not place revived kid into a wall/abiss
            if (!tile_is_floor(get_tile(Char.room, get_tile_div_mod_m7(Char.x + 8), Char.curr_row))) {
                Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
            } else {
                Char.x += 8;
            }
            play_seq();
            load_fram_det_col();
            set_start_pos();
        }
        if (check_sound_playing() && current_sound != sound_5_gate_opening) {
            return;
        }
        // custom logic - wait until the enhanced sound finishes playing unless it is shadow music
        if (ogg_fore_sound_playing &&
                curr_ogg_sound_id != sound_79_united_with_shadow &&
                curr_ogg_sound_id != sound_73_shadow_level_music) {
            return;
        }
        // custom logic - kid does not have enough lives to continue, do not restart the level
        if (delayed_sound.sound_id == sound_38_blink) {
            return;
        }
        is_show_time = 0;
        droppedout = 0;
        if (Char.alive < 0 || Char.alive >= 6) {
            if (Char.alive == 6) {
                if (is_sound_on) {
                    play_death_music();
                }
            } else {
                if (Char.alive != 7 || check_sound_playing() || ogg_fore_sound_playing) {
                    return;
                }
                if (current_level < settings->win_level && rem_min == 0) {
                    expired();
                }
                if (current_level == 0 && !play_demo_level) { // playing demo
                    display_text_bottom("OOPS. LET'S TRY AGAIN...");
                    text_time_remaining = text_time_total = 36;
                    level_restart_timer = 38; // restart automatically
                } else {
                    text_time_remaining = text_time_total = RESTART_AFTER_DEATH_DELAY;
                    display_press_button();
                }
            }
        }
        ++Char.alive;
    }
}

// seg006:0CD1
void control_kid(void) {
    // custom logic - dead kid should not keep running
    if (hitp_curr == 0 && (Char.frame == frame_7_run || Char.frame == frame_11_run)) {
        control_forward = release_arrows();
        seqtbl_offset_char(seq_13_stop_run); // stop run
    }
    if (Char.alive < 0 && hitp_curr == 0) {
        Char.alive = 0;
        // custom logic - stop feather fall music after kid dies
        if (feather_fall_timer > 1 && Char.action != actions_3_in_midair && Char.action != actions_4_in_freefall) {
            feather_fall_timer = 1;
        }
    }
    if (grab_timer != 0) {
        --grab_timer;
    }
    if (current_level == 0 && !play_demo_level) {
        do_paused();
        do_demo();
        control();
    } else {
        rest_ctrl_1();
        do_paused();
        read_user_control();
        user_control();
        save_ctrl_1();
    }
}

// seg006:0D49
void do_demo(void) {
    if (checkpoint == 3) {
        control_shift2 = release_arrows();
        do_auto_moves(settings->demo_moves_2);
    } else if (checkpoint > 0 && checkpoint < 3) {
        control_shift2 = release_arrows();
        // custom logic - align kit at a specific x-coordinate at the 2nd to last sheathe frame
        if (checkpoint == 1) {
            loadkid();
            if (Char.frame != frame_239_sheathe) {
                return;
            }
            Char.direction = dir_FF_left;
            Char.curr_col = 5;
            Char.x = x_bump[Char.curr_col + ONSCREEN_COLUMN_OFFSET] + CHAR_10_PX_OFFSET;
            Char.x = char_dx_forward(-1);
            savekid();
        }
        ++checkpoint;
    } else if (Char.sword == sword_2_drawn) {
        guard_skill = settings->demo_kid_skill; // kid
        autocontrol_opponent();
        guard_skill = settings->demo_guard_skill; // guard
    } else {
        control_shift2 = release_arrows();
        do_auto_moves(settings->demo_moves);
    }
}

// custom logic - enter dynamic guard after the first guard is killed
void enter_dynamic_guard(void) {
    word room_minus_1 = drawn_room - 1;
    word guard_room = dynamic_guard.direction == dir_0_right ? room_L : room_R;
    short guard_color;
    if (guard_room == 0) {
        return;
    }
    if (guard_room == falling_guard.room) {
        outbound_dead_guard3[guard_room - 1].room = guard_room;
        outbound_dead_guard3[guard_room - 1].direction = dir_56_none;
        return;
    }
    if (inactive_guard[room_minus_1].room == drawn_room && is_inactive_guard_ready(inactive_guard[room_minus_1])) {
        if (can_inactive_guard_see_kid) {
            return;
        }
        if (inactive_guard[room_minus_1].row == dynamic_guard.row) {
            level.guards_tile[room_minus_1] = 0xFF;
            draw_guard_hp(0, MAX_GUARD_HP);
            enter_guard();
            return;
        }
    }
    if (level.guards_tile[guard_room - 1] < MAX_TILES_PER_ROOM ||
            (Guard.room == guard_room && Guard.direction == dir_56_none) ||
            dead_guard1[guard_room - 1].room == guard_room ||
            dead_guard2[guard_room - 1].room == guard_room ||
            dead_guard3[guard_room - 1].room == guard_room ||
            (outbound_dead_guard3[guard_room - 1].room == guard_room && // fallen out guard
             outbound_dead_guard3[guard_room - 1].direction == dir_56_none)) {
        if (is_dynamic_guard_fight && Guard.alive >= 0 && Guard.direction != dir_56_none) {
            is_dynamic_guard_fight = 0;
            clear_char();
            level.guards_tile[room_minus_1] = 0xFF;
            on_guard_killed();
        }
        return;
    }
    if (Kid.curr_row < dynamic_guard.row || (Kid.curr_row - dynamic_guard.row) >= 2) {
        return;
    }
    dynamic_guard_count[room_minus_1]++;
    level.guards_seq_hi[room_minus_1] = 0;
    level.guards_dir[room_minus_1] = dynamic_guard.direction;
    level.guards_tile[room_minus_1] = dynamic_guard.row * 10;
    level.guards_x[room_minus_1] = dynamic_guard.x;
    if (settings->tbl_guard_type[current_level] == guard_type_guard) {
        if (dynamic_guard.color > 0 && dynamic_guard.color <= GUARD_COLORS) {
            guard_color = dynamic_guard.color;
        } else {
            guard_color = level.guards_color[room_minus_1] & 0x0F;
        }
    } else {
        guard_color = 0;
    }
    level.guards_color[room_minus_1] = guard_color;
    level.guards_color[room_minus_1] |= ((settings->tbl_guard_hp[current_level] - 1) << 4);
    if (dynamic_guard.frame2 < NUM_GUARD_SKILLS) {
        level.guards_skill[room_minus_1] = dynamic_guard.frame2; // frame2 stores the skill
    }
    droppedout = 0;
    set_sound_pos(drawn_room, get_tile_div_mod_m7(dynamic_guard.x), dynamic_guard.row);
    play_sound(sound_19_draw_sword);
    draw_guard_hp(0, MAX_GUARD_HP);
    enter_guard();
    check_can_guard_see_kid();
    if (Kid.sword == sword_0_sheathed &&
            can_guard_see_kid == 2 &&
            Kid.direction == Char.direction &&
            (Kid.frame == frame_15_stand ||
             (Kid.frame >= frame_45_turn && Kid.frame <= frame_52_turn) ||
             Kid.frame == frame_133_sheathe ||
             Kid.frame == frame_134_sheathe ||
             Kid.frame == frame_166_stand_inactive ||
             Kid.frame >= frame_230_sheathe)) {
        is_dynamic_guard_fight = 1;
        keep_sword_drawn = false;
        loadkid();
        draw_sword();
        savekid();
        loadshad();
    }
    dynamic_guard_timer = 24;
}

// custom logic - bump into inactive opponent
void bump_into_inactive_opponent(void) {
    char_type2 inactive;
    short distance;
    inactive = inactive_guard[Char.room - 1];
    if (inactive.room != Char.room || inactive.direction == dir_56_none) {
        return;
    }
    if (can_inactive_guard_see_kid >= 2 &&
            Char.sword == sword_0_sheathed && // Kid must not be in fighting pose
            inactive.frame != frame_166_stand_inactive && // but Guard must
            Char.direction != inactive.direction // must be facing toward each other
       ) {
        distance = inactive_guard_kid_dist();
        if (ABS(distance) <= 15) {
            if (Char.fall_y >= 22) {
                return; // don't bump
            }
            // don't bump guards in different rooms to avoid a room switching bug
            if (Char.curr_row < 0 || Char.curr_row > 2) {
                return;
            }
            // prevents kid from jumping over guards
            if ((Char.direction == dir_0_right && Char.x > inactive.x) ||
                    (Char.direction == dir_FF_left && Char.x < inactive.x)) {
                Char.x = inactive.x;
            }
            Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
            Char.fall_y = 0;
            seqtbl_offset_char(seq_47_bump); // bump into opponent
            play_seq();
        }
    }
}

// custom logic - if inactive guard can see kid but active guard cannot, swap them
void check_inactive_guard_proximity(void) {
    word room_minus_1;
    char_type2 inactive;
    int guard_color;
    int dist, inactive_dist;
    if (Kid.room == 0 || Kid.alive >= 0) {
        return;
    }
    // do not switch guards while kid is turning around
    if (Kid.frame >= frame_214_turn_engarde_1 && Kid.frame <= frame_216_turn_engarde_3) {
        return;
    }
    // do not switch guards while kid is climbing stairs
    if (Kid.frame >= frame_219_exit_stairs_3 && Kid.frame <= frame_228_exit_stairs_12) {
        return;
    }
    room_minus_1 = Kid.room - 1;
    inactive = inactive_guard[room_minus_1];
    if (inactive.room != Kid.room || inactive.direction == dir_56_none) {
        return;
    }
    check_can_inactive_guard_see_kid();
    check_can_guard_see_kid();
    // inactive guard is cannot fight kid
    if (can_inactive_guard_see_kid == 1 &&
            guardhp_curr > 0 &&
            Guard.alive < 0 &&
            Guard.direction != dir_56_none &&
            Guard.curr_row != Kid.curr_row &&
            Guard.curr_row >= 0 && Guard.curr_row <= 2 &&
            Guard.action <= actions_1_run_jump &&
            !guard_dropping_down) {
        goto loc_swap_guard;
    }
    dist = kid_guard_dist();
    inactive_dist = inactive_guard_kid_dist();
    // inactive guard is in front of kid while active guard is behind
    if (can_guard_see_kid &&
            can_inactive_guard_see_kid == can_guard_see_kid &&
            Guard.direction == ~inactive.direction &&
            Guard.curr_col >= -1 && Guard.curr_col < 10 &&
            dist < -14 && inactive_dist >= 0) {
        goto loc_swap_guard;
    }
    // inactive guard is in front of active guard facing kid
    if (can_guard_see_kid == 2 &&
            exit_room_timer == 0 &&
            can_inactive_guard_see_kid == can_guard_see_kid &&
            Guard.direction == inactive.direction &&
            Guard.direction == ~Kid.direction &&
            (Guard.frame == frame_158_stand_with_sword ||
             Guard.frame == frame_170_stand_with_sword ||
             Guard.frame == frame_171_stand_with_sword) &&
            dist - inactive_dist > 14) {
        goto loc_swap_guard;
    }
    // inactive guard cannot fight kid
    if (can_inactive_guard_see_kid != 2) {
        return;
    }
    // do not swap guards under various conditions
    if (exit_room_timer != 0 ||
            Guard.room != inactive.room ||
            can_guard_see_kid == 2 ||
            Guard.action > actions_1_run_jump ||
            Guard.frame < frame_150_parry ||
            Guard.frame > frame_171_stand_with_sword) {
        return;
    }
loc_swap_guard:
    if (guardhp_curr == 0 || Guard.alive >= 0 || Guard.direction == dir_56_none) {
        return;
    }
    // do not switch guards while kid is at a chomper
    if (get_tile(Kid.room, Kid.curr_col, Kid.curr_row) == tiles_18_chomper ||
            get_tile(Kid.room, Kid.curr_col + 1, Kid.curr_row) == tiles_18_chomper) {
        return;
    }
    if (Kid.action <= actions_1_run_jump || Kid.action == actions_7_turn) {
        droppedout = 0;
    }
    guard_color = Guard.color;
    guard_color |= (guardhp_curr << 4);
    level.guards_seq_hi[room_minus_1] = 0;
    level.guards_dir[room_minus_1] = inactive.direction;
    level.guards_tile[room_minus_1] = inactive.row * 10;
    level.guards_x[room_minus_1] = inactive.x;
    level.guards_color[room_minus_1] = inactive.color;
    level.guards_skill[room_minus_1] = inactive_guard_skill;
    inactive_guard[room_minus_1].direction = Guard.direction;
    inactive_guard[room_minus_1].x = Guard.x;
    inactive_guard[room_minus_1].row = MAX(Guard.curr_row, 0);
    inactive_guard[room_minus_1].color = guard_color;
    inactive_guard_skill = guard_skill;
    inactive_seq_ptr = inactive_seq_stand;
    Guard.direction = dir_56_none;
    draw_guard_hp(0, MAX_GUARD_HP);
    enter_guard();
}

// custom logic - checks whether inactive guard can strike kid
bool can_inactive_guard_strike(char_type2 inactive) {
    short distance;
    bool wall_behind;
    if (Kid.direction != inactive.direction && Kid.sword == sword_2_drawn && Guard.frame != frame_171_stand_with_sword) {
        return false;
    }
    if (Kid.direction != inactive.direction && (Kid.frame == frame_152_strike_2 || Kid.frame == frame_153_strike_3)) {
        return false;
    }
    if (Kid.frame >= frame_214_turn_engarde_1 && Kid.frame <= frame_216_turn_engarde_3) {
        return false;
    }
    wall_behind = get_tile(inactive.room, get_tile_div_mod_m7(inactive.x) + dir_behind[inactive.direction + 1], inactive.row) == tiles_20_wall;
    if (!wall_behind && Kid.action <= actions_1_run_jump) {
        if (inactive.direction == dir_FF_left && Kid.x > inactive.x_left_coll - 3) {
            return false;
        }
        if (inactive.direction == dir_0_right && Kid.x < inactive.x_right_coll + 3) {
            return false;
        }
    }
    if (Guard.alive < 0 &&
            Guard.direction != Kid.direction &&
            Guard.direction == inactive.direction &&
            Guard.curr_row == inactive.row &&
            Guard.frame >= frame_151_strike_1 && Guard.frame <= frame_154_poking) {
        return false;
    }
    distance = inactive_guard_kid_dist();
    if (distance > 0 &&
            Kid.frame >= frame_106_fall &&
            Kid.frame <= frame_108_fall_land_2) {
        distance += 7;
    }
    if (distance >= 7 && distance < 27 &&
            (Kid.action <= actions_1_run_jump ||
             Kid.action == actions_2_hang_climb ||
             Kid.action == actions_5_bumped ||
             Kid.action == actions_6_hang_straight) &&
            Kid.fall_y == 0) {
        return true;
    }
    return false;
}

// custom logic - checks whether inactive guard can block kid's strike
bool can_inactive_guard_block(char_type2 inactive) {
    short distance;
    if (Kid.direction == inactive.direction || Kid.frame != frame_154_poking) {
        return false;
    }
    if (inactive.direction == dir_FF_left && Kid.x > inactive.x_left_coll - 3) {
        return false;
    } else if (inactive.direction == dir_0_right && Kid.x < inactive.x_right_coll + 3) {
        return false;
    }
    distance = inactive_guard_kid_dist();
    return distance >= 4 && distance < 29;
}

// custom logic - inactive guard follows kid
void inactive_guard_follows_kid(char_type2 inactive) {
    get_tile(inactive.room, get_tile_div_mod_m7(inactive.x) + dir_front[inactive.direction + 1], inactive.row);
    if (// there is wall in front of Guard
        wall_type(curr_tile2) != 0 ||
        (!tile_is_floor(curr_tile2) && (
             (get_tile(curr_room, tile_col, ++tile_row) == tiles_2_spike ||
              // Guard would fall on loose floor
              curr_tile2 == tiles_11_loose ||
              // ... or wall (?)
              wall_type(curr_tile2) != 0 ||
              // ... or into a chasm
              !tile_is_floor(curr_tile2)) ||
             // ... or Kid is not below
             inactive.row + 1 != Kid.curr_row
         ))
    ) {
        // do nothing
    } else {
        // follow
        autocontrol_inactive_guard_advance(inactive);
    }
}

// custom logic - control inactive guard
void autocontrol_inactive_guard(void) {
    char_type2 inactive;
    word room_minus_1;
    short distance;
    bool is_guard_dying;
    // prevents inactive guard from striking kid too quickly
    if (inactive_refrac_timer > 0) {
        --inactive_refrac_timer;
    }
    if (Kid.room == 0) {
        should_inactive_turn_around = 0;
        return;
    }
    room_minus_1 = Kid.room - 1;
    inactive = inactive_guard[room_minus_1];
    if (inactive.room != Kid.room || inactive.direction == dir_56_none) {
        is_inactive_guard_notice = 0;
        should_inactive_turn_around = 0;
        return;
    }
    if (inactive_guard_skill >= NUM_GUARD_SKILLS) {
        inactive_guard_skill = level.guards_skill[room_minus_1];
    }
    if (inactive.row == Kid.curr_row - 1 && droppedout && can_guard_see_kid) {
        inactive_guard_follows_kid(inactive);
        return;
    }
    if (inactive_seq_ptr == inactive_seq_stand) {
        if ((can_inactive_guard_see_kid < 2 && inactive.frame != frame_166_stand_inactive) ||
                (can_inactive_guard_see_kid == 2 && inactive.frame == frame_166_stand_inactive)) {
            seqtbl_inactive_char(inactive_seq_stand);
            play_inactive_seq();
        }
    }
    distance = inactive_guard_kid_dist();
    if (((is_inactive_guard_notice > 0 || (can_inactive_guard_see_kid && hitp_curr)) &&
            (!is_kid_shadow || Guard.charid == charid_4_skeleton)) ||
            (Kid.room == inactive.room && is_inactive_guard_notice == 2)) {
        if (is_inactive_guard_notice == 2) {
            is_inactive_guard_notice = -10;
        }
        if ((!droppedout || Kid.curr_row <= inactive.row) && distance < 0 && (word)distance < (word) - 4) {
            if (inactive.frame == frame_166_stand_inactive || inactive.frame == frame_171_stand_with_sword) {
                int delta_x = Kid.x - inactive.x;
                if ((delta_x <= -7 && inactive.direction == dir_0_right) ||
                        (delta_x >= 7 && inactive.direction == dir_FF_left)) {
                    autocontrol_inactive_guard_turn(inactive);
                    return;
                }
            } else if (inactive.direction == Kid.direction) {
                autocontrol_inactive_guard_turn(inactive);
                return;
            }
        }
    }
    should_inactive_turn_around = 0;
    if (is_kid_shadow && Guard.charid != charid_4_skeleton) {
        return;
    }
    if (inactive.row != Kid.curr_row) {
        return;
    }
    if (can_inactive_guard_see_kid < 2) {
        return;
    }
    // check on floor
    if ((!tile_is_floor(get_tile(inactive.room, get_tile_div_mod_m7(inactive.x + dir_behind[inactive.direction + 1] * 13), inactive.row)) ||
            (curr_tile2 == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0)) &&
            (inactive.frame == frame_158_stand_with_sword ||
             inactive.frame == frame_170_stand_with_sword ||
             inactive.frame == frame_171_stand_with_sword) &&
            curr_tile2 != tiles_20_wall) {
        seqtbl_inactive_char(inactive_seq_advance);
        play_inactive_seq();
        return;
    }
    is_guard_dying = Guard.frame >= frame_179_collapse_1 &&
                     Guard.frame <= frame_183_collapse_5 &&
                     inactive.direction == Kid.direction;
    if ((Guard.alive >= 0 && !is_guard_dying) ||
            Guard.direction == dir_56_none ||
            Kid.alive >= 0 ||
            resurrect_time > 0) {
        seqtbl_inactive_char(inactive_seq_stand);
        play_inactive_seq();
        return;
    }
    if (can_guard_see_kid == 2 || is_guard_dying) {
        if (inactive.direction != Kid.direction || !can_inactive_guard_strike(inactive)) {
            autocontrol_inactive_guard_advance(inactive);
        }
    }
    if (can_inactive_guard_block(inactive)) {
        autocontrol_inactive_guard_hurt_or_block(inactive);
    } else if (inactive_refrac_timer == 0 && can_inactive_guard_strike(inactive)) {
        autocontrol_inactive_guard_strike(inactive);
    } else if (inactive.direction != Kid.direction && distance <= 14) {
        autocontrol_inactive_guard_retreat(inactive);
    }
}

// custom logic - inactive guard turns around
void autocontrol_inactive_guard_turn(char_type2 inactive) {
    // a guard that closes gates should not turn
    if (inactive_guard_skill == GATE_GUARDING_SKILL && (inactive_guard_kid_dist() >= 0 || Kid.curr_row != inactive.row)) {
        return;
    }
    if (can_inactive_guard_see_kid == 2) {
        seqtbl_inactive_char(inactive_seq_turn_engarde);
    } else {
        if (should_inactive_turn_around > 0) {
            --should_inactive_turn_around;
            return;
        }
        seqtbl_inactive_char(inactive_seq_turn);
        should_inactive_turn_around = 6;
    }
    play_inactive_seq();
}

// custom logic - advances inactive guard
void autocontrol_inactive_guard_advance(char_type2 inactive) {
    int inactive_tile, wall;
    if (inactive_seq_ptr != inactive_seq_stand &&
            (inactive_seq_ptr != inactive_seq_stop || (Kid.direction != inactive.direction && !droppedout))) {
        return;
    }
    inactive_tile = get_tile(inactive.room, get_tile_div_mod_m7(inactive.x), inactive.row);
    wall = wall_type2(inactive_tile);
    if (inactive.direction == dir_FF_left &&
            (wall == wall_type_2_left ||
             wall == wall_type_3_chomper ||
             wall == wall_type_4_wall)) {
        // standing at chomper, mirror or wall
        return;
    }
    if (inactive.direction == dir_0_right &&
            (wall == wall_type_1_right ||
             wall == wall_type_4_wall)) {
        // standing at closed gate, doortop or wall
        return;
    }
    inactive_tile = get_tile(inactive.room, get_tile_div_mod_m7(inactive.x) + dir_front[inactive.direction + 1], inactive.row);
    wall = wall_type2(inactive_tile);
    if (inactive.direction == dir_FF_left &&
            (wall == wall_type_1_right ||
             wall == wall_type_4_wall ||
             !tile_is_floor(inactive_tile) ||
             (inactive_tile == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0))
       ) {
        // behind closed gate, doortop, wall, loose tile or abiss
        return;
    }
    if (inactive.direction == dir_0_right &&
            (wall == wall_type_2_left ||
             wall == wall_type_3_chomper ||
             wall == wall_type_4_wall ||
             !tile_is_floor(inactive_tile) ||
             (inactive_tile == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0))
       ) {
        // behind chomper, mirror, wall, loose tile or abiss
        return;
    }
    if (Kid.curr_row == inactive.row + 1 && droppedout) {
        seqtbl_inactive_char(inactive_seq_advance);
    } else if (Kid.direction != inactive.direction) {
        if (inactive.direction == Guard.direction &&
                ABS(Kid.x - inactive.x) > ABS(Guard.x - inactive.x) &&
                ABS(Guard.x - inactive.x) > 14 &&
                char_opp_dist() > 23) {
            seqtbl_inactive_char(inactive_seq_advance);
        } else if (inactive.direction != Guard.direction &&
                   Kid.direction == Guard.direction &&
                   Kid.sword == sword_0_sheathed &&
                   inactive_guard_kid_dist() > 23) {
            seqtbl_inactive_char(inactive_seq_advance);
        } else {
            seqtbl_inactive_char(inactive_seq_stand);
        }
    } else if (can_inactive_guard_strike(inactive)) {
        seqtbl_inactive_char(inactive_seq_stand);
    } else {
        seqtbl_inactive_char(inactive_seq_advance);
    }
    play_inactive_seq();
}

// custom logic - makes inactive guard step back
void autocontrol_inactive_guard_retreat(char_type2 inactive) {
    int inactive_tile, wall;
    if (Kid.frame != frame_152_strike_2) {
        return;
    }
    inactive_tile = get_tile(inactive.room, get_tile_div_mod_m7(inactive.x), inactive.row);
    wall = wall_type2(inactive_tile);
    if (inactive.direction == dir_FF_left &&
            (wall == wall_type_1_right ||
             wall == wall_type_4_wall)) {
        // standing at closed gate, doortop or wall
        return;
    }
    if (inactive.direction == dir_0_right &&
            (wall == wall_type_2_left ||
             wall == wall_type_3_chomper ||
             wall == wall_type_4_wall)) {
        // standing at chomper, mirror or wall
        return;
    }
    inactive_tile = get_tile(inactive.room, get_tile_div_mod_m7(inactive.x) + dir_behind[inactive.direction + 1], inactive.row);
    wall = wall_type2(inactive_tile);
    if (inactive.direction == dir_FF_left &&
            (wall == wall_type_2_left ||
             wall == wall_type_3_chomper ||
             wall == wall_type_4_wall ||
             !tile_is_floor(inactive_tile) ||
             (inactive_tile == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0))
       ) {
        // in front of chomper, mirror, wall, loose tile or abiss
        return;
    }
    if (inactive.direction == dir_0_right &&
            (wall == wall_type_1_right ||
             wall == wall_type_4_wall ||
             !tile_is_floor(inactive_tile) ||
             (inactive_tile == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0))
       ) {
        // in front of closed gate, doortop, wall, loose tile or abiss
        return;
    }
    if (inactive.direction == dir_FF_left) {
        inactive_tile = get_tile(inactive.room, get_tile_div_mod_m7(inactive.x) + dir_behind[inactive.direction + 1] * 2, inactive.row);
        if (!tile_is_floor(inactive_tile) ||
                inactive_tile == tiles_18_chomper ||
                (inactive_tile == tiles_11_loose && get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0)
           ) {
            // in front of chomper, loose tile or abiss
            return;
        }
    }
    seqtbl_inactive_char(inactive_seq_retreat);
    play_inactive_seq();
}

// custom logic - inactive guard strikes kid in the back
void autocontrol_inactive_guard_strike(char_type2 inactive) {
    int tile_behind;
    if (Kid.direction == inactive.direction && inactive.frame == frame_171_stand_with_sword) {
        tile_behind = get_tile(inactive.room, get_tile_div_mod_m7(inactive.x) + dir_behind[inactive.direction + 1], inactive.row);
        if (Guard.direction != inactive.direction && inactive_guard_kid_dist() < 19) {
            if (tile_is_floor(tile_behind) &&
                    wall_type2(tile_behind) == wall_type_0_none &&
                    (tile_behind != tiles_11_loose || get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 1)) {
                inactive_guard[inactive.room - 1].x += (inactive.direction == dir_FF_left ? 7 : -7);
            } else {
                seqtbl_inactive_char(inactive_seq_advance);
                play_inactive_seq();
                return;
            }
        }
    }
    inactive_refrac_timer = 8;
    seqtbl_inactive_char(inactive_seq_strike);
    play_inactive_seq();
}

// custom logic - inactive guard takes or blocks kid's strikes
void autocontrol_inactive_guard_hurt_or_block(char_type2 inactive) {
    int tile_behind, room_minus_1, guard_type;
    byte color, hitp;
    bool is_floor_behind;
    if (Guard.action == actions_99_hurt || !have_sword) {
        return;
    }
    tile_behind = get_tile(inactive.room, get_tile_div_mod_m7(inactive.x) + dir_behind[inactive.direction + 1], inactive.row);
    room_minus_1 = inactive.room - 1;
    is_floor_behind = tile_is_floor(tile_behind) &&
                      wall_type2(tile_behind) == wall_type_0_none &&
                      (tile_behind != tiles_11_loose || get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 1);
    color = inactive.color & 0x0F;
    hitp = (inactive.color & 0xF0) >> 4;
    if (prandom(2) > (inactive_guard_skill >= 5 ? 1 : 0) &&
            Guard.direction != dir_56_none &&
            Guard.action != actions_3_in_midair &&
            Guard.action != actions_4_in_freefall) {
        // guard is getting struck by countering kid
        if (Guard.frame == frame_153_strike_3 || Guard.frame == frame_154_poking) {
            return;
        }
        // active guard is significantly ahead
        if (kid_guard_dist() - inactive_guard_kid_dist() < -7) {
            return;
        }
        guard_type = current_level <= 15 ? settings->tbl_guard_type[current_level] : -1;
        if (guard_type != guard_type_skeleton && guard_type != guard_type_Jaffar_vision) {
            inactive_guard[room_minus_1].color = color;
            if (hitp > 1) {
                inactive_guard[room_minus_1].color |= (hitp - 1) << 4;
            } else {
                inactive_guard[room_minus_1].color |= 1 << 4;
            }
            inactive_guard[room_minus_1].alt_frame = 0x56; // hurt splash
        }
        play_sound(sound_34_hit_by_sword);
        switch (guard_type) {
            case guard_type_skeleton:
                break;
            case guard_type_Jaffar_vision:
                delayed_sound_timer = 1;
                delayed_sound = (struct delayed_sound_type) {
                    sound_57_sword_striked, 0, 0, 0, 0, 0, 0
                };
                break;
            default:
                delayed_sound_timer = 1;
                delayed_sound = (struct delayed_sound_type) {
                    sound_12_guard_hurt, 0, 0, 0, 0, 0, 0
                };
                break;
        }
        if (is_floor_behind) {
            inactive_guard[room_minus_1].x -= (inactive.direction == dir_FF_left ? -4 : 4);
        }
        inactive_refrac_timer = 8;
        if (hitp > 1 || guard_type == guard_type_skeleton || guard_type == guard_type_Jaffar_vision) {
            seqtbl_inactive_char(inactive_seq_block_retreat);
        } else {
            seqtbl_inactive_char(inactive_seq_dropdead);
        }
    } else if (is_floor_behind && (prandom(1) == 0 || inactive_refrac_timer > 0)) {
        seqtbl_inactive_char(inactive_seq_block_retreat);
    } else {
        if (inactive_refrac_timer > 3) {
            return;
        }
        inactive_refrac_timer = 8;
        seqtbl_inactive_char(inactive_seq_counter);
    }
    play_inactive_seq();
}

// custom logic - called when inactive guard striked and hurt kid
void autocontrol_inactive_guard_hurt_kid(char_type2 inactive) {
    short distance;
    if (hitp_curr == 0 || Kid.alive >= 0 || Kid.curr_row != inactive.row) {
        return;
    }
    distance = inactive_guard_kid_dist();
    loadkid();
    if (Char.direction != inactive.direction) {
        distance -= 3;
    }
    if (distance > 27) {
        seqtbl_inactive_char(inactive_seq_stand);
        play_inactive_seq();
        return;
    }
    droppedout = 0;
    if (Char.direction != inactive.direction &&
            (Char.frame == frame_150_parry ||
             Char.frame == frame_161_parry ||
             Char.frame == frame_162_block_to_strike ||
             Char.frame == frame_167_blocked ||
             Char.frame == frame_169_begin_block ||
             (Char.frame == frame_158_stand_with_sword && ctrl1_up == control_m1_held_left))) {
        inactive_refrac_timer = 8;
        inactive_guard[drawn_room - 1].frame2 = frame_167_blocked;
        play_sound(sound_10_sword_vs_sword);
    } else {
        charid_hurt_by_sword = charid_0_kid;
        Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
        if (take_hp(Char.sword == sword_2_drawn ? 1 : 100)) {
            Char.x = char_dx_forward(4);
            seqtbl_offset_char(seq_85_stabbed_to_death);
        } else {
            Char.sword = sword_2_drawn;
            if (inactive.direction == Char.direction) {
                seqtbl_offset_char(seq_119_hit_by_sword_from_behind);
            } else {
                in_wall_engarde();
                seqtbl_offset_char(seq_74_hit_by_sword);
            }
        }
        play_seq();
        play_sound(sound_34_hit_by_sword);
        if (!is_kid_shadow) {
            delayed_sound_timer = 1;
            delayed_sound = (struct delayed_sound_type) {
                sound_13_kid_hurt, 0, 0, 0, 0, 0, 0
            };
        }
    }
    savekid();
}

// custom logic - simulates inactive guard's death
void inactive_guard_dead(void) {
    char_type2 inactive;
    inactive = inactive_guard[drawn_room - 1];
    if (inactive.room != drawn_room || inactive.direction == dir_56_none || inactive.frame == frame_185_dead) {
        return;
    }
    seqtbl_inactive_char(inactive_seq_dropdead);
}

// custom logic - restores collapsed guard
void inactive_guard_restore(void) {
    char_type2* inactive;
    short delta_x;
    inactive = &inactive_guard[drawn_room - 1];
    if (inactive->room != drawn_room || inactive->direction == dir_56_none) {
        return;
    }
    if (inactive->frame >= frame_179_collapse_1 && inactive->frame <= frame_185_dead) {
        switch (inactive->frame) {
            case frame_183_collapse_5:
                delta_x = -1;
                break;
            case frame_185_dead:
                delta_x = 3;
                break;
            default:
                delta_x = 0;
                break;
        }
        if (inactive->direction < dir_0_right) {
            delta_x = -delta_x;
        }
        inactive->x += delta_x;
        inactive->frame2 = frame_166_stand_inactive;
    }
}

// custom logic - leave dead guard in an available dead slot
void leave_dead_guard(word room_minus_1, byte char_frame) {
    const int outbound_left_x = 41;
    const int outbound_right_x = 215;
    const int outbound_delta = 35;
    char_type2* slot;
    slot = get_dead_guard_slot(room_minus_1);
    if (slot != NULL) {
        slot->room = room_minus_1 + 1;
        slot->direction = Char.direction;
        slot->row = Char.curr_row;
        slot->x = Char.x;
        slot->color = Char.color;
        slot->frame2 = char_frame;
        slot->alt_frame = Char.repeat;
        leave_outbound_guard(room_minus_1, slot->frame, slot->direction, slot->x, slot->row, slot->alt_frame, slot->color);
    }
    if ((dynamic_guard.x == outbound_left_x && Kid.x >= outbound_left_x + outbound_delta) ||
            (dynamic_guard.x == outbound_right_x && Kid.x <= outbound_right_x - outbound_delta)) {
        enter_dynamic_guard();
    } else {
        is_dynamic_guard_fight = 0;
        clear_char();
        level.guards_tile[room_minus_1] = 0xFF;
        on_guard_killed();
    }
}

// seg006:0D85
void play_guard(void) {
    word room_minus_1;
    room_minus_1 = drawn_room - 1;
    // custom logic - if inactive guard can see kid but active guard cannot, swap them
    check_inactive_guard_proximity();
    if (Char.charid == charid_24_mouse) {
        autocontrol_opponent();
    } else {
        // custom logic - do not retract spikes when guard died on them
        if (Char.frame >= frame_179_collapse_1 && Char.frame <= frame_185_dead && !is_disappearing_guard()) {
            load_frame_to_obj();
            set_char_collision();
            if (get_tile(Char.room, get_tile_div_mod_m7(char_x_left), Char.curr_row) == tiles_2_spike &&
                    curr_room_modif[curr_tilepos] != 0 &&
                    curr_room_modif[curr_tilepos] != 0xFF) {
                curr_room_modif[curr_tilepos] = 0xFF;
            }
            if (get_tile(Char.room, get_tile_div_mod_m7(char_x_right), Char.curr_row) == tiles_2_spike &&
                    curr_room_modif[curr_tilepos] != 0 &&
                    curr_room_modif[curr_tilepos] != 0xFF) {
                curr_room_modif[curr_tilepos] = 0xFF;
            }
        }
        if (Char.alive < 0) {
            if (guardhp_curr == 0 &&
                    Char.action != actions_3_in_midair &&
                    Char.action != actions_4_in_freefall &&
                    // bump fall
                    (Char.action != actions_5_bumped || Char.frame == frame_185_dead || can_guard_see_kid == 2)) {
                Char.alive = 0;
                music_suspended_timer = DROPDEAD_FRAMES + 1;
                if (dynamic_guard.room == drawn_room &&
                        (inactive_guard[room_minus_1].room == 0 || !can_inactive_guard_see_kid) &&
                        Kid.room == drawn_room &&
                        (level.guards_seq_hi[room_minus_1] == 0 ||
                         dynamic_guard_count[room_minus_1] == 0 ||
                         (dynamic_guard_count[room_minus_1] == 1 && dynamic_guard.frame > 0))) {
                    is_dynamic_guard_fight = 2;
                } else {
                    is_dynamic_guard_fight = dynamic_guard.room == drawn_room &&
                                             inactive_guard[room_minus_1].room == drawn_room &&
                                             is_inactive_guard_ready(inactive_guard[room_minus_1]) ?
                                             2 :
                                             0;
                    on_guard_killed();
                }
            } else {
                goto loc_control_guard;
            }
            // custom logic - dynamic guard logic
        } else if (Char.alive >= 0 &&
                   dynamic_guard.room == drawn_room &&
                   (inactive_guard[room_minus_1].room == 0 || !can_inactive_guard_see_kid) &&
                   Kid.room == drawn_room &&
                   Char.room == drawn_room &&
                   Char.alive < (DROPDEAD_FRAMES - 1) &&
                   Char.direction != dir_56_none) {
            // guard is still falling (spikes might not have a floor in the temple)
            if (Char.frame != frame_177_spiked && !tile_is_floor(get_tile_at_char())) {
                Char.alive = -1;
                return;
            }
            if (is_dead() || Char.curr_col < -1 || Char.curr_col > 10) {
                Char.alive = DROPDEAD_FRAMES - 1;
            } else {
                ++Char.alive;
            }
        } else if (Char.alive == (DROPDEAD_FRAMES - 1) &&
                   dynamic_guard.room == drawn_room &&
                   (inactive_guard[room_minus_1].room == 0 || !can_inactive_guard_see_kid) &&
                   Kid.room == drawn_room &&
                   Char.room == drawn_room &&
                   Char.direction != dir_56_none) {
            Char.alive = DROPDEAD_FRAMES;
            if (!is_kid_shadow && !(current_level == 10 && leveldoor_open >= DEAD_KID_LEVEL_10_TIMER)) {
                byte char_frame = is_dead() ? Char.frame : frame_185_dead;
                if (dynamic_guard_count[room_minus_1] == 0) {
                    ++Char.alive;
                    leave_dead_guard(room_minus_1, char_frame);
                } else if (dynamic_guard_count[room_minus_1] == 1 && dynamic_guard.frame > 0) { // frame is used to store the number of additional guards
                    ++Char.alive;
                    leave_dead_guard(room_minus_1, char_frame);
                } else if (level.guards_seq_hi[room_minus_1] == 0) {
                    is_dynamic_guard_fight = 0;
                    on_guard_killed();
                }
            } else if (level.guards_seq_hi[room_minus_1] == 0) {
                is_dynamic_guard_fight = 0;
                on_guard_killed();
            }
        }
        if (Char.charid == charid_1_shadow && guardhp_curr == 0) {
            if (Guard.room == Kid.room &&
                    Guard.direction != dir_56_none &&
                    Guard.curr_col >= -1 && Guard.curr_col < 10 &&
                    Guard.curr_row >= 0 && Guard.curr_row < 3) {
                hitp_delta = -hitp_curr;
                Kid.alive = 0;
            }
            clear_char();
        }
loc_control_guard:
        autocontrol_opponent();
        control();
    }
}

// seg006:0DC0
void user_control(void) {
    if (Char.direction >= dir_0_right) {
        flip_control_x();
        control();
        flip_control_x();
    } else {
        control();
    }
}

// seg006:0DDC
void flip_control_x(void) {
    byte temp;
    control_x = -control_x;
    temp = control_forward;
    control_forward = control_backward;
    control_backward = temp;
}

// seg006:0E00
int release_arrows(void) {
    control_backward = control_forward = control_up = control_down = control_0_released;
    return control_1_ignore;
}

// custom logic - reset ignored arrows when standing still
void reset_ignored_arrows(void) {
    if (control_backward == control_1_ignore) {
        control_backward = control_0_released;
    }
    if (control_forward == control_1_ignore) {
        control_forward = control_0_released;
    }
    if (control_up == control_1_ignore) {
        control_up = control_0_released;
    }
    if (control_down == control_1_ignore) {
        control_down = control_0_released;
    }
}

// seg006:0E12
void save_ctrl_1(void) {
    ctrl1_forward = control_forward;
    ctrl1_backward = control_backward;
    ctrl1_up = control_up;
    ctrl1_down = control_down;
    ctrl1_shift2 = control_shift2;
}

// seg006:0E31
void rest_ctrl_1(void) {
    control_forward = ctrl1_forward;
    control_backward = ctrl1_backward;
    control_up = ctrl1_up;
    control_down = ctrl1_down;
    control_shift2 = ctrl1_shift2;
}

// seg006:0E8E
void clear_saved_ctrl(void) {
    ctrl1_forward = ctrl1_backward = ctrl1_up = ctrl1_down = ctrl1_shift2 = control_0_released;
}

// seg006:0EAF
void read_user_control(void) {
    if (control_forward >= control_0_released) {
        if (control_x == control_m1_held_left) {
            if (control_forward == control_0_released) {
                control_forward = control_m1_held;
            }
        } else {
            control_forward = control_0_released;
        }
    }
    if (control_backward >= control_0_released) {
        if (control_x == control_1_held_right) {
            if (control_backward == control_0_released) {
                control_backward = control_m1_held;
            }
        } else {
            control_backward = control_0_released;
        }
    }
    if (control_up >= control_0_released) {
        if (control_y == control_m1_held_up) {
            if (control_up == control_0_released) {
                control_up = control_m1_held;
            }
        } else {
            control_up = control_0_released;
        }
    }
    if (control_down >= control_0_released) {
        if (control_y == control_1_held_down) {
            if (control_down == control_0_released) {
                control_down = control_m1_held;
            }
        } else {
            control_down = control_0_released;
        }
    }
    if (control_shift2 >= control_0_released) {
        if (control_shift == control_m1_held) {
            if (control_shift2 == control_0_released) {
                control_shift2 = control_m1_held;
            }
        } else {
            control_shift2 = control_0_released;
        }
    }
}

// seg006:0F55
int can_grab(void) {
    // Can char grab curr_tile2 through through_tile?
    byte modifier;
    modifier = curr_room_modif[curr_tilepos];
    // custom logic - fake walls can be grabbed
    if (is_fake_wall() == 1) {
        return 1;
    }
    // can't grab through wall
    if (through_tile == tiles_20_wall) return 0;
    // can't grab through a door top if looking right
    if (through_tile == tiles_12_doortop && Char.direction >= dir_0_right) return 0;
    // can't grab through floor
    if (tile_is_floor(through_tile)) return 0;
    // can't grab a shaking loose floor
    // Allow climbing onto a shaking loose floor if the delay is greater than the default. TODO: This should be a separate option.
    if (curr_tile2 == tiles_11_loose && modifier != 0 && !(settings->loose_floor_delay > 11)) return 0;
    // a doortop with floor can be grabbed only from the left (looking right)
    if (curr_tile2 == tiles_7_doortop_with_floor && Char.direction < dir_0_right) return 0;
    // custom logic - do not try to grab a chomper from the left unless prince turned into shadow
    if (curr_tile2 == tiles_18_chomper && Char.direction == dir_0_right && !is_kid_shadow) return 0;
    // can't grab something that has no floor
    if (!tile_is_floor(curr_tile2)) return 0;
    return 1;
}

// seg006:0FC3
int wall_type(byte tiletype) {
    // custom logic - fake walls are not walls
    if (is_fake_wall() == 1) {
        return wall_type_0_none;
    }
    // custom logic - gates on level 12B act inversely for kid
    else if (current_level == settings->inverse_level &&
             Char.charid == charid_0_kid &&
             tiletype == tiles_4_gate &&
             curr_room_modif[curr_tilepos] == 0) {
        return wall_type_0_none;
    }
    switch (tiletype) {
        case tiles_4_gate:
        case tiles_7_doortop_with_floor:
        case tiles_12_doortop:
            return wall_type_1_right;
        case tiles_13_mirror:
            return wall_type_2_left;
        case tiles_18_chomper:
            return wall_type_3_chomper;
        case tiles_20_wall:
            return wall_type_4_wall;
        default:
            return wall_type_0_none; // no wall
    }
}

// custom logic - does not consider an open gate as a wall
int wall_type2(byte tiletype) {
    if (tiletype == tiles_4_gate && !can_bump_into_gate()) {
        return wall_type_0_none;
    }
    return wall_type(tiletype);
}

// seg006:1005
int get_tile_above_char(void) {
    return get_tile(Char.room, Char.curr_col, Char.curr_row - 1);
}

// seg006:1020
int get_tile_behind_above_char(void) {
    return get_tile(Char.room, dir_behind[Char.direction + 1] + Char.curr_col, Char.curr_row - 1);
}

// seg006:1049
int get_tile_front_above_char(void) {
    return get_tile(Char.room, infrontx = dir_front[Char.direction + 1] + Char.curr_col, Char.curr_row - 1);
}

// seg006:1072
int back_delta_x(int delta_x) {
    if (Char.direction < dir_0_right) {
        // direction = left
        return delta_x;
    } else {
        // direction = right
        return -delta_x;
    }
}

// seg006:108A
void do_pickup(int obj_type) {
    bool is_flaming_sword;
    is_flaming_sword = obj_type == -1 &&
                       current_level == settings->flame_level &&
                       curr_room == settings->flaming_sword_room &&
                       curr_tilepos == settings->flaming_sword_tilepos;
    pickup_obj_type = obj_type;
    control_shift2 = control_1_ignore;
    // custom logic - kid dies when picking a flaming sword unless he has turned into shadow
    if (is_flaming_sword && !is_kid_shadow) {
        Char.direction = dir_FF_left;
        Char.x = x_bump[curr_tilepos % 10 + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
        Char.x = char_dx_forward(-3);
        seqtbl_offset_char(seq_50_crouch);
        play_seq();
        load_flame_sprite();
        stop_sounds();
        play_sound(sound_1_falling);
        take_hp(ALL_HP);
        holding_sword = 1;
        dead_kid_timer = -18;
        flash_color = color_12_brightred;
        flash_time = 12;
        return;
    }
    // erase picked up item
    curr_room_tiles[curr_tilepos] = tiles_1_floor;
    curr_room_modif[curr_tilepos] = 0;
    redraw_height = 35;
    set_wipe(curr_tilepos, 1);
    set_redraw_full(curr_tilepos, 1);
    // custom logic - redraw a tile to the left of flaming sword
    if (is_flaming_sword) {
        set_wipe(curr_tilepos - 1, 1);
        set_redraw_full(curr_tilepos - 1, 1);
    }
}

// seg006:10E6
void check_press(void) {
    short frame;
    short action;
    short room_minus_1;
    int tile_infront;
    frame = Char.frame;
    action = Char.action;
    // frames 87..99: hanging
    // frames 135..140: start climb up
    if ((frame >= frame_87_hanging_1 && frame <= frame_99_hanging_13) || (frame >= frame_135_climbing_1 && frame < frame_141_climbing_7)) {
        // the pressed tile is the one that the char is grabbing
        get_tile_above_char();
    } else if (action == actions_7_turn || action == actions_5_bumped || action < actions_2_hang_climb) {
        // frame 79: jumping up
        if (frame == frame_79_jumphang && super_jump_timer == 0) {
            tile_infront = get_tile_above_char();
            // custom logic - checks for the top loose tile more accurately
            if (tile_infront != tiles_11_loose) {
                tile_infront = get_tile(Char.room, get_tile_div_mod_m7(Char.x + dir_behind[Char.direction + 1]), Char.curr_row - 1);
            }
            // break a loose floor from above
            if (tile_infront == tiles_11_loose) {
                make_loose_fall(1);
            } else {
                return;
            }
        } else {
            // the pressed tile is the one that the char is standing on
            if (!(cur_frame.flags & FRAME_NEEDS_FLOOR)) {
                // custom logic - frame 59 should be checked when pressing tiles but not when in the air
                if (frame != frame_59_runturn) {
                    return;
                }
            }
            determine_col();
            get_tile_at_char();
        }
    } else {
        // custom logic - still check guards in other rooms for pressed tiles
        goto loc_other_room_press;
    }
    // custom logic - is the current character lightweight (in feather mode or shadow/mouse)
    bool lightweight_char = (Char.charid == charid_0_kid && (is_kid_shadow || is_kid_mouse())) || feather_fall_timer > 0;
    // custom logic - does the frame put extra load on tiles
    bool heavy_frame = frame == frame_17_standing_jump_2 ||
                       frame == frame_30_standing_jump_15 ||
                       frame == frame_37_start_run_jump_4 ||
                       frame == frame_43_running_jump_4 ||
                       frame == frame_68_start_jump_up_2 ||
                       frame == frame_85_hangdrop_5 ||
                       frame == frame_108_fall_land_2 ||
                       frame == frame_175_jumpfall_5 ||
                       frame == frame_183_collapse_5 ||
                       (frame >= frame_87_hanging_1 && frame <= frame_93_hanging_7) ||
                       (frame >= frame_135_climbing_1 && frame <= frame_141_climbing_7) ||
                       Char.charid == charid_1_shadow || // shadow (as a guard) should always press opener/closer tiles
                       (Char.charid == charid_2_guard && is_guarding_gate); // gate guarding guards have heavy feet
    if (curr_tile2 == tiles_15_opener || curr_tile2 == tiles_6_closer) {
        if (Char.alive < 0 || Char.charid == charid_0_kid) {
loc_press_button:
            // custom logic - put extra load to press tiles in feather mode
            if (!feather_fall_timer || heavy_frame) {
                trigger_button(1, 0, -1);
                // inactive guards can hear button tiles
                if (is_inactive_guard_notice >= 0 && is_kid_shadow && Char.charid == charid_0_kid) {
                    is_inactive_guard_notice = 2;
                }
            }
        } else {
            // custom logic - bowing guards should not make buttons stuck
            if (Char.frame >= frame_190_start_bowing_1 && Char.frame <= frame_202_bowing) {
                goto loc_press_button;
            } else {
                died_on_button(Char.charid);
            }
        }
    } else if (curr_tile2 == tiles_11_loose) {
        // custom logic - lightweight characters in most frames do not put enough pressure
        // on loose tiles
        if (!lightweight_char ||
                get_tile_ctrl_bits(curr_room, curr_tilepos, 1) == 1 || // totally loose tile
                heavy_frame) {
            make_guard_notice(false);
            make_loose_fall(1);
            // custom logic - indicates that a guard knocked the tile down
            guard_knocked_tile = Char.charid == charid_2_guard ? 1 : 0;
        }
    }
loc_other_room_press:
    // custom logic - press buttons in other rooms that guards are standing on
    if (Char.charid != charid_0_kid) { // no need to do this twice when a guard is in the room
        return;
    }
    for (room_minus_1 = 0; room_minus_1 < level.used_rooms; ++room_minus_1) {
        if (level.opener_col[room_minus_1] == 0 || level.opener_row[room_minus_1] == 0) {
            continue;
        }
        // clear values in the current room
        if (room_minus_1 == Char.room - 1) {
            level.opener_col[room_minus_1] = 0;
            level.opener_row[room_minus_1] = 0;
            continue;
        }
        // prison cell guard on level 1 should not activate the gate
        if (current_level == 1 && room_minus_1 == 17) {
            continue;
        }
        // do not press buttons in column 9 in the room on the left
        if (level.opener_col[room_minus_1] == 10 && level.roomlinks[room_minus_1].right == Char.room) {
            continue;
        }
        // opener_col/opener_row arrays default to 0. The stored value had been incremented by 1.
        get_tile(room_minus_1 + 1, level.opener_col[room_minus_1] - 1, level.opener_row[room_minus_1] - 1);
        if (curr_tile2 == tiles_15_opener) {
            trigger_button(0, 0, -1);
        }
    }
}

// custom logic - checks whether inactive guard is on a button
void check_inactive_press(void) {
    char_type2 inactive;
    short frame;
    frame_type inactive_frame;
    sbyte delta_x;
    short x_dx_weight;
    // do not press buttons during feather fall
    if (feather_fall_timer > 0) {
        return;
    }
    inactive = inactive_guard[Kid.room - 1];
    if (inactive.room != Kid.room || inactive.direction == dir_56_none) {
        return;
    }
    frame = inactive.frame;
    // load inactive guard frame
    if (frame >= 102 && frame < 107) frame += 70;
    inactive_frame = frame_tbl_guard[frame - 149];
    delta_x = inactive_frame.dx - (inactive_frame.flags & FRAME_WEIGHT_X);
    if (inactive.direction < dir_0_right) {
        delta_x = -delta_x;
    }
    // calculate inactive x + dx - frame weight
    x_dx_weight = inactive.x + delta_x;
    if (!(inactive_frame.flags & FRAME_NEEDS_FLOOR)) {
        return;
    }
    get_tile(inactive.room, get_tile_div_mod_m7(x_dx_weight), inactive.row);
    if (curr_tile2 == tiles_15_opener || curr_tile2 == tiles_6_closer) {
        if (inactive.frame == frame_185_dead) {
            died_on_button(charid_2_guard);
        } else {
            trigger_button(1, 0, -1);
        }
    }
}

// seg006:1199
void check_spike_near(void) {
    short not_finished;
    short room;
    short row;
    short col;
    short left_col;
    short right_col;
    short frame;
    bool is_spiked;
    frame = Char.frame;
    is_spiked = frame == frame_177_spiked;
    // custom logic - vertical spikes are only activated when the bottom of kid's sprite is at a spike tile
    if (is_temple()) {
        is_spiked = is_spiked || (Char.charid == charid_0_kid && frame == frame_184_spiked_hanging);
        load_fram_det_col();
        if (get_tile(Char.room, char_col_climbing_left(), Char.curr_row) == tiles_20_wall &&
                distance_to_edge_weight() < 4 &&
                Char.direction == dir_0_right &&
                frame >= frame_102_start_fall_1 && frame < frame_106_fall) {
            in_wall();
        }
        if (curr_tile2 == tiles_2_spike && y_land[Char.curr_row + ONSCREEN_ROW_OFFSET] <= Char.y + 20) {
            short spike_modif = curr_room_modif[curr_tilepos];
            short distance = distance_to_edge_weight();
            if (Char.direction == dir_FF_left && spike_modif == 0) {
                if ((frame >= frame_53_runturn && frame <= frame_65_runturn) ||
                        (frame >= frame_76_jumphang && frame <= frame_85_hangdrop_5)) {
                    distance += 4;
                }
                if (distance >= 12) {
                    return;
                }
            } else if (Char.direction == dir_0_right && spike_modif == 0) {
                if (Char.sword == sword_2_drawn) {
                    distance = MAX(distance - 4, 0);
                }
                if (distance <= 2 ||
                        (frame >= frame_70_jumphang && frame <= frame_99_hanging_13) ||
                        (frame >= frame_135_climbing_1 && frame <= frame_149_climbing_15)) {
                    return;
                }
            }
        } else if (Char.direction != dir_0_right ||
                   frame < frame_97_hanging_11 ||
                   frame > frame_99_hanging_13 ||
                   get_tile_behind_char() != tiles_2_spike) {
            return;
        }
    }
    // custom logic - disappearing guard does not stop spikes from retracting
    if (Char.charid == charid_2_guard && is_spiked && is_disappearing_guard()) {
        return;
    }
    row = Char.curr_row;
    room = Char.room;
    right_col = get_tile_div_mod_m7(char_x_right);
    if (room == 0 || right_col < 0) return;
    if (is_spiked) {
        left_col = right_col = Char.curr_col;
    } else {
        left_col = get_tile_div_mod_m7(char_x_left);
    }
    for (col = left_col; col <= right_col; ++col) {
        row = Char.curr_row;
        do {
            not_finished = 0;
            if (get_tile(room, col, row) == tiles_2_spike) {
                start_anim_spike(curr_room, curr_tilepos);
            } else if (
                !tile_is_floor(curr_tile2) &&
                curr_room != 0 &&
                curr_room == room &&
                row <= 1 + (Char.curr_row >= 2)
            ) {
                ++row;
                not_finished = 1;
            }
        } while(not_finished);
    }
}

// seg006:1231
void clip_char(void) {
    short frame;
    short room;
    short action;
    short col;
    short clipped_y;
    short row;
    short row_1;
    frame = Char.frame;
    action = Char.action;
    room = Char.room;
    row = Char.curr_row;
    reset_obj_clip();
    // custom logic - clip shadow moonwalking
    if (Char.charid == charid_1_shadow &&
            current_level == 9 &&
            Char.direction == 1 &&
            Char.room == 1 &&
            Char.curr_row == 1 &&
            Char.x < 190) {
        obj_clip_right = (col_xh[9] << 3) + 8;
        return;
    }
    // custom logic - clip kid during a super jump when jumping up into a tapestry with lattice
    if (settings->tbl_level_type[current_level] == PALACE_LEVEL &&
            settings->tbl_level_subtype[current_level] != TEMPLE_LEVEL) {
        if (frame == frame_79_jumphang || frame == frame_106_fall) {
            int top_left_tile = get_tile(room, char_col_left - 1, y_to_row_mod4(char_top_y + 10));
            if (top_left_tile == tiles_12_doortop && curr_room_modif[curr_tilepos] == 0 /* doortop with lattice */) {
                obj_clip_top = y_land[tile_row + ONSCREEN_ROW_OFFSET] - 22;
                return;
            }
        }
    }
    // custom logic - fixes a visual bug with a pillar when kid is falling/climbing on the left side
    if (frame >= frame_102_start_fall_1 && frame <= frame_106_fall && // falling
            char_top_row >= 0 &&
            (get_tile(room, char_col_left, char_top_row) == tiles_3_pillar ||
             get_tile(room, char_col_left, row) == tiles_3_pillar)) {
        int pillar_x = x_bump[tile_col + ONSCREEN_COLUMN_OFFSET] + 6; // 1 pixel to the left from the tile center line
        if (tile_row >= 0 && char_x_left <= pillar_x && char_x_right > pillar_x) {
            obj_clip_right = (col_xh[tile_col] << 3) + 29;
            obj_clip_top = y_clip[tile_row];
            return;
        }
    }
    // frames 217..228: going up the level door
    if (frame >= frame_224_exit_stairs_8 && frame < frame_229_found_sword) {
        // custom logic - clip kid correctly when there are multiple open doors or a balcony
        if (get_tile(Char.room, char_col_left - 2, row) == tiles_16_level_door_left) {
            obj_clip_right = (col_xh[tile_col + 1] << 3) + 48;
            if (settings->tbl_level_type[current_level]) {
                obj_clip_right += 8;
            }
            obj_clip_top = leveldoor_ybottom + 1;
        } else if (get_tile(Char.room, char_col_left - 2, row) == tiles_23_balcony_left) {
            obj_clip_right = col_xh[tile_col] << 3;
            obj_clip_top = 65 * tile_row + 63;
        }
    } else {
        if (get_tile(room, char_col_left, char_top_row) == tiles_20_wall ||
                (tile_is_floor(curr_tile2) && (Char.direction != dir_0_right || Char.action != actions_6_hang_straight)) ||
                // custom logic - fixed clipping when hanging on a large pillar
                (Char.direction == dir_0_right &&
                 curr_tile2 == tiles_9_bigpillar_top &&
                 frame >= frame_93_hanging_7 &&
                 frame <= frame_99_hanging_13)) {
            // frame 79: jump up, frame 81: grab
            if ((action == actions_0_stand && (frame == frame_79_jumphang || frame == frame_81_hangdrop_1)) ||
                    // custom logic - fixed clipping when hanging
                    (Char.direction == dir_0_right && frame >= frame_87_hanging_1 && frame <= frame_88_hanging_2) ||
                    get_tile(room, char_col_right, char_top_row) == tiles_20_wall ||
                    tile_is_floor(curr_tile2)) {
                row_1 = row + 1;
                if (row_1 == 1 ||
                        ((clipped_y = y_clip[row_1]) < obj_y && clipped_y - 15 < char_top_y)
                   ) {
                    obj_clip_top = char_top_y = y_clip[row_1];
                }
            }
        }
        col = get_tile_div_mod(char_x_left_coll - 4);
        if (get_tile(room, col + 1, row) == tiles_7_doortop_with_floor ||
                curr_tile2 == tiles_12_doortop
           ) {
            obj_clip_right = (tile_col << 5) + 32;
        } else {
            if ((get_tile(room, col, row) != tiles_7_doortop_with_floor &&
                    curr_tile2 != tiles_12_doortop) ||
                    action == actions_3_in_midair ||
                    (action == actions_4_in_freefall && frame == frame_106_fall) ||
                    (action == actions_5_bumped && frame == frame_107_fall_land_1) ||
                    (Char.direction < dir_0_right && (
                         action == actions_2_hang_climb ||
                         action == actions_6_hang_straight ||
                         (action == actions_1_run_jump &&
                          frame >= frame_137_climbing_3 && frame < frame_140_climbing_6)
                     ))
               ) {
                if (get_tile(room, col = get_tile_div_mod(char_x_right_coll), y_to_row_mod4(Char.y)) == tiles_20_wall &&
                        get_tile(room, col, y_to_row_mod4(char_top_y)) == tiles_20_wall &&
                        room == curr_room) {
                    obj_clip_right = tile_col << 5;
                }
                // custom logic - fine tuned mirror clipping
                if (get_tile(room, col = get_tile_div_mod(char_x_right_coll), y_to_row_mod4(Char.y)) == tiles_13_mirror &&
                        Char.direction == dir_0_right &&
                        room == curr_room) {
                    obj_clip_right = (tile_col << 5) + 6;
                }
            } else {
                // custom logic - do not clip kid on the right side of a doortop
                if (char_x_left < x_bump[tile_col + ONSCREEN_COLUMN_OFFSET] + 13 &&
                        // or when climbing up facing left
                        (Char.direction != dir_FF_left || frame != frame_140_climbing_6)) {
                    obj_clip_right = (tile_col << 5) + 32;
                }
            }
        }
    }
}

// seg006:13E6
void stuck_lower(void) {
    if (get_tile_at_char() == tiles_5_stuck) {
        ++Char.y;
    }
}

// seg006:13F3
void set_objtile_at_char(void) {
    short char_frame;
    short char_action;
    short char_id;
    char_frame = Char.frame;
    char_action = Char.action;
    char_id = Char.charid;
    // custom logic - flame silhouette should be drawn away on z-axis
    if (char_frame == frame_212_flame_silhouette) {
        obj_tilepos = 30;
        return;
    }
    // custom logic - dead character positioning on z-axis
    if (Char.alive >= 0 && (is_dead() || (char_id == charid_2_guard && char_frame == frame_202_bowing))) {
        // draw dead kid body further away on z-axis
        if (is_kid_shadow &&
                char_id == charid_0_kid &&
                ABS(Char.curr_col - Kid.curr_col) <= 1 &&
                Char.curr_row == Kid.curr_row &&
                Kid.frame >= frame_179_collapse_1 &&
                Kid.frame <= frame_185_dead) { // kid uniting with shadow
            obj_tilepos = 30;
            // draw dead characters closer on z-axis unless there is dead kid body nearby
        } else if (Kid.alive < 0 && hitp_curr > 0 &&
                   !(char_id > charid_0_kid &&
                     Dead.room == Char.room &&
                     ABS(Char.x - Dead.x) <= 14 &&
                     Char.curr_row == Dead.row)) {
            obj_tilepos = -1;
        }
        // kid's chomped frame needs to be redrawn
        if (char_id != charid_0_kid || char_frame != frame_178_chomped) {
            return;
        }
        // custom logic - kid as shadow should appear behind the guard
    } else if (is_kid_shadow &&
               !is_shadow_flame &&
               char_id == charid_0_kid &&
               Guard.alive < 0 &&
               Guard.direction != dir_56_none &&
               ABS(Char.curr_col - Guard.curr_col) <= 1 &&
               Char.curr_row == Guard.curr_row) {
        obj_tilepos = 30;
        return;
    }
    if (char_action == actions_1_run_jump) {
        tile_row = char_bottom_row;
        tile_col = char_col_left;
    } else {
        tile_row = Char.curr_row;
        tile_col = Char.curr_col;
    }
    // frame 135..148: climbing
    if ((char_frame >= frame_135_climbing_1 && char_frame < frame_149_climbing_15) ||
            char_action == actions_2_hang_climb ||
            char_action == actions_3_in_midair ||
            char_action == actions_4_in_freefall ||
            char_action == actions_6_hang_straight
       ) {
        --tile_col;
    }
    obj_tilepos = get_tilepos_nominus(tile_col, tile_row);
}

// custom logic - reset shadow/flame palette on level 13
void reset_shadow_palette(void) {
    if (current_level != settings->flame_level) {
        return;
    }
    if (is_kid_shadow && is_shadow_flame) {
        load_flame_sprite();
    } else {
        free_chtab(chtab_addrs[id_chtab_2_kid]);
        chtab_addrs[id_chtab_2_kid] = NULL;
        load_kid_sprite();
    }
}

// seg006:1463
void proc_get_object(void) {
    bool picked_by_shadow;
    byte tile_col_num;
    picked_by_shadow = Char.charid == charid_1_shadow;
    if (Char.charid != charid_0_kid && !picked_by_shadow) {
        return;
    }
    if (pickup_obj_type == -1) {
        have_sword = 1;
        // custom logic - do not fall off into an abyss while picking up sword (potions use a different logic)
        if (!tile_is_floor(get_tile_behind_char())) {
            short distance = distance_to_edge_weight();
            if (distance > 8) {
                Char.x = char_dx_forward(distance - 8);
            }
        }
        // custom logic - shadow picks up the sword and becomes flame
        if (current_level == settings->flame_level && is_kid_shadow && !is_shadow_flame) {
            is_shadow_flame = true;
            enable_lighting = 0;
            reset_shadow_palette();
            hitp_curr = hitp_max;
            clear_kid_hp(hitp_max);
            draw_kid_hp(hitp_curr, hitp_max);
        }
        stop_sounds();
        if (current_level == settings->flame_level && is_kid_shadow && is_shadow_flame) {
            // custom logic - get flaming sword
            play_sound(is_positional_audio ? sound_79_united_with_shadow : sound_37_victory);
        } else {
            play_sound(is_positional_audio ? sound_77_get_sword : sound_37_victory); // get sword
        }
        flash_color = color_14_brightyellow;
        flash_time = 8;
    } else {
        // custom logic - potions have no affect on kid turned shadow except large potions
        if (Char.charid == charid_0_kid && is_kid_shadow) {
            if (pickup_obj_type == potion_2_life) {
                if (hitp_max < settings->max_hitp_allowed) {
                    ++shadow_potion_counter;
                }
                if (hitp_curr < hitp_max) {
                    hitp_delta = 1;
                }
            }
            return;
        }
        switch (pickup_obj_type) {
            // custom logic - an empty potion on different levels has different effects
            case potion_0_custom:
                switch (current_level) {
                    // large white potion on level 3 (drops loose tile on closer tile)
                    case 3:
                        get_tile(8, 4, 0);
                        trigger_button(0, tiles_6_closer, -1);
                        break;
                    // large cyan potion on level 5 (the antidote)
                    case 5:
                        set_health_life();
                        took_antidote = 2;
                        // set huge poison potion upright
                        get_tile(15, 0, 1);
                        curr_room_modif[curr_tilepos] = 2;
                        // poisoned guard wakes up
                        stop_sounds();
                        set_sound_pos(15, 1, 1);
                        if (is_positional_audio) {
                            play_sound(sound_76_skel_presentation);
                        } else {
                            play_sound(sound_33_small_potion);
                        }
                        flash_color = color_11_brightcyan;
                        flash_time = 10;
                        level.guards_seq_hi[14] = 0;
                        break;
                    // small yellow potion on levels 7-9 (makes torches brighter/dimmer)
                    case 7:
                    case 8:
                    case 9:
                        enable_lighting = 1 - enable_lighting;
                        need_full_redraw = 1;
                        break;
                    // large white potion on level 10 (creates floor tiles)
                    case 10:
                        is_shadow_text_shown = 1; // tracks whether the dead kid image should appear
                        // right side of room 19
                        delayed_sound_timer = 1;
                        delayed_sound = (struct delayed_sound_type) {
                            sound_20_loose_shake_1, 19, 6, 2, 0, 0, 3 /* repeat count */
                        };
                        // left side of room 19
                        redraw_height = 24;
                        for (tile_col_num = 0; tile_col_num <= 9; tile_col_num++) {
                            get_tile(19, tile_col_num, 2);
                            if (tile_col_num < 8) {
                                curr_room_tiles[curr_tilepos] = tiles_11_loose;
                            } else {
                                curr_room_tiles[curr_tilepos] = tiles_1_floor;
                            }
                            curr_room_modif[curr_tilepos] = 0;
                        }
                        // room 18
                        for (tile_col_num = 1; tile_col_num <= 9; tile_col_num++) {
                            get_tile(18, tile_col_num, 2);
                            curr_room_tiles[curr_tilepos] = tiles_11_loose;
                            curr_room_modif[curr_tilepos] = 0;
                        }
                        break;
                    case 12:
                        // large white potion on level 12 has mouse in it
                        if (potion_room == settings->mouse_2_room && potion_pos == get_tilepos(settings->mouse_2_potion_column, settings->mouse_2_potion_row)) {
                            if (leveldoor_open == 0) {
                                if (is_positional_audio) {
                                    stop_sounds();
                                    play_sound(sound_70_mouse_squeak);
                                }
                                leveldoor_open = 1;
                                potion_timer += 2;
                            }
                            // small yellow potion on level 12
                        } else {
                            enable_lighting = 1 - enable_lighting;
                            need_full_redraw = 1;
                        }
                        break;
                    // large red potion on level 12B (kills kid)
                    case 13:
                        stop_sounds();
                        play_sound(sound_1_falling);
                        take_hp(ALL_HP);
                        Char.x = char_dx_forward(-4);
                        seqtbl_offset_char(seq_2_stand);
                        play_seq();
                        break;
                    // small light grey potion turns kid into mouse
                    case 14:
                        set_health_life();
                        potion_timer += 6;
                        stop_sounds();
                        if (is_positional_audio) {
                            set_sound_pos(Char.room, Char.curr_col, Char.curr_row);
                            play_sound(sound_70_mouse_squeak);
                        }
                        seqtbl_offset_kid_char(seq_106_mouse);
                        play_kid_seq();
                        flash_color = color_7_lightgray;
                        flash_time = 6;
                        break;
                }
                break;
            case potion_1_health:
                if (hitp_curr != hitp_max) {
                    stop_sounds();
                    play_sound(sound_33_small_potion); // small potion
                    flash_color = color_15_brightwhite;
                    flash_time = 1;
                    hitp_delta = 1;
                }
                // custom logic - do not put down health potions on level 6
                if (current_level == 6) {
                    Char.x = char_dx_forward(-4);
                    seqtbl_offset_char(seq_2_stand);
                    play_seq();
                }
                break;
            case potion_2_life:
                // custom logic - counting the number of potions shadow drank
                if (picked_by_shadow && shadow_potion_counter < 2) {
                    ++shadow_potion_counter;
                    return;
                }
                if (hitp_curr != hitp_max || hitp_max < settings->max_hitp_allowed) {
                    // custom logic - life potion stops feather mode
                    if (feather_fall_timer > 0) {
                        feather_fall_timer = 0;
                        music_suspended_timer = 0;
                    }
                    stop_sounds();
                    play_sound(sound_30_big_potion); // big potion
                    flash_color = color_15_brightwhite;
                    flash_time = 4;
                    add_life();
                }
                break;
            case potion_3_feather:
                feather_fall();
                break;
            case potion_4_invert:
                upside_down_timer = 1080; // 90 seconds
                toggle_upside();
                break;
            case potion_5_hurt:
                // custom logic - hurt potion stops upside down mode
                if (upside_down_timer > 0) {
                    upside_down_timer = 0;
                    toggle_upside();
                    break;
                }
                // custom logic - hurt potion stops feather mode
                if (feather_fall_timer > 4) {
                    feather_fall_timer = 4;
                }
                stop_sounds();
                play_sound(sound_13_kid_hurt); // Kid hurt (by potion)
                // custom logic - drop potion quicker
                potion_timer += 6;
                Char.x = char_dx_forward(-4);
                seqtbl_offset_char(seq_2_stand);
                play_seq();
                hitp_delta = -1;
                break;
            case potion_6_open:
                get_tile(8, 0, 0);
                trigger_button(0, 0, -1);
                // custom logic - changed logic for mouse to appear
                if (current_level == 8 && leveldoor_open == 1) {
                    ++leveldoor_open;
                }
                break;
        }
    }
}

// seg006:1599
int is_dead(void) {
    // custom logic - shadow disappears when it dies
    if (is_kid_shadow && Char.charid == charid_0_kid && Char.alive >= 0) {
        return 1;
    }
    // custom logic - kid has a different spiked frame for vertical spikes in the temple
    if (Char.charid == charid_0_kid && Char.frame == frame_184_spiked_hanging) {
        return 1;
    }
    return Char.frame == frame_177_spiked || Char.frame == frame_178_chomped || Char.frame == frame_185_dead;
}

// seg006:15B5
void play_death_music(void) {
    word sound_id;
    if (holding_sword) {
        sound_id = sound_28_death_in_fight; // death in fight
    } else {
        sound_id = sound_24_death_regular; // death not in fight
    }
    play_sound(sound_id);
}

// seg006:15E8
void on_guard_killed(void) {
    word room_minus_1 = drawn_room - 1;
    dynamic_guard_timer = 0;
    // custom logic - activate inactive guard
    if (inactive_guard[room_minus_1].room == drawn_room && level.guards_seq_hi[room_minus_1] == 0) {
        if (Guard.alive < 0) {
            Guard.alive = 1;
        }
        inactive_guard_timer = DROPDEAD_FRAMES;
        return;
    }
    droppedout = 0;
    if (current_level == 0 && !play_demo_level) {
        // demo level: after killing Guard, run out of room
        checkpoint = 1;
        demo_index = demo_time = 0;
    }
    // custom logic - change the level palette back when kid stops hallucinating on level 6B
    if (current_level == 6 && falling_exit_checkpoint >= 4) {
        load_lev_spr(current_level);
        dead_guard3[16].room = 0; // remove dead fat guard
        leveldoor_open = 1;
        need_full_redraw = 1;
        draw_kid_hp(hitp_curr, hitp_max);
    }
    // custom logic - play spiked instead of victory sound for the spiked guard on level 1, room 1
    if (current_level == 1 && Guard.room == 1 && Guard.frame == frame_177_spiked) {
        if (is_positional_audio) {
            stop_all_sounds();
            set_sound_pos(Guard.room, Guard.curr_col, Guard.curr_row);
            play_sound(sound_48_spiked); // something spiked
        }
        if (need_level1_music == 0) {
            delayed_sound_timer = 3;
            delayed_sound = (struct delayed_sound_type) {
                sound_37_victory, 0, 0, 0, 0, 0, 0
            };
        }
        return;
    } else if (current_level == settings->win_level && Guard.charid != charid_0_kid) {
        // Jaffar's level: flash
        flash_color = color_15_brightwhite;
        flash_time = /*18*/ settings->jaffar_victory_flash_time;
        is_show_time = 1;
        leveldoor_open = 2;
        stop_ambient_music();
        music_suspended_timer = 18;
        // custom logic - turn flame into shadow
        dead_kid_timer = -18;
        keep_sword_drawn = false;
        holding_sword = 0;
        delayed_sound_timer = 4;
        delayed_sound = (struct delayed_sound_type) {
            sound_43_victory_Jaffar, 0, 0, 0, 0, 0, 0
        };
    } else if (Guard.charid != charid_1_shadow) {
        // custom logic - killing the last guard would stop the game
        // since there is no way to get to level 11
        if (current_level == 10 &&
                ((Guard.room == 8 && Guard.curr_row >= 2) || Guard.room == 22 || level.guards_tile[7] >= 30 /* no guard in room 8 */) &&
                level_restart_timer == 0 &&
                leveldoor_open == 0) {
            is_show_time = 0;
            level_restart_timer = DEAD_KID_LEVEL_10_TIMER_P12; // restart level after music stops playing
            display_text_bottom("THERE IS NO WAY OUT");
            text_time_total = 48;
            text_time_remaining = 48;
            if (is_sound_on) {
                if (Guard.action == actions_4_in_freefall) {
                    delayed_sound_timer = 24;
                    level_restart_timer += delayed_sound_timer;
                    delayed_sound = (struct delayed_sound_type) {
                        sound_36_out_of_time, 0, 0, 0, 0, 0, 0
                    };
                } else {
                    stop_sounds();
                    play_sound(sound_36_out_of_time);
                }
            }
            return;
        }
        // custom logic - kid turned into shadow
        if (is_kid_shadow && Guard.charid == charid_0_kid) {
            play_sound(sound_25_presentation);
            return;
        }
        // custom logic - do not play death music for bowing guards on level 10
        if (current_level == 10 && leveldoor_open >= DEAD_KID_LEVEL_10_TIMER) {
            return;
        }
        stop_ambient_music();
        // custom logic - play delayed victory sound
        if (delayed_sound_timer == 0 || delayed_sound.play_victory_sound == 0) {
            determine_col();
            bool guard_on_loose_tile = get_tile_at_char() == tiles_11_loose || (get_tile_behind_char() == tiles_11_loose && distance_to_edge_weight() >= 8);
            delayed_sound_timer = music_suspended_timer = guard_on_loose_tile ? 4 : 1;
            delayed_sound = (struct delayed_sound_type) {
                sound_37_victory, 0, 0, 0, 0, 0, 0
            };
        }
    }
}

// seg006:1634
void clear_char(void) {
    Char.direction = dir_56_none;
    Char.alive = 0;
    Char.action = 0;
    draw_guard_hp(0, guardhp_curr);
    guardhp_curr = 0;
}

// data:42EC
byte obj2_tilepos;
// data:34A6
word obj2_x;
// data:34A8
byte obj2_y;
// data:599E
sbyte obj2_direction;
// data:5948
byte obj2_id;
// data:42BE
byte obj2_chtab;
// data:4D90
short obj2_clip_top;
// data:460C
short obj2_clip_bottom;
// data:4C94
short obj2_clip_left;
// data:4CDE
short obj2_clip_right;

short obj2_color;

// seg006:1654
void save_obj(void) {
    obj2_tilepos = obj_tilepos;
    obj2_x = obj_x;
    obj2_y = obj_y;
    obj2_direction = obj_direction;
    obj2_id = obj_id;
    obj2_chtab = obj_chtab;
    obj2_clip_top = obj_clip_top;
    obj2_clip_bottom = obj_clip_bottom;
    obj2_clip_left = obj_clip_left;
    obj2_clip_right = obj_clip_right;
    // custom logic - saving the guard's color
    obj2_color = obj_color;
}

// seg006:1691
void load_obj(void) {
    obj_tilepos = obj2_tilepos;
    obj_x = obj2_x;
    obj_y = obj2_y;
    obj_direction = obj2_direction;
    obj_id = obj2_id;
    obj_chtab = obj2_chtab;
    obj_clip_top = obj2_clip_top;
    obj_clip_bottom = obj2_clip_bottom;
    obj_clip_left = obj2_clip_left;
    obj_clip_right = obj2_clip_right;
    // custom logic - loading the guard's color
    obj_color = obj2_color;
}

// seg006:16CE
void draw_hurt_splash(void) {
    short frame;
    frame = Char.frame;
    // custom logic - draw the same red splash during sword fighting
    if (frame != frame_178_chomped && frame != frame_177_spiked) {
        save_obj();
        obj_tilepos = -1;
        if (frame == frame_185_dead || (frame >= frame_106_fall && frame < frame_111_stand_up_from_crouch_2)) {
            obj_y += 4;
            obj_dx_forward(5);
        } else {
            obj_y -= ((Char.charid == charid_0_kid) << 2) + 11;
            obj_dx_forward(5);
        }
        obj_chtab = id_chtab_2_kid; // use kid's sprite for the red splash
        obj_id = 218; // splash!
        reset_obj_clip();
        add_objtable(objtype_5_hurt_splash);
        load_obj();
    }
}

// seg006:1798
void add_sword_to_objtable(void) {
    short frame;
    short sword_frame;
    sword_table_type entry;
    short delta = 0;
    // custom logic - do not draw sword in fist fight mode
    if (Char.charid == charid_0_kid && !have_sword) {
        return;
    }
    frame = Char.frame;
    // custom logic - tall and fat guard on level 8 requires an offset sword position
    if (current_level == 8 && Char.charid == charid_2_guard && Char.color == 0) {
        delta = -2;
    }
    if ((frame >= frame_229_found_sword && frame < frame_238_sheathe) || // found sword + put sword away
            Char.sword != sword_0_sheathed ||
            (Char.charid == charid_2_guard && Char.alive < 0)
       ) {
        sword_frame = cur_frame.sword & 0x3F;
        if (sword_frame) {
            // custom logic - skeletons use a different sword table
            entry = Char.charid == charid_4_skeleton ? sword_skel_tbl[sword_frame] : sword_tbl[sword_frame];
            obj_id = entry.id;
            if (obj_id != 0xFF) {
                obj_x = calc_screen_x_coord(obj_x);
                obj_dx_forward(entry.x + (delta >> 2));
                obj_y += entry.y + delta;
                // custom logic - certain purple guard images/frames require an adjustment for the sword
                if (Char.charid == charid_2_guard && settings->tbl_guard_type[current_level] == guard_type_purple) {
                    switch (sword_frame) {
                        case 6:
                        case 7:
                            obj_dx_forward(-3);
                            break;
                        case 9:
                            obj_dx_forward(4);
                            obj_y += 2;
                            break;
                        case 10:
                            obj_y += 2;
                            break;
                        case 49:
                            obj_dx_forward(5);
                            obj_y -= 5;
                            break;
                    }
                }
                // custom logic - skeletons have a different sword
                if (Char.charid > charid_0_kid && settings->tbl_guard_type[current_level] == guard_type_skeleton) {
                    obj_chtab = is_dungeon_2() ? id_chtab_13_alt_skel_sword : id_chtab_12_skel_sword;
                    // custom logic - Jaffar's vision has a sword without a partial hand
                } else if (Char.charid == charid_2_guard && settings->tbl_guard_type[current_level] == guard_type_Jaffar_vision) {
                    obj_chtab = id_chtab_11_sword_no_hand;
                } else if (Char.charid == charid_0_kid && is_kid_shadow && is_shadow_flame) {
                    obj_chtab = id_chtab_14_flaming_sword;
                } else {
                    obj_chtab = id_chtab_0_sword;
                }
                add_objtable(objtype_3_sword);
            }
        }
    }
}

// seg006:1827
void control_guard_inactive(void) {
    if (Char.frame == frame_166_stand_inactive && control_down < control_0_released) {
        if (control_forward < control_0_released) {
            draw_sword();
        } else {
            control_down = control_1_ignore;
            seqtbl_offset_char(seq_80_stand_flipped); // stand flipped
        }
    }
}

// seg006:1852
int char_opp_dist(void) {
    // >0 if Opp is in front of char
    // <0 if Opp is behind char
    short distance;
    if (Char.room != Opp.room) {
        return 999;
    }
    distance = Opp.x - Char.x;
    if (Char.direction < dir_0_right) {
        distance = -distance;
    }
    if (distance >= 0 && Char.direction != Opp.direction) {
        distance += 13;
    }
    return distance;
}

// custom logic - returns fighting distance between kid and guard
int kid_guard_dist(void) {
    short distance;
    if (Kid.room != Guard.room) {
        return 999;
    }
    distance = Guard.x - Kid.x;
    if (Kid.direction < dir_0_right) {
        distance = -distance;
    }
    if (distance >= 0 && Kid.direction != Guard.direction) {
        distance += 13;
    }
    return distance;
}

// custom logic - returns absolute distance between guard and kid
int guard_kid_dist(void) {
    if (Kid.room != Guard.room || Kid.curr_row != Guard.curr_row) {
        return 999;
    }
    return ABS(Guard.x - Kid.x);
}

// custom logic - returns absolute distance between kid and separated shadow
int shad_kid_dist(void) {
    if (Kid.room != Dead.room || Kid.curr_row != Dead.row) {
        return 999;
    }
    return ABS(Dead.x - Kid.x);
}

// custom logic - returns distance between kid and inactive guard (from guard's position)
int inactive_guard_kid_dist(void) {
    char_type2 inactive;
    short distance;
    inactive = inactive_guard[Kid.room - 1];
    if (inactive.room != Kid.room) {
        return 999;
    }
    distance = Kid.x - inactive.x;
    if (inactive.direction < dir_0_right) {
        distance = -distance;
    }
    if (distance >= 0 && inactive.direction != Kid.direction) {
        distance += 13;
    }
    return distance;
}

// custom logic - returns distance between inactive guard and kid (from kid's position)
int kid_inactive_guard_dist(void) {
    char_type2 inactive;
    short distance;
    inactive = inactive_guard[Kid.room - 1];
    if (Kid.room != inactive.room) {
        return 999;
    }
    distance = inactive.x - Kid.x;
    if (Kid.direction < dir_0_right) {
        distance = -distance;
    }
    if (distance >= 0 && Kid.direction != inactive.direction) {
        distance += 13;
    }
    return distance;
}

// seg006:189B
void inc_curr_row(void) {
    ++Char.curr_row;
}

// custom logic - returns whether a bit has been set in the tile control bits
// represented by the upper 3 bits of the tile byte.
byte get_tile_ctrl_bits(short room, short tilepos, byte bit_num) {
    // bit_num: 0 = tile bit 6, 1 = tile bit 7, 2 = tile bit 8, 3 - all 3 bits
    byte* room_tiles;
    byte ctrl_bits, value, result;
    room_tiles = &level.fg[(room - 1) * 30];
    ctrl_bits = room_tiles[tilepos] >> 5;
    if (bit_num > 2) {
        return ctrl_bits;
    }
    value = ctrl_bits & (1 << bit_num);
    result = value != 0;
    if (result != 0 && value != ctrl_bits) {
        ++result;
    }
    return result;
}

// custom logic - returns whether the room is not linked to any other room
bool is_room_unlinked(short room) {
    link_type room_links;
    room_links = level.roomlinks[room - 1];
    return !room_links.up && !room_links.down && !room_links.left && !room_links.right;
}
