/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"

// seg007:0000
void process_trobs(void) {
    word need_delete;
    word index;
    word new_index;
    need_delete = 0;
    if (trobs_count == 0) return;
    for (index = 0; index < trobs_count; ++index) {
        trob = trobs[index];
        animate_tile();
        trobs[index].type = trob.type;
        if (trob.type < 0) {
            need_delete = 1;
        }
    }
    if (need_delete) {
        for (index = new_index = 0; index < trobs_count; ++index) {
            if (trobs[index].type >= 0) {
                trobs[new_index++] = trobs[index];
            }
        }
        trobs_count = new_index;
    }
}

// seg007:00AF
void animate_tile(void) {
    get_room_address(trob.room);
    switch (get_curr_tile(trob.tilepos)) {
        case tiles_19_torch:
        case tiles_30_torch_with_debris:
            animate_torch();
            break;
        // custom logic - tile 26 is a torch without a floor in dungeons
        case tiles_26_torch_empty:
            if (settings->tbl_level_type[current_level] == DUNGEON_LEVEL) {
                animate_torch();
            }
            break;
        case tiles_6_closer:
        case tiles_15_opener:
            animate_button();
            break;
        case tiles_2_spike:
            animate_spike();
            break;
        case tiles_11_loose:
            animate_loose();
            break;
        case tiles_0_empty:
            animate_empty();
            break;
        case tiles_18_chomper:
            animate_chomper();
            break;
        case tiles_4_gate:
            animate_gate();
            break;
        case tiles_16_level_door_left:
            animate_leveldoor();
            break;
        case tiles_10_potion:
            animate_potion();
            break;
        case tiles_22_sword:
            // custom logic - sword on level 13 is flaming
            if (current_level == settings->flame_level &&
                    drawn_room == settings->flaming_sword_room &&
                    trob.tilepos == settings->flaming_sword_tilepos) {
                animate_torch();
            } else {
                animate_sword();
            }
            break;
        case tiles_23_balcony_left:
            // custom logic - animate balcony stars
            animate_balcony();
            break;
        default:
            trob.type = -1;
            break;
    }
    curr_room_modif[trob.tilepos] = curr_modifier;
}

// seg007:0166
short is_trob_in_drawn_room(void) {
    if (trob.room != drawn_room) {
        trob.type = -1;
        return 0;
    } else {
        return 1;
    }
}

// seg007:017E
void set_redraw_anim_right(void) {
    set_redraw_anim(get_trob_right_pos_in_drawn_room(), 1);
}

// seg007:018C
void set_redraw_anim_curr(void) {
    set_redraw_anim(get_trob_pos_in_drawn_room(), 1);
}

// seg007:019A
void redraw_at_trob(void) {
    word tilepos;
    redraw_height = 63;
    tilepos = get_trob_pos_in_drawn_room();
    set_redraw_full(tilepos, 1);
    set_wipe(tilepos, 1);
}

// seg007:01C5
void redraw_21h(void) {
    redraw_height = 0x21;
    redraw_tile_height();
}

// seg007:01D0
void redraw_11h(void) {
    redraw_height = 0x11;
    redraw_tile_height();
}

// seg007:01DB
void redraw_20h(void) {
    redraw_height = 0x20;
    redraw_tile_height();
}

// seg007:01E6
void draw_trob(void) {
    word trob_pos;
    trob_pos = get_trob_right_pos_in_drawn_room();
    set_redraw_anim(trob_pos, 1);
    set_redraw_fore(trob_pos, 1);
    set_redraw_anim(get_trob_right_above_pos_in_drawn_room(), 1);
}

// seg007:0218
void redraw_tile_height(void) {
    short tilepos;
    tilepos = get_trob_pos_in_drawn_room();
    set_redraw_full(tilepos, 1);
    set_wipe(tilepos, 1);
    tilepos = get_trob_right_pos_in_drawn_room();
    set_redraw_full(tilepos, 1);
    set_wipe(tilepos, 1);
}

// seg007:0258
short get_trob_pos_in_drawn_room(void) {
    short tilepos;
    tilepos = trob.tilepos;
    if (trob.room == room_A) {
        if (tilepos >= 20 && tilepos < 30) {
            // 20..29 -> -1..-10
            tilepos = 19 - tilepos;
        } else {
            tilepos = 30;
        }
    } else {
        if (trob.room != drawn_room) {
            tilepos = 30;
        }
    }
    return tilepos;
}

// seg007:029D
short get_trob_right_pos_in_drawn_room(void) {
    word tilepos;
    tilepos = trob.tilepos;
    if (trob.room == drawn_room) {
        if (tilepos % 10 != 9) {
            ++tilepos;
        } else {
            tilepos = 30;
        }
    } else if (trob.room == room_L) {
        if (tilepos % 10 == 9) {
            tilepos -= 9;
        } else {
            tilepos = 30;
        }
    } else if (trob.room == room_A) {
        if (tilepos >= 20 && tilepos < 29) {
            // 20..28 -> -2..-10
            tilepos = 18 - tilepos;
        } else {
            tilepos = 30;
        }
    } else if (trob.room == room_AL && tilepos == 29) {
        tilepos = -1;
    } else {
        tilepos = 30;
    }
    return tilepos;
}

// seg007:032C
short get_trob_right_above_pos_in_drawn_room(void) {
    word tilepos;
    tilepos = trob.tilepos;
    if (trob.room == drawn_room) {
        if (tilepos % 10 != 9) {
            if (tilepos < 10) {
                // 0..8 -> -2..-10
                tilepos = -(tilepos + 2);
            } else {
                tilepos -= 9;
            }
        } else {
            tilepos = 30;
        }
    } else if (trob.room == room_L) {
        if (tilepos == 9) {
            tilepos = -1;
        } else {
            if (tilepos % 10 == 9) {
                tilepos -= 19;
            } else {
                tilepos = 30;
            }
        }
    } else if (trob.room == room_B) {
        if (tilepos < 9) {
            tilepos += 21;
        } else {
            tilepos = 30;
        }
    } else if (trob.room == room_BL && tilepos == 9) {
        tilepos = 20;
    } else {
        tilepos = 30;
    }
    return tilepos;
}

// seg007:03CF
void animate_torch(void) {
    //if (is_trob_in_drawn_room()) {
    // Keep animating torches in the rightmost column of the left-side room as well, because they are visible in the current room.
    if (trob.room == drawn_room || (trob.room == room_L && (trob.tilepos % 10) == 9) ) {
        curr_modifier = get_torch_frame(curr_modifier);
        set_redraw_anim_right();
    } else {
        trob.type = -1;
    }
}

// seg007:03E9
void animate_potion(void) {
    word type;
    if (trob.type >= 0 && is_trob_in_drawn_room()) {
        // custom logic - do not animate empty/spilled potions
        if ((curr_modifier >> 3) >= 7) {
            return;
        }
        type = curr_modifier & 0xF8;
        curr_modifier = bubble_next_frame(curr_modifier & 0x07) | type;
        redraw_at_trob();
    }
}

// seg007:0425
void animate_sword(void) {
    if (is_trob_in_drawn_room()) {
        --curr_modifier;
        if (curr_modifier == 0) {
            curr_modifier = (prandom(255) & 0x3F) + 0x28;
        }
        redraw_at_trob();
    }
}

// seg007:0448
void animate_chomper(void) {
    word blood;
    word frame;
    word chomper_delay;
    short chomper_col, kid_col;
    // custom logic - determine chomper speed
    byte chomper_speed_bits = get_tile_ctrl_bits(trob.room, trob.tilepos, 3) >> 1;
    // custom logic - chompers can have different delay based on the upper 2 bits of the tile byte
    // 0x00 - normal speed with a random delay (default)
    // 0x01 - slow chomper
    // 0x02 - fast chomper
    // 0x03 - normal speed without a random delay
    switch (chomper_speed_bits) {
        case 0x01:
            chomper_delay = 42;
            break;
        case 0x02:
            chomper_delay = 8;
            break;
        case 0x03:
            chomper_delay = 15;
            break;
        default:
            chomper_delay = 15 + prandom(5);
            break;
    }
    if (trob.type >= 0) {
        blood = curr_modifier & 0x80;
        frame = (curr_modifier & 0x7F) + 1; // setting to the next frame (+1)
        // custom logic - additional behavior when the chomper is open
        if (frame == 1 || frame > 6) { // the actual chomper frame numbers are 0 or greater than 5
            chomper_col = trob.tilepos % 10;
            kid_col = Kid.curr_col;
            // in the temple chompers sense when kid is behind a wall or an abyss
            if (is_temple() && Kid.room == trob.room) {
                if (Kid.curr_row == trob.tilepos / 10) {
                    short left_col = MIN(chomper_col, kid_col);
                    short right_col = MAX(chomper_col, kid_col);
                    for (short col = left_col; col <= right_col; col++) {
                        // do not chomp if there is a wall/tapestry/closed gate between the kid and the chomper
                        short walltype = wall_type2(get_tile(trob.room, col, Kid.curr_row));
                        // reset the current room after the tile lookup
                        if (curr_room != trob.room) {
                            curr_room = trob.room;
                            get_room_address(curr_room);
                        }
                        if (walltype != wall_type_0_none && walltype != wall_type_3_chomper) {
                            // do not count mirror as a wall if kid is to the left of the chomper
                            bool skip_left_tile = walltype == wall_type_2_left && kid_col == left_col && col == left_col;
                            // do not count gates or tapestries as walls if kid is to the right of the chomper
                            bool skip_right_tile = walltype == wall_type_1_right && kid_col == right_col && col == right_col;
                            if (!skip_left_tile && !skip_right_tile) {
                                frame = MIN(frame - 1, chomper_delay - 1);
                                break;
                            }
                            // or there is an abyss between them
                        } else if (tile_is_empty(curr_tile2) && ABS(chomper_col - kid_col) > 1) {
                            frame = MIN(frame - 1, chomper_delay - 1);
                            break;
                        }
                    }
                } else if (Kid.curr_row != trob.tilepos / 10) {
                    frame = MIN(frame - 1, chomper_delay - 1);
                }
                // temple chompers are very likely to kill running and jumping kid
                if (Kid.frame == frame_39_start_run_jump_6 && // in the middle of the jump
                        get_tile_div_mod_m7(Kid.x + dir_front[Kid.direction + 1] * 12 /* dx(12) */) == chomper_col) {
                    frame = 1; // chomper is half-open, ready to chomp
                }
            }
        }
        if (frame > chomper_delay) {
            frame = 1;
        }
        curr_modifier = blood | frame;
        if (frame == 2 && trob.room == Kid.room) {
            set_sound_pos(trob.room, trob.tilepos % 10, trob.tilepos / 10);
            play_sound(sound_47_chomper); // chomper
            // custom logic - an activated chomper behind guard should make him notice kid
            if (trob.tilepos / 10 == Kid.curr_row &&
                    Guard.room == Kid.room &&
                    Guard.curr_row == Kid.curr_row) {
                int guard_pos = get_tilepos(Guard.curr_col, Guard.curr_row);
                if ((Guard.direction == dir_0_right && guard_pos > trob.tilepos) ||
                        (Guard.direction == dir_FF_left && guard_pos < trob.tilepos)) {
                    guard_notice_timer = 3;
                }
            }
        }
        // If either:
        // - Kid left this room
        // - Kid left this row
        // - Kid died but not in this chomper
        // and chomper is past frame 6
        // then stop.
        if ((trob.room != drawn_room ||
                (trob.tilepos / 10 != Kid.curr_row &&
                 // custom logic - an active guard triggers chompers
                 (droppedout == 0 ||
                  trob.tilepos / 10 != Guard.curr_row ||
                  Guard.room != drawn_room ||
                  Guard.alive >= 0 ||
                  Guard.sword == sword_0_sheathed)
                ) ||
                (Kid.alive >= 0 && blood == 0)) && (curr_modifier & 0x7F) >= 6
           ) {
            trob.type = -1;
        }
    }
    if ((curr_modifier & 0x7F) < 6) {
        redraw_at_trob();
    }
}

// seg007:04D3
void animate_spike(void) {
    if (trob.type >= 0) {
        // 0xFF means a disabled spike.
        if (curr_modifier == 0xFF) return;
        if (curr_modifier & 0x80) { // modifier is between to 0x8F and 0x80
            --curr_modifier;
            if (curr_modifier & 0x7F) return; // decrementing modifier until 0x80, spike stays extended
            curr_modifier = 6; // retracting
        } else {
            ++curr_modifier;
            if (curr_modifier == 5) {
                curr_modifier = 0x8F; // spike is fully extended at modifier 5
            } else if (curr_modifier == 9) {
                curr_modifier = 0;
                trob.type = -1;
            }
        }
    }
    // custom logic - play a sound when a spike retracts
    bool visible_spike = trob.room == drawn_room || (trob.room == room_L && trob.tilepos % 10 == 9);
    if (curr_modifier == 6) {
        if (visible_spike) {
            set_sound_pos(trob.room, trob.tilepos % 10, trob.tilepos / 10);
            play_sound(sound_60_spikes_in);
        } else if (is_positional_audio) {
            // do not interrupt current room spike sounds
            if (check_sound_playing() &&
                    (current_sound == sound_49_spikes_out ||
                     (current_sound == sound_60_spikes_in && current_sound_distance_loss >= 1.0f) ||
                     is_exclusive_sound(current_sound))) {
                goto redraw_spike;
            }
            // spikes near the flaming sword cannot be heard in the room above
            if (current_level != settings->flame_level || trob.room != 7 || drawn_room != 3) {
                set_sound_pos_with_dist_loss(trob.room, trob.tilepos % 10, trob.tilepos / 10, 20.0f);
                play_sound(sound_60_spikes_in);
            }
        }
    }
redraw_spike:
    // custom logic - vertical spikes in the temple need to be redrawn at full height
    if (is_temple()) {
        redraw_height = 63;
        redraw_tile_height();
    } else {
        redraw_21h();
    }
}

// data:27B2
const byte gate_close_speeds[] = {0, 0, 0, 20, 40, 60, 80, 100, 120};
// data:27C0
const byte gate_anim_delta[] = {-2, 4, 4};
// seg007:0522
void animate_gate(void) {
    /*
    Possible values of anim_type:
    0: closing
    1: open
    2: permanent open
    3,4,5,6,7,8: fast closing with speeds 20,40,60,80,100,120 /4 pixel/frame
    */
    sbyte anim_type;
    anim_type = trob.type;
    if (anim_type >= 0) {
        if (anim_type >= 3) {
            // closing fast
            if (anim_type < 8) {
                ++anim_type;
                trob.type = anim_type;
            }
            bool slam_gate = anim_type > 4 || curr_modifier > 18;
            short new_mod = curr_modifier - gate_close_speeds[anim_type];
            curr_modifier = new_mod;
            if (new_mod < 0) {
                curr_modifier = 0;
                trob.type = -1;
                set_sound_pos(trob.room, trob.tilepos % 10 + 1, trob.tilepos / 10);
                if (slam_gate) {
                    if (is_positional_audio) {
                        stop_sounds();
                    }
                    play_sound(sound_6_gate_closing_fast); // gate closing fast
                } else {
                    // custom logic - closing gate does not make a slamming from low heights
                    play_sound(sound_7_gate_stop);
                }
                // custom logic - kid says whew on level 1 when a gate closes quickly before the guard
                say_whew();
            }
        } else {
            if (curr_modifier != 0xFF) {
                // 0xFF means permanently open.
                byte delta = gate_anim_delta[anim_type];
                // custom logic - last control byte represents a slowly closing gate
                if (anim_type == 0 && (curr_modifier > 200 || get_tile_ctrl_bits(trob.room, trob.tilepos, 2) != 0)) {
                    // gates close faster after kid kills Jaffar
                    if (!(current_level == settings->win_level && leveldoor_open > 2 && trob.room != settings->win_room && curr_modifier > 16)) {
                        ++delta;
                    }
                }
                curr_modifier += delta;
                if (anim_type == 0) {
                    // closing
                    if (curr_modifier != 0) {
                        if (curr_modifier < 188) {
                            byte sound_delay = ABS((sbyte)delta) * 3;
                            if ((curr_modifier & sound_delay) == sound_delay) {
                                play_door_sound_if_visible(sound_4_gate_closing); // gate closing
                            }
                        }
                    } else {
                        gate_stop();
                    }
                } else {
                    // opening
                    if (curr_modifier < 188) {
                        if ((curr_modifier & 7) == 0) {
                            set_sound_pos(trob.room, trob.tilepos % 10 + 1, trob.tilepos / 10);
                            play_sound(sound_5_gate_opening); // gate opening
                        }
                    } else {
                        // stop
                        if (anim_type < 2) {
                            // after regular open
                            curr_modifier = 238;
                            trob.type = 0; // closing
                            set_sound_pos(trob.room, trob.tilepos % 10 + 1, trob.tilepos / 10);
                            play_sound(sound_7_gate_stop); // gate stop (after opening)
                        } else {
                            // after permanent open
                            curr_modifier = 0xFF; // keep open
                            gate_stop();
                        }
                    }
                }
            } else {
                gate_stop();
            }
        }
    }
    draw_trob();
}

// seg007:05E3
void gate_stop(void) {
    trob.type = -1;
    // custom logic - do not play stop gate sound after on level 12B as gates are reversed
    if (current_level != settings->inverse_level) {
        play_door_sound_if_visible(sound_7_gate_stop); // gate stop (after closing)
    }
}

// data:27B8
const byte leveldoor_close_speeds[] = {0, 5, 17, 99, 0};
// seg007:05F1
void animate_leveldoor(void) {
    /*
    Possible values of trob_type:
    0: open
    1: open (with button)
    2: open
    3,4,5,6: fast closing with speeds 0,5,17,99 pixel/frame
    */
    word trob_type;
    trob_type = trob.type;
    if (trob.type >= 0) {
        if (trob_type >= 3) {
            // closing
            ++trob.type;
            curr_modifier -= leveldoor_close_speeds[trob.type - 3];
            if ((sbyte)curr_modifier < 0) {
                curr_modifier = 0;
                trob.type = -1;
                set_sound_pos(trob.room, trob.tilepos % 10, trob.tilepos / 10);
                play_sound(sound_14_leveldoor_closing); // level door closing
            } else {
                if (trob.type == 4) {
                    sound_interruptible[sound_15_leveldoor_sliding] = 1;
                    set_sound_pos(trob.room, trob.tilepos % 10, trob.tilepos / 10);
                    play_sound(sound_15_leveldoor_sliding); // level door sliding (closing)
                }
            }
        } else {
            // opening
            ++curr_modifier;
            if (curr_modifier > KID_HEIGHT_DOOR_MODIFIER) {
                // custom logic - the level door is considered open 10 pixels away from its top position
                if (curr_modifier >= 43) {
                    trob.type = -1;
                    if (check_sound_playing() && current_sound != sound_41_end_level_music) {
                        stop_sounds();
                    }
                }
                if (leveldoor_open == 0 || leveldoor_open == 2) {
                    leveldoor_open = 1;
                    // custom logic - add wall before the gate on level 4, room 20
                    if (current_level == 4) {
                        get_tile(20, 5, 1);
                        curr_room_tiles[curr_tilepos] = tiles_20_wall;
                        curr_room_modif[curr_tilepos] = 1;
                        get_tile(20, 4, 1);
                        curr_room_tiles[curr_tilepos] = tiles_7_doortop_with_floor;
                        curr_room_modif[curr_tilepos] = 1;
                    }
                    // custom logic - horse statue warns kid
                    if (is_positional_audio &&
                            current_level == 10 &&
                            curr_room == 20 &&
                            leveldoor_open < DEAD_KID_LEVEL_10_TIMER) {
                        delayed_sound_timer = 2;
                        delayed_sound = (struct delayed_sound_type) {
                            sound_71_horse_neigh, 4, 5, 2, 0, 0, 0
                        };
                    }
                    // custom logic - level 2 has two level doors
                } else if (leveldoor_open == 1 && current_level == 2) {
                    ++leveldoor_open;
                }
            } else {
                // custom logic - close the level door on level 12B
                if (current_level == settings->inverse_level) {
                    if (leveldoor_open == 0) {
                        if (curr_modifier == 1) {
                            byte* imgdoor_room_modif = &level.bg[(12 /* door room */ - 1) * 30];
                            imgdoor_room_modif[13/* door left tile position */] = 0;
                        } else if (curr_modifier == 2) {
                            stop_sounds();
                            sound_interruptible[sound_15_leveldoor_sliding] = 1;
                            set_sound_pos(trob.room, trob.tilepos % 10, trob.tilepos / 10);
                            play_sound(sound_15_leveldoor_sliding); // level door sliding (closing)
                        } else if (curr_modifier == 4) {
                            leveldoor_open = 1;
                            set_sound_pos(trob.room, trob.tilepos % 10, trob.tilepos / 10);
                            play_sound(sound_14_leveldoor_closing); // door slammed closed sound
                            redraw_screen(0);
                        }
                    } else if (curr_modifier == 18) {
                        if (drawn_room != Kid.room) {
                            if (Kid.curr_row == 2 && Kid.direction == dir_0_right) {
                                Kid.x = char_dx_forward(-10);
                            }
                            next_room = drawn_room = Kid.room;
                            load_room_links();
                            anim_tile_modif();
                            find_all_rooms_from(next_room);
                            show_black_screen(6, 1);
                            redraw_screen(1);
                        }
                    }
                } else {
                    set_sound_pos(trob.room, trob.tilepos % 10, trob.tilepos / 10);
                    sound_interruptible[sound_15_leveldoor_sliding] = 0;
                    play_sound(sound_15_leveldoor_sliding); // level door sliding (opening)
                    // custom logic - play an additional POP2 sound
                    if (leveldoor_open != 2 &&
                            curr_modifier == 1 &&
                            is_positional_audio &&
                            !ogg_fore_sound_playing &&
                            current_level != 10 &&
                            current_level < 14) {
                        play_ogg_fore_sound(sound_80_leveldoor_opening, 0, true);
                    }
                }
            }
        }
    }
    set_redraw_anim_right();
}

// custom logic - animate balcony stars
void animate_balcony(void) {
    byte modifier;
    if (trob.room == drawn_room || (trob.room == room_L && (trob.tilepos % 10) == 8)) {
        modifier = curr_modifier >> 4;
        if (modifier == 0) {
            modifier = prandom(8) + 1;
        } else {
            --modifier;
        }
        curr_modifier = modifier << 4;
        if (modifier == 1) {
            animate_balcony_stars(trob.tilepos, trob.room == room_L);
        }
    }
}

// seg007:06AD
short bubble_next_frame(short curr) {
    short next;
    next = curr + 1;
    if (next >= 8) next = 1;
    return next;
}

// seg007:06CD
short get_torch_frame(short curr) {
    short next;
    next = prandom(255);
    if (next != curr) {
        if (next < 9) {
            return next;
        } else {
            next = curr;
        }
    }
    ++next;
    if (next >= 9) next = 0;
    return next;
}

// seg007:070A
void set_redraw_anim(short tilepos, byte frames) {
    if (tilepos < MAX_TILES_PER_ROOM) {
        if (tilepos < 0) {
            ++tilepos;
            redraw_frames_above[-tilepos] = frames;
            // or simply: ~tilepos
        } else {
            redraw_frames_anim[tilepos] = frames;
        }
    }
}

// seg007:0738
void set_redraw2(short tilepos, byte frames) {
    if (tilepos < MAX_TILES_PER_ROOM) {
        if (tilepos < 0) {
            // trying to draw a mob at a negative tilepos, in the range -1 .. -10
            // used e.g. when the kid is climbing up to the room above
            // however, loose tiles falling out of the room end up with a negative tilepos {-2 .. -11} !
            tilepos = (-tilepos) - 1;
            if (tilepos > 9) tilepos = 9; // prevent array index out of bounds!
            redraw_frames_above[tilepos] = frames;
        } else {
            redraw_frames2[tilepos] = frames;
        }
    }
}

// seg007:0766
void set_redraw_floor_overlay(short tilepos, byte frames) {
    if (tilepos < MAX_TILES_PER_ROOM) {
        if (tilepos < 0) {
            ++tilepos;
            redraw_frames_above[-tilepos] = frames;
            // or simply: ~tilepos
        } else {
            redraw_frames_floor_overlay[tilepos] = frames;
        }
    }
}

// seg007:0794
void set_redraw_full(short tilepos, byte frames) {
    if (tilepos < MAX_TILES_PER_ROOM) {
        if (tilepos < 0) {
            ++tilepos;
            redraw_frames_above[-tilepos] = frames;
            // or simply: ~tilepos
        } else {
            redraw_frames_full[tilepos] = frames;
        }
    }
}

// seg007:07C2
void set_redraw_fore(short tilepos, byte frames) {
    if (tilepos < MAX_TILES_PER_ROOM && tilepos >= 0) {
        redraw_frames_fore[tilepos] = frames;
    }
}

// seg007:07DF
void set_wipe(short tilepos, byte frames) {
    if (tilepos < MAX_TILES_PER_ROOM && tilepos >= 0) {
        if (wipe_frames[tilepos] != 0) {
            redraw_height = MAX(wipe_heights[tilepos], redraw_height);
        }
        wipe_heights[tilepos] = redraw_height;
        wipe_frames[tilepos] = frames;
    }
}

// seg007:081E
void start_anim_torch(short room, short tilepos) {
    curr_room_modif[tilepos] = prandom(8);
    add_trob(room, tilepos, 1);
}

// seg007:0847
void start_anim_potion(short room, short tilepos) {
    word pot_type;
    // custom logic - do not animate empty/spilled potions
    pot_type = curr_room_modif[tilepos] >> 3;
    if (pot_type < 7) {
        curr_room_modif[tilepos] &= 0xF8;
        curr_room_modif[tilepos] |= prandom(6) + 1;
    }
    add_trob(room, tilepos, 1);
    // custom logic - redraw spilled potion
    if (pot_type >= 9) {
        redraw_at_trob();
    }
}

// seg007:087C
void start_anim_sword(short room, short tilepos) {
    curr_room_modif[tilepos] = prandom(0xFF) & 0x1F;
    add_trob(room, tilepos, 1);
}

// seg007:08A7
void start_anim_chomper(short room, short tilepos, byte modifier) {
    short old_modifier;
    old_modifier = curr_room_modif[tilepos];
    if (old_modifier == 0 || old_modifier >= 6) {
        curr_room_modif[tilepos] = modifier;
        add_trob(room, tilepos, 1);
    }
}

// seg007:08E3
void start_anim_spike(short room, short tilepos) {
    sbyte old_modifier;
    old_modifier = curr_room_modif[tilepos];
    if (old_modifier <= 0) {
        if (old_modifier == 0) {
            add_trob(room, tilepos, 1);
            set_sound_pos(trob.room, trob.tilepos % 10, trob.tilepos / 10);
            play_sound(sound_49_spikes_out);
            guard_notice_timer = 1;
        } else {
            // 0xFF means a disabled spike.
            if (old_modifier != (sbyte)0xFF) {
                curr_room_modif[tilepos] = 0x8F;
            }
        }
    }
}

// custom logic - start animating balcony stars
void start_anim_balcony(short room, short tilepos) {
    // temple has different balconies
    if (is_temple()) {
        return;
    }
    if ((curr_room_modif[tilepos] & 0x0F) == 0) {
        curr_room_modif[tilepos] = (prandom(8) + 1) << 4;
        add_trob(room, tilepos, 1);
    }
}

// seg007:092C
short trigger_gate(short room, short tilepos, short button_type) {
    byte modifier = curr_room_modif[tilepos];
    if (button_type == tiles_15_opener) {
        // If the gate is permanently open, don't to anything.
        if (modifier == 0xFF) return -1;
        if (modifier >= 188) { // if it's already open
            curr_room_modif[tilepos] = 238; // keep it open for a while
            return -1;
        }
        curr_room_modif[tilepos] = (modifier + 3) & 0xFC;
        return 1; // regular open
    } else if (button_type == tiles_14_debris) {
        // If it's not fully open:
        if (modifier < 188) return 2; // permanent open
        curr_room_modif[tilepos] = 0xFF; // keep open
        return -1;
        // custom logic - kid gets resurrected or united with shadow
    } else if (button_type == tiles_5_stuck) {
        if (modifier >= 188) { // if it's already open
            curr_room_modif[tilepos] = 238; // keep it open for a while
            add_trob(room, tilepos, 0); // change to the closing state
            return -1;
        }
        curr_room_modif[tilepos] = (modifier + 3) & 0xFC;
        return 1; // regular open
    } else {
        if (modifier != 0) {
            // custom logic - find the current gate trob
            short trob_index;
            short trob_room = trob.room;
            short trob_tilepos = trob.tilepos;
            trob.room = room;
            trob.tilepos = tilepos;
            trob_index = find_trob();
            trob.room = trob_room;
            trob.tilepos = trob_tilepos;
            // gate is already closing fast
            if (trob_index >= 0 && trobs[trob_index].type >= 3) {
                return -1;
            }
            return 3; // close fast
        } else {
            return -1; // already closed
        }
    }
}

// seg007:0999
short trigger_tile(short target_type, short room, short tilepos, short button_type) {
    short result;
    result = -1;
    if (target_type == tiles_4_gate) {
        result = trigger_gate(room, tilepos, button_type);
    } else if (target_type == tiles_16_level_door_left) {
        if (curr_room_modif[tilepos] != 0) {
            result = -1;
        } else {
            result = 1;
        }
    } else { // custom logic - allow triggering any tile
        result = 1;
    }
    return result;
}

// seg007:09E5
void do_trigger_events(short first_event, short button_type) {
    word room;
    word tilepos;
    byte target_type;
    sbyte trigger_result;
    while (1) {
        room = get_doorlink_room(first_event);
        get_room_address(room);
        tilepos = get_doorlink_tile(first_event);
        target_type = curr_room_tiles[tilepos] & 0x1F;
        trigger_result = trigger_tile(target_type, room, tilepos, button_type);
        if (trigger_result >= 0) {
            add_trob(room, tilepos, trigger_result);
        }
        // the event is not linked to the next event
        if (get_doorlink_next(first_event++) == 0) {
            break;
        }
    }
}

// seg007:0A5A
void add_trob(byte room, byte tilepos, sbyte type) {
    short found;
    if (trobs_count >= MAX_TROBS) {
        show_dialog("Trobs Overflow", default_dialog);
        return; // added
    }
    trob.room = room;
    trob.tilepos = tilepos;
    trob.type = type;
    found = find_trob();
    if (found == -1) {
        // add new
        if (trobs_count == MAX_TROBS) return;
        trobs[trobs_count++] = trob;
    } else {
        // change existing
        trobs[found].type = trob.type;
    }
}

// custom logic - find trob at a specific room location
short find_trob_at_loc(byte room, byte tilepos) {
    short index;
    for (index = 0; index < trobs_count; ++index) {
        if (trobs[index].tilepos == tilepos && trobs[index].room == room) {
            return index;
        }
    }
    return -1;
}

// seg007:0ACA
short find_trob(void) {
    short index;
    for (index = 0; index < trobs_count; ++index) {
        if (trobs[index].tilepos == trob.tilepos && trobs[index].room == trob.room) {
            return index;
        }
    }
    return -1;
}

// seg007:0B0A
void clear_tile_wipes(void) {
    memset(redraw_frames_full, 0, sizeof(redraw_frames_full));
    memset(wipe_frames, 0, sizeof(wipe_frames));
    memset(wipe_heights, 0, sizeof(wipe_heights));
    memset(redraw_frames_anim, 0, sizeof(redraw_frames_anim));
    memset(redraw_frames_fore, 0, sizeof(redraw_frames_fore));
    memset(redraw_frames2, 0, sizeof(redraw_frames2));
    memset(redraw_frames_floor_overlay, 0, sizeof(redraw_frames_floor_overlay));
    memset(tile_object_redraw, 0, sizeof(tile_object_redraw));
    memset(redraw_frames_above, 0, sizeof(redraw_frames_above));
}

// seg007:0BB6
short get_doorlink_timer(short index) {
    // bits 0-4 as number is the pressure plate (opener/closer tile) timer
    // max value of 31 represents a jammed event
    return doorlink_map[index] & 0x1F;
}

// seg007:0BCD
short set_doorlink_timer(short index, byte value) {
    doorlink_map[index] &= 0xE0; // set bits 0-4 to 0
    doorlink_map[index] |= value & 0x1F; // store only bits 0-4 of value
    return doorlink_map[index];
}

// seg007:0BF2
short get_doorlink_tile(short index) {
    // bits 0-4 as number is the tile position (0-29) of the activated trob (typically gate/level door)
    return doorlink_loc[index] & 0x1F;
}

// seg007:0C09
short get_doorlink_next(short index) {
    // bit 7 = 1 represents an event that is linked to another event stored under index + 1
    return !(doorlink_loc[index] & 0x80);
}

// seg007:0C26
short get_doorlink_room(short index) {
    return
        // bits 5-6 as a number
        ((doorlink_loc[index] & 0x60) >> 5) +
        // and bits 5-7 as a number multiplied by 8
        ((doorlink_map[index] & 0xE0) >> 3);
    // are added together to calculate the activated trob room number
}

// seg007:0C53
void trigger_button(int playsound, int button_type, int modifier) {
    sbyte link_timer;
    get_curr_tile(curr_tilepos);
    if (button_type == 0) {
        // 0 means currently selected
        button_type = curr_tile;
    }
    if (modifier == -1) {
        // -1 means currently selected
        modifier = curr_modifier;
    }
    link_timer = get_doorlink_timer(modifier);
    // is the event jammed?
    if (link_timer != 0x1F) {
        // sets the timer for the pressure plate event with index equal to the modifier of the button
        set_doorlink_timer(modifier, 5);
        if (link_timer < 2) {
            add_trob(curr_room, curr_tilepos, 1);
            redraw_11h();
            if (playsound) {
                make_guard_notice(false);
                // custom logic - chomper activator button plays a chomper sound
                if (current_level == 3 &&
                        trob.room == 15 &&
                        trob.tilepos == 22) {
                    delayed_sound_timer = 6;
                    delayed_sound = (struct delayed_sound_type) {
                        sound_47_chomper, 16, 3, 0, 0, 0, 1
                    };
                }
                // custom logic - hide button pressed dropping sound in unlinked rooms
                if (!is_room_unlinked(trob.room)) {
                    set_sound_pos(trob.room, trob.tilepos % 10, trob.tilepos / 10);
                    play_sound(sound_3_button_pressed); // button pressed
                }
            }
            // custom logic - buttons with tile control bit 7 are sticky
            if (get_tile_ctrl_bits(trob.room, trob.tilepos, 1) != 0) {
                // the same logic as when a character dies on button
                curr_room_tiles[curr_tilepos] = tiles_1_floor;
                curr_room_modif[curr_tilepos] = 0;
                button_type = tiles_14_debris; // force permanent open
                redraw_11h();
            }
            // custom logic - level door tile on level 12B should show the closing level door
            if (current_level == settings->inverse_level &&
                    trob.room == 1 &&
                    trob.tilepos == 18) {
                if (leveldoor_open == 0 && drawn_room == trob.room) {
                    // switch to the level door room
                    leave_guard();
                    draw_guard_hp(0, MAX_GUARD_HP);
                    next_room = room_L;
                    show_black_screen(6, 0);
                }
            }
        }
        do_trigger_events(modifier, button_type);
    }
}

// seg007:0CD9
void died_on_button(word charid) {
    word button_type;
    word modifier;
    button_type = get_curr_tile(curr_tilepos);
    modifier = curr_modifier;
    // custom logic - save data to restore button if kid units with shadow or resurrects
    if (charid == charid_0_kid || charid == charid_1_shadow) {
        dead_kid_tile = (struct stuck_button_type) {
            curr_tilepos, curr_room, curr_room_tiles[curr_tilepos], modifier
        };
    }
    if (curr_tile == tiles_15_opener) {
        curr_room_tiles[curr_tilepos] = tiles_1_floor;
        curr_room_modif[curr_tilepos] = 0;
        // custom logic - dead kid should not jam the gate
        if (charid > charid_0_kid) {
            button_type = tiles_14_debris; // force permanent open
        }
    } else {
        curr_room_tiles[curr_tilepos] = tiles_5_stuck;
    }
    trigger_button(1, button_type, modifier);
}

// seg007:0D3A
void animate_button(void) {
    word timer_value;
    if (trob.type >= 0) {
        // decrements the tile timer by 1
        set_doorlink_timer(curr_modifier, timer_value = get_doorlink_timer(curr_modifier) - 1);
        if (timer_value < 2) {
            trob.type = -1;
            redraw_11h();
        }
    }
}

// seg007:0D72
void start_level_door(short room, short tilepos) {
    curr_room_modif[tilepos] = 43; // start fully open
    add_trob(room, tilepos, 3);
}

// seg007:0D93
void animate_empty(void) {
    trob.type = -1;
    redraw_20h();
}

// data:2284
const word y_loose_land[] = {2, 65, 128, 191, 254};
// seg007:0D9D
void animate_loose(void) {
    word room;
    word row;
    word tilepos;
    byte shake_count;
    short anim_type;
    bool is_lightweight_char;
    bool totally_loose_tile;
    byte delay;
    anim_type = trob.type;
    shake_count = loose_tile_counter[trob.room - 1][trob.tilepos];
    // custom logic - are there any totally loose tiles in the room
    is_lightweight_char = feather_fall_timer || is_kid_shadow || is_kid_mouse();
    totally_loose_tile = !is_lightweight_char && get_tile_ctrl_bits(trob.room, trob.tilepos, 1) == 1;
    if (anim_type >= 0) {
        ++curr_modifier;
        if ((curr_modifier & 0x80) &&
                !totally_loose_tile &&
                // custom logic - shake several times before falling
                (is_lightweight_char || shake_count < settings->loose_shake_count)) {
            // just shaking
            // custom logic - don't shake falling tiles on level 12B but shake the rest of loose tiles
            if (current_level == settings->loose_tiles_level &&
                    ((drawn_room == settings->loose_tiles_room_1 && trob.room != settings->loose_tiles_room_1) ||
                     (drawn_room == settings->loose_tiles_room_2 && trob.room != settings->loose_tiles_room_2))) {
                return;
            }
            // custom logic - do not shake stuck loose tiles
            if (get_tile_ctrl_bits(trob.room, trob.tilepos, 0) == 1) {
                curr_modifier = 0;
                trob.type = -1;
                redraw_20h();
                return;
            }
            if (curr_modifier == 0x82) {
                // custom logic - increment loose tile shake counter
                if (++shake_count == settings->loose_shake_count) {
                    ++curr_modifier;
                }
                ++loose_tile_counter[trob.room - 1][trob.tilepos];
            } else if (curr_modifier >= 0x84) {
                curr_modifier = 0;
                trob.type = -1;
            }
            loose_shake(!curr_modifier);
        } else {
            // custom logic - totally loose tiles drop immediately
            if (totally_loose_tile) {
                goto loc_drop_tile;
            }
            delay = settings->loose_floor_delay;
            // custom logic - dead guard causes a loose tile to fall immediately
            if (Guard.alive >= 0 && Guard.room == trob.room && trob.tilepos == get_tilepos(Guard.curr_col, Guard.curr_row)) {
                delay = 1;
            }
            // something is on the floor
            // should it fall already?
            if (curr_modifier >= delay) {
loc_drop_tile:
                room = trob.room;
                tilepos = trob.tilepos;
                curr_modifier = remove_loose(tilepos);
                trob.type = -1;
                curmob.xh = (tilepos % 10) << 2;
                row = tilepos / 10;
                curmob.y = y_loose_land[row + 1];
                curmob.room = room;
                curmob.speed = 0;
                curmob.type = 0;
                curmob.row = row;
                add_mob();
            } else {
                loose_shake(0);
            }
        }
    }
    redraw_20h();
}

// data:2734
const byte loose_sound[] = {0, 1, 1, 1, 0, 1, 0, 0, 1, 0, 0, 0};
// seg007:0E55
void loose_shake(int sound) {
    word sound_id;
    if (sound || loose_sound[curr_modifier & 0x7F]) {
        do {
            // Sounds 20,21,22: loose floor shaking
            sound_id = prandom(2) + sound_20_loose_shake_1;
        } while(sound_id == last_loose_sound);
        last_loose_sound = sound_id;
        // custom logic - hide loose tile shaking sound in unlinked rooms
        if (!is_room_unlinked(trob.room)) {
            set_sound_pos(trob.room, trob.tilepos % 10, trob.tilepos / 10);
            play_sound(sound_id);
        }
    }
}

// seg007:0EB8
int remove_loose(int tilepos) {
    curr_room_tiles[tilepos] = tiles_0_empty;
    // note: the level type is used to determine the modifier of the empty space left behind
    return settings->tbl_level_type[current_level];
}

// custom logic - kid should fall with a loose tile
void kid_loose_fall(void) {
    if (Kid.frame == frame_185_dead &&
            Kid.room == curr_room &&
            Kid.curr_col == tile_col &&
            Kid.curr_row == tile_row) {
        curr_room_modif[curr_tilepos] = 10;
        char_type old_char = Char;
        loadkid();
        start_fall();
        savekid();
        if (Char.charid != old_char.charid) {
            Char = old_char;
        }
    }
}

// custom logic - guard should fall with a loose tile
void guard_loose_fall(void) {
    int guard_room, guard_col;
    guard_room = Guard.room;
    guard_col = Guard.curr_col;
    if (Guard.curr_col < 0) {
        guard_room = level.roomlinks[Guard.room - 1].left;
        guard_col += 10;
    } else if (Guard.curr_col > 9) {
        guard_room = level.roomlinks[Guard.room - 1].right;
        guard_col -= 10;
    }
    if ((Guard.frame == frame_185_dead || Guard.fall_y >= 11) &&
            guard_room == curr_room &&
            guard_col == tile_col &&
            Guard.curr_row == tile_row) {
        curr_room_modif[curr_tilepos] = 10;
        char_type old_char = Char;
        loadshad();
        start_fall();
        saveshad();
        if (Char.charid != old_char.charid) {
            Char = old_char;
        }
    }
}

// seg007:0ED5
void make_loose_fall(byte modifier) {
    // custom logic - falling sword makes a stuck loose tile fall on level 6
    bool is_falling_sword = current_level == 6 && Kid.room == level.start_room && potion_timer >= 1 && potion_timer < 42;
    // custom logic - is this a stuck loose tile?
    if (get_tile_ctrl_bits(curr_room, curr_tilepos, 0) == 0 || is_falling_sword) {
        if ((sbyte)curr_room_modif[curr_tilepos] <= 0) {
            curr_room_modif[curr_tilepos] = modifier;
            add_trob(curr_room, curr_tilepos, 0);
            redraw_20h();
        }
        // custom logic - kid should fall with a loose tile
        kid_loose_fall();
        // custom logic - guard should fall with a loose tile
        guard_loose_fall();
    }
}

// custom logic - active guards can also trigger chompers
void start_chompers_at_char(char_type curr_char) {
    short timing;
    short modifier;
    short tilepos;
    short column;
    bool random_chomper;
    random_chomper = true;
    timing = prandom(9) + 6;
    if ((byte)curr_char.curr_row < 3) {
        get_room_address(curr_char.room);
        for (column = 0, tilepos = tbl_line[curr_char.curr_row]; column < 10; ++column, ++tilepos) {
            if (get_curr_tile(tilepos) == tiles_18_chomper) {
                // custom logic - chompers with tile bits 7 and 8 set to 1 follow the original timing
                // the rest of the chompers are random
                if (random_chomper && (curr_room_tiles[tilepos] >> 6) == 0x03) {
                    timing = 15;
                    random_chomper = false;
                }
                modifier = curr_modifier & 0x7F;
                if (modifier == 0 || modifier >= 6) {
                    start_anim_chomper(curr_char.room, tilepos, timing | (curr_modifier & 0x80));
                    timing = next_chomper_timing(timing, random_chomper);
                }
            }
        }
    }
}

// seg007:0F13
void start_chompers(void) {
    start_chompers_at_char(Char);
    // custom logic - active guards can also trigger chompers
    if (droppedout && Opp.alive < 0 && Opp.room == Char.room && Opp.sword == sword_2_drawn) {
        start_chompers_at_char(Opp);
    }
}

// custom logic - added support for random chomper order
short next_chomper_timing(byte timing, bool random_chomper) {
    short prev_timing;
    if (random_chomper) {
        prev_timing = timing;
        while (timing == prev_timing) {
            timing = prandom(9) + 6; // 6-15
        }
    } else {
        // 15,12,9,6,13,10,7,14,11,8,repeat
        timing -= 3;
        if (timing < 6) {
            timing += 10;
        }
    }
    return timing;
}

// seg007:0FB4
void loose_make_shake(void) {
    // custom logic - shake tiles on all levels including level 12B
    if (curr_room_modif[curr_tilepos] == 0) {
        curr_room_modif[curr_tilepos] = 0x80;
        add_trob(curr_room, curr_tilepos, 1);
    }
}

// seg007:0FE0
void do_knock(int room, int tile_row, int impact_col) {
    short tile_col;
    int adj_room;
    // custom logic - loose tiles that are separated by empty tiles from
    // the shook tile should not shake
    if (knock &&
            Char.frame == frame_79_jumphang &&
            !tile_is_floor(get_tile(room, impact_col, tile_row)) && // kid hit the very edge of the tile behind
            curr_tile2 != tiles_20_wall) {
        impact_col += dir_behind[Char.direction + 1];
    }
    for (tile_col = impact_col; tile_col >= -1; --tile_col) {
        if (get_tile(room, tile_col, tile_row) == tiles_11_loose) {
            loose_make_shake();
            if (Kid.room == room) {
                if (tile_col == -1) {
                    adj_room = level.roomlinks[room - 1].left;
                    if (adj_room > 0 && get_tile(adj_room, 8, tile_row) == tiles_11_loose) {
                        do_knock(adj_room, tile_row, 8);
                    }
                } else if (tile_col == 9) {
                    adj_room = level.roomlinks[room - 1].right;
                    if (adj_room > 0 && get_tile(adj_room, 0, tile_row) == tiles_11_loose) {
                        do_knock(adj_room, tile_row, 0);
                    }
                }
            }
        } else if (tile_is_empty(curr_tile2)) {
            break;
        }
    }
    for (tile_col = impact_col + 1; tile_col < 10; ++tile_col) {
        if (get_tile(room, tile_col, tile_row) == tiles_11_loose) {
            loose_make_shake();
            if (Kid.room == room && tile_col == 9) {
                adj_room = level.roomlinks[room - 1].right;
                if (adj_room > 0 && get_tile(adj_room, 0, tile_row) == tiles_11_loose) {
                    do_knock(adj_room, tile_row, 0);
                }
            }
        } else if (tile_is_empty(curr_tile2)) {
            break;
        }
    }
}

// seg007:1010
void add_mob(void) {
    if (mobs_count >= MAX_MOBS) {
        show_dialog("Mobs Overflow", default_dialog);
        return; // added
    }
    mobs[mobs_count++] = curmob;
}

// seg007:1041
short get_curr_tile(short tilepos) {
    curr_modifier = curr_room_modif[tilepos];
    return curr_tile = curr_room_tiles[tilepos] & 0x1F;
}

// data:43DC
word curmob_index;

// seg007:1063
void do_mobs(void) {
    short n_mobs;
    short index;
    short new_index;
    n_mobs = mobs_count;
    for (curmob_index = 0; n_mobs > curmob_index; ++curmob_index) {
        curmob = mobs[curmob_index];
        move_mob();
        check_loose_fall_on_kid();
        mobs[curmob_index] = curmob;
    }
    new_index = 0;
    for (index = 0; index < mobs_count; ++index) {
        if (mobs[index].speed != -1) {
            mobs[new_index++] = mobs[index];
        }
    }
    mobs_count = new_index;
}

// seg007:110F
void move_mob(void) {
    if (curmob.type == 0) {
        move_loose();
    }
    if (curmob.speed <= 0) {
        ++curmob.speed;
    }
}

// data:227A
const short mob_y_land[] = {-1, 62, 125, 188, 25};
// data:594A
word curr_tile_temp;
// seg007:1126
void move_loose(void) {
    short curmob_col;
    if (curmob.speed < 0) return;
    // custom logic - replace fake pillar with a floor tile in room 17 of level 13
    if (current_level == 14 && curr_room == 17 && curr_room_tiles[0] != tiles_1_floor) {
        curr_room_tiles[0] = tiles_1_floor;
        redraw_height = 63;
        set_redraw_full(0, 1);
        set_wipe(0, 1);
        set_redraw_full(1, 1);
        set_wipe(1, 1);
    }
    if (curmob.speed < 29) curmob.speed += 3;
    curmob.y += curmob.speed;
    if (curmob.room == 0) {
        if (curmob.y < 210) {
            return;
        } else {
            curmob.speed = -2;
            return;
        }
    }
    curmob_col = curmob.xh >> 2;
    if (curmob.y < 226 && mob_y_land[curmob.row + 1] <= curmob.y) {
        // fell into a different row
        curr_tile_temp = get_tile(curmob.room, curmob_col, curmob.row);
        if (curr_tile_temp == tiles_11_loose) {
            // custom logic - a falling tile landing on a loose tile should make a crashing sound
            set_sound_pos(curmob.room, curmob_col, curmob.row);
            play_sound(sound_2_tile_crashing);
            loose_fall();
        }
        if (tile_is_empty(curr_tile_temp) ||
                curr_tile_temp == tiles_11_loose ||
                // custom logic - spikes in the temple can have no floor
                (curr_tile_temp == tiles_2_spike &&
                 is_temple() &&
                 get_tile_ctrl_bits(curmob.room, curr_tilepos, 3) == 2)) {
            mob_down_a_row();
            // custom logic - floor spikes activate when a loose tile is landing on them
            if (!is_temple() &&
                    (tile_is_empty(get_tile(curmob.room, curmob_col, curmob.row)) || curr_tile2 == tiles_11_loose || curr_tile2 == tiles_2_spike) &&
                    (curr_tile2 == tiles_2_spike || get_tile(curmob.room, curmob_col, curmob.row + 1) == tiles_2_spike)) {
                start_anim_spike(curr_room, curr_tilepos);
            }
            return;
        }
        // custom logic - hide tile crashing sound in unlinked rooms
        if (!is_room_unlinked(curmob.room)) {
            set_sound_pos(curmob.room, curmob_col, curmob.row);
            play_sound(sound_2_tile_crashing); // tile crashing
        }
        do_knock(curmob.room, curmob.row, curmob_col);
        curmob.y = mob_y_land[curmob.row + 1];
        curmob.speed = -2;
        loose_land();
    }
}

// custom logic - landed loose tile changes dead guard frame to dead
bool loose_land_spiked_dead(char_type2* dead, word curmob_room, byte curmob_col, byte row) {
    char_type2 dead_guard = *dead;
    byte room_minus_1 = dead_guard.room - 1;
    if (dead_guard.room == curmob_room &&
            dead_guard.row == row &&
            (dead_guard.frame != frame_185_dead || dead_guard.frame2 != frame_185_dead)) {
        short dead_guard_col = get_tile_div_mod_m7(dead_guard.x);
        sbyte dead_guard_dir = dead_guard.direction;
        if ((dead_guard_dir == dir_0_right && curmob_col == dead_guard_col) ||
                (dead_guard_dir == dir_FF_left && curmob_col == MIN(dead_guard_col + 1, 9))) {
            dead->frame2 = frame_185_dead;
            // guard facing left might end up in abyss when spiked or laying close to a ledge
            if (!tile_is_floor(get_tile(dead_guard.room, dead_guard_col, dead_guard.row)) &&
                    curr_tile2 != tiles_20_wall) {
                if (falling_guard.room == 0 || falling_guard.direction == dir_56_none) {
                    dead->room = 0;
                    dead->direction = dir_56_none;
                    falling_guard.room = dead_guard.room;
                    falling_guard.direction = dead_guard_dir;
                    falling_guard.x = dead_guard.x;
                    falling_guard.y = y_land[dead_guard.row + 1];
                    falling_guard.fall_y = 0;
                    falling_guard.skill = guard_skill;
                    falling_guard.color = dead_guard.color;
                    falling_guard.start_fall = true;
                    falling_guard.hitp = 0;
                    return false;
                } else {
                    dead->x += 4;
                }
            }
            if (dead_guard1[room_minus_1].room == dead_guard.room &&
                    dead_guard1[room_minus_1].x == dead_guard.x &&
                    dead_guard1[room_minus_1].row == dead_guard.row &&
                    dead_guard1[room_minus_1].direction != dir_56_none) {
                dead->x -= 2;
            } else if (dead_guard2[room_minus_1].room == dead_guard.room &&
                       dead_guard2[room_minus_1].x == dead_guard.x &&
                       dead_guard2[room_minus_1].row == dead_guard.row &&
                       dead_guard2[room_minus_1].direction != dir_56_none) {
                dead->x += 2;
            }
            return true;
        }
    }
    return false;
}

// custom logic - landed loose tile changes dead guard frame in the next room to dead
bool loose_land_spiked_dead_left_room(char_type2* dead, word left_room, byte row) {
    char_type2 dead_guard = *dead;
    if (dead_guard.room == left_room &&
            dead_guard.row == row &&
            (dead_guard.frame == frame_177_spiked || dead_guard.frame2 == frame_177_spiked) &&
            dead_guard.x > DEADGUARD_OFFSET_RIGHT
       ) {
        dead->frame2 = frame_185_dead;
        return true;
    }
    return false;
}

// custom logic - landed loose tile changes inactive guard to dead
void loose_land_inactive(word room_minus_1, word curmob_room, byte curmob_col, byte row) {
    char_type2 inactive = inactive_guard[room_minus_1];
    if (inactive.room == curmob_room &&
            inactive.row == row &&
            (inactive.frame != frame_185_dead || inactive.frame2 != frame_185_dead)) {
        short inactive_col = get_tile_div_mod_m7(inactive.x);
        if (get_tile(curmob.room, inactive_col, curmob.row) == tiles_20_wall) {
            if (tile_is_floor(get_tile(curmob.room, dir_behind[inactive.direction + 1] + inactive_col, curmob.row))) {
                inactive_col += inactive.direction == dir_0_right ? -1 : 1;
            } else if (tile_is_floor(get_tile(curmob.room, dir_front[inactive.direction + 1] + inactive_col, curmob.row))) {
                inactive_col += inactive.direction == dir_0_right ? 1 : -1;
            }
        }
        if ((inactive.direction == dir_0_right && curmob_col == inactive_col) ||
                (inactive.direction == dir_FF_left && curmob_col == MIN(inactive_col + 1, 9))) {
            char_type2* slot = get_dead_guard_slot(room_minus_1);
            if (slot != NULL) {
                slot->room = inactive.room;
                slot->color = inactive.color;
                slot->direction = inactive.direction;
                slot->x = inactive.x;
                slot->row = inactive.row;
                slot->frame = inactive.frame;
                slot->frame2 = frame_185_dead;
                if (slot == &dead_guard1[room_minus_1] &&
                        Guard.room == inactive.room &&
                        Guard.curr_row == inactive.row &&
                        ABS(Guard.x - inactive.x) <= 1) {
                    slot->x += dir_front[inactive.direction + 1] * 2;
                }
            }
            inactive_guard[room_minus_1].room = 0;
            is_inactive_guard_notice = 0;
        }
    }
}

// seg007:11E8
void loose_land(void) {
    short button_type;
    short tiletype;
    button_type = 0;
    tiletype = get_tile(curmob.room, curmob.xh >> 2, curmob.row);
    switch (tiletype) {
        case tiles_15_opener:
            curr_room_tiles[curr_tilepos] = tiles_14_debris;
            button_type = tiles_14_debris;
        // fallthrough!
        case tiles_6_closer:
            trigger_button(1, button_type, -1);
            tiletype = get_tile(curmob.room, curmob.xh >> 2, curmob.row);
        // fallthrough!
        case tiles_1_floor:
        case tiles_2_spike:
        case tiles_5_stuck: // custom logic - stuck button should also be changed to debris
        case tiles_10_potion:
        case tiles_19_torch:
        case tiles_22_sword: // custom logic - sword should be also be changed to debris
        case tiles_30_torch_with_debris:
            if (tiletype == tiles_19_torch ||
                    tiletype == tiles_30_torch_with_debris
               ) {
                curr_room_tiles[curr_tilepos] = tiles_30_torch_with_debris;
                // custom logic - spikes in the temple use the first control bit to change the floor to debris
            } else if (tiletype == tiles_2_spike && is_temple()) {
                curr_room_tiles[curr_tilepos] = tiles_2_spike | 0x20;
                redraw_at_cur_mob();
                break;
            } else {
                curr_room_tiles[curr_tilepos] = tiles_14_debris;
            }
            redraw_at_cur_mob();
            if (tile_col != 0) {
                set_redraw_full(curr_tilepos - 1, 1);
            }
        // fallthrough!
        default:
            // custom logic - guards can be killed by fallen loose tiles
            fell_on_his_head();
            break;
    }
}

// seg007:12CB
void loose_fall(void) {
    curr_room_modif[curr_tilepos] = remove_loose(curr_tilepos);
    curmob.speed >>= 1;
    mobs[curmob_index] = curmob;
    curmob.y += 6;
    mob_down_a_row();
    add_mob();
    curmob = mobs[curmob_index];
    redraw_at_cur_mob();
}

// seg007:132C
void redraw_at_cur_mob(void) {
    if (curmob.room == drawn_room) {
        redraw_height = 0x20;
        set_redraw_full(curr_tilepos, 1);
        set_wipe(curr_tilepos, 1);
        // Redraw tile to the right only if it's in the same room.
        if ((curr_tilepos % 10) + 1 < 10) { // changed
            set_redraw_full(curr_tilepos + 1, 1);
            set_wipe(curr_tilepos + 1, 1);
        }
        // custom logic - redraw floor in the room on the left
    } else if (curmob.room == level.roomlinks[drawn_room - 1].left &&
               curr_tilepos % 10 == 9) {
        need_full_redraw = 1;
        // custom logic - redraw ceiling in the room at the top
    } else if (curmob.room == level.roomlinks[drawn_room - 1].up &&
               curr_tilepos / 10 >= 2) {
        need_full_redraw = 1;
    }
}

// seg007:1387
void mob_down_a_row(void) {
    ++curmob.row;
    if (curmob.row >= 3) {
        curmob.y -= 192;
        curmob.row = 0;
        curmob.room = level.roomlinks[curmob.room - 1].down;
    }
}

// seg007:13AE
void draw_mobs(void) {
    short index;
    for (index = 0; index < mobs_count; ++index) {
        curmob = mobs[index];
        draw_mob();
    }
}

// seg007:13E5
void draw_mob(void) {
    short tile_row;
    short ypos;
    short top_row;
    short tilepos;
    short tile_col;
    ypos = curmob.y;
    tile_col = curmob.xh >> 2;
    if (curmob.room == drawn_room) {
        if (curmob.y >= 210) return;
    } else if (curmob.room == room_B) {
        // 18 is a generous estimation of the falling tile height
        if (ABS((sbyte)curmob.y) >= 18) return;
        curmob.y += 192;
        ypos = curmob.y;
    } else if (curmob.room == room_A) {
        if (curmob.y < 174) return;
        ypos = curmob.y - 189;
        // custom logic - loose tiles in the room on the left at column 9 should be visible
    } else if (curmob.room == room_L && tile_col == 9) {
        if (curmob.y >= 210) return;
        curmob.xh = -4;
        tile_col = -1;
    } else {
        return;
    }
    tile_row = y_to_row_mod4(ypos);
    obj_tilepos = get_tilepos_nominus(tile_col, tile_row);
    ++tile_col;
    tilepos = get_tilepos(tile_col, tile_row);
    set_redraw2(tilepos, 1);
    // custom logic - an extra redraw frame prevents black rectagles near a wall
    set_redraw_fore(tilepos, 2);
    top_row = y_to_row_mod4(ypos - 18);
    if (top_row != tile_row) {
        tilepos = get_tilepos(tile_col, top_row);
        set_redraw2(tilepos, 1);
        set_redraw_fore(tilepos, 2);
    }
    add_mob_to_objtable(ypos);
}

// seg007:14DE
void add_mob_to_objtable(int ypos) {
    word index;
    objtable_type* curr_obj;
    index = objtable_count++;
    curr_obj = &objtable[index];
    curr_obj->obj_type = curmob.type | 0x80;
    curr_obj->xh = curmob.xh;
    curr_obj->xl = 0;
    curr_obj->y = ypos;
    curr_obj->chtab_id = id_chtab_6_environment;
    curr_obj->id = 10;
    curr_obj->clip.top = 0;
    curr_obj->clip.left = 0;
    curr_obj->clip.right = 40;
    mark_obj_tile_redraw(index);
}

// seg007:1556
int is_spike_harmful(void) {
    // custom logic - shadow/mouse does not die on spikes
    if (Char.charid == charid_0_kid && (is_kid_shadow || is_kid_mouse())) {
        return 0;
    }
    // custom logic - skeletons in dungeon 2 are not spikable
    if (Char.charid == charid_4_skeleton && is_dungeon_2()) {
        return 0;
    }
    sbyte modifier;
    modifier = curr_room_modif[curr_tilepos];
    // custom logic - regular spikes that impaled someone can still impale
    // another character when the fall speed is high enough
    if (!is_temple() && Char.fall_y > 22 && modifier == -1) {
        return 1;
    }
    if (modifier == 0 || modifier == -1) {
        return 0;
    } else if (modifier < 0) {
        return 1;
    } else if (modifier < 5) {
        return 2;
    } else {
        return 0;
    }
}

// seg007:1591
void check_loose_fall_on_kid(void) {
    sbyte char_col;
    byte room;
    loadkid();
    char_col = char_col_climbing_left();
    room = Char.room;
    // custom logic - kid can get hurt by a tile in the room on the left at column 9
    if (char_col < 0) {
        room = level.roomlinks[room - 1].left;
        char_col += 10;
    }
    if (room == curmob.room &&
            char_col == curmob.xh >> 2 &&
            curmob.y < Char.y &&
            Char.y - 30 < curmob.y
       ) {
        fell_on_your_head();
        savekid();
    }
}

// seg007:15D3
void fell_on_your_head(void) {
    // custom logic - loose floors do not hurt kid as shadow
    if (is_kid_shadow) {
        return;
    }
    if (Char.alive >= 0) {
        return;
    }
    short frame;
    short action;
    bool is_bumped_frame;
    frame = Char.frame;
    action = Char.action;
    // custom logic - do not affect kid when hanging off vertical spikes in the temple
    if (frame == frame_184_spiked_hanging || frame == frame_100_spiking_hanging || frame == frame_100_spiking_hanging + 1 /* facing right */) {
        return;
    }
    // custom logic - lose floors do hurt kid when he stands up from crouch or bumping into a wall
    // but not if he stays crouched
    is_bumped_frame = (frame >= frame_45_turn && frame <= frame_65_runturn) ||
                      (frame >= frame_110_stand_up_from_crouch_1 && frame <= frame_119_stand_up_from_crouch_10);
    // custom logic - tiles hurt kid when hanging
    // loose floors only hurt kid when falling from 2+ stories while running or crouching
    if ((frame != frame_109_crouch || curmob.speed >= 22) &&
            (curmob.speed >= 22 || (frame < frame_5_start_run || frame >= frame_15_stand)) &&
            (action <= actions_2_hang_climb ||
             (action == actions_5_bumped && is_bumped_frame) ||
             action == actions_7_turn ||
             action == actions_6_hang_straight)
       ) {
        Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
        Char.sword = sword_0_sheathed;
        // custom logic - do not take 2 hit points during a single animation
        if (delayed_sound_timer > 0 &&
                delayed_sound.sound_id == sound_13_kid_hurt &&
                delayed_sound.button_type == tiles_11_loose) {
            delayed_sound_timer = 0;
            goto loc_fell_on_head;
        }
        if (take_hp(is_kid_mouse() ? 100 : 1) && tile_is_floor(get_tile_at_char())) {
            seqtbl_offset_char(seq_22_crushed); // dead (because of loose floor)
            if (frame == frame_177_spiked) { // spiked
                Char.frame = frame_185_dead;
                Char.x = char_dx_forward(-12);
            }
        } else {
loc_fell_on_head:
            if (get_tile_behind_char() == 0) {
                Char.x = char_dx_forward(-2);
            }
            if (!is_kid_shadow) {
                seqtbl_offset_char(seq_52_loose_floor_fell_on_kid); // loose floor fell on Kid
                // custom logic - kid screams when a loose tile hits him
                delayed_sound_timer = 6;
                delayed_sound = (struct delayed_sound_type) {
                    sound_13_kid_hurt, 0, 0, 0, 0, 0, 0
                };
            }
        }
    }
}

// custom logic - kill guard if a loose fell on him
void fell_on_his_head(void) {
    short curmob_col;
    short guard_col;
    sbyte guard_dir;
    word crushed_seq;
    word spiked_seq;
    word room_minus_1;
    bool dead_guard_crushed;
    byte curr_guard_knocked_tile;
    curmob_col = curmob.xh >> 2;
    spiked_seq = seqtbl_offset_get_seq(seq_51_spiked);
    // opponent must be a guard (not a skeleton, shadow or mouse)
    if (Guard.charid != charid_2_guard) {
        return;
    }
    // guard turn around when loose tiles land
    if (Guard.alive < 0 && Guard.room == drawn_room && Guard.room == curr_room && !can_guard_see_kid) {
        if ((Guard.direction == dir_0_right && Guard.curr_col > curmob_col) ||
                (Guard.direction == dir_FF_left && Guard.curr_col < curmob_col)) {
            guard_notice_timer = 1;
        }
    }
    // kid dies when shadow drops a loose tile on him
    if (is_kid_shadow &&
            Dead.room == curmob.room &&
            Dead.row == curmob.row &&
            (curmob_col == get_tile_div_mod_m7(Dead.x + 7))
       ) {
        Dead.color = 0;
        Kid.alive = 0;
        hitp_curr = 0;
    }
    curr_guard_knocked_tile = guard_knocked_tile;
    guard_knocked_tile = 0;
    room_minus_1 = curmob.room - 1;
    // kill inactive guard
    loose_land_inactive(room_minus_1, curmob.room, curmob_col, curmob.row);
    // change frame for spiked guards except for the temple level that has vertical spikes
    if (!is_temple()) {
        dead_guard_crushed = loose_land_spiked_dead(&dead_guard1[room_minus_1], curmob.room, curmob_col, curmob.row);
        dead_guard_crushed = dead_guard_crushed || loose_land_spiked_dead(&dead_guard2[room_minus_1], curmob.room, curmob_col, curmob.row);
        dead_guard_crushed = dead_guard_crushed || loose_land_spiked_dead(&dead_guard3[room_minus_1], curmob.room, curmob_col, curmob.row);
        dead_guard_crushed = dead_guard_crushed || loose_land_spiked_dead(&outbound_dead_guard1[room_minus_1], curmob.room, curmob_col, curmob.row);
        dead_guard_crushed = dead_guard_crushed || loose_land_spiked_dead(&outbound_dead_guard2[room_minus_1], curmob.room, curmob_col, curmob.row);
        dead_guard_crushed = dead_guard_crushed || loose_land_spiked_dead(&outbound_dead_guard3[room_minus_1], curmob.room, curmob_col, curmob.row);
        if (dead_guard_crushed && curmob_col <= 0) {
            word left_room = level.roomlinks[room_minus_1].left;
            if (left_room) {
                room_minus_1 = left_room - 1;
                loose_land_spiked_dead_left_room(&dead_guard1[room_minus_1], left_room, curmob.row);
                loose_land_spiked_dead_left_room(&dead_guard2[room_minus_1], left_room, curmob.row);
                loose_land_spiked_dead_left_room(&dead_guard3[room_minus_1], left_room, curmob.row);
                loose_land_spiked_dead_left_room(&outbound_dead_guard1[room_minus_1], left_room, curmob.row);
                loose_land_spiked_dead_left_room(&outbound_dead_guard2[room_minus_1], left_room, curmob.row);
                loose_land_spiked_dead_left_room(&outbound_dead_guard3[room_minus_1], left_room, curmob.row);
            }
            // no active guard in the room
            return;
        }
    }
    // kill guards in the drawn room
    if (drawn_room == curmob.room) {
        guard_col = get_tile_div_mod_m7(Guard.x);
        if (Guard.charid == charid_2_guard &&
                Guard.curr_row == curmob.row &&
                ((Guard.direction == dir_0_right && curmob_col == guard_col) ||
                 (Guard.direction == dir_FF_left && curmob_col == MIN(guard_col + 1, 9)))
           ) {
            loadshad();
            if (Char.alive < 0) {
                // guard should not die if he knocked the tile down
                if (!curr_guard_knocked_tile) {
                    draw_guard_hp(0, guardhp_max);
                    guardhp_curr = guardhp_delta = 0;
                    Char.frame = frame_185_dead;
                    goto loc_fell_on_guard;
                }
            } else if (Char.frame != frame_202_bowing &&
                       (dynamic_guard.room != drawn_room || Char.alive >= DROPDEAD_FRAMES)) {
loc_fell_on_guard:
                Char.action = actions_1_run_jump;
                Char.x = x_bump[guard_col + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
                Char.y = y_land[Char.curr_row + ONSCREEN_ROW_OFFSET];
                Char.fall_y = 0;
                seqtbl_offset_char(seq_127_dead);
                saveshad();
                play_seq();
            }
        }
        // kill guards in a different room
    } else {
        guard_col = get_tile_div_mod_m7(level.guards_x[room_minus_1]);
        guard_dir = level.guards_dir[room_minus_1];
        // an edge case where guard tile ends up inside a wall
        if (get_tile(curmob.room, guard_col, curmob.row) == tiles_20_wall) {
            if (tile_is_floor(get_tile(curmob.room, dir_behind[guard_dir + 1] + guard_col, curmob.row))) {
                guard_col += guard_dir == dir_0_right ? -1 : 1;
            } else if (tile_is_floor(get_tile(curmob.room, dir_front[guard_dir + 1] + guard_col, curmob.row))) {
                guard_col += guard_dir == dir_0_right ? 1 : -1;
            }
        }
        if ((!curr_guard_knocked_tile || (Guard.room == Kid.room && Guard.direction != dir_56_none)) &&
                (level.guards_seq_hi[room_minus_1] == 0 ||
                 (level.guards_seq_lo[room_minus_1] == (byte)spiked_seq && level.guards_seq_hi[room_minus_1] == (spiked_seq >> 8))) &&
                level.guards_tile[room_minus_1] / 10 == curmob.row &&
                (((level.guards_seq_hi[room_minus_1] != 0 || guard_dir == dir_0_right) && curmob_col == guard_col) ||
                 (level.guards_seq_hi[room_minus_1] == 0 && guard_dir == dir_FF_left && curmob_col == MIN(guard_col + 1, 9)))
           ) {
            short guardtype = settings->tbl_guard_type[current_level];
            if (guardtype == guard_type_guard || guardtype == guard_type_fat || guardtype == guard_type_bird) {
                if (inactive_guard[room_minus_1].room == 0) {
                    crushed_seq = seqtbl_offset_get_seq(seq_22_crushed);
                    level.guards_seq_lo[room_minus_1] = crushed_seq;
                    level.guards_seq_hi[room_minus_1] = crushed_seq >> 8;
                    level.guards_x[room_minus_1] = x_bump[guard_col + ONSCREEN_COLUMN_OFFSET] + CHAR_OFFSET;
                } else {
                    char_type2* slot = get_dead_guard_slot(room_minus_1);
                    if (slot != NULL) {
                        slot->room = room_minus_1 + 1;
                        slot->direction = level.guards_dir[room_minus_1];
                        slot->row = curmob.row;
                        slot->x = level.guards_x[room_minus_1];
                        slot->color = level.guards_color[room_minus_1];
                        slot->frame2 = frame_185_dead;
                    }
                    level.guards_seq_hi[room_minus_1] = 0;
                    level.guards_dir[room_minus_1] = inactive_guard[room_minus_1].direction;
                    level.guards_x[room_minus_1] = inactive_guard[room_minus_1].x;
                    level.guards_tile[room_minus_1] = inactive_guard[room_minus_1].row * 10;
                    level.guards_color[room_minus_1] = inactive_guard[room_minus_1].color;
                    if (inactive_guard_skill < NUM_GUARD_SKILLS) {
                        level.guards_skill[room_minus_1] = inactive_guard_skill;
                        inactive_guard_skill = 0xFF;
                    }
                    inactive_guard[room_minus_1].room = 0;
                }
            }
        }
    }
}

// seg007:1669
void play_door_sound_if_visible(int sound_id) {
    word tilepos;
    word gate_room;
    sbyte has_sound_condition;
    bool is_loud;
    tilepos = trob.tilepos;
    gate_room = trob.room;
    // custom logic - a gate that can be heard throughout the level
    is_loud = get_tile_ctrl_bits(gate_room, tilepos, 1) != 0;
    // custom logic - all gates can be heard in positional audio mode
    if (is_loud) {
        has_sound_condition = 1;
    } else if (is_positional_audio) {
        // custom logic - do not interrupt exclusive sounds or gate sounds from the current room
        if (gate_room != drawn_room && (gate_room != room_L || tilepos % 10 != 9)) {
            if (ogg_fore_sound_playing) {
                return;
            }
            if (check_sound_playing() &&
                    (current_sound == sound_4_gate_closing ||
                     current_sound == sound_5_gate_opening ||
                     is_exclusive_sound(current_sound))) {
                return;
            }
        }
        has_sound_condition = 1;
        // custom logic - gates near the flaming sword cannot be heard in the rooms above
        if (current_level == settings->flame_level &&
                (gate_room == 7 || gate_room == 9 || gate_room == 10) &&
                drawn_room != 7 && drawn_room != 9 && drawn_room != 10) {
            has_sound_condition = 0;
        }
    } else {
        has_sound_condition = (gate_room == room_L && tilepos % 10 == 9) ||
                              (gate_room == drawn_room && tilepos % 10 != 9);
    }
    if (!has_sound_condition) {
        return;
    }
    // Special event: sound of closing gates
    if (is_loud) {
        set_sound_pos(trob.room, trob.tilepos % 10 + 1, trob.tilepos / 10);
    } else {
        // regular gates sound quieter from a distance when closing
        set_sound_pos_with_dist_loss(trob.room, trob.tilepos % 10 + 1, trob.tilepos / 10, 12.0f);
    }
    play_sound(sound_id);
}

// custom logic - animate moving wall on level 12B
void animate_moving_wall(void) {
    short wall_pos, wall_pos_m10, wall_col, pos_offset, delta_x;
    word level, room, row, left_col, right_col;
    bool kid_on_wall, kid_infrontof_wall, is_torch;
    level = settings->moving_wall_level;
    room = settings->moving_wall_room;
    row = settings->moving_wall_row;
    left_col = settings->moving_wall_col_left;
    right_col = settings->moving_wall_col_right;
    if (current_level != level) {
        return;
    }
    wall_col = -1;
    pos_offset = tbl_line[row];
    get_room_address(room);
    for (wall_pos = left_col + pos_offset; wall_pos <= right_col + pos_offset; ++wall_pos) {
        if (curr_room_tiles[wall_pos] == tiles_20_wall) {
            wall_col = wall_pos % 10;
            break;
        }
    }
    if (wall_col < 0) {
        return;
    }
    if (wall_col == left_col) {
        moving_wall_direction = dir_0_right;
    } else if (wall_col == right_col) {
        moving_wall_direction = dir_FF_left;
    }
    loadkid();
    if (Char.room == room && Char.curr_row == row - 1 &&
            Char.y < y_land[Char.curr_row + ONSCREEN_ROW_OFFSET] && (Char.frame < frame_67_start_jump_up_1 || Char.frame > frame_85_hangdrop_5)) {
        return;
    }
    kid_on_wall = Char.curr_row == 0 &&
                  Char.curr_col == wall_col &&
                  (Char.frame <= frame_86_test_foot || (Char.frame >= frame_109_crouch && Char.frame <= frame_132_stepping_12));
    kid_infrontof_wall = (Char.curr_row == 0 || Char.curr_row == 1) &&
                         Char.curr_col == wall_col &&
                         ((Char.frame >= frame_87_hanging_1 && Char.frame <= frame_99_hanging_13) ||
                          (Char.frame >= frame_135_climbing_1 && Char.frame <= frame_149_climbing_15));
    if (Char.room == room && (kid_on_wall || kid_infrontof_wall)) {
        delta_x = Char.direction == moving_wall_direction ? 14 : -14;
        Char.x = char_dx_forward(delta_x);
        determine_col();
        savekid();
    } else if (wall_col == left_col || wall_col == right_col ||
               ((Char.action == actions_3_in_midair || Char.action == actions_4_in_freefall) && Char.room == room && Char.curr_row <= row)) {
        return;
    } else {
        moving_wall_direction = dir_FF_left;
    }
    if (moving_wall_direction == dir_0_right) {
        ++wall_col;
    } else {
        --wall_col;
    }
    for (wall_pos = left_col + pos_offset; wall_pos <= right_col + pos_offset; ++wall_pos) {
        wall_pos_m10 = wall_pos - 10;
        is_torch = wall_pos_m10 % 10 == 7;
        if (wall_pos == wall_col + pos_offset) {
            curr_room_tiles[wall_pos] = tiles_20_wall;
            curr_room_modif[wall_pos] = 0;
            curr_room_tiles[wall_pos_m10] = is_torch ? tiles_19_torch : tiles_1_floor;
        } else {
            curr_room_tiles[wall_pos] = tiles_0_empty;
            curr_room_modif[wall_pos] = 2;
            curr_room_tiles[wall_pos_m10] = is_torch ? tiles_26_torch_empty : tiles_0_empty;
        }
        if (is_torch && curr_room_modif[wall_pos_m10] == 0) {
            start_anim_torch(room, 7);
        }
    }
    need_full_redraw = 1;
    set_sound_pos(7, wall_col, 1);
    play_sound(sound_51_wall_sliding);
}

// *INDENT-OFF*
// custom logic - balcony star offset table
#define N_BALCONY_STARS 5
const custom_image_type balcony_stars[] = {
    // star id, x offset, y offset
    { 0, 53, -2 },
    { 1, 55,  4 },
    { 2, 56, 17 },
    { 3, 72, -2 },
    { 4, 71, 10 },
};
// *INDENT-ON*

// custom logic - animate balcony stars
void animate_balcony_stars(short tilepos, bool left_room) {
    word which;
    rect_type rect;
    custom_image_type star_image;
    which = prandom(N_BALCONY_STARS - 1);
    if (left_room && which < 3) {
        return;
    }
    star_image = balcony_stars[which];
    if (left_room) {
        star_image.x -= col_xh[10 - (tilepos % 10)] * 8;
    } else {
        star_image.x += col_xh[tilepos % 10] * 8;
    }
    star_image.y += 63 * (tilepos / 10);
    rect = (struct rect_type) {
        star_image.y, star_image.x, star_image.y + 1, star_image.x + 1
    };
    draw_star_rect(&rect, which, 1);
}
