/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"

// data:27E0
add_table_type ptr_add_table = add_backtable;

// *INDENT-OFF*
// data:259C
const piece tile_table[31] = {
    {   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0}, // 0x00 empty
    {  41,   1,   0,  42,   1,   2, 145,   0,  43,   0,   0,   0}, // 0x01 floor
    { 127,   1,   0, 133,   1,   2, 145,   0,  43,   0,   0,   0}, // 0x02 spike
    {  92,   1,   0,  93,   1,   2,   0,  94,  43,  95,   1,   0}, // 0x03 pillar
    {  46,   1,   0,  47,   1,   2,   0,  48,  43,  49,   3,   0}, // 0x04 door
    {  41,   1,   1,  35,   1,   3, 145,   0,  36,   0,   0,   0}, // 0x05 stuck floor
    {  41,   1,   0,  42,   1,   2, 145,   0,  96,   0,   0,   0}, // 0x06 close button
    {  46,   1,   0,   0,   0,   2,   0,   0,  43,  49,   3,   0}, // 0x07 door top with floor
    {  86,   1,   0,  87,   1,   2,   0,   0,  43,  88,   1,   0}, // 0x08 big pillar bottom
    {   0,   0,   0,  89,   0,   3,   0,  90,   0,  91,   1,   3}, // 0x09 big pillar top
    {  41,   1,   0,  42,   1,   2, 145,   0,  43,  12,   2,  -3}, // 0x0A potion
    {   0,   1,   0,   0,   0,   0, 145,   0,   0,   0,   0,   0}, // 0x0B loose floor
    {   0,   0,   0,   0,   0,   2,   0,   0,  85,  49,   3,   0}, // 0x0C door top
    {  75,   1,   0,  42,   1,   2,   0,   0,  43,  77,   0,   0}, // 0x0D mirror
    {  97,   1,   0,  98,   1,   2, 145,   0,  43, 100,   0,   0}, // 0x0E debris
    { 147,   1,   0,  42,   1,   1, 145,   0, 149,   0,   0,   0}, // 0x0F open button
    {  41,   1,   0,  37,   0,   0,   0,  38,  43,   0,   0,   0}, // 0x10 leveldoor left
    {   0,   0,   0,  39,   1,   2,   0,  40,  43,   0,   0,   0}, // 0x11 leveldoor right
    {   0,   0,   0,  42,   1,   2, 145,   0,  43,   0,   0,   0}, // 0x12 chomper
    {  41,   1,   0,  42,   1,   2,   0,   0,  43,   0,   0,   0}, // 0x13 torch
    {   0,   0,   0,   1,   1,   2,   0,   2,   0,   0,   0,   0}, // 0x14 wall
    {  30,   1,   0,  31,   1,   2,   0,   0,  43,   0,   0,   0}, // 0x15 skeleton
    {  41,   1,   0,  42,   1,   2, 145,   0,  43,   0,   0,   0}, // 0x16 sword
    {  41,   1,   0,  10,   0,   0,   0,  11,  43,   0,   0,   0}, // 0x17 balcony left
    {   0,   0,   0,  12,   1,   2,   0,  13,  43,   0,   0,   0}, // 0x18 balcony right
    {  92,   1,   0,  42,   1,   2, 145,   0,  43,  95,   1,   0}, // 0x19 lattice pillar
    {   1,   0,   0,   0,   0,   0,   0,   0,   2,   9,   0, -53}, // 0x1A lattice down
    {   3,   0, -10,   0,   0,   0,   0,   0,   0,   9,   0, -53}, // 0x1B lattice small
    {   4,   0, -10,   0,   0,   0,   0,   0,   0,   9,   0, -53}, // 0x1C lattice left
    {   5,   0, -10,   0,   0,   0,   0,   0,   0,   9,   0, -53}, // 0x1D lattice right
    {  97,   1,   0,  98,   1,   2,   0,   0,  43, 100,   0,   0}, // 0x1E debris with torch
};
// *INDENT-ON*

// data:4334
short drawn_row;
// data:4352
short draw_bottom_y;
// data:4326
short draw_main_y;
// data:34D0
short drawn_col;

// data:6592
byte tile_left;
// data:4CCC
byte modifier_left;

// seg008:0006
void redraw_room(void) {
    free_peels();
    memset(table_counts, 0, sizeof(table_counts));
    reset_obj_clip();
    draw_room();
    clear_tile_wipes();
}

// seg008:0035
void load_room_links(void) {
    room_BR = 0;
    room_BL = 0;
    room_AR = 0;
    room_AL = 0;
    if (drawn_room) {
        get_room_address(drawn_room);
        room_L = level.roomlinks[drawn_room - 1].left;
        room_R = level.roomlinks[drawn_room - 1].right;
        room_A = level.roomlinks[drawn_room - 1].up;
        room_B = level.roomlinks[drawn_room - 1].down;
        if (room_A) {
            room_AL = level.roomlinks[room_A - 1].left;
            room_AR = level.roomlinks[room_A - 1].right;
        } else {
            if (room_L) {
                room_AL = level.roomlinks[room_L - 1].up;
            }
            if (room_R) {
                room_AR = level.roomlinks[room_R - 1].up;
            }
        }
        if (room_B) {
            room_BL = level.roomlinks[room_B - 1].left;
            room_BR = level.roomlinks[room_B - 1].right;
        } else {
            if (room_L) {
                room_BL = level.roomlinks[room_L - 1].down;
            }
            if (room_R) {
                room_BR = level.roomlinks[room_R - 1].down;
            }
        }
    } else {
        room_B = 0;
        room_A = 0;
        room_R = 0;
        room_L = 0;
    }
}

// seg008:0125
void draw_room(void) {
    word saved_room;
    load_leftroom();
    leveldoor_ybottom = 0;
    for (drawn_row = 3; drawn_row--; ) { /*2,1,0*/
        load_rowbelow();
        draw_bottom_y = 63 * drawn_row + 65;
        draw_main_y = draw_bottom_y - 3;
        for (drawn_col = 0; drawn_col < 10; drawn_col++) {
            load_curr_and_left_tile();
            draw_tile();
        }
    }
    saved_room = drawn_room;
    drawn_room = room_A;
    load_room_links();
    load_leftroom();
    drawn_row = 2;
    load_rowbelow();
    for (drawn_col = 0; drawn_col < 10; ++drawn_col) {
        load_curr_and_left_tile();
        draw_main_y = -1;
        draw_bottom_y = 2;
        draw_tile_aboveroom(saved_room);
    }
    drawn_room = saved_room;
    load_room_links();
}

// seg008:01C7
void draw_tile(void) {
    draw_tile_floorright();
    draw_tile_anim_topright();
    draw_tile_right();
    draw_tile_anim_right();
    draw_tile_bottom(0);
    draw_loose();
    draw_tile_base();
    draw_tile_anim();
    draw_tile_fore();
}

// custom logic - do not draw tiles above rooms 6 and 23 on level 5
#define IS_FAKE_CEILING(level, room) level == 5 && (room == 6 || room == 23)

// seg008:01F2
void draw_tile_aboveroom(word saved_room) {
    // custom logic - check for fake ceiling
    if (IS_FAKE_CEILING(current_level, saved_room)) {
        return;
    }
    draw_tile_floorright();
    draw_tile_anim_topright();
    draw_tile_right();
    draw_tile_bottom(0);
    draw_loose();
    draw_tile_fore();
}

// seg008:0211
void redraw_needed(short tilepos) {
    if (wipe_frames[tilepos]) {
        --wipe_frames[tilepos];
        draw_tile_wipe(wipe_heights[tilepos]);
    }
    if (redraw_frames_full[tilepos]) {
        --redraw_frames_full[tilepos];
        draw_tile();
    } else {
        if (redraw_frames_anim[tilepos]) {
            --redraw_frames_anim[tilepos];
            draw_tile_anim_topright();
            draw_tile_anim_right();
            draw_tile_anim();
            draw_tile_fore();
            draw_tile_bottom(0);
        }
    }
    if (redraw_frames2[tilepos]) {
        --redraw_frames2[tilepos];
        draw_other_overlay();
    } else {
        if (redraw_frames_floor_overlay[tilepos]) {
            --redraw_frames_floor_overlay[tilepos];
            draw_floor_overlay();
        }
    }
    if (tile_object_redraw[tilepos]) {
        if (tile_object_redraw[tilepos] == 0xFF) {
            draw_objtable_items_at_tile(tilepos - 1);
        }
        draw_objtable_items_at_tile(tilepos);
        tile_object_redraw[tilepos] = 0;
    }
    if (redraw_frames_fore[tilepos]) {
        --redraw_frames_fore[tilepos];
        draw_tile_fore();
    }
}

// seg008:02C1
void redraw_needed_above(int column) {
    if (redraw_frames_above[column] != 0) {
        --redraw_frames_above[column];
        // custom logic - jumping up near a big pillar with the top
        // part in another room creates black squares on the screencreates
        if (curr_tile != tiles_9_bigpillar_top) {
            draw_tile_wipe(3);
            draw_tile_floorright();
        }
        draw_tile_anim_topright();
        draw_tile_right();
        draw_tile_bottom(1);
        draw_loose();
        draw_tile_fore();
    }
}

// seg008:02FE
int get_tile_to_draw(int room, int column, int row, byte *ptr_tiletype, byte *ptr_modifier, byte tile_room0) {
    word tilepos = tbl_line[row] + column;
    if (column == -1) {
        *ptr_tiletype = leftroom_[row].tiletype;
        *ptr_modifier = leftroom_[row].modifier;
    } else if (room) {
        *ptr_tiletype = curr_room_tiles[tilepos] & 0x1F;
        *ptr_modifier = curr_room_modif[tilepos];
    } else {
        *ptr_modifier = 0;
        *ptr_tiletype = tile_room0;
    }
    // Is this a pressed button?
    byte tiletype = (*ptr_tiletype) & 0x1F;
    byte modifier = *ptr_modifier;
    if (tiletype == tiles_6_closer) {
        if (get_doorlink_timer(modifier) > 1) {
            *ptr_tiletype = tiles_5_stuck;
        }
    } else if (tiletype == tiles_15_opener) {
        if (get_doorlink_timer(modifier) > 1) {
            *ptr_modifier = 0;
            *ptr_tiletype = tiles_1_floor;
        }
    } else if (tiletype == tiles_11_loose) {
        if ((*ptr_modifier & 0x7F) == 0) {
            *ptr_tiletype = tiles_1_floor;
        }
    } else if (tiletype == tiles_23_balcony_left) {
        // custom logic - reset upper 4 bits used for animating stars
        *ptr_modifier = modifier & 0x0F;
    }
    return *ptr_tiletype;
}

// seg008:03BB
void load_curr_and_left_tile(void) {
    word tiletype;
    tiletype = tiles_20_wall;
    if (drawn_row == 2) {
        // custom logic - do not draw tiles above rooms 6 and 23 on level 5
        // This fix prevents the dropped loose tile to appear at the top of the room.
        if (IS_FAKE_CEILING(current_level, Kid.room)) {
            tiletype = tiles_0_empty;
        } else {
            tiletype = tiles_1_floor;
        }
    }
    get_tile_to_draw(drawn_room, drawn_col, drawn_row, &curr_tile, &curr_modifier, tiletype);
    get_tile_to_draw(drawn_room, drawn_col - 1, drawn_row, &tile_left, &modifier_left, tiletype);
    draw_xh = col_xh[drawn_col];
}

// seg008:041A
void load_leftroom(void) {
    word row;
    get_room_address(room_L);
    for (row = 0; row < 3; ++row) {
        // wall at left of level
        get_tile_to_draw(room_L, 9, row, &leftroom_[row].tiletype, &leftroom_[row].modifier, tiles_20_wall);
    }
}

// seg008:0460
void load_rowbelow(void) {
    word row_below;
    word column;
    word room;
    word room_left;
    if (drawn_row == 2) {
        room = room_B;
        room_left = room_BL;
        row_below = 0;
    } else {
        room = drawn_room;
        room_left = room_L;
        row_below = drawn_row + 1;
    }
    get_room_address(room);
    for (column = 1; column < 10; ++column) {
        get_tile_to_draw(room, column - 1, row_below, &row_below_left_[column].tiletype, &row_below_left_[column].modifier, tiles_0_empty);
    }
    get_room_address(room_left);
    // wall at left of level
    get_tile_to_draw(room_left, 9, row_below, &row_below_left_[0].tiletype, &row_below_left_[0].modifier, tiles_20_wall);
    get_room_address(drawn_room);
}

// seg008:04FA
void draw_tile_floorright(void) {
    if (can_see_bottomleft() == 0) return;
    draw_tile_topright();
    if (tile_table[tile_left].floor_right == 0) return;
    add_backtable(id_chtab_6_environment, 42 /*floor right part*/, draw_xh, 0, tile_table[tiles_1_floor].right_y + draw_main_y, blitters_9_black, 1);
}

// seg008:053A
int can_see_bottomleft(void) {
    return curr_tile == tiles_0_empty ||
           curr_tile == tiles_9_bigpillar_top ||
           curr_tile == tiles_12_doortop ||
           curr_tile == tiles_26_lattice_down ||
           // custom logic - spikes in the temple with control bits set to 010 do not have a floor
           (curr_tile == tiles_2_spike &&
            is_temple() &&
            get_tile_ctrl_bits(drawn_room, get_tilepos(drawn_col, drawn_row), 3) == 2);
}

const byte doortop_fram_top[] = {0, 81, 83, 0};

// seg008:055A
void draw_tile_topright(void) {
    byte tiletype;
    byte tile_modifier;
    tiletype = row_below_left_[drawn_col].tiletype;
    if (tiletype == tiles_7_doortop_with_floor || tiletype == tiles_12_doortop) {
        if (settings->tbl_level_type[current_level] == DUNGEON_LEVEL) return;
        // custom logic - do not draw the top part of a doortop at row 0 in the temple
        if (drawn_row == 2 && tile_is_empty(tile_left) && is_temple()) {
            return;
        }
        tile_modifier = row_below_left_[drawn_col].modifier;
        if (tile_modifier > 10) {
            tile_modifier -= 10;
        }
        add_backtable(id_chtab_6_environment, doortop_fram_top[tile_modifier], draw_xh, 0, draw_bottom_y, blitters_10h_transp, 0);
    } else if (tiletype == tiles_20_wall) {
        add_backtable(id_chtab_7_environmentwall, 2, draw_xh, 0, draw_bottom_y, blitters_10h_transp, 0);
    } else {
        int id = tile_table[tiletype].topright_id;
        // custom logic - balcony with modifier of 1 on left tile should be replaced with a huge poison potion
        if ((tiletype == tiles_23_balcony_left && row_below_left_[drawn_col].modifier > 0) ||
                (tiletype == tiles_24_balcony_right && row_below_left_[drawn_col - 1].modifier > 0)) {
            id += 4; // transparent sprite for poison potion
        }
        add_backtable(id_chtab_6_environment, id, draw_xh, 0, draw_bottom_y, blitters_10h_transp, 0);
    }
}

const byte door_fram_top[] = {60, 61, 62, 63, 64, 65, 66, 67};

// seg008:05D1
void draw_tile_anim_topright(void) {
    word modifier;
    if ((curr_tile == tiles_0_empty ||
            curr_tile == tiles_9_bigpillar_top ||
            curr_tile == tiles_12_doortop) &&
            row_below_left_[drawn_col].tiletype == tiles_4_gate
       ) {
        add_backtable(id_chtab_6_environment, 68 /*gate top mask*/, draw_xh, 0, draw_bottom_y, blitters_40h_mono, 0);
        modifier = row_below_left_[drawn_col].modifier;
        if (modifier > 188) modifier = 188;
        add_backtable(id_chtab_6_environment, door_fram_top[(modifier >> 2) % 8], draw_xh, 0, draw_bottom_y, blitters_10h_transp, 0);
    }
}

const byte blueline_fram1[] = {0, 124, 125, 126};
const sbyte blueline_fram_y[] = {0, -20, -20, 0};
const byte blueline_fram3[] = {44, 44, 45, 45};
const byte doortop_fram_bot[] = {78, 80, 82, 0};

// seg008:066A
void draw_tile_right(void) {
    byte id;
    byte tile_modifier;
    if (curr_tile == tiles_20_wall) return;
    switch (tile_left) {
        default:
            id = tile_table[tile_left].right_id;
            // custom logic - balcony tile with modifier of 1 or 2 should be replaced with a huge poison potion
            // 1 = spilled, 2 = upright
            if (tile_left == tiles_23_balcony_left && modifier_left > 0) {
                id += 4; // transparent sprite for poison potion
                if (modifier_left == 2) { // upright
                    id += modifier_left;
                }
            } else if (tile_left == tiles_24_balcony_right) {
                // Read the modifier of the tile two places to the left.
                // It's not prefetched in advance as in draw_tile_topright().
                byte tile_left_by_2;
                byte modifier_left_by_2;
                get_tile_to_draw(drawn_room, drawn_col - 2, drawn_row, &tile_left_by_2, &modifier_left_by_2, tiles_20_wall);
                if (modifier_left_by_2 > 0) {
                    id += 4; // transparent sprite for poison potion
                }
            } else if (tile_left == tiles_9_bigpillar_top) {
                tile_modifier = modifier_left;
                if (tile_modifier == 1) {
                    add_backtable(id_chtab_6_environment, blueline_fram3[tile_modifier], draw_xh, 0, draw_main_y - 20, blitters_0_no_transp, 0);
                }
            }
            if (id) {
                if (tile_left == tiles_5_stuck) {
                    if (curr_tile == tiles_0_empty || curr_tile == tiles_5_stuck || !tile_is_floor(curr_tile)) {
                        id = 42; /* floor B (right side) */
                    }
                    // custom logic - vertical spikes in the temple have regular floors
                    // Setting the lower control byte to 1 changes the floor to debris.
                } else if (tile_left == tiles_2_spike && is_temple()) {
                    byte control_bits = get_tile_ctrl_bits(drawn_room, get_tilepos(drawn_col - 1, drawn_row), 3);
                    if (control_bits == 1) {
                        id = 98; /* debris (right side) */
                    } else if (control_bits == 2) {
                        id = 0; /* no floor */
                    }
                }
                add_backtable(id_chtab_6_environment, id, draw_xh, 0, tile_table[tile_left].right_y + draw_main_y, blitters_10h_transp, 0);
            }
            if (settings->tbl_level_type[current_level] != DUNGEON_LEVEL) {
                add_backtable(id_chtab_6_environment, tile_table[tile_left].stripe_id, draw_xh, 0, draw_main_y - 27, blitters_10h_transp, 0);
            }
            if (tile_left == tiles_19_torch || tile_left == tiles_30_torch_with_debris) {
                add_backtable(id_chtab_6_environment, 146 /*torch base*/, draw_xh, 0, draw_bottom_y - 28, blitters_0_no_transp, 0);
            }
            break;
        case tiles_0_empty:
            if (modifier_left > 3) return;
            add_backtable(id_chtab_6_environment, blueline_fram1[modifier_left], draw_xh, 0, blueline_fram_y[modifier_left] + draw_main_y, blitters_10h_transp, 0);
            break;
        case tiles_1_floor:
            ptr_add_table(id_chtab_6_environment, 42 /*floor B*/, draw_xh, 0, tile_table[tile_left].right_y + draw_main_y, blitters_10h_transp, 0);
            tile_modifier = modifier_left;
            if (tile_modifier > 3) tile_modifier = 0;
            if (tile_modifier == !!settings->tbl_level_type[current_level]) return;
            add_backtable(id_chtab_6_environment, blueline_fram3[tile_modifier], draw_xh, 0, draw_main_y - 20, blitters_0_no_transp, 0);
            break;
        case tiles_13_mirror: // custom logic - draw wall stripe for the tile after the mirror when the modifier is 1
            ptr_add_table(id_chtab_6_environment, 42 /*floor B*/, draw_xh, 0, tile_table[tile_left].right_y + draw_main_y, blitters_10h_transp, 0);
            tile_modifier = modifier_left;
            if (tile_modifier == 1) {
                add_backtable(id_chtab_6_environment, blueline_fram3[tile_modifier], draw_xh, 0, draw_main_y - 20, blitters_0_no_transp, 0);
            }
            break;
        case tiles_7_doortop_with_floor:
        case tiles_12_doortop:
            if (settings->tbl_level_type[current_level] == DUNGEON_LEVEL) return;
            tile_modifier = modifier_left;
            // custom logic - on some levels draw the wall stripe for doortops without it
            if (tile_modifier > 10) {
                tile_modifier -= 10;
                add_backtable(id_chtab_6_environment, 84 /*wall stripe*/, draw_xh + 3, 0, draw_main_y - 27, blitters_0_no_transp, 0);
            }
            add_backtable(id_chtab_6_environment, doortop_fram_bot[tile_modifier], draw_xh, 0, tile_table[tile_left].right_y + draw_main_y, blitters_10h_transp, 0);
            break;
        case tiles_20_wall:
            if (settings->tbl_level_type[current_level] && (modifier_left & 0x80) == 0) {
                add_backtable(id_chtab_6_environment, 84 /*wall stripe*/, draw_xh + 3, 0, draw_main_y - 27, blitters_0_no_transp, 0);
            }
            short wall_id = 1; /* wall (the right side) */
            // custom logic - walls next to spikes in the temple have holes in them
            if (curr_tile == tiles_2_spike && is_temple()) wall_id = 18; /* wall with vertical spike holes */
            add_backtable(id_chtab_7_environmentwall, wall_id, draw_xh, 0, tile_table[tile_left].right_y + draw_main_y, blitters_10h_transp, 0);
            break;
        // custom logic - tile 26 is a torch without a floor in dungeons
        case tiles_26_torch_empty:
            if (settings->tbl_level_type[current_level] == DUNGEON_LEVEL) {
                add_backtable(id_chtab_6_environment, 146 /*torch base*/, draw_xh, 0, draw_bottom_y - 28, blitters_0_no_transp, 0);
            }
            break;
    }
}

const sbyte spikes_delta_y[] = {0, -7, -3, 0, 0, -1, 0, -3, -7, 0};
const byte spikes_fram_right[] = {0, 134, 135, 136, 137, 138, 137, 135, 134, 0};
const byte loose_fram_right[] = {42, 71, 42, 72, 72, 42, 42, 42, 72, 72, 72, 0};

// seg008:08A0
int get_spike_frame(byte modifier) {
    if (modifier & 0x80) {
        return 5;
    } else {
        return modifier;
    }
}

// seg008:08B5
void draw_tile_anim_right(void) {
    int* torch_colors = enable_lighting ? settings->torch_level_colors_dimmed : settings->torch_level_colors;
    switch (tile_left) {
        case tiles_2_spike: {
            // custom logic - spikes in the temple retract vertically
            short spike_frame = get_spike_frame(modifier_left);
            if (is_temple()) {
                add_backtable(id_chtab_6_environment, spikes_fram_right[spike_frame], draw_xh - 3, 5, draw_bottom_y - 22 + spikes_delta_y[spike_frame], blitters_10h_transp, 0);
            } else {
                add_backtable(id_chtab_6_environment, spikes_fram_right[spike_frame], draw_xh, 0, draw_main_y - 7, blitters_10h_transp, 0);
            }
            break;
        }
        case tiles_4_gate:
            draw_gate_back();
            break;
        case tiles_11_loose:
            add_backtable(id_chtab_6_environment, loose_fram_right[get_loose_frame(modifier_left)], draw_xh, 0, draw_bottom_y - 1, blitters_10h_transp, 0);
            break;
        case tiles_16_level_door_left:
            draw_leveldoor();
            break;
        case tiles_19_torch:
        case tiles_30_torch_with_debris:
        // custom logic - tile 26 is a torch without a floor in dungeons
        case tiles_26_torch_empty:
            if (settings->tbl_level_type[current_level] != DUNGEON_LEVEL && tile_left == tiles_26_torch_empty) {
                break;
            }
            if (modifier_left < 9) {
                int blit = blitters_0_no_transp;
                // custom logic - add colored torches
                int color = torch_colors[current_level];
                if (current_level == 6 &&
                        (level_6_sequence_timer != 0 ||
                         (falling_exit_checkpoint >= 4 && leveldoor_open == 1))
                   ) {
                    color = torch_colors[current_level - 1];
                }
                // custom logic - special torch colors on level 8, room 19
                if (current_level == 8 && drawn_room == 19 && drawn_row == 0) {
                    if (drawn_col == 7) {
                        color = 2; // blue
                    } else if (drawn_col == 9) {
                        color = 32; // red
                    }
                }
                // custom logic - do not dim lights on level 13, room 23
                if (current_level == 14 && drawn_room == 23) {
                    color = 11; // argon blue
                }
                if (color > 0) {
                    blit = blitters_colored_flame + (color & 0x3F);
                }
                // images 1..9 are the flames
                add_backtable(id_chtab_1_flameswordpotion, modifier_left + 1, draw_xh + 1, 0, draw_main_y - 40, blit, 0);
            }
            break;
        case tiles_22_sword:
            // custom logic - sword on level 13 is flaming
            if (current_level == settings->flame_level &&
                    drawn_room == settings->flaming_sword_room &&
                    get_tilepos(drawn_col - 1, drawn_row) == settings->flaming_sword_tilepos &&
                    modifier_left < 9) {
                // clear the old animation (needed because the flame is drawn in the foretable)
                redraw_height = 35;
                set_redraw_full(settings->flaming_sword_tilepos - 1, 1);
                set_wipe(settings->flaming_sword_tilepos - 1, 1);
                set_redraw_full(settings->flaming_sword_tilepos, 1);
                set_wipe(settings->flaming_sword_tilepos, 1);
                // images 1..9 are the flames
                add_foretable(id_chtab_1_flameswordpotion, modifier_left + 1, draw_xh - 5, 1, draw_main_y - 7, blitters_colored_flaming_sword, 0);
            }
            break;
    }
}

const byte wall_fram_bottom[] = {7, 9, 5, 3};

// seg008:0971
void draw_tile_bottom(word redraw) {
    word chtab_id;
    byte id;
    byte blit;
    id = 0;
    blit = blitters_0_no_transp;
    chtab_id = id_chtab_6_environment;
    switch (curr_tile) {
        case tiles_20_wall:
            if (settings->tbl_level_type[current_level] == DUNGEON_LEVEL || settings->tbl_level_wda[current_level] == 1) {
                id = wall_fram_bottom[curr_modifier & 0x7F];
            }
            chtab_id = id_chtab_7_environmentwall;
            break;
        // custom logic - spikes in the temple with control bits set to 010 do not have a floor
        case tiles_2_spike:
            if (is_temple() && get_tile_ctrl_bits(drawn_room, get_tilepos(drawn_col, drawn_row), 3) == 2) {
                id = 0; /* no floor */
                break;
            }
        // fallthrough!
        case tiles_12_doortop:
            blit = blitters_10h_transp;
        // fallthrough!
        default:
            id = tile_table[curr_tile].bottom_id;
            break;
    }
    if (ptr_add_table(chtab_id, id, draw_xh, 0, draw_bottom_y, blit, 0) && redraw) {
        add_foretable(chtab_id, id, draw_xh, 0, draw_bottom_y, blit, 0);
    }
    if (chtab_id == id_chtab_7_environmentwall) {
        wall_pattern(0, 0);
    }
}

const byte loose_fram_bottom[] = {43, 73, 43, 74, 74, 43, 43, 43, 74, 74, 74, 0};

// seg008:0A38
void draw_loose(void) {
    word id;
    if (curr_tile == tiles_11_loose) {
        id = loose_fram_bottom[get_loose_frame(curr_modifier)];
        add_backtable(id_chtab_6_environment, id, draw_xh, 0, draw_bottom_y, blitters_0_no_transp, 0);
        add_foretable(id_chtab_6_environment, id, draw_xh, 0, draw_bottom_y, blitters_0_no_transp, 0);
    }
}

const byte loose_fram_left[] = {41, 69, 41, 70, 70, 41, 41, 41, 70, 70, 70, 0};

// seg008:0A8E
void draw_tile_base(void) {
    word ybottom;
    word id;
    ybottom = draw_main_y;
    // custom logic - lattice tiles are drawn in the draw_tile_fore() method
    if (settings->tbl_level_type[current_level] == PALACE_LEVEL) {
        if ((curr_tile >= tiles_26_lattice_down && curr_tile <= tiles_29_lattice_right) ||
                (tile_left == tiles_26_lattice_down && curr_tile == tiles_12_doortop)) {
            return;
        }
    }
    if (curr_tile == tiles_11_loose) {
        id = loose_fram_left[get_loose_frame(curr_modifier)];
    } else if (curr_tile == tiles_15_opener && tile_left == tiles_0_empty && settings->tbl_level_type[current_level] == DUNGEON_LEVEL) {
        id = 148; /* left half of open button with no floor to the left */
        // custom logic - vertical spikes in the temple have regular floors
        // Setting the lower control byte to 1 changes the floor to debris.
    } else if (curr_tile == tiles_2_spike && is_temple()) {
        byte control_bits = get_tile_ctrl_bits(drawn_room, get_tilepos(drawn_col, drawn_row), 3);
        if (control_bits == 1) {
            id = 97; /* debris (left side) */
        } else if (control_bits == 2) {
            id = 0; /* no floor */
        } else {
            id = tile_table[curr_tile].base_id;
        }
    } else {
        id = tile_table[curr_tile].base_id;
    }
    ptr_add_table(id_chtab_6_environment, id, draw_xh, 0, tile_table[curr_tile].base_y + ybottom, blitters_10h_transp, 0);
}

const byte spikes_fram_left[] = {0, 128, 129, 130, 131, 132, 131, 129, 128, 0};
const byte potion_fram_bubb[] = {0, 16, 17, 18, 19, 20, 21, 22};
const byte chomper_fram1[] = {3, 2, 0, 1, 4, 3, 3, 0};
const byte chomper_fram_bot[] = {101, 102, 103, 104, 105, 0};
const byte chomper_fram_top[] = {0, 0, 111, 112, 113, 0};
const byte chomper_fram_y[] = {0, 0, 0x25, 0x2F, 0x32};

// custom logic - alternating potion color flag
byte alt_color = 0;

// seg008:0B2B
void draw_tile_anim(void) {
    short color;
    word potion_type;
    word pot_size;
    word chomper_frame;
    pot_size = 0;
    // custom logic - the default color is red except for 12B where red potions are blue
    color = current_level == settings->inverse_level ? color_9_brightblue : color_12_brightred;
    switch (curr_tile) {
        case tiles_2_spike: {
            // custom logic - spikes in the temple retract vertically
            short spike_frame = get_spike_frame(curr_modifier);
            if (is_temple()) {
                ptr_add_table(id_chtab_6_environment, spikes_fram_left[spike_frame], draw_xh + 1, 0, draw_bottom_y - 22 + spikes_delta_y[spike_frame], blitters_10h_transp, 0);
            } else {
                ptr_add_table(id_chtab_6_environment, spikes_fram_left[spike_frame], draw_xh, 0, draw_main_y - 2, blitters_10h_transp, 0);
            }
            break;
        }
        case tiles_10_potion:
            potion_type = (curr_modifier & 0xF8) >> 3;
            // custom logic - mouse potion
            if (potion_type == potion_0_custom &&
                    current_level == settings->mouse_2_level &&
                    drawn_room == settings->mouse_2_room &&
                    drawn_col == settings->mouse_2_potion_column &&
                    drawn_row == settings->mouse_2_potion_row) {
                color = color_15_brightwhite;
                pot_size = 1;
                if (mouse_potion_timer < 20) {
                    ++mouse_potion_timer;
                } else {
                    --pot_size; // bubble animation is lower
                }
            } else {
                switch (potion_type) {
                    case potion_0_custom:
                        color = settings->tbl_empty_potion_color[current_level];
                        // custom logic - magenta color alternates with green and yellow
                        if (color == color_13_brightmagenta) {
                            if (alt_color) {
                                color = prandom(1) ? color_10_brightgreen : color_14_brightyellow;
                            }
                            alt_color = 1 - alt_color;
                        }
                        if (color > -1) {
                            pot_size = settings->tbl_empty_potion_size[current_level];
                            break;
                        }
                        return;
                    // custom logic - empty potion after kid drinks them
                    case potion_7_empty_small:
                    case potion_8_empty_small:
                        return;
                    // custom logic - dropped empty potion
                    case potion_11_spilled_left_large:
                    case potion_12_spilled_right_large:
                        pot_size = 1; // large empty potions
                    // fallthrough!
                    case potion_9_spilled_left_small:
                    case potion_10_spilled_right_small:
                        // custom logic - determine spilled potion color
                        switch (curr_modifier & 0x07) {
                            case 0:
                                color = settings->tbl_empty_potion_color[current_level];
                                if (color <= -1) {
                                    return;
                                }
                                break;
                            case 3:
                            case 4:
                                color = color_10_brightgreen;
                                break;
                            case 5:
                            case 6:
                                color = current_level == settings->inverse_level ? color_12_brightred : color_9_brightblue;
                                break;
                        }
                        break;
                    case potion_5_hurt:
                    case potion_6_open:
                        color = current_level == settings->inverse_level ? color_12_brightred : color_9_brightblue;
                        break;
                    case potion_3_feather:
                    case potion_4_invert:
                        color = color_10_brightgreen;
                    // fallthrough!
                    case potion_2_life:
                        pot_size = 1;
                        break;
                }
            }
            // custom logic - potion spilled to the left
            if (potion_type == potion_9_spilled_left_small || potion_type == potion_11_spilled_left_large) {
                add_backtable(id_chtab_1_flameswordpotion, 28, draw_xh + 2, 1, draw_main_y - 3, color + blitters_40h_mono, 0);
                // custom logic - potion spilled to the right
            } else if (potion_type == potion_10_spilled_right_small || potion_type == potion_12_spilled_right_large) {
                add_backtable(id_chtab_1_flameswordpotion, 29, draw_xh + 3, 1 + (pot_size << 2), draw_main_y - 3, color + blitters_40h_mono, 0);
                // upright potion
            } else {
                add_backtable(id_chtab_1_flameswordpotion, 23 /*bubble mask*/, draw_xh + 3, 1, draw_main_y - (pot_size << 2) - 14, blitters_40h_mono, 0);
                add_foretable(id_chtab_1_flameswordpotion, potion_fram_bubb[curr_modifier & 0x7], draw_xh + 3, 1, draw_main_y - (pot_size << 2) - 14, color + blitters_40h_mono, 0);
            }
            break;
        case tiles_22_sword:
            // custom logic - do not draw regular sword in the flaming sword room
            if (current_level == settings->flame_level && drawn_room == settings->flaming_sword_room && get_tilepos(drawn_col, drawn_row) == settings->flaming_sword_tilepos) {
                break;
            }
            // custom logic - draw the sword on level 6B in the background to avoid clashing with characters
            if (current_level == 6 && drawn_room == 17 && drawn_col == 3 && drawn_row == 2) {
                add_backtable(id_chtab_15_pop2_sword, (curr_modifier == 1) + 1, draw_xh, 0, draw_main_y - 8, blitters_10h_transp, curr_modifier == 1);
            } else {
                // custom logic - shift the sword to the right during the initial level 6 sequence
                byte seq_delta = current_level == 6 && drawn_room == 24 && drawn_col == 4 && drawn_row == 2 ? 1 : 0;
                add_midtable(id_chtab_15_pop2_sword, (curr_modifier == 1) + 1, draw_xh + seq_delta, 0, draw_main_y - 3, blitters_10h_transp, curr_modifier == 1);
            }
            break;
        case tiles_18_chomper:
            chomper_frame = chomper_fram1[MIN(curr_modifier & 0x7F, 6)];
            add_backtable(id_chtab_6_environment, chomper_fram_bot[chomper_frame], draw_xh, 0, draw_main_y, blitters_10h_transp, 0);
            if (curr_modifier & 0x80) { // blood
                add_backtable(id_chtab_6_environment, chomper_frame + 114, draw_xh + 1, 4, draw_main_y - 6, blitters_4Ch_mono_12, 0);
            }
            add_backtable(id_chtab_6_environment, chomper_fram_top[chomper_frame], draw_xh, 0, draw_main_y - chomper_fram_y[chomper_frame], blitters_10h_transp, 0);
            break;
    }
}

const byte spikes_fram_fore[] = {0, 139, 140, 141, 142, 143, 142, 140, 139, 0};
const byte chomper_fram_for[] = {106, 107, 108, 109, 110, 0};
const byte wall_fram_main[] = {8, 10, 6, 4};

// seg008:0D15
void draw_tile_fore(void) {
    word ybottom;
    byte xh;
    word potion_type;
    word id;
    word chomper_frame;
    if (tile_left == tiles_4_gate &&
            (modifier_left != 0 || !can_guard_see_kid || ABS(Guard.curr_col - Kid.curr_col) > 1) && // prevents drawing gate over a standing guard
            Kid.curr_row == drawn_row &&
            Kid.curr_col == drawn_col - 1 &&
            Kid.room != room_R) {
        draw_gate_fore();
    }
    switch (curr_tile) {
        case tiles_2_spike: {
            // custom logic - spikes in the temple retract vertically
            short spike_frame = get_spike_frame(curr_modifier);
            if (is_temple()) {
                add_foretable(id_chtab_6_environment, spikes_fram_fore[spike_frame], draw_xh + 1, 0, draw_bottom_y - 22 + spikes_delta_y[spike_frame], blitters_10h_transp, 0);
            } else {
                add_foretable(id_chtab_6_environment, spikes_fram_fore[spike_frame], draw_xh, 0, draw_main_y - 2, blitters_10h_transp, 0);
            }
            break;
        }
        case tiles_18_chomper:
            chomper_frame = chomper_fram1[MIN(curr_modifier & 0x7F, 6)];
            add_foretable(id_chtab_6_environment, chomper_fram_for[chomper_frame], draw_xh, 0, draw_main_y, blitters_10h_transp, 0);
            if (curr_modifier & 0x80) {
                add_foretable(id_chtab_6_environment, chomper_frame + 119, draw_xh + 1, 4, draw_main_y - 6, blitters_4Ch_mono_12, 0);
            }
            break;
        case tiles_20_wall:
            if (settings->tbl_level_type[current_level] == DUNGEON_LEVEL || settings->tbl_level_wda[current_level] == 1) {
                add_foretable(id_chtab_7_environmentwall, wall_fram_main[curr_modifier & 0x7F], draw_xh, 0, draw_main_y, blitters_0_no_transp, 0);
            }
            wall_pattern(1, 1);
            break;
        // custom logic - draw front parts of lattice in the foretable
        // so prince is not in front of it while jumping
        case tiles_26_lattice_down:
        case tiles_27_lattice_small:
        case tiles_28_lattice_left:
        case tiles_29_lattice_right:
            if (settings->tbl_level_type[current_level] == PALACE_LEVEL) {
                add_foretable(id_chtab_6_environment, tile_table[curr_tile].base_id, draw_xh, 0, tile_table[curr_tile].base_y + draw_main_y, blitters_10h_transp, 0);
            }
        // fallthrough!
        default:
            if (settings->tbl_level_type[current_level] == PALACE_LEVEL && tile_left == tiles_26_lattice_down && curr_tile == tiles_12_doortop) {
                add_foretable(id_chtab_6_environment, 6, draw_xh, 0, tile_table[curr_tile].base_y + draw_main_y + 3, blitters_10h_transp, 0);
            }
            id = tile_table[curr_tile].fore_id;
            if (id == 0) return;
            xh = tile_table[curr_tile].fore_x + draw_xh;
            bool mouse_potion = current_level == settings->mouse_2_level &&
                                drawn_room == settings->mouse_2_room &&
                                drawn_col == settings->mouse_2_potion_column &&
                                drawn_row == settings->mouse_2_potion_row;
            if (curr_tile == tiles_10_potion) {
                // large pots are drawn for potion types 2, 3, 4
                potion_type = (curr_modifier & 0xF8) >> 3;
                // custom logic - mouse potion
                if (mouse_potion) {
                    id = 13;
                    // custom logic - "empty" non-standard potions
                } else if (potion_type == potion_0_custom && settings->tbl_empty_potion_size[current_level]) {
                    id = 13;
                } else if (potion_type >= potion_2_life && potion_type < potion_5_hurt) {
                    id = 13; // small pot = 12, large pot = 13
                    // custom logic - empty potion type 8 is in a large pot
                } else if (potion_type == potion_8_empty_small) {
                    id = 13;
                    // custom logic - small dropped potion turned left
                } else if (potion_type == potion_9_spilled_left_small) {
                    id = 24;
                    // custom logic - small dropped potion turned right
                } else if (potion_type == potion_10_spilled_right_small) {
                    id = 26;
                    --xh;
                    // custom logic - large dropped potion turned left
                } else if (potion_type == potion_11_spilled_left_large) {
                    id = 25;
                    // custom logic - large dropped potion turned right
                } else if (potion_type == potion_12_spilled_right_large) {
                    id = 27;
                    --xh;
                }
            }
            ybottom = tile_table[curr_tile].fore_y + draw_main_y;
            if (curr_tile == tiles_10_potion) {
                // potions look different in the dungeon and the palace
                if (settings->tbl_level_type[current_level] != DUNGEON_LEVEL && potion_type < potion_9_spilled_left_small) id += 2;
                // custom logic - change mouse potion to the dropped frame after prince drinks it
                if (mouse_potion && leveldoor_open > 0) {
                    id += 12;
                    xh -= 2;
                }
                add_foretable(id_chtab_1_flameswordpotion, id, xh, 6, ybottom, blitters_10h_transp, 0);
            } else {
                if (curr_tile >= tiles_27_lattice_small && curr_tile < tiles_30_torch_with_debris) {
                    add_foretable(id_chtab_6_environment, id, xh, 0, ybottom, blitters_0_no_transp, 0);
                    // custom logic - pillar front in the dungeon 2 levels is 4 pixels wider and needs to be centered
                } else if (curr_tile == tiles_3_pillar && is_dungeon_2()) {
                    add_foretable(id_chtab_6_environment, id, xh, -2, ybottom, blitters_10h_transp, 0);
                } else {
                    add_foretable(id_chtab_6_environment, id, xh, 0, ybottom, blitters_10h_transp, 0);
                }
            }
            break;
    }
}

// seg008:0FF6
int get_loose_frame(byte modifier) {
    // Don't display garbled tiles if the delay is greater than the default.
    if ((modifier & 0x80) || settings->loose_floor_delay > 11) {
        modifier &= 0x7F;
        if (modifier > 10) {
            return 1;
        }
    }
    return modifier;
}

// Get an image, with index and NULL checks.
image_type* get_image(short chtab_id, int id, int color) {
    if (chtab_id < 0 || (size_t) chtab_id > COUNT(chtab_addrs)) {
        printf("Tried to use chtab %d not in 0..%d\n", chtab_id, (int)COUNT(chtab_addrs));
        return NULL;
    }
    chtab_type* chtab;
    // custom logic - palette is getting changed for every guard instead of once per room
    if (chtab_id == 5 && color != last_guard_color && color > 0 && color <= GUARD_COLORS) {
        set_chtab_palette(chtab_addrs[chtab_id], &guard_palettes[0x30 * color - 0x30], 0x10);
        last_guard_color = color;
    }
    chtab = chtab_addrs[chtab_id];
    if (chtab == NULL) {
        return NULL;
    }
    if (id < 0 || id >= chtab->n_images) {
        if (id != 255) printf("Tried to use image %d of chtab %d, not in 0..%d\n", id, chtab_id, chtab->n_images - 1);
        return NULL;
    }
    return chtab->images[id];
}

// seg008:10A8
int add_backtable(short chtab_id, int id, sbyte xh, sbyte xl, int ybottom, int blit, byte peel) {
    word index;
    (void) peel; // gets rid of the parameter not used warning - the parameter is needed for the pointer compatibility
    if (id == 0) {
        return 0;
    }
    index = backtable_count;
    if (index >= 200) {
        show_dialog("BackTable Overflow", default_dialog);
        return 0; // added
    }
    back_table_type* backtable_item = &backtable[index];
    backtable_item->xh = xh;
    backtable_item->xl = xl;
    backtable_item->chtab_id = chtab_id;
    backtable_item->id = id - 1;
    image_type* image = get_image(chtab_id, id - 1, 0);
    if (image == NULL) {
        return 0;
    }
    backtable_item->y = ybottom - image->h/*height*/ + 1;
    backtable_item->blit = blit;
    ++backtable_count;
    return 1;
}

// seg008:1017
int add_foretable(short chtab_id, int id, sbyte xh, sbyte xl, int ybottom, int blit, byte peel) {
    word index;
    (void) peel; // gets rid of the parameter not used warning - the parameter is needed for the pointer compatibility
    if (id == 0) {
        return 0;
    }
    index = foretable_count;
    if (index >= 200) {
        show_dialog("ForeTable Overflow", default_dialog);
        return 0; // added
    }
    back_table_type* foretable_item = &foretable[index];
    foretable_item->xh = xh;
    foretable_item->xl = xl;
    foretable_item->chtab_id = chtab_id;
    foretable_item->id = id - 1;
    image_type* image = get_image(chtab_id, id - 1, 0);
    if (image == NULL) {
        return 0;
    }
    foretable_item->y = ybottom - image->h/*height*/ + 1;
    foretable_item->blit = blit;
    ++foretable_count;
    return 1;
}

// seg008:113A
int add_midtable(short chtab_id, int id, sbyte xh, sbyte xl, int ybottom, int blit, byte peel) {
    word index;
    if (id == 0) {
        return 0;
    }
    index = midtable_count;
    if (index >= 50) {
        show_dialog("MidTable Overflow", default_dialog);
        return 0; // added
    }
    midtable_type* midtable_item = &midtable[index];
    midtable_item->xh = xh;
    midtable_item->xl = xl;
    midtable_item->chtab_id = chtab_id;
    midtable_item->id = id - 1;
    image_type* image = get_image(chtab_id, id - 1, 0);
    if (image == NULL) {
        return 0;
    }
    midtable_item->y = ybottom - image->h/*height*/ + 1;
    if (obj_direction == dir_0_right && chtab_flip_clip[chtab_id] != 0) {
        blit += 0x80;
    }
    midtable_item->blit = blit;
    midtable_item->peel = peel;
    midtable_item->clip.left = obj_clip_left;
    midtable_item->clip.right = obj_clip_right;
    midtable_item->clip.top = obj_clip_top;
    midtable_item->clip.bottom = obj_clip_bottom;
    // custom logic - storing guard's color in the midtable
    midtable_item->color = obj_color;
    ++midtable_count;
    return 1;
}

// seg008:1208
void add_peel(int left, int right, int top, int height) {
    rect_type rect;
    if (peels_count >= MAX_PEELS) {
        show_dialog("Peels OverFlow", default_dialog);
        return; // added
    }
    rect.left = left;
    rect.right = right;
    rect.top = top;
    rect.bottom = top + height;
    peels_table[peels_count++] = read_peel_from_screen(&rect);
}

// seg008:1254
void add_wipetable(sbyte layer, short left, short bottom, sbyte height, short width, sbyte color) {
    word index;
    index = wipetable_count;
    if (index >= 300) {
        show_dialog("WipeTable Overflow", default_dialog);
        return; // added
    }
    wipetable_type* wipetable_item = &wipetable[index];
    wipetable_item->left = left;
    wipetable_item->bottom = bottom + 1;
    wipetable_item->height = height;
    wipetable_item->width = width;
    wipetable_item->color = color;
    wipetable_item->layer = layer;
    ++wipetable_count;
}

// seg008:12BB
void draw_table(int which_table) {
    short index;
    short count;
    count = table_counts[which_table];
    for (index = 0; index < count; ++index) {
        if (which_table == 3) {
            draw_mid(index);
        } else {
            draw_back_fore(which_table, index);
        }
    }
}

// seg008:12FE
void draw_wipes(int which) {
    word index;
    word count;
    count = wipetable_count;
    for (index = 0; index < count; ++index) {
        if (which == wipetable[index].layer) {
            draw_wipe(index);
        }
    }
}

// seg008:133B
void draw_back_fore(int which_table, int index) {
    image_type* image;
    image_type* mask;
    back_table_type* table_entry;
    if (which_table == 0) {
        table_entry = &backtable[index];
    } else {
        table_entry = &foretable[index];
    }
    image = mask = get_image(table_entry->chtab_id, table_entry->id, 0);
    draw_image(image, mask, table_entry->xh * 8 + table_entry->xl, table_entry->y, table_entry->blit);
}

SDL_Surface* hflip(SDL_Surface* input) {
    int width = input->w;
    int height = input->h;
    int source_x, target_x;

    // The simplest way to create a surface with same format as input:
    SDL_Surface* output = SDL_ConvertSurface(input, input->format, 0);
    SDL_SetSurfacePalette(output, input->format->palette);
    // The copied image will be overwritten anyway.
    if (output == NULL) {
        sdlperror("SDL_ConvertSurface");
        quit(1);
    }

    SDL_SetSurfaceBlendMode(input, SDL_BLENDMODE_NONE);
    // Temporarily turn off alpha and colorkey on input. So we overwrite the output image.
    SDL_SetColorKey(input, SDL_FALSE, 0);
    SDL_SetColorKey(output, SDL_FALSE, 0);
    SDL_SetSurfaceAlphaMod(input, 255);

    for (source_x = 0, target_x = width - 1; source_x < width; ++source_x, --target_x) {
        SDL_Rect srcrect = {source_x, 0, 1, height};
        SDL_Rect dstrect = {target_x, 0, 1, height};
        if (SDL_BlitSurface(input/*32*/, &srcrect, output, &dstrect) != 0) {
            sdlperror("SDL_BlitSurface");
            quit(1);
        }
    }

    return output;
}

// seg008:140C
void draw_mid(int index) {
    word need_free_mask;
    word image_id;
    image_type* mask;
    word chtab_id;
    word blit_flip;
    short ypos;
    short xpos;
    midtable_type* midtable_entry;
    word blit;
    word need_free_image;
    image_type* image;

    blit_flip = 0;
    need_free_image = 0;
    need_free_mask = 0;
    midtable_entry = &midtable[index];
    image_id = midtable_entry->id;
    chtab_id = midtable_entry->chtab_id;
    image = mask = get_image(chtab_id, image_id, midtable_entry->color);
    xpos = midtable_entry->xh * 8 + midtable_entry->xl;
    ypos = midtable_entry->y;
    blit = midtable_entry->blit;
    if (blit & 0x80) {
        blit_flip = 0x8000;
        blit &= 0x7F;
    }

    if (chtab_flip_clip[chtab_id]) {
        set_clip_rect(&midtable_entry->clip);
        if (chtab_screen_x[chtab_id]) {
            xpos = calc_screen_x_coord(xpos);
        }
    }
    if (blit_flip) {
        xpos -= image->w/*width*/;
        // for this version:
        need_free_image = 1;
        image = hflip(image);
    }

    if (midtable_entry->peel) {
        add_peel(round_xpos_to_byte(xpos), round_xpos_to_byte(image->w/*width*/ + xpos), ypos, image->h/*height*/);
    }

    draw_image(image, mask, xpos, ypos, blit);

    if (chtab_flip_clip[chtab_id]) {
        reset_clip_rect();
    }
    if (need_free_image) {
        //free(image);
        SDL_FreeSurface(image);
    }
    if (need_free_mask) {
        free(mask);
    }
}

// seg008:167B
void draw_image(image_type *image, image_type *mask, int xpos, int ypos, int blit) {
    rect_type rect;
    switch (blit) {
        case blitters_9_black:
            method_6_blit_img_to_scr(mask, xpos, ypos, blitters_9_black);
            break;
        case blitters_0_no_transp:
        case blitters_3_xor:
        case blitters_10h_transp:
            method_6_blit_img_to_scr(image, xpos, ypos, blit);
            break;
        default:
            if (blit >= 0x100) {
                method_6_blit_img_to_scr(mask, xpos, ypos, blit);
            } else {
                method_3_blit_mono(image, xpos, ypos, 0, blit & 0xBF);
            }
            break;
    }
    if (need_drects) {
        rect.left = rect.right = xpos;
        rect.right += image->w/*width*/;
        rect.top = rect.bottom = ypos;
        rect.bottom += image->h/*height*/;
        add_drect(&rect);
    }
}

// seg008:1730
void draw_wipe(int index) {
    rect_type rect;
    wipetable_type* ptr;
    ptr = &wipetable[index];
    rect.left = rect.right = ptr->left;
    rect.right += ptr->width;
    rect.bottom = rect.top = ptr->bottom;
    rect.top -= ptr->height;
    draw_rect(&rect, ptr->color);
    if (need_drects) {
        add_drect(&rect);
    }
}

// data:4E8C
word gate_top_y;
// data:4CB6
word gate_openness;
// data:436C
word gate_bottom_y;

// seg008:178E
void calc_gate_pos(void) {
    gate_top_y = draw_bottom_y - 62;
    gate_openness = (MIN(modifier_left, 188) >> 2) + 1;
    gate_bottom_y = draw_main_y - gate_openness;
}

// data:2785
const byte door_fram_slice[] = {67, 59, 58, 57, 56, 55, 54, 53, 52};
// seg008:17B7
void draw_gate_back(void) {
    short ybottom;
    word gate_bottom;
    calc_gate_pos();
    if (gate_bottom_y + 12 < draw_main_y) {
        add_backtable(id_chtab_6_environment, 50 /*gate bottom with B*/, draw_xh, 0, gate_bottom_y, blitters_0_no_transp, 0);
    } else {
        // custom logic - fix a drawing bug when a pillar or a gate is below a gate in palace
        // pillar with the 010 control byte value is exempt from redrawing
        if (curr_room_tiles[get_tilepos(drawn_col - 1 /* left tile */, drawn_row + 1 /* row below */)] != (tiles_3_pillar | 0x40)) {
            // The following line (erroneously) erases the top-right of the tile below-left (because it is drawn non-transparently).
            // -- But it draws something that was already drawn! (in draw_tile_right()).
            add_backtable(id_chtab_6_environment, tile_table[tiles_4_gate].right_id, draw_xh, 0,
                          tile_table[tiles_4_gate].right_y + draw_main_y, blitters_0_no_transp, 0);
            // And this line tries to fix it. But it fails if it was a gate or a pillar.
            if (can_see_bottomleft()) draw_tile_topright();
            // trying to fix gate under another gate (can cause another bug when a gate opens - do not stack gates)
            draw_tile_anim_topright(); // redraw the erased top-right section of the gate below-left
        }
        // The following 3 lines draw things that are drawn after this anyway.
        draw_tile_bottom(0);
        draw_loose();
        draw_tile_base();
        add_backtable(id_chtab_6_environment, 51 /*gate bottom*/, draw_xh, 0, gate_bottom_y - 2, blitters_10h_transp, 0);
    }
    ybottom = gate_bottom_y - 12;
    if (ybottom < 192) {
        for (; ybottom >= 0 && ybottom > 7 && ybottom - 7 > gate_top_y; ybottom -= 8) {
            add_backtable(id_chtab_6_environment, 52 /*gate slice 8px*/, draw_xh, 0, ybottom, blitters_0_no_transp, 0);
        }
    }
    gate_bottom = ybottom - gate_top_y + 1;
    if (gate_bottom > 0 && gate_bottom < 9) {
        add_backtable(id_chtab_6_environment, door_fram_slice[gate_bottom], draw_xh, 0, ybottom, blitters_0_no_transp, 0);
    }
}

// seg008:18BE
void draw_gate_fore(void) {
    short ybottom;
    calc_gate_pos();
    add_foretable(id_chtab_6_environment, 51 /*gate bottom*/, draw_xh, 0, gate_bottom_y - 2, blitters_10h_transp, 0);
    ybottom = gate_bottom_y - 12;
    if (ybottom < 192) {
        for (; ybottom >= 0 && ybottom > 7 && ybottom - 7 > gate_top_y; ybottom -= 8) {
            add_foretable(id_chtab_6_environment, 52 /*gate slice 8px*/, draw_xh, 0, ybottom, blitters_10h_transp, 0);
        }
    }
}

// seg008:1937
void alter_mods_allrm(void) {
    word tilepos;
    word room;
    // level.used_rooms is 25 on some levels. Limit it to the actual number of rooms.
    if (level.used_rooms > 24) level.used_rooms = 24;
    for (room = 1; room <= level.used_rooms; room++) {
        get_room_address(room);
        room_L = level.roomlinks[room - 1].left;
        room_R = level.roomlinks[room - 1].right;
        for (tilepos = 0; tilepos < MAX_TILES_PER_ROOM; tilepos++) {
            load_alter_mod(tilepos);
        }
    }
}

// seg008:198E
void load_alter_mod(int tilepos) {
    word wall_to_right;
    word tiletype;
    word wall_to_left;
    byte* curr_tile_modif;
    curr_tile_modif = tilepos + curr_room_modif;
    tiletype = curr_room_tiles[tilepos] & 0x1F;
    switch (tiletype) {
        case tiles_4_gate:
            if (*curr_tile_modif == 1) {
                *curr_tile_modif = 188;
            } else {
                *curr_tile_modif = 0;
            }
            break;
        case tiles_11_loose:
            *curr_tile_modif = 0;
            break;
        case tiles_10_potion:
            *curr_tile_modif <<= 3;
            break;
        case tiles_20_wall: {
            byte stored_modif = *curr_tile_modif;
            if (stored_modif == 1) *curr_tile_modif = 0x80;
            else *curr_tile_modif = (stored_modif << 4);
        }
            // fallthrough: not done yet, just moving to another scope
        goto label_wall_continued;

label_wall_continued: {
            // Here, ahead of time, we define the condition for a neighbouring tile to be seen as a wall connection
#define read_adj_tile_modif_in_curr_room()      // nop; access to modifier of adjacent tiles is not needed
#define read_adj_tile_modif_in_external_room()  // nop
#define WALL_CONNECTION_CONDITION (adj_tile == tiles_20_wall)
            wall_to_right = 1;
            wall_to_left = 1;
            int adj_tile_index, adj_tile;
            if (tilepos % 10 == 0) {
                if (room_L) {
                    adj_tile_index = 30 * (room_L - 1) + tilepos + 9;
                    adj_tile = (level.fg[adj_tile_index] & 0x1F);
                    read_adj_tile_modif_in_external_room();
                    wall_to_left = WALL_CONNECTION_CONDITION;
                }
            } else {
                adj_tile_index = tilepos - 1;
                adj_tile = (curr_room_tiles[adj_tile_index] & 0x1F);
                read_adj_tile_modif_in_curr_room();
                wall_to_left = WALL_CONNECTION_CONDITION;
            }
            if (tilepos % 10 == 9) {
                if (room_R) {
                    adj_tile_index = 30 * (room_R - 1) + tilepos - 9;
                    adj_tile = (level.fg[adj_tile_index] & 0x1F);
                    read_adj_tile_modif_in_external_room();
                    wall_to_right = WALL_CONNECTION_CONDITION;
                }
            } else {
                adj_tile_index = tilepos + 1;
                adj_tile = (curr_room_tiles[adj_tile_index] & 0x1F);
                read_adj_tile_modif_in_curr_room();
                wall_to_right = WALL_CONNECTION_CONDITION;
            }
            if (wall_to_left && wall_to_right) {
                *curr_tile_modif |= 3;
            } else if (wall_to_left) {
                *curr_tile_modif |= 2;
            } else if (wall_to_right) {
                *curr_tile_modif |= 1;
            }
        }
        break;
    }
}

// seg008:1AF8
void draw_moving(void) {
    draw_mobs();
    draw_people();
    redraw_needed_tiles();
}

// seg008:1B06
void redraw_needed_tiles(void) {
    word saved_drawn_room;
    load_leftroom();
    draw_objtable_items_at_tile(30);
    for (drawn_row = 3; drawn_row--; ) {
        load_rowbelow();
        draw_bottom_y = 63 * drawn_row + 65;
        draw_main_y = draw_bottom_y - 3;
        for (drawn_col = 0; drawn_col < 10; ++drawn_col) {
            load_curr_and_left_tile();
            redraw_needed(tbl_line[drawn_row] + drawn_col);
        }
    }
    saved_drawn_room = drawn_room;
    drawn_room = room_A;
    load_room_links();
    load_leftroom();
    drawn_row = 2;
    load_rowbelow();
    for (drawn_col = 0; drawn_col < 10; ++drawn_col) {
        // custom logic - check for fake ceiling on top row
        if (IS_FAKE_CEILING(current_level, saved_drawn_room) && Kid.curr_row == 0) {
            continue;
        }
        load_curr_and_left_tile();
        draw_main_y = -1;
        draw_bottom_y = 2;
        redraw_needed_above(drawn_col);
    }
    drawn_room = saved_drawn_room;
    load_room_links();
    draw_objtable_items_at_tile(-1);
}

// seg008:1BCB
void draw_tile_wipe(byte height) {
    add_wipetable(0, draw_xh * 8, draw_bottom_y, height, 4 * 8, 0);
}

// seg008:1BEB
void draw_tables(void) {
    drects_count = 0;
    current_target_surface = offscreen_surface;
    if (is_blind_mode) {
        draw_rect(&rect_top, color_0_black);
    }
    restore_peels();
    draw_wipes(0);
    draw_table(0); // backtable
    draw_custom_graphics_backtable();
    draw_table(3); // midtable
    draw_wipes(1);
    draw_table(1); // foretable
    draw_custom_graphics_foretable();
    // custom logic - prevent black rectangles when guard is starting to fall close to a wall
    if (Char.charid > charid_0_kid && Char.room == drawn_room && (Char.action == actions_3_in_midair || Char.action == actions_4_in_freefall)) {
        if (get_tile_at_char() == tiles_20_wall || get_tile_infrontof_char() == tiles_20_wall) {
            redraw_at_char();
            redraw_at_char2();
        }
    }
    current_target_surface = onscreen_surface_;
}

// seg008:1C4E
void restore_peels(void) {
    peel_type* peel;
    while (peels_count--) {
        peel = peels_table[peels_count];
        if (need_drects) {
            add_drect(&peel->rect); // ?
        }
        restore_peel(peel);
    }
    peels_count = 0;
}

// seg008:1C8F
void add_drect(rect_type *source) {
    rect_type* current_drect;
    short index;
    rect_type target_rect;
    for (index = 0; index < drects_count; ++index) {
        if (intersect_rect(&target_rect, shrink2_rect(&target_rect, source, -1, -1), &drects[index])) {
            current_drect = &drects[index];
            union_rect(current_drect, current_drect, source);
            return;
        }
    }
    if (drects_count >= MAX_DRECTS) {
        show_dialog("DRects Overflow", default_dialog);
        return; // added
    }
    drects[drects_count++] = *source;
}

// seg008:1D29
void draw_leveldoor(void) {
    short door_bottom, current_door_bottom, ybottom;
    ybottom = draw_main_y - 13;
    add_backtable(id_chtab_6_environment, 99 /*leveldoor stairs bottom*/, draw_xh + 1, 0, ybottom, blitters_0_no_transp, 0);
    if (modifier_left) {
        if (level.start_room != drawn_room) {
            short stairs_id = 144;
            // custom logic - in a dungeon 1 do not draw stairs window for the right door
            // when there are 2 level doors next to each other
            if (settings->tbl_level_type[current_level] == DUNGEON_LEVEL &&
                    settings->tbl_level_subtype[current_level] == DUNGEON_OR_PALACE_1_LEVEL &&
                    drawn_col >= 3 &&
                    (curr_room_tiles[get_tilepos(drawn_col - 3, drawn_row)] & 0x1F) == tiles_17_level_door_right) {
                stairs_id += 8;
            }
            add_backtable(id_chtab_6_environment, stairs_id, draw_xh + 1, 0, ybottom - 4, blitters_0_no_transp, 0);
        } else {
            short leveldoor_width = (settings->tbl_level_type[current_level] == DUNGEON_LEVEL) ? 39 : 48;
            sbyte x_low = (settings->tbl_level_type[current_level] == DUNGEON_LEVEL) ? 2 : 0; // dungeon level doors are shifted 2px to the right
            add_wipetable(0, 8 * (draw_xh + 1) + x_low, ybottom - 4, 45, leveldoor_width, 0);
        }
    }
    current_door_bottom = ybottom - (modifier_left & 3) - 48;
    do { // runs at least once
        door_bottom = ybottom - modifier_left;
        add_backtable(id_chtab_6_environment, 33 /*level door bottom*/, draw_xh + 1, 0, current_door_bottom, blitters_0_no_transp, 0);
        if (door_bottom > current_door_bottom) {
            current_door_bottom += 4;
        } else {
            break;
        }
    } while (true);
    add_backtable(id_chtab_6_environment, 34 /*level door top*/, draw_xh + 1, 0, draw_main_y - 64, blitters_0_no_transp, 0);
    // custom logic - set the leveldoor_ybottom to the open door if there are multiple doors
    if (leveldoor_ybottom == 0 || current_door_bottom < leveldoor_ybottom) {
        leveldoor_ybottom = current_door_bottom;
    }
}

byte placeholder_room_tiles[30] = {[0 ... 29] = 20};
byte placeholder_room_modif[30] = {0};

// seg008:1E0C
void get_room_address(int room) {
    if (room == PLACEHOLDER_ROOM) {
        curr_room_tiles = placeholder_room_tiles;
        curr_room_modif = placeholder_room_modif;
    } else if (room) {
        curr_room_tiles = &level.fg[(room - 1) * 30];
        curr_room_modif = &level.bg[(room - 1) * 30];
    }
}

// data:286A
const word floor_left_overlay[] = {32, 151, 151, 150, 150, 151, 32, 32};

// seg008:1E3A
void draw_floor_overlay(void) {
    if (tile_left != tiles_0_empty && (tile_left != tiles_26_torch_empty || settings->tbl_level_type[current_level] != DUNGEON_LEVEL)) {
        // Bug: When climbing up to a floor with a big pillar top behind, turned right, Kid sees through floor.
        // The bigpillar_top tile should be treated similarly to an empty tile here.
        if (tile_left != tiles_9_bigpillar_top) {
            return;
        }
    }
    if (curr_tile == tiles_1_floor ||
            curr_tile == tiles_2_spike ||
            curr_tile == tiles_3_pillar ||
            curr_tile == tiles_5_stuck ||
            curr_tile == tiles_8_bigpillar_bottom ||
            curr_tile == tiles_10_potion ||
            curr_tile == tiles_14_debris ||
            curr_tile == tiles_16_level_door_left ||
            curr_tile == tiles_18_chomper ||
            curr_tile == tiles_19_torch ||
            curr_tile == tiles_22_sword ||
            curr_tile == tiles_23_balcony_left ||
            curr_tile == tiles_25_lattice_pillar ||
            curr_tile == tiles_30_torch_with_debris ||
            (Kid.direction == dir_0_right && (curr_tile == tiles_4_gate || curr_tile == tiles_7_doortop_with_floor))
       ) {
        // frames 137..144: climb
        if (Kid.frame >= frame_137_climbing_3 && Kid.frame <= frame_144_climbing_10) {
            add_midtable(id_chtab_6_environment, floor_left_overlay[Kid.frame - 137], draw_xh, 0, (curr_tile == tiles_5_stuck) + draw_main_y, blitters_10h_transp, 0);
        }
        ptr_add_table = &add_midtable;
        draw_tile_bottom(0);
        ptr_add_table = &add_backtable;
    } else {
        draw_other_overlay();
    }
}

// seg008:1EB5
void draw_other_overlay(void) {
    byte tiletype;
    byte modifier;
    if (tile_is_empty(tile_left)) {
        ptr_add_table = &add_midtable;
        draw_tile2();
    } else if (curr_tile != tiles_0_empty && drawn_col > 0 &&
               get_tile_to_draw(drawn_room, drawn_col - 2, drawn_row, &tiletype, &modifier, tiles_0_empty) == tiles_0_empty
              ) {
        ptr_add_table = &add_midtable;
        draw_tile2();
        ptr_add_table = &add_backtable;
        draw_tile2();
        tile_object_redraw[tbl_line[drawn_row] + drawn_col] = 0xFF;
    }
    ptr_add_table = &add_backtable;
}

// seg008:1F48
void draw_tile2(void) {
    draw_tile_right();
    draw_tile_anim_right();
    draw_tile_base();
    draw_tile_anim();
    draw_tile_bottom(0);
    draw_loose();
}

// seg008:1F67
void draw_objtable_items_at_tile(byte tilepos) {
    short obj_count;
    short obj_index;
    obj_count = objtable_count;
    if (obj_count) {
        for (obj_index = obj_count - 1, n_curr_objs = 0; obj_index >= 0; --obj_index) {
            if (objtable[obj_index].tilepos == tilepos) {
                curr_objs[n_curr_objs++] = obj_index;
            }
        }
        if (n_curr_objs) {
            sort_curr_objs();
            for (obj_index = 0; obj_index < n_curr_objs; ++obj_index) {
                draw_objtable_item(curr_objs[obj_index]);
            }
        }
    }
}

// seg008:1FDE
void sort_curr_objs(void) {
    short swapped;
    short temp;
    short last;
    short index;
    // bubble sort
    last = n_curr_objs - 1;
    do {
        for (swapped = index = 0; index < last; ++index) {
            if (compare_curr_objs(index, index + 1)) {
                temp = curr_objs[index];
                curr_objs[index] = curr_objs[index + 1];
                curr_objs[index + 1] = temp;
                swapped = 1;
            }
        }
        // --last ?
    } while (swapped);
}

// seg008:203C
int compare_curr_objs(int index1, int index2) {
    short obj_index1;
    short obj_index2;
    obj_index1 = curr_objs[index1];
    if (objtable[obj_index1].obj_type == objtype_1_shadow) {
        return 1;
    }
    obj_index2 = curr_objs[index2];
    if (objtable[obj_index2].obj_type == objtype_1_shadow) {
        return 0;
    }
    if (objtable[obj_index1].obj_type == objtype_128_loose_floor &&
            objtable[obj_index2].obj_type == objtype_128_loose_floor
       ) {
        return (objtable[obj_index1].y < objtable[obj_index2].y);
    } else {
        return (objtable[obj_index1].y > objtable[obj_index2].y);
    }
    return 1;
}

// custom logic - calculating the distance to the wall tile to avoid graphical imperfections
// since "distance_to_edge(x)" modifies obj_xh and obj_xl
sbyte get_kid_delta_x(short xpos) {
    short calc_xl = (xpos - 65) % 14; // (x - 7) % 14, accounting for the 58 out-of-bound pixels
    if (calc_xl < 0) {
        calc_xl += 14; // xl is always positive, xh would have been decremented
    }
    // calc_xl should be 1 or 2 on the left side
    switch (calc_xl) {
        case 1:
        // fallthrough!
        case 2: {
            sbyte delta_x;
            if (!is_kid_shadow || is_shadow_flame) {
                delta_x = calc_xl * -2;
            } else {
                delta_x = (calc_xl == 1) * 2;
            }
            if (delta_x < 0) {
                short tilepos = get_tilepos_nominus(char_col_right, char_top_row);
                if (tilepos < 30) {
                    short tile_type = curr_room_tiles[tilepos] & 0x1F;
                    if (tile_type == tiles_5_stuck || tile_type == tiles_6_closer) {
                        delta_x += 2;
                    }
                }
            }
            return delta_x;
        }
        default:
            return 0;
    }
}

// custom logic - shift shadow image 1 pixel to avoid graphical imperfections
sbyte get_shadow_delta_x(void) {
    sbyte delta_x = 0;
    // shadow gets a partial kid's face when standing on specific horizontal x-coordinates
    if (obj_id == 14) { // standing
        short adjusted_x = obj_xh * 8 + obj_xl + delta_x;
        if (obj_direction == dir_FF_left) {
            adjusted_x -= 6;
        } else {
            ++adjusted_x;
        }
        if (adjusted_x > 0 && adjusted_x % 14 == 0) {
            --delta_x;
        }
        // shadow gets shifted 1 pixel back when climbing on the left side
    } else if (obj_direction == dir_0_right && obj_id >= 145 && obj_id <= 159) { // climbing right
        --delta_x;
    }
    return delta_x;
}

// seg008:20CA
void draw_objtable_item(int index) {
    int curr_obj = load_obj_from_objtable(index);
    sbyte delta_x = 0;
    switch (curr_obj) {
        case objtype_0_kid:
            // custom logic - makes kid's hand visible when hanging from a wall facing right
            if (Kid.direction == dir_0_right && Kid.action == actions_6_hang_straight) {
                delta_x = get_kid_delta_x(Kid.x);
            }
            // custom logic - kid has turned into shadow (not flame) that uses the XOR biltter
            if (is_kid_shadow && !is_shadow_flame && !is_cutscene) {
                delta_x += get_shadow_delta_x();
                goto shadow;
            }
        // fallthrough!
        case objtype_4_mirror_image:
            if (obj_id == 0xFF) return;
            // the Kid blinks a bit after uniting with shadow
            if (united_with_shadow && (united_with_shadow % 2) == 0) goto shadow;
            // custom logic - see shadow in the mirror reflection
            if (curr_obj == objtype_4_mirror_image) goto shadow;
        // fallthrough!
        case objtype_2_guard:
        case objtype_3_sword:
        case objtype_5_hurt_splash:
            add_midtable(obj_chtab, obj_id + 1, obj_xh, obj_xl + delta_x, obj_y, blitters_10h_transp, 1);
            break;
        case objtype_1_shadow:
shadow:
            add_midtable(obj_chtab, obj_id + 1, obj_xh, obj_xl + delta_x, obj_y, blitters_10h_transp, 1);
            add_midtable(obj_chtab, obj_id + 1, obj_xh, obj_xl + delta_x + 1, obj_y, blitters_3_xor, 1);
            break;
        case objtype_128_loose_floor:
            obj_direction = dir_FF_left;
            add_midtable(obj_chtab, loose_fram_left[obj_id], obj_xh, obj_xl, obj_y - 3, blitters_10h_transp, 1);
            add_midtable(obj_chtab, loose_fram_bottom[obj_id], obj_xh, obj_xl, obj_y, 0, 1);
            add_midtable(obj_chtab, loose_fram_right[obj_id], obj_x + 4, obj_xl, obj_y - 1, blitters_10h_transp, 1);
            break;
    }
}

// seg008:2228
int load_obj_from_objtable(int index) {
    objtable_type* curr_obj;
    curr_obj = &objtable[index];
    obj_xh = obj_x = curr_obj->xh;
    obj_xl = curr_obj->xl;
    obj_y = curr_obj->y;
    obj_id = curr_obj->id;
    obj_chtab = curr_obj->chtab_id;
    obj_direction = curr_obj->direction;
    obj_clip_top = curr_obj->clip.top;
    obj_clip_bottom = curr_obj->clip.bottom;
    obj_clip_left = curr_obj->clip.left;
    obj_clip_right = curr_obj->clip.right;
    // custom logic - getting guard's color from the object table
    obj_color = curr_obj->color;
    return curr_obj->obj_type;
}

// custom logic - dead kid fell out due to a fallen loose tile
int dead_fallout(sbyte col, char_type2 *dead) {
    word room;
    sbyte row;
    int dead_tile;
    room = dead->room;
    row = dead->row;
    while (room) {
        row++;
        if (row >= 3) {
            room = level.roomlinks[room - 1].down;
            row = 0;
            continue;
        }
        dead_tile = get_tile(room, col, row);
        if (tile_is_floor(dead_tile)) {
            dead->room = room;
            dead->row = row;
            dead->color = 0;
            Kid.alive = 0;
            hitp_curr = 0;
            return dead_tile;
        }
    }
    return 0;
}

// custom logic - draws falling guard
void draw_falling_guard(falling_guard_type guard) {
    if (guard.room == 0 || guard.direction == dir_56_none) {
        return;
    }
    byte guard_room = guard.room;
    byte guard_x = guard.x;
    byte guard_y = guard.y;
    int guard_col = get_tile_div_mod_m7(guard.x);
    if (guard_col < 0) {
        guard_room = level.roomlinks[guard_room - 1].left;
        guard_x += 140;
        guard_col += 10;
    } else if (guard_col > 9) {
        guard_room = level.roomlinks[guard_room - 1].right;
        guard_x -= 140;
        guard_col -= 10;
    }
    if (guard_room != 0 && guard_room != drawn_room && guard_y > 191) {
        guard_room = level.roomlinks[guard_room - 1].down;
        guard_y -= 189;
    }
    if (guard_room != drawn_room) {
        return;
    }
    char_type prev_char = Char;
    Char = (char_type) {
        frame_106_fall,
        guard_x,
        guard_y,
        guard.direction,
        guard_col,
        y_to_row_mod4(guard_y),
        actions_4_in_freefall,
        0, // fall x
        guard.fall_y, // fall y
        guard_room,
        0, // repeat
        charid_2_guard,
        sword_0_sheathed,
        -1, // alive
        0,
        guard.color & 0x0F
    };
    load_fram_det_col();
    load_frame_to_obj();
    stuck_lower();
    set_char_collision();
    set_objtile_at_char();
    redraw_at_char();
    redraw_at_char2();
    clip_char();
    add_objtable(objtype_2_guard);
    add_sword_to_objtable();
    Char = prev_char;
}

// custom logic - draws kid after shadow separation, dead guards and inactive guards
void draw_dead_or_inactive(char_type2* dead, byte charid, sbyte alive) {
    char_type2 dead_char;
    byte dead_frame;
    sbyte dead_col;
    int dead_tile;
    int delta_x, delta_y;
    bool is_alt_frame;
    delta_x = 0;
    delta_y = 0;
    dead_char = *dead;
    if (dead_char.room == 0 || dead_char.direction == dir_56_none) {
        return;
    }
    guard_overlap = charid == charid_2_guard &&
                    Guard.direction == dead_char.direction &&
                    ABS(Guard.x - dead_char.x) <= 28 &&
                    (Guard.curr_row == dead_char.row ||
                     (Guard.curr_row > dead_char.row && Guard.action >= actions_3_in_midair && Guard.action <= actions_4_in_freefall));
    // offset inactive guard from active guard
    if (dead->alt_frame == 0x05) {
        delta_y = -5;
    } else if (guard_overlap) {
        if (Guard.frame <= frame_171_stand_with_sword &&
                (Guard.frame == dead_char.frame || Guard.frame == dead_char.frame2) &&
                Guard.x == dead_char.x) {
            delta_x = dir_behind[dead_char.direction + 1];
        }
        if (can_inactive_guard_see_kid == 2 && dead == &inactive_guard[drawn_room - 1]) {
            delta_y = Guard.alive < 0 && dead->frame2 == frame_185_dead ? -5 : -2;
        }
    }
    dead_col = get_tile_div_mod_m7(dead_char.x);
    dead_tile = get_tile(dead_char.room, dead_col, dead_char.row);
    if (dead_char.room != drawn_room && tile_is_floor(dead_tile)) {
        return;
    }
    // check for a floor tile below when shadow got separated on a loose tile
    // that has fallen
    if (dead == &Dead && !tile_is_floor(dead_tile)) {
        dead_char.frame2 = 0;
        dead_tile = dead_fallout(dead_col, dead);
    }
    if (dead->room != drawn_room) {
        return;
    }
    if (!dead_char.frame2) {
        dead_frame = frame_185_dead;
        switch (dead_tile) {
            case 0: // nowhere to fall
                dead->direction = dir_56_none;
                return;
            case tiles_2_spike:
                if (charid != charid_0_kid && curr_room_modif[curr_tilepos] != 0xFF) {
                    dead_frame = frame_177_spiked;
                    // make spikes harmless (stuck)
                    curr_room_modif[curr_tilepos] = 0xFF;
                }
                break;
            case tiles_18_chomper:
                if (charid != charid_0_kid) {
                    dead_frame = frame_178_chomped;
                    // show blood on chompers
                    curr_room_modif[curr_tilepos] |= 0x80;
                }
                break;
        }
        dead->frame = dead_frame;
    } else {
        dead->frame = dead_char.frame2;
    }
    char_type prev_char = Char;
    is_alt_frame = charid == charid_2_guard &&
                   (settings->tbl_guard_type[current_level] == guard_type_guard ||
                    settings->tbl_guard_type[current_level] == guard_type_bird) &&
                   dead_char.alt_frame == 0xFF;
    Char = (char_type) {
        dead->frame,
             dead->x + delta_x,
             y_land[dead_char.row + ONSCREEN_ROW_OFFSET] + delta_y,
             dead_char.direction,
             dead_col,
             dead_char.row,
             actions_0_stand,
             0, // fall x
             0, // fall y
             dead_char.room,
             is_alt_frame ? 0xFF : 0, // repeat
             charid,
             sword_0_sheathed,
             alive,
             0,
             charid == charid_2_guard ? dead_char.color & 0x0F : 0
    };
    load_fram_det_col();
    load_frame_to_obj();
    if (alive < 0) {
        inactive_obj_x = obj_x;
    }
    stuck_lower();
    set_char_collision();
    dead->x_left_coll = char_x_left_coll;
    dead->x_right_coll = char_x_right_coll;
    if (delta_y < 0) { // inactive guard is next to an active guard
        obj_tilepos = 30; // draw further from the screen on z-axis
    } else if (Guard.alive < 0 && Guard.charid == charid_1_shadow && Guard.curr_row == Char.curr_row) {
        obj_tilepos = -1; // draw closer to the screen on z-axis
    } else {
        set_objtile_at_char();
    }
    redraw_at_char();
    redraw_at_char2();
    clip_char();
    // dead kid is also added to object table as guard to avoid shimmering when playing as shadow
    // on level 6B the dead character is shadow
    add_objtable(objtype_2_guard);
    if (alive && charid == charid_2_guard) {
        // draw splash from guard tab if inactive guard is hurt
        if (dead_char.alt_frame == 0x56) {
            dead->alt_frame = 0;
            save_obj();
            obj_tilepos = -1;
            obj_y -= 11;
            obj_dx_forward(2);
            // show a guard color splash when inactive guard is fighting next to active guard
            if (can_guard_see_kid == 2) {
                obj_chtab = current_level == settings->fat_guard_level && Char.color == 0
                            ? id_chtab_10_fat_guard
                            : id_chtab_5_guard;
                obj_id = 1;
            } else {
                obj_chtab = id_chtab_2_kid;
                obj_id = 218;
            }
            reset_obj_clip();
            add_objtable(objtype_5_hurt_splash);
            load_obj();
        }
        // add sword to inactive guard
        add_sword_to_objtable();
    }
    // press the button when the kid/guard is on
    if (get_tile_at_char() == tiles_15_opener || curr_tile2 == tiles_6_closer) {
        trigger_button(1, 0, -1);
    }
    Char = prev_char;
}

// custom logic - draw inactive guard
void draw_inactive_guard(void) {
    inactive_obj_x = 0;
    if (inactive_guard[drawn_room - 1].room == drawn_room) {
        draw_dead_or_inactive(&inactive_guard[drawn_room - 1], charid_2_guard, -1);
    }
}

// custom logic - draw dead kid
void draw_dead_kid(void) {
    draw_dead_or_inactive(&Dead, charid_0_kid, 0);
}

// custom logic - draw dead guards
void draw_dead_guards(void) {
    draw_dead_or_inactive(&outbound_dead_guard1[drawn_room - 1], charid_2_guard, 0);
    draw_dead_or_inactive(&outbound_dead_guard2[drawn_room - 1], charid_2_guard, 0);
    draw_dead_or_inactive(&outbound_dead_guard3[drawn_room - 1], charid_2_guard, 0);
    draw_dead_or_inactive(&dead_guard1[drawn_room - 1], charid_2_guard, 0);
    draw_dead_or_inactive(&dead_guard2[drawn_room - 1], charid_2_guard, 0);
    draw_dead_or_inactive(&dead_guard3[drawn_room - 1], charid_2_guard, 0);
}

// custom logic - draw falling guards
void draw_falling_guards(void) {
    draw_falling_guard(falling_guard);
}

// seg008:228A
void draw_people(void) {
    check_mirror();
    draw_dead_kid();
    draw_kid();
    draw_inactive_guard();
    draw_falling_guards();
    draw_dead_guards();
    draw_guard();
    reset_obj_clip();
    draw_hp();
}

// seg008:22A2
void draw_kid(void) {
    if (Kid.room != 0 && Kid.direction != dir_56_none && Kid.room == drawn_room) {
        add_kid_to_objtable();
        // custom logic - kid is hurt and has not separated as shadow
        if (charid_hurt_by_sword == charid_0_kid &&
                !is_kid_shadow &&
                hitp_delta < 0) {
            draw_hurt_splash();
            charid_hurt_by_sword = 0xFF;
        }
        add_sword_to_objtable();
        // custom logic - render flame silhouette image
        if (current_level == settings->flame_level &&
                is_kid_shadow &&
                is_shadow_flame &&
                Kid.frame >= frame_229_found_sword && Kid.frame <= frame_230_sheathe) {
            byte kid_frame = Kid.frame;
            Kid.frame = frame_212_flame_silhouette;
            add_kid_to_objtable();
            Kid.frame = kid_frame;
        }
    }
}

// seg008:22C9
void draw_guard(void) {
    if (Guard.direction != dir_56_none && Guard.room == drawn_room) {
        add_guard_to_objtable();
        // custom logic - guard is hurt and is not Jaffar's vision
        if (charid_hurt_by_sword == charid_2_guard &&
                settings->tbl_guard_type[current_level] != guard_type_Jaffar_vision &&
                guardhp_delta < 0) {
            draw_hurt_splash();
            charid_hurt_by_sword = 0xFF;
        }
        add_sword_to_objtable();
    }
}

// seg008:22F0
void add_kid_to_objtable(void) {
    loadkid();
    load_fram_det_col();
    load_frame_to_obj();
    stuck_lower();
    set_char_collision();
    set_objtile_at_char();
    redraw_at_char();
    redraw_at_char2();
    clip_char();
    add_objtable(objtype_0_kid);
}

// seg008:2324
void add_guard_to_objtable(void) {
    word obj_type;
    bool is_jaffars_vision;
    // custom logic - Jaffar's vision has a sword trick when standing on the row 2
    is_jaffars_vision = settings->tbl_guard_type[current_level] == guard_type_Jaffar_vision && Guard.curr_row == 2;
    loadshad();
    load_fram_det_col();
    load_frame_to_obj();
    stuck_lower();
    set_char_collision();
    set_objtile_at_char();
    redraw_at_char();
    redraw_at_char2();
    if (!is_jaffars_vision) {
        clip_char();
    }
    if (Char.charid == charid_1_shadow) {
        // Special event: shadow is clipped: may appear only right from the mirror
        if (current_level == settings->mirror_level && Char.room == settings->mirror_room) {
            obj_clip_left = 137;
            obj_clip_left += (settings->mirror_column - 4) * 32; // added
        }
        obj_type = objtype_1_shadow;
    } else {
        obj_type = objtype_2_guard;
        // custom logic - Jaffar's vision is shown 2 rows above the sword
        if (is_jaffars_vision) {
            obj_y -= 126;
        }
    }
    add_objtable(obj_type);
    // custom logic - the sword is shown 2 rows below when fighting
    if (is_jaffars_vision && Char.frame != frame_166_stand_inactive) {
        obj_y += 126;
    }
}

// seg008:2388
void add_objtable(byte obj_type) {
    word index;
    objtable_type* entry_addr;
    index = objtable_count++;
    if (index >= 50) {
        show_dialog("ObjTable Overflow", default_dialog);
        return; // added
    }
    entry_addr = &objtable[index];
    entry_addr->obj_type = obj_type;
    x_to_xh_and_xl(obj_x, &entry_addr->xh, &entry_addr->xl);
    entry_addr->y = obj_y;
    entry_addr->clip.top = obj_clip_top;
    entry_addr->clip.bottom = obj_clip_bottom;
    entry_addr->clip.left = obj_clip_left;
    entry_addr->clip.right = obj_clip_right;
    entry_addr->chtab_id = obj_chtab;
    entry_addr->id = obj_id;
    // custom logic - setting guard's color in the object table
    entry_addr->direction = obj_direction;
    entry_addr->color = obj_color;
    mark_obj_tile_redraw(index);
}

// seg008:2423
void mark_obj_tile_redraw(int index) {
    objtable[index].tilepos = obj_tilepos;
    if (obj_tilepos < MAX_TILES_PER_ROOM) {
        tile_object_redraw[obj_tilepos] = 1;
    }
}

// seg008:2448
void load_frame_to_obj(void) {
    word chtab_base;
    short guard_type;
    chtab_base = id_chtab_2_kid;
    reset_obj_clip();
    load_frame();
    obj_direction = Char.direction;
    guard_type = current_level <= 15 ? settings->tbl_guard_type[current_level] : -1;
    // custom logic - fat guard has no color on level 8
    if (current_level == settings->fat_guard_level && Char.charid == charid_2_guard && Char.color == 0) {
        guard_type = guard_type_fat;
    }
    // custom logic - regular or bird guard can have a different death frame image
    if (current_level <= 15 &&
            Char.charid == charid_2_guard &&
            Char.frame == frame_185_dead &&
            Char.repeat == 0xFF) {
        if (guard_type == guard_type_guard || guard_type == guard_type_bird) {
            ++cur_frame.image;
        }
    }
    obj_id = cur_frame.image;
    // top 6 bits of sword are the chtab
    obj_chtab = chtab_base + (cur_frame.sword >> 6);
    // custom logic - fat guard has a different chtab
    if (obj_chtab == id_chtab_5_guard && guard_type == guard_type_fat) {
        obj_chtab = id_chtab_10_fat_guard;
    }
    // (Char.x + frame x-offset) * 2 - (58 * 2)
    // The x value is scaled 1:2. 0-58 are offscreen pixels that represent the rightmost
    // 5 tiles (4 tiles and 2 pixels) of the room on the left. That makes obj_x to start
    // from 0. But the sword coordinate may have negative values when facing left.
    // obj_x is a short integer instead of a byte because the values are doubled.
    // It is split into xh and xl sbyte values in the object table.
    obj_x = (char_dx_forward(cur_frame.dx) << 1) - 116;
    // The y value is scaled 1:1 and includes the frame y-offset.
    obj_y = cur_frame.dy + Char.y;
    if ((sbyte)(cur_frame.flags ^ obj_direction) >= 0) {
        // even/odd (0x80) pixel frames are shifted 1 pixel when facing left
        // frames without this flag are shifted 1 pixel with facing right
        ++obj_x;
    }
    // custom logic - position dead fat guard on level 6 slightly lower
    if (current_level == 6 && Char.charid == charid_2_guard && (Char.frame == 191 || Char.frame == 192) /* dead fat guard */) {
        ++obj_y;
    }
    // custom logic - set the assigned color for each separate regular guard
    if (Char.charid == charid_2_guard && guard_type == guard_type_guard) {
        obj_color = Char.color;
    } else {
        obj_color = 0;
    }
}

// seg008:24A8
void show_time(void) {
    char sprintf_temp[40];
    word rem_sec;
    bool time_stopped;
    // custom logic - time stops when kid meets Jaffar on level 13 or turns into flame
    time_stopped = current_level < 1 ||
                   current_level > settings->win_level ||
                   (current_level == settings->win_level && (jaffar_presented == 7 || is_shadow_flame));
    if (
        (Kid.alive < 0 ||
         // custom logic - kid does not have enough lives to continue, the remaining time needs to run out
         delayed_sound.sound_id == sound_38_blink) &&
        next_level == current_level &&
        rem_min != 0 &&
        !time_stopped
    ) {
        // Time passes
        --rem_tick;
        if (rem_tick == 0) {
            rem_tick = 719; // 720=12*60 ticks = 1 minute
            --rem_min;
            if (rem_min > 0 && (rem_min <= 5 || rem_min % 5 == 0)) {
                is_show_time = 1;
            } else if (rem_min < 0) {
                is_show_time = ((~rem_min) % 5 == 0 ) ? 1 : 0;
            }
        } else {
            if (rem_min == 1 && rem_tick % 12 == 0) {
                is_show_time = 1;
                text_time_remaining = 0;
            }
        }
    }
    if (is_show_time && text_time_remaining == 0 && !time_stopped && text_time_total != RESTART_NOT_ENOUGH_HITPOINTS) {
        text_time_remaining = text_time_total = 24;
        if (rem_min > 0) {
            if (rem_min == 1) {
                rem_sec = (rem_tick + 1) / 12;
                if (rem_sec == 1) {
                    strncpy(sprintf_temp, "1 SECOND LEFT", sizeof(sprintf_temp));
                    text_time_remaining = text_time_total = 12;
                } else {
                    snprintf(sprintf_temp, sizeof(sprintf_temp), "%d SECONDS LEFT", rem_sec);
                }
                // custom logic - do not display remaining seconds on level 10, room 8, row 2
                if (current_level != 10 || Kid.room != 8 || Kid.curr_row != 2) {
                    display_text_bottom(sprintf_temp);
                }
            } else {
                if (cheats_enabled) {
                    snprintf(sprintf_temp, sizeof(sprintf_temp), "%d MINUTES LEFT (CHEATING)", rem_min);
                } else if (max_saves_per_level == 0) {
                    snprintf(sprintf_temp, sizeof(sprintf_temp), "%d MINUTES LEFT", rem_min);
                } else {
                    snprintf(sprintf_temp, sizeof(sprintf_temp), "%d MINUTES LEFT (%d SAVES)", rem_min, total_saves);
                }
                display_text_bottom(sprintf_temp);
            }
        } else {
            display_text_bottom("TIME HAS EXPIRED!");
        }
        is_show_time = 0;
    }
}

// seg008:25A8
void show_level(void) {
    byte disp_level;
    text_time_remaining = text_time_total = 0;
    disp_level = current_level;
    if (disp_level > 0 && disp_level < 15) {
        // custom logic - display demo level message only when playing demo
        if (disp_level == 0 && !play_demo_level) {
            return;
        }
        // custom logic - levels 6B-13 have an incremented index value
        if ((disp_level == 6 && falling_exit_checkpoint >= 4) || disp_level > 6) {
            ++disp_level;
        }
        text_time_remaining = text_time_total = 60;
        // custom logic - use predefined level names
        display_text_bottom(settings->level_names[disp_level]);
        is_show_time = 1;
    }
    seamless = 0;
}

// seg008:2602
short calc_screen_x_coord(short logical_x) {
    return logical_x * 320 / 280;
}

// seg008:2627
void free_peels(void) {
    while (peels_count > 0) {
        --peels_count;
        free_peel(peels_table[peels_count]);
    }
}

// seg008:2644
void display_text_bottom(const char *text) {
    erase_bottom_text(0);
    // custom logic - use PoP2 yellowish text
    show_text_with_color(&rect_bottom_text, text_align_center, text_align_bottom, text, color_20_yellowish_text);
}

// seg008:266D
void erase_bottom_text(int clear_text) {
    draw_rect(&rect_bottom_text, color_0_black);
    if (clear_text) {
        text_time_total = 0;
        text_time_remaining = 0;
    }
}

// Dungeon wall drawing algorithm by HTamas

#define RES_WALL_FACE_MAIN     1    // face stack main
#define RES_WALL_FACE_TOP      2    // face stack top
#define RES_WALL_CENTRE_BASE   3    // centre stack base
#define RES_WALL_CENTRE_MAIN   4    // centre stack main
#define RES_WALL_RIGHT_BASE    5    // right stack base
#define RES_WALL_RIGHT_MAIN    6    // right stack main
#define RES_WALL_SINGLE_BASE   7    // single stack base
#define RES_WALL_SINGLE_MAIN   8    // single stack main
#define RES_WALL_LEFT_BASE     9    // left stack base
#define RES_WALL_LEFT_MAIN    10    // left stack main
#define RES_WALL_DIVIDER1     11    // the broad divider
#define RES_WALL_DIVIDER2     12    // the narrow divider
#define RES_WALL_RNDBLOCK     13    // random block.bmp
#define RES_WALL_MARK_TL      14    // top left mark
#define RES_WALL_MARK_BL      15    // bottom left mark
#define RES_WALL_MARK_TR      16    // top right mark
#define RES_WALL_MARK_BR      17    // bottom right mark

#define WALL_MODIFIER_NO_WALLS    0
#define WALL_MODIFIER_RIGHT_WALL  1
#define WALL_MODIFIER_LEFT_WALL   2
#define WALL_MODIFIER_BOTH_WALLS  3

// seg008:268F
void wall_pattern(int which_part, int which_table) {
    // which_part: 0 = bottom part, 1 = foreground part
    // which_table: 0 = background table, 1 = foreground table
    // local variables
    add_table_type saved_sim;
    word middle_brick_divider;
    byte middle_brick_offset;
    word bottom_brick_divider;
    byte bottom_brick_offset;
    byte bg_modifier;
    dword saved_prng_state;
    word is_dungeon;
    // save the value for the sprite insertion method, so that it can be restored
    saved_sim = ptr_add_table;
    // set the sprite insertion method based on the arguments
    if (which_table == 0) {
        ptr_add_table = &add_backtable;
    } else {
        ptr_add_table = &add_foretable;
    }
    // save the state of the pseudorandom number generator
    saved_prng_state = random_seed;
    // set the new seed
    random_seed = drawn_room + tbl_line[drawn_row] + drawn_col;
    // custom logic - placeholder room randomization
    if (drawn_room == PLACEHOLDER_ROOM) {
        random_seed += placeholder_id;
    }
    prandom(1); // fetch a random number and discard it
    is_dungeon = (settings->tbl_level_type[current_level] < PALACE_LEVEL) || settings->tbl_level_wda[current_level] == 1;
    if (!is_dungeon) {
        if (which_part) {
            // draw solid colours for the bricks
            add_wipetable(which_table, 8 * (draw_xh), draw_main_y - 40, 20, 4 * 8, palace_wall_colors[44 * drawn_row +      drawn_col]);
            add_wipetable(which_table, 8 * (draw_xh), draw_main_y - 19, 21, 2 * 8, palace_wall_colors[44 * drawn_row + 11 + drawn_col]);
            add_wipetable(which_table, 8 * (draw_xh + 2), draw_main_y - 19, 21, 2 * 8, palace_wall_colors[44 * drawn_row + 12 + drawn_col]);
            add_wipetable(which_table, 8 * (draw_xh), draw_main_y, 19, 1 * 8, palace_wall_colors[44 * drawn_row + 22 + drawn_col]);
            add_wipetable(which_table, 8 * (draw_xh + 1), draw_main_y, 19, 3 * 8, palace_wall_colors[44 * drawn_row + 23 + drawn_col]);
            // draw the dividing decals between bricks
            ptr_add_table(id_chtab_7_environmentwall, prandom(2) +  3, draw_xh + 3, 0, draw_main_y - 53, blitters_46h_mono_6, 0);
            ptr_add_table(id_chtab_7_environmentwall, prandom(2) +  6, draw_xh, 0, draw_main_y - 34, blitters_46h_mono_6, 0);
            ptr_add_table(id_chtab_7_environmentwall, prandom(2) +  9, draw_xh, 0, draw_main_y - 13, blitters_46h_mono_6, 0);
            ptr_add_table(id_chtab_7_environmentwall, prandom(2) + 12, draw_xh, 0, draw_main_y, blitters_46h_mono_6, 0);
        }
        add_wipetable(which_table, 8 * draw_xh, draw_bottom_y,  3, 4 * 8, palace_wall_colors[44 * drawn_row + 33 + drawn_col]);
        ptr_add_table(id_chtab_7_environmentwall, prandom(2) + 15, draw_xh, 0, draw_bottom_y, blitters_46h_mono_6, 0);
    } else {
        middle_brick_divider = prandom(1);
        middle_brick_offset = prandom(4);
        bottom_brick_divider = prandom(1);
        bottom_brick_offset = prandom(4);
        // store the background modifier for the current tile in a local variable
        // for walls, the modifier stores whether there are adjacent walls
        bg_modifier = curr_modifier & 0x7F;
        switch (bg_modifier) {
            // walls on both sides
            case WALL_MODIFIER_BOTH_WALLS:
                if (which_part != 0) {
                    if (prandom(4) == 0) {
                        ptr_add_table(id_chtab_7_environmentwall, RES_WALL_RNDBLOCK, draw_xh, 0, draw_bottom_y - 42, blitters_0_no_transp, 0);
                    }
                    ptr_add_table(id_chtab_7_environmentwall, RES_WALL_DIVIDER1 + middle_brick_divider, draw_xh + 1, middle_brick_offset, draw_bottom_y - 21, blitters_10h_transp, 0);
                }
                ptr_add_table(id_chtab_7_environmentwall, RES_WALL_DIVIDER1 + bottom_brick_divider, draw_xh, bottom_brick_offset, draw_bottom_y, blitters_10h_transp, 0);
                if (which_part != 0) {
                    if (is_dungeon) {
                        if (prandom(4) == 0) {
                            draw_right_mark(prandom(3), middle_brick_offset);
                        }
                        if (prandom(4) == 0) {
                            draw_left_mark(prandom(4), middle_brick_offset - middle_brick_divider, bottom_brick_offset - bottom_brick_divider);
                        }
                    }
                }
                break;
            // no walls on either side
            case WALL_MODIFIER_NO_WALLS:
                if (is_dungeon) {
                    if (which_part != 0) {
                        if (prandom(6) == 0) {
                            draw_left_mark(prandom(1), middle_brick_offset - middle_brick_divider, bottom_brick_offset - bottom_brick_divider);
                        }
                    }
                }
                break;
            // wall on the right, but not on the left
            case WALL_MODIFIER_RIGHT_WALL:
                if (which_part != 0) {
                    if (prandom(4) == 0) {
                        ptr_add_table(id_chtab_7_environmentwall, RES_WALL_RNDBLOCK, draw_xh, 0, draw_bottom_y - 42, blitters_0_no_transp, 0);
                    }
                    ptr_add_table(id_chtab_7_environmentwall, RES_WALL_DIVIDER1 + middle_brick_divider, draw_xh + 1 /*fix*/, middle_brick_offset, draw_bottom_y - 21, blitters_10h_transp, 0);
                    if (is_dungeon) {
                        if (prandom(4) == 0) {
                            draw_right_mark(prandom(3), middle_brick_offset);
                        }
                        if (prandom(4) == 0) {
                            draw_left_mark(prandom(3), middle_brick_offset - middle_brick_divider, bottom_brick_offset - bottom_brick_divider);
                        }
                    }
                }
                break;
            // wall on the left, but not on the right
            case WALL_MODIFIER_LEFT_WALL:
                if (which_part != 0) {
                    ptr_add_table(id_chtab_7_environmentwall, RES_WALL_DIVIDER1 + middle_brick_divider, draw_xh + 1, middle_brick_offset, draw_bottom_y - 21, blitters_10h_transp, 0);
                }
                ptr_add_table(id_chtab_7_environmentwall, RES_WALL_DIVIDER1 + bottom_brick_divider, draw_xh, bottom_brick_offset, draw_bottom_y, blitters_10h_transp, 0);
                if (which_part != 0) {
                    if (is_dungeon) {
                        if (prandom(4) == 0) {
                            draw_right_mark(prandom(1) + 2, middle_brick_offset);
                        }
                        if (prandom(4) == 0) {
                            draw_left_mark(prandom(4), middle_brick_offset - middle_brick_divider, bottom_brick_offset - bottom_brick_divider);
                        }
                    }
                }
                break;
        }
    }
    random_seed = saved_prng_state;
    ptr_add_table = saved_sim;
}

void draw_left_mark(word decal_type, word middle_xl, word bottom_xl) {
    static const word y_offset[] = {58, 41, 37, 20, 16}; // y offset for all brick decal variants
    word image_id = RES_WALL_MARK_TL;
    word xl = 0;
    if (decal_type % 2) {
        image_id = RES_WALL_MARK_BL;
    }
    if (decal_type > 3) {
        xl = bottom_xl + 6;
    } else if (decal_type > 1) {
        xl = middle_xl + 6;
    }
    ptr_add_table(id_chtab_7_environmentwall, image_id, draw_xh + (decal_type == 2 || decal_type == 3), xl, draw_bottom_y - y_offset[decal_type], blitters_10h_transp, 0);
}

void draw_right_mark(word decal_type, word xl) {
    static const word y_offset[] = {52, 42, 31, 21}; // y offset for all brick decal variants
    word image_id = RES_WALL_MARK_TR;
    if (decal_type % 2) {
        image_id = RES_WALL_MARK_BR;
    }
    if (decal_type < 2) {
        xl = 24;
    } else {
        xl -= 3;
    }
    ptr_add_table(id_chtab_7_environmentwall, image_id, draw_xh + (decal_type > 1), xl, draw_bottom_y - y_offset[decal_type], blitters_10h_transp, 0);
}

// custom logic - loads custom images from PRINCE15.DAT
image_type* load_custom_graphics_image(int resource) {
    int size;
    byte* image_data;
    SDL_RWops* rw;
    image_type* image = NULL;
    image_data = load_from_opendats_alloc(resource, "bin", NULL, &size);
    rw = SDL_RWFromConstMem(image_data, size);
    if (rw == NULL) {
        printf("Could not load resource %d from PRINCE15.DAT", resource);
        quit(1);
    }
    image = IMG_Load_RW(rw, 0);
    if (image == NULL) {
        printf("Could not load resource %d from PRINCE15.DAT", resource);
        quit(1);
    }
    if (SDL_RWclose(rw) != 0) {
        sdlperror("SDL_RWclose");
    }
    if (image_data != NULL) free(image_data);
    return image;
}

// custom logic - initialize custom graphic resources from PRINCE15.DAT
void init_custom_graphics(void) {
    dat_type* dathandle;
    dathandle = open_dat("PRINCE15.DAT");
    byte* n_image_data = load_from_opendats_alloc(4000, "bin", NULL, NULL);
    if (n_image_data == NULL) {
        printf("Could not load resource 4000 from PRINCE15.DAT");
        quit(1);
    }
    int n_images = *n_image_data;
    size_t alloc_size = sizeof(chtab_type) + sizeof(void *) * n_images;
    custom_graphics_tab = (chtab_type*) malloc(alloc_size);
    memset(custom_graphics_tab, 0, alloc_size);
    custom_graphics_tab->n_images = n_images;
    for (int i = 0; i < n_images; i++) {
        custom_graphics_tab->images[i] = load_custom_graphics_image(4007 + i);
    }
    close_dat(dathandle);
}

// *INDENT-OFF*
// custom logic - custom image position table
const custom_image_type custom_image_table[] = {
    // resource id, screen x, screen y
    {  0,   0,   0 },
    {  1,  68,  65 },
    {  2, 102,  65 },
    {  3,  68,  65 },
    {  4, 148,  70 },
    {  5, 224,  53 },
    {  6, 192,   3 },
    {  7,  27,  73 },
    {  8,  23,  70 },
    {  9,  -1,  -1 },
    { 10,   0,   0 },
    { 11, 139,   4 },
    { 12, 153, 101 },
    { 13, 142,  71 },
    { 14,   5,  22 },
    { 15, 281,  63 },    
};
// *INDENT-ON*

// custom logic - draws custom graphics in foretable
void draw_custom_graphics_foretable(void) {
    custom_image_type pos;
    if (is_cutscene || quick_loading || is_blind_mode) {
        return;
    }
    if (current_level == 4 &&
            drawn_room == 23 &&
            curr_room == drawn_room &&
            get_curr_tile(5) == tiles_6_closer &&
            get_doorlink_timer(curr_modifier) > 1) { // pressed
        // hide an artifact with a black pixel
        rect_type pixel_rect = (struct rect_type) {
            63, 160, 64, 161
        };
        draw_rect(&pixel_rect, color_0_black);
    } else if (current_level == 10 && drawn_room == 4) {
        // show the image of the horse statue
        pos = custom_image_table[image_12_horse_statue];
        draw_image(custom_graphics_tab->images[image_12_horse_statue], NULL, pos.x, pos.y, blitters_10h_transp);
    } else if (current_level == 10 && drawn_room == 10) {
        // show the image of the missing gate frame piece
        pos = custom_image_table[image_15_gate_frame_piece];
        draw_image(custom_graphics_tab->images[image_15_gate_frame_piece], NULL, pos.x, pos.y, blitters_0_no_transp);
    } else if (current_level == settings->flame_level &&
               drawn_room == settings->flaming_sword_room &&
               (curr_room_tiles[settings->flaming_sword_tilepos] & 0x1F) == tiles_22_sword) {
        // show flaming sword base on level 13
        pos = custom_image_table[image_5_flaming_sword];
        draw_image(custom_graphics_tab->images[image_5_flaming_sword], NULL, pos.x, pos.y, blitters_10h_transp);
    }
}

// custom logic - draws custom graphics in backtable
void draw_custom_graphics_backtable(void) {
    custom_image_type pos;
    if (is_cutscene || quick_loading || is_blind_mode) {
        return;
    }
    // show Jaffar in the mirror room of level 6
    if (current_level == 6 &&
            drawn_room == 1 &&
            leveldoor_open == 0x4D) {
        pos = custom_image_table[image_7_jaffar_vision + jumped_through_mirror];
        draw_image(custom_graphics_tab->images[image_7_jaffar_vision + jumped_through_mirror], NULL, pos.x, pos.y, blitters_10h_transp);
        // show the image of flaming shadow on level 10
    } else if (current_level == 10 &&
               drawn_room == 8 &&
               (Kid.alive < 0 || holding_sword) &&
               (level_restart_timer == 0 || level_restart_timer > DEAD_KID_LEVEL_10_TIMER_P12)) {
        pos = custom_image_table[image_4_fire_kid];
        draw_image(custom_graphics_tab->images[image_4_fire_kid], NULL, pos.x, pos.y, blitters_0_no_transp);
        // show the message explaining that kid needs to get killed by the last guard on level 10
        // if he has not merged with the shadow
    } else if (current_level == 10 &&
               drawn_room == 19 &&
               is_shadow_text_shown &&
               shadow_initialized == 0 &&
               (level_restart_timer == 0 || level_restart_timer > DEAD_KID_LEVEL_10_TIMER_P12)) {
        pos = custom_image_table[image_2_dead_kid];
        draw_image(custom_graphics_tab->images[image_2_dead_kid], NULL, pos.x, pos.y, blitters_10h_transp);
        // show the birdman picture above the start level door on level 10
    } else if (current_level == 10 && drawn_room == 1) {
        pos = custom_image_table[image_11_birdman];
        draw_image(custom_graphics_tab->images[image_11_birdman], NULL, pos.x, pos.y, blitters_0_no_transp);
        // show the image of the serpent symbol above the level door that cannot be opened
    } else if (current_level == 10 && drawn_room == 20) {
        pos = custom_image_table[image_13_serpent_symbol];
        draw_image(custom_graphics_tab->images[image_13_serpent_symbol], NULL, pos.x, pos.y, blitters_10h_transp);
        // show the image of the bird guard statue
    } else if (current_level == 10 && drawn_room == 14) {
        pos = custom_image_table[image_14_birdman_statue];
        draw_image(custom_graphics_tab->images[image_14_birdman_statue], NULL, pos.x, pos.y, blitters_10h_transp);
        // show the message explaining that kid needs separate from the shadow to pass through the last 2 guards on level 11
    } else if (current_level == 11 && drawn_room == 20) {
        pos = custom_image_table[image_3_turned_kid];
        draw_image(custom_graphics_tab->images[image_3_turned_kid], NULL, pos.x, pos.y, blitters_10h_transp);
        // show dragon banner on in last room of level 13
    } else if (current_level == 14 && drawn_room == 23) {
        pos = custom_image_table[image_6_dragon_wall];
        draw_image(custom_graphics_tab->images[image_6_dragon_wall], NULL, pos.x, pos.y, blitters_10h_transp);
    }
}

// *INDENT-OFF*
// custom logic - initial level 6 sword x/y frame offsets
const int sword_x_offset[] = {
     0,    0,    0,    2,
     2,    2,    2,    0,
    -4,   -3,   -5,   -5,
    -5,   -8,   -9,  -10,
   -10,  -10,  -10,  -10,
    -5,   -4,    1,    2,
    -5,  -12,  -27,  -32,
   -41,  -38,  -44,  -50,
};

const int sword_y_offset[] = {
    10,   8,   6,   4,
     2,   2,   2,   2,
     2,   2,   2,   2,
     3,   3,   4,   5,
     5,   5,   5,   4,
     3,   1,   2,   4,
     1,  12,  19,  24,
    36,  48,  64,  80,
};
// *INDENT-ON*

// custom logic - initial level 6 last sword frame number
int last_sword_frame = 0;
bool is_sword_flipped = 0;

// custom logic - draws initial level 6 sword animation
void draw_level_6_sword_animation(void) {
    short sword_frame;
    short sword_tilepos;
    short crouch_x_offset;
    short crouch_y_offset;
    // show moving sword in the initial sequence on level 6
    if (is_level_6_sequence() && had_sword && !is_menu_shown) {
        crouch_y_offset = 0;
        crouch_x_offset = 0;
        if (level_6_sequence_timer > 0 && level_6_sequence_timer <= 13) {
            sword_frame = 4;
            if (level_6_sequence_timer == 1) {
                crouch_y_offset = 11;
            } else if (level_6_sequence_timer == 2) {
                crouch_x_offset = -2;
                crouch_y_offset = 11;
            } else if (level_6_sequence_timer < 6) {
                crouch_y_offset = (6 - level_6_sequence_timer) * 3;
            }
            goto loc_draw_sword;
        } else if (potion_timer > 0 && potion_timer < 32 && last_sword_frame != potion_timer) {
            sword_frame = potion_timer;
loc_draw_sword:
            if (is_sword_flipped && (sword_frame < 24 || sword_frame > 28)) {
                flip_screen(custom_graphics_tab->images[image_9_sword]);
                is_sword_flipped = 0;
            }
            last_sword_frame = sword_frame <= 30 ? sword_frame : 0;
            redraw_height = 63;
            for (sword_tilepos = 14; sword_tilepos < 17; ++sword_tilepos) {
                set_redraw_full(sword_tilepos, 1);
                set_wipe(sword_tilepos, 1);
            }
            get_tile(24, 4, 2);
            set_redraw_full(curr_tilepos, 1);
            set_wipe(curr_tilepos, 1);
            draw_image(custom_graphics_tab->images[image_9_sword],
                       NULL,
                       settings->level_6_seq_sword_left_x + sword_x_offset[sword_frame] + crouch_x_offset,
                       settings->level_6_seq_sword_top_y + sword_y_offset[sword_frame] + crouch_y_offset,
                       blitters_10h_transp);
            if (sword_frame == 24 || sword_frame == 28) {
                flip_screen(custom_graphics_tab->images[image_9_sword]);
                is_sword_flipped = 1 - is_sword_flipped;
            }
        }
    }
}
