/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#include "common.h"

// *INDENT-OFF*
const word seqtbl_offsets[] = {
    running,          startrun,         stand,            standjump,
    runjump,          turn,             runturn,          stepfall,
    jumphangMed,      hang,             climbup,          hangdrop,
    freefall,         runstop,          jumpup,           fallhang,
    jumpbackhang,     softland,         jumpfall,         stepfall2,
    medland,          rjumpfall,        hardland,         hangfall,
    jumphangLong,     hangstraight,     rdiveroll,        sdiveroll,
    highjump,         step1,            step2,            step3,
    step4,            step5,            step6,            step7,
    step8,            step9,            step10,           step11,
    step12,           step13,           step14,           turnrun,
    testfoot,         bumpfall,         hardbump,         bump,
    superhijump,      standup,          stoop,            impale,
    crush,            deadfall,         halve,            engarde,
    advance,          retreat,          strike,           flee,
    turnengarde,      striketoblock,    readyblock,       landengarde,
    bumpengfwd,       bumpengback,      blocktostrike,    strikeadv,
    climbdown,        blockedstrike,    climbstairs,      dropdead,
    stepback,         climbfail,        stabbed,          faststrike,
    strikeret,        alertstand,       drinkpotion,      crawl,
    alertturn,        fightfall,        efightfall,       efightfallfwd,
    running,          stabkill,         fastadvance,      goalertstand,
    arise,            turndraw,         guardengarde,     pickupsword,
    resheathe,        fastsheathe,      Pstand,           Vstand,
    Vwalk,            Vstop,            Palert,           Pstepback,
    Vexit,            Mclimb,           Vraise,           Plie,
    patchfall,        Mscurry,          Mstop,            Mleave,
    Pembrace,         Pwaiting,         Pstroke,          Prise,
    Pcrouch,          Pslump,           Mraise,           balcony,
    crawlrepeat,      putdownpotion,    landjump,         stabbedfrombehind,
    drinkmousepotion, hangstraightdrop, chompersprint,    birdbow,
    birdturnbow,      cellstoop,        cellcrouch,       dropdead_dead,
    bumpcloseddoor,   kidturnengarde,   kidturnwithsword, Pdead,
    impalehanging,    birdimpalehang,   stepcrouch6,      stepcrouch7,
    stepcrouch8,      stepcrouch9,      stepcrouch10,     stepcrouch11,
    runjumphang,      collapse,         shadowrise,       stepbackhang4,
    stepbackhang5,    stepbackhang6,    stepbackhang7,    stepbackhang8,
    stepbackjumphang, stepfwdhang9,     stepfwdhang8,     stepfwdhang7,
    stepfwdhang6,     stepfwdhang5,     stepfwdhang4,     startfistfight,
    stopfistfight,    fistfighthit,     stabfallfar,      stabfalldead,
    mscurrymanual,    retreatstep,      retreatkid,       advancekid,
    fistfightadv,     fistfightkick,
};
// *INDENT-ON*

// data:196E
byte seqtbl[] = {
    LABEL(running) // running
    act(actions_1_run_jump), jmp(runcyc1), // goto running: frame 7

    LABEL(startrun) // startrun
    act(actions_1_run_jump), LABEL(runstt1) frame_1_start_run,
    frame_2_start_run, frame_3_start_run, LABEL(runstt4) frame_4_start_run,
    dx(8), frame_5_start_run,
    dx(3), frame_6_start_run,
    dx(3), LABEL(runcyc1) frame_7_run,
    dx(5), frame_8_run,
    dx(1), snd(SND_FOOTSTEP), frame_9_run,
    dx(2), frame_10_run,
    dx(4), frame_11_run,
    dx(5), frame_12_run,
    dx(2), LABEL(runcyc7) snd(SND_FOOTSTEP), frame_13_run,
    dx(3), frame_14_run,
    dx(4), jmp(runcyc1),

    LABEL(stand) // stand
    act(actions_0_stand), frame_15_stand,
    jmp(stand), // goto "stand"

    LABEL(goalertstand) // alert stand
    act(actions_1_run_jump), LABEL(alertstand) frame_166_stand_inactive,
    jmp(alertstand), // goto "alertstand"

    LABEL(arise) // arise (skeleton)
    act(actions_5_bumped), dx(10), frame_177_spiked,
    frame_177_spiked,
    dx(-7), dy(-2), frame_178_chomped,
    dx(5), dy(2), frame_166_stand_inactive,
    dx(-1), jmp(ready), // goto "ready"

    // guard engarde
    LABEL(guardengarde)
    jmp(ready), // goto "ready"

    // engarde
    LABEL(engarde)
    act(actions_1_run_jump),
    dx(2), frame_207_draw_1,
    snd(SND_DRAW_SWORD), frame_208_draw_2,
    dx(2), frame_209_draw_3,
    dx(2), frame_210_draw_4,
    dx(3), LABEL(ready) act(actions_1_run_jump), snd(SND_SILENT), frame_158_stand_with_sword,
    frame_170_stand_with_sword,
    LABEL(ready_loop) frame_171_stand_with_sword,
    jmp(ready_loop), // goto ":loop"

    LABEL(stabbed)// stabbed
    act(actions_5_bumped), set_fall(-1, 0), frame_172_jumpfall_2,
    dx(-1), dy(1), frame_173_jumpfall_3,
    dx(-1), frame_174_jumpfall_4,
    dx(-1), dy(2),
    dx(-2), dy(1),
    dx(-5), dy(-4),
    jmp(guy8), // goto "guy8"

    // strike - advance
    LABEL(strikeadv)
    act(actions_1_run_jump), set_fall(1, 0), frame_155_post_strike_1,
    dx(2), frame_165_walk_with_sword,
    dx(-2), jmp(ready), // goto "ready"

    LABEL(strikeret)// strike - retreat
    act(actions_1_run_jump), set_fall(-1, 0), frame_155_post_strike_1,
    frame_156_post_strike_2,
    frame_157_walk_with_sword,
    frame_158_stand_with_sword,
    jmp(retreat), // goto "retreat"

    LABEL(advance) // advance
    act(actions_1_run_jump), set_fall(1, 0),
    dx(2), frame_163_fighting,
    dx(4), frame_164_fighting,
    frame_165_walk_with_sword,
    jmp(ready), // goto "ready"

    LABEL(fastadvance) // fast advance
    set_fall(1, 0), dx(4), frame_164_fighting,
    frame_165_walk_with_sword, dx(2),
    jmp(ready), // goto "ready"

    LABEL(retreat) // retreat
    act(actions_1_run_jump), set_fall(-1, 0), dx(-3), frame_160_fighting,
    dx(-2), frame_157_walk_with_sword,
    jmp(ready), // goto "ready"

    LABEL(strike) // strike
    act(actions_1_run_jump), set_fall(-1, 0), frame_168_back,
    LABEL(faststrike) act(actions_1_run_jump), frame_151_strike_1,
    LABEL(guy4) act(actions_1_run_jump), frame_152_strike_2,
    frame_153_strike_3,
    frame_154_poking,
    LABEL(guy7) act(actions_5_bumped), frame_155_post_strike_1,
    LABEL(guy8) act(actions_1_run_jump), frame_156_post_strike_2,
    frame_157_walk_with_sword,
    jmp(ready), // goto "ready"

    LABEL(blockedstrike)// blocked strike
    act(actions_1_run_jump), frame_167_blocked,
    jmp(guy7), // goto "guy7"

    // block to strike
    LABEL(blocktostrike)// "blocktostrike"
    frame_162_block_to_strike,
    jmp(guy4), // goto "guy4"

    LABEL(readyblock) // ready block
    frame_169_begin_block,
    LABEL(blocking) frame_150_parry,
    jmp(ready), // goto "ready"

    LABEL(striketoblock) // strike to block
    frame_159_fighting,
    frame_160_fighting,
    jmp(blocking), // goto "blocking"

    LABEL(landengarde) // land en garde
    act(actions_1_run_jump), SEQ_KNOCK_DOWN, jmp(ready), // goto "ready"

    LABEL(bumpengfwd) // bump en garde (forward)
    act(actions_5_bumped), dx(-2), jmp(ready), // goto "ready"

    LABEL(bumpengback) // bump en garde (back)
    act(actions_5_bumped), frame_160_fighting,
    frame_157_walk_with_sword,
    jmp(ready), // goto "ready"

    LABEL(flee) // flee
    act(actions_7_turn), dx(-8), jmp(turn), // goto "turn"

    LABEL(turnengarde) // turn en garde
    act(actions_5_bumped), SEQ_FLIP, dx(5), jmp(retreat), // goto "retreat"

    LABEL(alertturn) // alert turn (for enemies)
    act(actions_5_bumped), SEQ_FLIP, dx(18), jmp(goalertstand), // goto "goalertstand"

    LABEL(standjump) // standing jump
    act(actions_1_run_jump), frame_16_standing_jump_1,
    frame_17_standing_jump_2,
    dx(2), frame_18_standing_jump_3,
    dx(2), frame_19_standing_jump_4,
    dx(2), frame_20_standing_jump_5,
    dx(2), frame_21_standing_jump_6,
    dx(2), frame_22_standing_jump_7,
    dx(7), frame_23_standing_jump_8,
    dx(9), frame_24_standing_jump_9,
    dx(5), dy(-6), /* "sjland" */ LABEL(sjland) frame_25_standing_jump_10,
    dx(1), dy(6), frame_26_standing_jump_11,
    dx(4), SEQ_KNOCK_DOWN, snd(SND_FOOTSTEP), frame_27_standing_jump_12,
    dx(-3), frame_28_standing_jump_13,
    dx(5), frame_29_standing_jump_14,
    snd(SND_FOOTSTEP), frame_30_standing_jump_15,
    frame_31_standing_jump_16,
    frame_32_standing_jump_17,
    frame_33_standing_jump_18,
    dx(1), jmp(stand), // goto "stand"

    LABEL(runjump) // running jump
    act(actions_1_run_jump), snd(SND_FOOTSTEP), frame_34_start_run_jump_1,
    dx(5), frame_35_start_run_jump_2,
    dx(6), frame_36_start_run_jump_3,
    dx(3), frame_37_start_run_jump_4,
    dx(5), snd(SND_FOOTSTEP), frame_38_start_run_jump_5,
    dx(7), frame_39_start_run_jump_6,
    dx(12), dy(-3), frame_40_running_jump_1,
    dx(8), dy(-9), frame_41_running_jump_2,
    dx(8), dy(-2), frame_42_running_jump_3,
    dx(4), dy(11), frame_43_running_jump_4,
    dx(4), dy(3), /* "rjlandrun" */ LABEL(rjlandrun) frame_44_running_jump_5,
    dx(5), SEQ_KNOCK_DOWN, snd(SND_FOOTSTEP), jmp(runcyc1), // goto "runcyc1"

    LABEL(rdiveroll) // run dive roll
    act(actions_1_run_jump), dx(1), frame_107_fall_land_1,
    dx(2), dx(2), frame_108_fall_land_2,
    dx(2), frame_109_crouch,
    dx(2), frame_109_crouch,
    dx(2), /* ":crouch" */ LABEL(rdiveroll_crouch) frame_109_crouch,
    jmp(rdiveroll_crouch), // goto ":crouch"

    LABEL(sdiveroll)
    0x00, // stand dive roll; not implemented

    LABEL(crawl) // crawl
    act(actions_1_run_jump), dx(1), frame_110_stand_up_from_crouch_1,
    frame_111_stand_up_from_crouch_2,
    dx(2), frame_112_stand_up_from_crouch_3,
    dx(2), frame_108_fall_land_2,
    dx(2), /* ":crouch" */ LABEL(crawl_crouch) frame_109_crouch,
    jmp(crawl_crouch), // goto ":crouch"

    LABEL(turndraw) // turn draw
    act(actions_7_turn), SEQ_FLIP, dx(6), frame_45_turn,
    dx(1), frame_46_turn,
    jmp(engarde), // goto "engarde"

    LABEL(turn) // turn
    act(actions_7_turn), SEQ_FLIP, dx(6), frame_45_turn,
    dx(1), frame_46_turn,
    dx(2), frame_47_turn,
    dx(-1), /* "finishturn" */ frame_48_turn,
    dx(1), frame_49_turn,
    dx(-2), frame_50_turn,
    frame_51_turn,
    frame_52_turn,
    jmp(stand), // goto "stand"

    LABEL(turnrun) // turnrun (from frame 48)
    act(actions_1_run_jump), dx(-1), jmp(runstt1), // goto "runstt1"

    LABEL(runturn) // runturn
    act(actions_1_run_jump), dx(1), frame_53_runturn,
    dx(1), snd(SND_FOOTSTEP), frame_54_runturn,
    dx(8), frame_55_runturn,
    snd(SND_FOOTSTEP), frame_56_runturn,
    dx(7), frame_57_runturn,
    dx(3), frame_58_runturn,
    dx(1), frame_59_runturn,
    frame_60_runturn,
    dx(2), frame_61_runturn,
    dx(-1), frame_62_runturn,
    frame_63_runturn,
    frame_64_runturn,
    dx(-1), frame_65_runturn,
    dx(-14), SEQ_FLIP, jmp(runcyc7), // goto "runcyc7"

    LABEL(fightfall) // fightfall (backward)
    act(actions_3_in_midair), dy(-1), frame_102_start_fall_1,
    dx(-2), dy(6), frame_103_start_fall_2,
    dx(-2), dy(9), frame_104_start_fall_3,
    dx(-1), dy(12), frame_105_start_fall_4,
    dx(-3), set_fall(0, 15), jmp(freefall), // goto "freefall"

    LABEL(efightfall) // enemy fight fall
    act(actions_3_in_midair), dy(-1), dx(-2), frame_102_start_fall_1,
    dx(-3), dy(6), frame_103_start_fall_2,
    dx(-3), dy(9), frame_104_start_fall_3,
    dx(-2), dy(12), frame_105_start_fall_4,
    dx(-3), set_fall(0, 15), jmp(freefall), // goto "freefall"

    LABEL(efightfallfwd) // enemy fight fall forward
    act(actions_3_in_midair), dx(1), dy(-1), frame_102_start_fall_1,
    dx(2), dy(6), frame_103_start_fall_2,
    dx(-1), dy(9), frame_104_start_fall_3,
    dy(12), frame_105_start_fall_4,
    dx(-2), set_fall(1, 15), jmp(freefall), // goto "freefall"

    LABEL(stepfall) // stepfall
    act(actions_3_in_midair), dx(1), dy(3), jmp_if_shad_feather(stepfloat), // goto "stepfloat"
    /* "fall1" */ LABEL(fall1) frame_102_start_fall_1,
    dx(2), dy(6), frame_103_start_fall_2,
    dx(-1), dy(9), frame_104_start_fall_3,
    dy(12), frame_105_start_fall_4,
    dx(-2), set_fall(1, 15), jmp(freefall), // goto "freefall"

    LABEL(patchfall) // patchfall
    dx(-1), dy(-3), jmp(fall1), // goto "fall1"

    LABEL(stepfall2) // stepfall2 (from frame 12)
    dx(1), jmp(stepfall), // goto "stepfall"

    LABEL(stepfloat) // stepfloat
    frame_102_start_fall_1,
    dx(2), dy(3), frame_103_start_fall_2,
    dx(-1), dy(4), frame_104_start_fall_3,
    dy(5), frame_105_start_fall_4,
    dx(-2), set_fall(1, 6), jmp(freefall), // goto "freefall"

    LABEL(jumpfall) // jump fall (from standing jump)
    act(actions_3_in_midair), dx(1), dy(3), frame_102_start_fall_1,
    dx(2), dy(6), frame_103_start_fall_2,
    dx(1), dy(9), frame_104_start_fall_3,
    dx(2), dy(12), frame_105_start_fall_4,
    set_fall(2, 15), jmp(freefall), // goto "freefall"

    LABEL(rjumpfall) // running jump fall
    act(actions_3_in_midair), dx(1), dy(3), frame_102_start_fall_1,
    dx(3), dy(6), frame_103_start_fall_2,
    dx(1), dy(9), frame_104_start_fall_3,
    dx(3), dy(12), frame_105_start_fall_4,
    set_fall(3, 15), jmp(freefall), // goto "freefall"

    LABEL(jumphangMed) // jumphang (medium: DX = 0)
    act(actions_1_run_jump), frame_67_start_jump_up_1,
    frame_68_start_jump_up_2, frame_69_start_jump_up_3, frame_70_jumphang, frame_71_jumphang,
    frame_72_jumphang, frame_73_jumphang, frame_74_jumphang, frame_75_jumphang, frame_76_jumphang,
    frame_77_jumphang,
    act(actions_2_hang_climb), frame_78_jumphang, frame_79_jumphang, frame_80_jumphang,
    jmp(hang), // goto "hang"

    LABEL(jumphangLong)// jumphang (long: DX = 4)
    act(actions_1_run_jump), frame_67_start_jump_up_1,
    frame_68_start_jump_up_2, frame_69_start_jump_up_3, frame_70_jumphang, frame_71_jumphang,
    frame_72_jumphang, frame_73_jumphang, frame_74_jumphang, frame_75_jumphang, frame_76_jumphang,
    frame_77_jumphang,
    act(actions_2_hang_climb), dx(1), frame_78_jumphang,
    dx(2), frame_79_jumphang,
    dx(1), frame_80_jumphang,
    jmp(hang), // goto "hang"

    LABEL(jumpbackhang) // jumpbackhang
    act(actions_1_run_jump), frame_67_start_jump_up_1,
    frame_68_start_jump_up_2, frame_69_start_jump_up_3, frame_70_jumphang, frame_71_jumphang,
    frame_72_jumphang, frame_73_jumphang, frame_74_jumphang, frame_75_jumphang, frame_76_jumphang,
    dx(-1), frame_77_jumphang,
    act(actions_2_hang_climb), dx(-2), frame_78_jumphang,
    dx(-1), frame_79_jumphang,
    dx(-1), frame_80_jumphang,
    jmp(hang), // goto "hang"

    LABEL(hang) // hang
    act(actions_2_hang_climb), frame_91_hanging_5,
    /* "hang1" */ LABEL(hang1) frame_90_hanging_4, frame_89_hanging_3, frame_88_hanging_2,
    frame_87_hanging_1, frame_87_hanging_1, frame_87_hanging_1, frame_88_hanging_2,
    frame_89_hanging_3, frame_90_hanging_4, frame_91_hanging_5, frame_92_hanging_6,
    frame_93_hanging_7, frame_94_hanging_8, frame_95_hanging_9, frame_96_hanging_10,
    frame_97_hanging_11, frame_98_hanging_12, frame_99_hanging_13, frame_97_hanging_11,
    frame_96_hanging_10, frame_95_hanging_9, frame_94_hanging_8, frame_93_hanging_7,
    frame_92_hanging_6, frame_91_hanging_5, frame_90_hanging_4, frame_89_hanging_3,
    frame_88_hanging_2, frame_87_hanging_1, frame_88_hanging_2, frame_89_hanging_3,
    frame_90_hanging_4, frame_91_hanging_5, frame_92_hanging_6, frame_93_hanging_7,
    frame_94_hanging_8, frame_95_hanging_9, frame_96_hanging_10, frame_95_hanging_9,
    frame_94_hanging_8, frame_93_hanging_7, frame_92_hanging_6,
    jmp_if_shad_feather(hangdrop), // goto "hangdrop"

    LABEL(hangstraight) // hangstraight
    act(actions_6_hang_straight), frame_92_hanging_6, // Apple II source has a bump sound here
    frame_93_hanging_7, frame_93_hanging_7, frame_92_hanging_6, frame_92_hanging_6,
    /* ":loop" */ LABEL(hangstraight_loop) frame_91_hanging_5,
    jmp(hangstraight_loop), // goto ":loop"

    LABEL(climbfail) // climbfail
    frame_135_climbing_1, frame_136_climbing_2, frame_137_climbing_3, frame_137_climbing_3,
    frame_138_climbing_4, frame_138_climbing_4, frame_138_climbing_4, frame_138_climbing_4,
    frame_137_climbing_3, frame_136_climbing_2, frame_135_climbing_1,
    dx(-7), jmp(hangdrop), // goto "hangdrop"

    LABEL(climbdown) // climbdown
    act(actions_1_run_jump), frame_148_climbing_14,
    frame_145_climbing_11, frame_144_climbing_10, frame_143_climbing_9, frame_142_climbing_8,
    frame_141_climbing_7,
    dx(-5), dy(63), SEQ_DOWN, act(actions_3_in_midair), frame_140_climbing_6,
    frame_138_climbing_4, frame_136_climbing_2,
    frame_91_hanging_5,
    act(actions_2_hang_climb), jmp(hang1), // goto "hang1"

    LABEL(climbup) // climbup
    act(actions_1_run_jump), frame_135_climbing_1,
    frame_136_climbing_2, frame_137_climbing_3, frame_138_climbing_4,
    frame_139_climbing_5, frame_140_climbing_6,
    dx(5), dy(-63), SEQ_UP, frame_141_climbing_7,
    frame_142_climbing_8, frame_143_climbing_9, frame_144_climbing_10, frame_145_climbing_11,
    frame_146_climbing_12, frame_147_climbing_13, frame_148_climbing_14,
    act(actions_5_bumped), // to clear flags
    frame_149_climbing_15,
    act(actions_1_run_jump), frame_118_stand_up_from_crouch_9, frame_119_stand_up_from_crouch_10,
    dx(1), jmp(stand), // goto "stand"

    LABEL(hangdrop) // hangdrop
    frame_81_hangdrop_1, frame_82_hangdrop_2,
    act(actions_5_bumped), frame_83_hangdrop_3,
    act(actions_1_run_jump), SEQ_KNOCK_DOWN, snd(SND_FOOTSTEP),
    frame_84_hangdrop_4, frame_85_hangdrop_5,
    dx(3), jmp(stand), // goto "stand"

    LABEL(hangfall) // hangfall
    act(actions_3_in_midair), frame_81_hangdrop_1,
    dy(6), frame_81_hangdrop_1,
    dy(9), frame_81_hangdrop_1,
    dy(12), dx(2), set_fall(0, 12), jmp(freefall), // goto "freefall"

    LABEL(freefall) // freefall
    act(actions_4_in_freefall), /* ":loop" */ LABEL(freefall_loop) frame_106_fall,
    jmp(freefall_loop), // goto :loop

    LABEL(runstop) // runstop
    act(actions_1_run_jump), frame_53_runturn,
    dx(2), snd(SND_FOOTSTEP), frame_54_runturn,
    dx(7), frame_55_runturn,
    snd(SND_FOOTSTEP), frame_56_runturn,
    dx(2), frame_49_turn,
    dx(-2), frame_50_turn,
    frame_51_turn, frame_52_turn,
    jmp(stand), // goto "stand"

    LABEL(jumpup) // jump up (and touch ceiling)
    act(actions_1_run_jump), frame_67_start_jump_up_1,
    frame_68_start_jump_up_2, frame_69_start_jump_up_3, frame_70_jumphang, frame_71_jumphang,
    frame_72_jumphang, frame_73_jumphang, frame_74_jumphang, frame_75_jumphang,
    frame_76_jumphang, frame_77_jumphang, frame_78_jumphang,
    act(actions_0_stand), SEQ_KNOCK_UP, frame_79_jumphang,
    jmp(landjump), // goto "landjump"

    LABEL(highjump) // highjump (no ceiling above)
    act(actions_1_run_jump), frame_67_start_jump_up_1,
    frame_68_start_jump_up_2, frame_69_start_jump_up_3, frame_70_jumphang, frame_71_jumphang,
    frame_72_jumphang, frame_73_jumphang, frame_74_jumphang, frame_75_jumphang,
    frame_76_jumphang, frame_77_jumphang, frame_78_jumphang, frame_79_jumphang,
    dy(-4), frame_79_jumphang,
    dy(-2), frame_79_jumphang,
    frame_79_jumphang,
    dy(2), frame_79_jumphang,
    dy(4), jmp(landjump), // goto "landjump"

    LABEL(superhijump) // superhijump (when weightless)
    frame_67_start_jump_up_1, frame_68_start_jump_up_2, frame_69_start_jump_up_3, frame_70_jumphang,
    frame_71_jumphang, frame_72_jumphang, frame_73_jumphang, frame_74_jumphang,
    frame_75_jumphang, frame_76_jumphang,
    dy(-1), frame_77_jumphang,
    dy(-3), frame_78_jumphang,
    dy(-4), frame_79_jumphang,
    dy(-10), frame_79_jumphang,
    dy(-8), frame_79_jumphang,
    dy(-6), frame_79_jumphang,
    dy(-5), frame_79_jumphang,
    dy(-4), frame_79_jumphang,
    dy(-4), frame_79_jumphang,
    dy(-4), frame_79_jumphang,
    dy(-3), frame_79_jumphang,
    dy(-2), frame_79_jumphang,
    dy(-2), frame_79_jumphang,
    dy(-1), frame_79_jumphang,
    dy(-1), frame_79_jumphang,
    dy(-1), frame_79_jumphang,
    frame_79_jumphang, frame_79_jumphang, frame_79_jumphang,
    dy(1), frame_79_jumphang,
    dy(1), frame_79_jumphang,
    dy(2), frame_79_jumphang,
    dy(2), frame_79_jumphang,
    dy(3), frame_79_jumphang,
    dy(4), frame_79_jumphang,
    dy(5), frame_79_jumphang,
    dy(6), frame_79_jumphang,
    set_fall(0, 6), jmp(freefall), // goto "freefall"

    LABEL(fallhang) // fall hang
    act(actions_3_in_midair), frame_80_jumphang,
    jmp(hang), // goto "hang"

    LABEL(bump) // bump
    act(actions_5_bumped), dx(-4), frame_50_turn,
    frame_51_turn, frame_52_turn,
    jmp(stand), // goto "stand"

    LABEL(bumpfall) // bumpfall
    /* action is patched to actions_3_in_midair */
    act(actions_3_in_midair), dx(1), dy(3), jmp_if_shad_feather(bumpfloat),
    frame_102_start_fall_1,
    dx(2), dy(6), frame_103_start_fall_2,
    dx(-1), dy(9), frame_104_start_fall_3,
    dy(12), frame_105_start_fall_4,
    dx(-2), set_fall(0, 15), jmp(freefall), // goto "freefall"

    LABEL(bumpfloat) // bumpfloat
    frame_102_start_fall_1,
    dx(2), dy(3), frame_103_start_fall_2,
    dx(-1), dy(4), frame_104_start_fall_3,
    dy(5), frame_105_start_fall_4,
    dx(-2), set_fall(0, 6), jmp(freefall), // goto "freefall"

    LABEL(hardbump) // hard bump
    act(actions_5_bumped), dx(-1), dy(-4), frame_102_start_fall_1,
    dx(-1), dy(3), dx(-3), dy(1), SEQ_KNOCK_DOWN,
    dx(1), snd(SND_FOOTSTEP), frame_107_fall_land_1,
    dx(2), frame_108_fall_land_2,
    snd(SND_FOOTSTEP), frame_109_crouch,
    jmp(standup), // goto "standup"

    LABEL(testfoot) // test foot
    frame_121_stepping_1,
    dx(1), frame_122_stepping_2,
    frame_123_stepping_3,
    dx(2), frame_124_stepping_4,
    dx(4), frame_125_stepping_5,
    dx(3), frame_126_stepping_6,
    dx(-4), frame_86_test_foot,
    snd(SND_FOOTSTEP), SEQ_KNOCK_DOWN, dx(-4), frame_116_stand_up_from_crouch_7,
    dx(-2), frame_117_stand_up_from_crouch_8,
    frame_118_stand_up_from_crouch_9,
    frame_119_stand_up_from_crouch_10,
    jmp(stand), // goto "stand"

    LABEL(stepback) // step back
    dx(-5), jmp(stand), // goto "stand"

    LABEL(step14) // step forward 14 pixels
    act(actions_1_run_jump), frame_121_stepping_1,
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(3), frame_124_stepping_4,
    dx(4), frame_125_stepping_5,
    dx(3), frame_126_stepping_6,
    dx(-1), dx(3), frame_127_stepping_7,
    frame_128_stepping_8, frame_129_stepping_9, frame_130_stepping_10,
    frame_131_stepping_11, frame_132_stepping_12,
    jmp(stand), // goto "stand"

    LABEL(step13) // step forward 13 pixels
    act(actions_1_run_jump), frame_121_stepping_1,
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(3), frame_124_stepping_4,
    dx(4), frame_125_stepping_5,
    dx(3), frame_126_stepping_6,
    dx(-1), dx(2), frame_127_stepping_7,
    frame_128_stepping_8, frame_129_stepping_9, frame_130_stepping_10,
    frame_131_stepping_11, frame_132_stepping_12,
    jmp(stand), // goto "stand"

    LABEL(step12) // step forward 12 pixels
    act(actions_1_run_jump), frame_121_stepping_1,
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(3), frame_124_stepping_4,
    dx(4), frame_125_stepping_5,
    dx(3), frame_126_stepping_6,
    dx(-1), dx(1), frame_127_stepping_7,
    frame_128_stepping_8, frame_129_stepping_9, frame_130_stepping_10,
    frame_131_stepping_11, frame_132_stepping_12,
    jmp(stand), // goto "stand"

    LABEL(step11) // step forward 11 pixels (normal step)
    act(actions_1_run_jump), frame_121_stepping_1,
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(3), frame_124_stepping_4,
    dx(4), frame_125_stepping_5,
    dx(3), frame_126_stepping_6,
    dx(-1), frame_127_stepping_7,
    frame_128_stepping_8, frame_129_stepping_9, frame_130_stepping_10,
    frame_131_stepping_11, frame_132_stepping_12,
    jmp(stand), // goto "stand"

    LABEL(step10) // step forward 10 pixels
    act(actions_1_run_jump), frame_121_stepping_1,
    dx(1), /* "step10a "*/ LABEL(step10a) frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(3), frame_124_stepping_4,
    dx(4), frame_125_stepping_5,
    dx(3), frame_126_stepping_6,
    dx(-2), frame_128_stepping_8,
    frame_129_stepping_9, frame_130_stepping_10, frame_131_stepping_11, frame_132_stepping_12,
    jmp(stand), // goto "stand"

    LABEL(step9) // step forward 9 pixels
    act(actions_1_run_jump), frame_121_stepping_1,
    jmp(step10a), // goto "step10a"

    LABEL(step8) // step forward 8 pixels
    act(actions_1_run_jump), frame_121_stepping_1,
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(3), frame_124_stepping_4,
    dx(4), frame_125_stepping_5,
    dx(-1), frame_127_stepping_7,
    frame_128_stepping_8, frame_129_stepping_9, frame_130_stepping_10, frame_131_stepping_11,
    frame_132_stepping_12,
    jmp(stand), // goto "stand"

    LABEL(step7) // step forward 7 pixels
    act(actions_1_run_jump), frame_121_stepping_1,
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(3), frame_124_stepping_4,
    dx(2), frame_129_stepping_9,
    frame_130_stepping_10, frame_131_stepping_11, frame_132_stepping_12,
    jmp(stand), // goto "stand"

    LABEL(step6) // step forward 6 pixels
    act(actions_1_run_jump), frame_121_stepping_1,
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(2), frame_124_stepping_4,
    dx(2), frame_129_stepping_9,
    frame_130_stepping_10, frame_131_stepping_11, frame_132_stepping_12,
    jmp(stand), // goto "stand"

    LABEL(step5) // step forward 5 pixels
    act(actions_1_run_jump), frame_121_stepping_1,
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(2), frame_124_stepping_4,
    dx(1), frame_129_stepping_9,
    frame_130_stepping_10, frame_131_stepping_11, frame_132_stepping_12,
    jmp(stand), // goto "stand"

    LABEL(step4) // step forward 4 pixels
    act(actions_1_run_jump), frame_121_stepping_1,
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(2), frame_131_stepping_11,
    frame_132_stepping_12,
    jmp(stand), // goto "stand"

    LABEL(step3) // step forward 3 pixels
    act(actions_1_run_jump), frame_121_stepping_1,
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(1), frame_131_stepping_11,
    frame_132_stepping_12,
    jmp(stand), // goto "stand"

    LABEL(step2) // step forward 2 pixels
    act(actions_1_run_jump), frame_121_stepping_1,
    dx(1), frame_122_stepping_2,
    dx(1), frame_132_stepping_12,
    jmp(stand), // goto "stand"

    LABEL(step1) // step forward 1 pixel
    act(actions_1_run_jump), frame_121_stepping_1,
    dx(1), frame_132_stepping_12,
    jmp(stand), // goto "stand"

    LABEL(stoop) // stoop
    act(actions_1_run_jump), dx(1), frame_107_fall_land_1,
    dx(2), frame_108_fall_land_2,
    /* ":crouch" */ LABEL(stoop_crouch) frame_109_crouch,
    jmp(stoop_crouch), // goto ":crouch

    LABEL(standup) // stand up
    act(actions_5_bumped), dx(1), frame_110_stand_up_from_crouch_1,
    frame_111_stand_up_from_crouch_2,
    dx(2), frame_112_stand_up_from_crouch_3,
    frame_113_stand_up_from_crouch_4,
    dx(1), frame_114_stand_up_from_crouch_5,
    frame_115_stand_up_from_crouch_6, frame_116_stand_up_from_crouch_7,
    dx(-4), frame_117_stand_up_from_crouch_8,
    frame_118_stand_up_from_crouch_9, frame_119_stand_up_from_crouch_10,
    jmp(stand), // goto "stand"

    LABEL(pickupsword) // pick up sword
    act(actions_1_run_jump), SEQ_GET_ITEM, 1, frame_229_found_sword,
    frame_229_found_sword, frame_229_found_sword, frame_229_found_sword, frame_229_found_sword,
    frame_229_found_sword, frame_230_sheathe, frame_231_sheathe, frame_232_sheathe,
    jmp(pickupsheathe), // goto "pickupsheathe"

    LABEL(resheathe) // resheathe
    act(actions_1_run_jump), dx(-5), frame_233_sheathe,
    frame_234_sheathe, frame_235_sheathe, frame_236_sheathe, frame_237_sheathe,
    frame_238_sheathe, frame_239_sheathe, frame_240_sheathe, frame_133_sheathe,
    frame_133_sheathe, frame_134_sheathe, frame_134_sheathe, frame_134_sheathe,
    frame_48_turn,
    dx(1), frame_49_turn,
    dx(-2), act(actions_5_bumped), frame_50_turn,
    act(actions_1_run_jump), frame_51_turn,
    frame_52_turn,
    jmp(stand), // goto "stand"

    LABEL(fastsheathe) // fast sheathe
    act(actions_1_run_jump), dx(-5), frame_234_sheathe,
    frame_236_sheathe, frame_238_sheathe, frame_240_sheathe, frame_134_sheathe,
    dx(-1), jmp(stand), // goto "stand"

    LABEL(drinkpotion) // drink potion
    act(actions_1_run_jump), dx(4), frame_191_drink,
    frame_192_drink, frame_193_drink, frame_194_drink, frame_195_drink,
    frame_196_drink, frame_197_drink, snd(SND_DRINK),
    frame_198_drink, frame_199_drink, frame_200_drink, frame_201_drink,
    frame_202_drink, frame_203_drink, frame_204_drink, frame_205_drink,
    frame_205_drink, frame_205_drink,
    SEQ_GET_ITEM, 1, frame_205_drink,
    frame_205_drink, frame_201_drink, frame_198_drink,
    dx(0), jmp(putdownpotion), // goto "putdownpotion"

    LABEL(softland)// soft land
    act(actions_5_bumped), SEQ_KNOCK_DOWN, dx(1), frame_107_fall_land_1,
    dx(2), frame_108_fall_land_2,
    act(actions_1_run_jump), /* ":crouch" */ LABEL(softland_crouch) frame_109_crouch,
    jmp(softland_crouch), // goto ":crouch"

    LABEL(landrun) // land run
    act(actions_1_run_jump), dy(-2), dx(1), frame_107_fall_land_1,
    dx(2), frame_108_fall_land_2,
    frame_109_crouch,
    dx(1), frame_110_stand_up_from_crouch_1,
    frame_111_stand_up_from_crouch_2,
    dx(2), frame_112_stand_up_from_crouch_3,
    frame_113_stand_up_from_crouch_4,
    dx(1), dy(1), frame_114_stand_up_from_crouch_5,
    dy(1), frame_115_stand_up_from_crouch_6,
    dx(-2), jmp(runstt4), // goto "runstt4"

    LABEL(medland) // medium land (1.5 - 2 stories)
    act(actions_5_bumped), SEQ_KNOCK_DOWN, dy(0), dx(1), dx(2), frame_108_fall_land_2,
    frame_109_crouch, frame_109_crouch, frame_109_crouch, frame_109_crouch,
    frame_109_crouch, frame_109_crouch, frame_109_crouch, frame_109_crouch,
    frame_109_crouch, frame_109_crouch, frame_109_crouch, frame_109_crouch,
    frame_109_crouch, frame_109_crouch, frame_109_crouch, frame_109_crouch,
    frame_109_crouch, frame_109_crouch, frame_109_crouch, frame_109_crouch,
    frame_109_crouch, frame_109_crouch, frame_109_crouch, frame_109_crouch,
    frame_109_crouch, frame_109_crouch, frame_109_crouch, frame_109_crouch,
    frame_109_crouch,
    dx(1), frame_110_stand_up_from_crouch_1,
    frame_110_stand_up_from_crouch_1, frame_110_stand_up_from_crouch_1, frame_111_stand_up_from_crouch_2,
    dx(2), frame_112_stand_up_from_crouch_3,
    frame_113_stand_up_from_crouch_4,
    dx(1), dy(0), frame_114_stand_up_from_crouch_5,
    dy(0), frame_115_stand_up_from_crouch_6,
    frame_116_stand_up_from_crouch_7,
    dx(-4), frame_117_stand_up_from_crouch_8,
    frame_118_stand_up_from_crouch_9, frame_119_stand_up_from_crouch_10,
    jmp(stand), // goto "stand"

    LABEL(hardland) // hard land (splat!; >2 stories)
    act(actions_5_bumped), SEQ_KNOCK_DOWN, dy(0), dx(3), frame_185_dead,
    SEQ_DIE, /* ":dead" */ LABEL(hardland_dead) frame_185_dead,
    jmp(hardland_dead), // goto ":dead"

    LABEL(stabkill) // stabkill
    act(actions_5_bumped), jmp(dropdead), // goto "dropdead"

    LABEL(dropdead) // dropdead
    act(actions_1_run_jump), SEQ_DIE, frame_179_collapse_1,
    frame_180_collapse_2, frame_181_collapse_3, frame_182_collapse_4,
    dx(1), frame_183_collapse_5,
    dx(-4), /* ":dead" */ LABEL(dropdead_dead) frame_185_dead,
    jmp(dropdead_dead), // goto ":dead"

    LABEL(impale) // impale
    act(actions_1_run_jump), SEQ_KNOCK_DOWN, dx(4), frame_177_spiked,
    SEQ_DIE, /* ":dead" */ LABEL(impale_dead) frame_177_spiked,
    jmp(impale_dead), // goto ":dead"

    LABEL(halve) // halve
    act(actions_1_run_jump), frame_178_chomped,
    SEQ_DIE, /* ":dead" */ LABEL(halve_dead) frame_178_chomped,
    jmp(halve_dead), // goto ":dead"

    LABEL(crush) // crush
    jmp(medland), // goto "medland"

    LABEL(deadfall) // deadfall
    set_fall(0, 0), act(actions_4_in_freefall), /* ":loop"*/ LABEL(deadfall_loop) frame_185_dead,
    jmp(deadfall_loop), // goto ":loop"

    LABEL(climbstairs) // climb stairs
    act(actions_5_bumped),
    dx(-5), dy(-1), snd(SND_FOOTSTEP), frame_217_exit_stairs_1,
    frame_218_exit_stairs_2, frame_219_exit_stairs_3,
    dx(1), frame_220_exit_stairs_4,
    dx(-4), dy(-3), snd(SND_FOOTSTEP), frame_221_exit_stairs_5,
    dx(-4), dy(-2), frame_222_exit_stairs_6,
    dx(-2), dy(-3), frame_223_exit_stairs_7,
    dx(-3), dy(-8), snd(SND_LEVEL), snd(SND_FOOTSTEP), frame_224_exit_stairs_8,
    dx(-1), dy(-1), frame_225_exit_stairs_9,
    dx(-3), dy(-4), frame_226_exit_stairs_10,
    dx(-1), dy(-5), snd(SND_FOOTSTEP), frame_227_exit_stairs_11,
    dx(-2), dy(-1), frame_228_exit_stairs_12,
    frame_0,
    snd(SND_FOOTSTEP), frame_0, frame_0, frame_0, // these footsteps are only heard when the music is off
    snd(SND_FOOTSTEP), frame_0, frame_0, frame_0,
    snd(SND_FOOTSTEP), frame_0, frame_0, frame_0,
    snd(SND_FOOTSTEP), SEQ_END_LEVEL, /* ":loop" */ LABEL(climbstairs_loop) frame_0,
    jmp(climbstairs_loop), // goto ":loop"

    LABEL(Vstand) // Vizier: stand
    frame_54_vizier_stand_1,
    jmp(Vstand), // goto "Vstand"

    LABEL(Vraise) // Vizier: raise arms
    frame_85_vizier_cast_19, frame_67_vizier_cast_1, frame_67_vizier_cast_1, frame_67_vizier_cast_1,
    frame_67_vizier_cast_1, frame_67_vizier_cast_1, frame_67_vizier_cast_1, frame_67_vizier_cast_1,
    frame_67_vizier_cast_1, frame_67_vizier_cast_1, frame_67_vizier_cast_1, frame_68_vizier_cast_2,
    frame_69_vizier_cast_3, frame_70_vizier_cast_4, frame_71_vizier_cast_5, frame_72_vizier_cast_6,
    frame_73_vizier_cast_7, frame_74_vizier_cast_8, frame_75_vizier_cast_9, frame_83_vizier_cast_17,
    frame_84_vizier_cast_18, /* ":loop" */ LABEL(Vraise_loop) frame_76_vizier_cast_10,
    jmp(Vraise_loop), // goto ":loop"

    LABEL(Vwalk) // Vizier: walk
    dx(1), /* "Vwalk1" */ LABEL(Vwalk1) frame_48_vizier_walk_1,
    dx(2), /* "Vwalk2" */ LABEL(Vwalk2) frame_49_vizier_walk_2,
    dx(6), frame_50_vizier_walk_3,
    dx(1), frame_51_vizier_walk_4a,
    dx(-1), frame_52_vizier_walk_5,
    dx(1), frame_53_vizier_walk_6,
    dx(1), jmp(Vwalk1), // goto "Vwalk1"

    LABEL(Vstop) // Vizier: stop
    dx(1), frame_55_vizier_stand_2,
    frame_56_vizier_stand_3,
    jmp(Vstand),

    LABEL(Vexit) // Vizier: lower arms, turn & exit ("Vexit")
    frame_77_vizier_cast_11, frame_78_vizier_cast_12, frame_79_vizier_cast_13, frame_80_vizier_cast_14,
    frame_81_vizier_cast_15, frame_82_vizier_cast_16,
    dx(1), frame_54_vizier_stand_1,
    frame_54_vizier_stand_1, frame_54_vizier_stand_1, frame_54_vizier_stand_1, frame_54_vizier_stand_1,
    frame_54_vizier_stand_1, frame_57_vizier_turn_1, frame_58_vizier_turn_2, frame_59_vizier_turn_3,
    frame_60_vizier_turn_4, frame_61_vizier_turn_5,
    dx(2), frame_62_vizier_turn_6,
    dx(-1), frame_63_vizier_turn_7,
    dx(-3), frame_64_vizier_turn_8,
    frame_65_vizier_turn_9,
    dx(-1), frame_66_vizier_turn_10,
    SEQ_FLIP, dx(16), dx(3), jmp(Vwalk2), // goto "Vwalk2"

    // Princess: stand
    LABEL(Pstand) frame_11_princess_stand,
    jmp(Pstand), // goto "Pstand"

    LABEL(Palert) // Princess: alert
    frame_2_princess_turn_1, frame_3_princess_turn_2, frame_4_princess_turn_3, frame_5_princess_turn_4,
    frame_6_princess_turn_5, frame_7_princess_turn_6, frame_8_princess_turn_7, frame_9_princess_turn_8,
    SEQ_FLIP, dx(8), frame_11_princess_stand,
    jmp(Pstand),

    LABEL(Pstepback) // Princess: step back
    SEQ_FLIP, dx(11), frame_12_princess_step_back,
    dx(1), frame_13_princess_step_back,
    dx(1), frame_14_princess_step_back,
    dx(3), frame_15_princess_step_back,
    dx(1), frame_16_princess_step_back,
    /* ":loop" */ LABEL(Pstepback_loop) frame_17_princess_step_back_stand,
    jmp(Pstepback_loop), // goto ":loop"

    LABEL(Plie) // Princess lying on cushions ("Plie")
    frame_19_princess_lie,
    jmp(Plie), // goto "Plie"

    LABEL(Pwaiting) // Princess: waiting
    /* ":loop" */ frame_20_embrace_1,
    jmp(Pwaiting), // goto ":loop"

    LABEL(Pembrace)// Princess: embrace
    frame_21_embrace_2,
    dx(1), frame_22_embrace_3,
    frame_23_embrace_4, frame_24_embrace_5,
    dx(1), frame_25_embrace_6,
    dx(-3), frame_26_embrace_7,
    dx(-2), frame_27_embrace_8,
    dx(-4), frame_28_embrace_9,
    dx(-3), frame_29_embrace_10,
    dx(-2), frame_30_embrace_11,
    dx(-3), frame_31_embrace_12,
    dx(-1), frame_32_embrace_13,
    /* ":loop" */ LABEL(Pembrace_loop) frame_33_embrace_14,
    jmp(Pembrace_loop), // goto ":loop"

    LABEL(Pstroke) // Princess: stroke mouse
    /* ":loop" */ frame_37_princess_rise_4,
    jmp(Pstroke), // goto ":loop"

    LABEL(Prise) // Princess: rise
    frame_37_princess_rise_4, frame_38_princess_rise_5, frame_39_princess_rise_6, frame_40_princess_rise_7,
    frame_41_princess_rise_8, frame_42_princess_rise_9, frame_43_princess_rise_10, frame_44_princess_rise_11,
    frame_45_princess_rise_12, frame_46_princess_rise_13, frame_47_princess_rise_14,
    SEQ_FLIP, dx(12), /* ":loop" */ LABEL(Prise_loop) frame_11_princess_stand,
    jmp(Prise_loop), // goto ":loop"

    LABEL(Pcrouch) // Princess: crouch & stroke mouse
    frame_11_princess_stand, frame_11_princess_stand,
    SEQ_FLIP, dx(13), frame_47_princess_rise_14,
    frame_46_princess_rise_13, frame_45_princess_rise_12, frame_44_princess_rise_11, frame_43_princess_rise_10,
    frame_42_princess_rise_9, frame_41_princess_rise_8, frame_40_princess_rise_7, frame_39_princess_rise_6,
    frame_38_princess_rise_5, frame_37_princess_rise_4,
    frame_36_princess_rise_3, frame_36_princess_rise_3, frame_36_princess_rise_3,
    frame_35_princess_rise_2, frame_35_princess_rise_2, frame_35_princess_rise_2,
    frame_34_princess_rise_1, frame_34_princess_rise_1, frame_34_princess_rise_1,
    frame_34_princess_rise_1, frame_34_princess_rise_1, frame_34_princess_rise_1, frame_34_princess_rise_1,
    frame_35_princess_rise_2, frame_35_princess_rise_2,
    frame_36_princess_rise_3, frame_36_princess_rise_3, frame_36_princess_rise_3,
    frame_35_princess_rise_2, frame_35_princess_rise_2, frame_35_princess_rise_2,
    frame_34_princess_rise_1, frame_34_princess_rise_1, frame_34_princess_rise_1,
    frame_34_princess_rise_1, frame_34_princess_rise_1, frame_34_princess_rise_1, frame_34_princess_rise_1,
    frame_35_princess_rise_2, frame_35_princess_rise_2,
    frame_36_princess_rise_3, frame_36_princess_rise_3, frame_36_princess_rise_3,
    frame_35_princess_rise_2, frame_35_princess_rise_2, frame_35_princess_rise_2,
    frame_34_princess_rise_1, frame_34_princess_rise_1, frame_34_princess_rise_1,
    frame_34_princess_rise_1, frame_34_princess_rise_1, frame_34_princess_rise_1,
    frame_34_princess_rise_1, frame_34_princess_rise_1, frame_34_princess_rise_1,
    frame_35_princess_rise_2, frame_35_princess_rise_2, frame_35_princess_rise_2,
    /* ":loop" */ LABEL(Pcrouch_loop) frame_36_princess_rise_3,
    jmp(Pcrouch_loop), // goto ":loop"

    LABEL(Pslump) // Princess: slump shoulders
    frame_1_princess_slump_1, /* ":loop" */ LABEL(Pslump_loop) frame_18_princess_slump_2,
    jmp(Pslump_loop), // goto ":loop"

    LABEL(Mscurry) // Mouse: scurry
    act(actions_1_run_jump), /* "Mscurry1" */ LABEL(Mscurry1) frame_186_mouse_1,
    dx(5), frame_186_mouse_1,
    dx(3), frame_187_mouse_2,
    dx(4), jmp(Mscurry1), // goto "Mscurry1"

    LABEL(Mstop) // Mouse: stop
    /* ":loop" */ frame_186_mouse_1,
    jmp(Mstop), // goto ":loop"

    LABEL(Mraise) // Mouse: raise head
    /* ":loop" */ frame_188_mouse_stand,
    jmp(Mraise), // goto ":loop"

    LABEL(Mleave) // Mouse: leave
    act(actions_0_stand), frame_186_mouse_1,
    frame_186_mouse_1, frame_186_mouse_1, frame_188_mouse_stand, frame_188_mouse_stand,
    frame_188_mouse_stand, frame_188_mouse_stand, frame_188_mouse_stand, frame_188_mouse_stand,
    frame_188_mouse_stand, frame_188_mouse_stand,
    SEQ_FLIP, dx(8), jmp(Mscurry1), // goto "Mscurry1"

    LABEL(Mclimb) // Mouse: climb
    frame_186_mouse_1, frame_186_mouse_1, /* ":loop" */ LABEL(Mclimb_loop) frame_188_mouse_stand,
    jmp(Mclimb_loop), // goto ":loop"

    // custom logic - additional sequences

    LABEL(balcony) // running off a balcony/teleport based on stairs
    act(actions_5_bumped),
    dx(-5), dy(-1), snd(SND_FOOTSTEP), frame_217_exit_stairs_1,
    frame_218_exit_stairs_2, frame_219_exit_stairs_3,
    dx(1), frame_220_exit_stairs_4,
    dx(-4), dy(-3), snd(SND_FOOTSTEP), frame_221_exit_stairs_5,
    frame_0,
    snd(SND_SILENT), frame_0, frame_0, frame_0,
    snd(SND_SILENT), frame_0, frame_0, frame_0,
    snd(SND_SILENT), frame_0, frame_0, frame_0,
    snd(SND_SILENT), SEQ_END_LEVEL, tiles_23_balcony_left,
    /* ":loop" */ LABEL(balcony_loop) frame_0,
    jmp(balcony_loop), // goto ":loop"

    LABEL(crawlrepeat) // crawl/crouch repeatedly
    act(actions_1_run_jump), frame_109_crouch,
    jmp(crawlrepeat), // goto "crawlrepeat"

    LABEL(putdownpotion) // put down potion (no bottle swallowing)
    frame_196_drink, frame_194_drink, frame_192_drink,
    dx(-4), jmp(stand), // goto "stand"

    LABEL(landjump) // jump up only 1 pixel forward to make dropping tiles easier
    dx(-1), frame_81_hangdrop_1, frame_82_hangdrop_2,
    act(actions_5_bumped), frame_83_hangdrop_3,
    act(actions_1_run_jump), SEQ_KNOCK_DOWN, snd(SND_FOOTSTEP),
    frame_84_hangdrop_4, frame_85_hangdrop_5,
    dx(2), jmp(stand), // goto "stand"

    LABEL(stabbedfrombehind) // stabbed by sword from behind by the second guard
    act(actions_5_bumped), set_fall(-1, 0),
    dx(7), frame_172_jumpfall_2,
    dx(-1), dy(1), frame_173_jumpfall_3,
    dx(-1), frame_174_jumpfall_4,
    dx(-1), dy(2),
    dx(-1), dy(1),
    dx(0), dy(-4),
    jmp(guy8), // goto "guy8"

    LABEL(drinkmousepotion) // kid drinks the mouse potion
    act(actions_1_run_jump), dx(4), frame_191_drink,
    frame_192_drink, frame_193_drink, frame_194_drink, frame_195_drink,
    frame_196_drink, frame_197_drink, snd(SND_DRINK),
    frame_198_drink, frame_199_drink, frame_200_drink, frame_201_drink,
    frame_202_drink, frame_203_drink, frame_204_drink, frame_205_drink,
    frame_205_drink, frame_205_drink,
    SEQ_GET_ITEM, 1, frame_205_drink,
    frame_205_drink, frame_201_drink, frame_198_drink,
    dx(-4), jmp(stoop), // goto "crouch"

    LABEL(hangstraightdrop) // fall after hanging straight for a while
    act(actions_6_hang_straight), frame_92_hanging_6, // Apple II source has a bump sound here
    frame_93_hanging_7, frame_93_hanging_7, frame_92_hanging_6, frame_92_hanging_6,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    LABEL(hangstraight1)
    frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5, frame_91_hanging_5,
    jmp_if_shad_feather(hangfall), // goto "hangfall"

    LABEL(pickupsheathe) // pick up and sheathe sword quicker without moving back half a tile
    act(actions_1_run_jump),
    frame_234_sheathe,
    dx(1), frame_236_sheathe,
    dx(-2), frame_238_sheathe,
    frame_239_sheathe, frame_240_sheathe,
    frame_133_sheathe, frame_133_sheathe, frame_134_sheathe,
    act(actions_5_bumped) /* release arrows */, frame_134_sheathe,
    jmp(stand), // goto "stand"

    LABEL(chompersprint) // kid sprints through chomper then stops
    act(actions_5_bumped), // release arrows
    frame_1_start_run, frame_2_start_run,
    frame_3_start_run, frame_4_start_run,
    dx(8), frame_5_start_run,
    snd(SND_FOOTSTEP),
    dx(3), frame_6_start_run,
    act(actions_1_run_jump),
    frame_56_runturn,
    dx(2), frame_49_turn,
    dx(-2), frame_50_turn,
    frame_51_turn,
    jmp(stand), // goto "stand"

    LABEL(birdbow) // bird guard bowing
    act(actions_1_run_jump),
    frame_190_start_bowing_1, frame_191_start_bowing_2, frame_192_start_bowing_3,
    frame_193_start_bowing_4, frame_194_start_bowing_5, frame_195_start_bowing_6,
    frame_196_start_bowing_7, frame_197_start_bowing_8, frame_198_start_bowing_9,
    frame_199_start_bowing_10, frame_200_start_bowing_11, frame_201_start_bowing_12,
    LABEL(birdbowing) frame_202_bowing,
    jmp(birdbowing), // goto "birdbowing"

    LABEL(birdturnbow) // bird guard turning then bowing
    act(actions_5_bumped), SEQ_FLIP, dx(8),
    jmp(birdbow), // goto "birdbow"

    LABEL(cellstoop) // sitting in a cell (level 14)
    act(actions_8_sit), frame_211_sit_in_cell,
    jmp(cellstoop), // goto "cellstoop"

    LABEL(cellcrouch) // crouching in a cell
    act(actions_1_run_jump),
    dx(-1), frame_107_fall_land_1,
    dx(1), frame_108_fall_land_2,
    jmp(cellstoop), // goto "cellstoop"

    LABEL(bumpcloseddoor) // bump into a closed level door
    act(actions_5_bumped),
    dx(-11), dy(-1), frame_217_exit_stairs_1,
    frame_218_exit_stairs_2, snd(SND_BUMP), frame_219_exit_stairs_3,
    SEQ_FLIP, dy(1), dx(5), frame_217_exit_stairs_1,
    jmp(stand), // goto "stand"

    LABEL(kidturnengarde) // kid turns during a fight
    act(actions_5_bumped),
    frame_214_turn_engarde_1, frame_215_turn_engarde_2,
    SEQ_FLIP, LABEL(kidturning) frame_216_turn_engarde_3,
    act(actions_1_run_jump),
    jmp(ready_loop), // goto "ready_loop"

    LABEL(kidturnwithsword) // kid turns with sword (shift-back arrow)
    act(actions_5_bumped),
    frame_214_turn_engarde_1, frame_215_turn_engarde_2,
    SEQ_FLIP, dx(4),
    jmp(kidturning),

    LABEL(Pdead) // Princess is dead laying on cushions ("Pdead")
    frame_86_dead_princess,
    jmp(Pdead), // goto "Pdead"

    LABEL(impalehanging) // kid impaled on vertical spikes
    act(actions_1_run_jump), dx(4), frame_100_spiking_hanging,
    LABEL(impalehangdead) frame_184_spiked_hanging,
    jmp(impalehangdead), // goto ":dead"

    LABEL(birdimpalehang) // bird guard impaled on vertical spikes
    act(actions_1_run_jump), dx(4), frame_184_spiked_hanging, frame_177_spiked,
    jmp(impale_dead), // goto ":dead"

    LABEL(stepcrouch11) // step forward 11 pixels then crouch
    act(actions_1_run_jump),
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(3), frame_124_stepping_4,
    dx(4), frame_125_stepping_5,
    dx(3), frame_126_stepping_6,
    dx(-1), frame_127_stepping_7,
    frame_128_stepping_8,
    dx(1), frame_107_fall_land_1,
    dx(2), frame_108_fall_land_2,
    jmp(stoop_crouch), // goto "stoop_crouch"

    LABEL(stepcrouch10) // step forward 10 pixels then crouch
    act(actions_1_run_jump),
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(3), frame_124_stepping_4,
    dx(4), frame_125_stepping_5,
    dx(3), frame_126_stepping_6,
    dx(-2), frame_128_stepping_8,
    dx(1), frame_107_fall_land_1,
    dx(2), frame_108_fall_land_2,
    jmp(stoop_crouch), // goto "stoop_crouch"

    LABEL(stepcrouch9) // step forward 9 pixels then crouch
    act(actions_1_run_jump),
    frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(3), frame_124_stepping_4,
    dx(4), frame_125_stepping_5,
    dx(3), frame_126_stepping_6,
    dx(-2), frame_128_stepping_8,
    dx(1), frame_107_fall_land_1,
    dx(2), frame_108_fall_land_2,
    jmp(stoop_crouch), // goto "stoop_crouch"

    LABEL(stepcrouch8) // step forward 8 pixels then crouch
    act(actions_1_run_jump),
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(3), frame_124_stepping_4,
    dx(4), frame_125_stepping_5,
    dx(-1), frame_127_stepping_7,
    frame_128_stepping_8,
    dx(1), frame_107_fall_land_1,
    dx(2), frame_108_fall_land_2,
    jmp(stoop_crouch), // goto "stoop_crouch"

    LABEL(stepcrouch7) // step forward 7 pixels then crouch
    act(actions_1_run_jump),
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(3), frame_124_stepping_4,
    dx(2), frame_129_stepping_9,
    frame_130_stepping_10,
    dx(1), frame_107_fall_land_1,
    dx(2), frame_108_fall_land_2,
    jmp(stoop_crouch), // goto "stoop_crouch"

    LABEL(stepcrouch6) // step forward 6 pixels then crouch
    act(actions_1_run_jump),
    dx(1), frame_122_stepping_2,
    dx(1), frame_123_stepping_3,
    dx(2), frame_124_stepping_4,
    dx(2), frame_129_stepping_9,
    frame_130_stepping_10,
    dx(1), frame_107_fall_land_1,
    dx(2), frame_108_fall_land_2,
    jmp(stoop_crouch), // goto "stoop_crouch"

    LABEL(runjumphang) // hanging after a running jump
    act(actions_2_hang_climb), frame_98_hanging_12,
    jmp(hang1), // goto "hang1"

    LABEL(collapse) // skeleton collapses into a pile of bones
    act(actions_5_bumped),
    frame_171_stand_with_sword,
    dx(-4), frame_178_chomped,
    dx(-2), dy(-2), frame_177_spiked,
    dy(2), act(actions_1_run_jump),
    jmp(dropdead_dead), // goto "dropdead_dead"

    LABEL(shadowrise) // shadow rises from kid's body
    act(actions_5_bumped),
    dx(4), frame_183_collapse_5,
    dx(-1), frame_182_collapse_4,
    frame_181_collapse_3,
    frame_117_stand_up_from_crouch_8,
    frame_118_stand_up_from_crouch_9,
    frame_119_stand_up_from_crouch_10,
    jmp(stand), // goto "stand"

    LABEL(stepbackhang4) // step back 4 pixels, jump and hang
    act(actions_1_run_jump),
    frame_132_stepping_12, frame_131_stepping_11, frame_130_stepping_10,
    frame_129_stepping_9, dx(-4),
    jmp(jumphangMed), // goto "jumphangMed"

    LABEL(stepbackhang5) // step back 5 pixels, jump and hang
    act(actions_1_run_jump),
    frame_132_stepping_12, frame_131_stepping_11, frame_130_stepping_10,
    frame_129_stepping_9, dx(-5),
    jmp(jumphangMed), // goto "jumphangMed"

    LABEL(stepbackhang6) // step back 6 pixels, jump and hang
    act(actions_1_run_jump),
    frame_131_stepping_11, frame_130_stepping_10,
    frame_129_stepping_9, dx(-6),
    jmp(jumphangMed), // goto "jumphangMed"

    LABEL(stepbackhang7) // step back 7 pixels, jump and hang
    act(actions_1_run_jump),
    frame_130_stepping_10, frame_129_stepping_9,
    frame_128_stepping_8, dx(-7),
    jmp(jumphangMed), // goto "jumphangMed"

    LABEL(stepbackhang8) // step back 8 pixels, jump and hang
    act(actions_1_run_jump),
    frame_131_stepping_11, frame_130_stepping_10, frame_129_stepping_9,
    frame_128_stepping_8, dx(-8),
    jmp(jumphangMed), // goto "jumphangMed"

    LABEL(stepbackjumphang) // step back 4 pixels, jump back and hang
    act(actions_1_run_jump),
    frame_132_stepping_12, frame_131_stepping_11, frame_130_stepping_10,
    frame_129_stepping_9, dx(-4),
    jmp(jumpbackhang), // goto "jumpbackhang"

    LABEL(stepfwdhang9) // step forward 9 pixels, jump and hang
    act(actions_1_run_jump),
    frame_121_stepping_1, dx(1),
    frame_122_stepping_2, dx(1),
    frame_123_stepping_3, dx(4),
    frame_124_stepping_4, dx(4),
    frame_129_stepping_9, dx(-1),
    jmp(jumphangLong), // goto "jumphangLong"

    LABEL(stepfwdhang8) // step forward 8 pixels, jump and hang
    act(actions_1_run_jump),
    frame_121_stepping_1, dx(1),
    frame_122_stepping_2, dx(1),
    frame_123_stepping_3, dx(3),
    frame_124_stepping_4, dx(4),
    frame_129_stepping_9, dx(-1),
    jmp(jumphangLong), // goto "jumphangLong"

    LABEL(stepfwdhang7) // step forward 7 pixels, jump and hang
    act(actions_1_run_jump),
    frame_121_stepping_1, dx(1),
    frame_122_stepping_2, dx(1),
    frame_123_stepping_3, dx(3),
    frame_124_stepping_4, dx(3),
    frame_129_stepping_9, dx(-1),
    jmp(jumphangLong), // goto "jumphangLong"

    LABEL(stepfwdhang6) // step forward 6 pixels, jump and hang
    act(actions_1_run_jump),
    frame_121_stepping_1, dx(1),
    frame_122_stepping_2, dx(1),
    frame_123_stepping_3, dx(2),
    frame_124_stepping_4, dx(3),
    frame_129_stepping_9, dx(-1),
    jmp(jumphangLong), // goto "jumphangLong"

    LABEL(stepfwdhang5) // step forward 5 pixels, jump and hang
    act(actions_1_run_jump),
    frame_122_stepping_2, dx(1),
    frame_123_stepping_3, dx(2),
    frame_124_stepping_4, dx(3),
    frame_130_stepping_10, dx(-1),
    jmp(jumphangLong), // goto "jumphangLong"

    LABEL(stepfwdhang4) // step forward 4 pixels, jump and hang
    act(actions_1_run_jump),
    frame_122_stepping_2, dx(1),
    frame_123_stepping_3, dx(2),
    frame_124_stepping_4, dx(2),
    frame_130_stepping_10, dx(-1),
    jmp(jumphangLong), // goto "jumphangLong"

    LABEL(stopfistfight) // stop fist fighting
    act(actions_5_bumped), dx(-4),
    frame_122_stepping_2, dx(-2),
    jmp(stand), // goto "stand"

    LABEL(startfistfight) // start fist fighting
    act(actions_1_run_jump), dx(2),
    frame_122_stepping_2, dx(4),
    jmp(ready), // goto "ready"

    LABEL(fistfighthit) // hit in fist fight
    act(actions_5_bumped), set_fall(-1, 0), frame_172_jumpfall_2,
    dx(-1), dy(1), frame_173_jumpfall_3,
    dx(-1), dy(2),
    dx(-1), dy(1),
    dx(-2), dy(-4),
    jmp(guy8), // goto "guy8"

    LABEL(stabfallfar) // stab and push opponent far (with 1 hitpoint left)
    act(actions_5_bumped), dx(-16),
    frame_172_jumpfall_2, dy(1), dx(-6),
    jmp(fightfall), // goto "fightfall"

    LABEL(stabfalldead) // guard lands then collapses to the ground
    act(actions_1_run_jump), frame_180_collapse_2,
    frame_182_collapse_4, dx(1),
    frame_183_collapse_5, dx(-4),
    jmp(dropdead_dead), // goto "dropdead_dead"

    LABEL(mscurrymanual) // mouse stand up in manual mode after pressing up arrow key
    act(actions_1_run_jump),
    /* ":loop" */ LABEL(Mscurry2) frame_186_mouse_1,
    dx(5), frame_186_mouse_1,
    dx(4), frame_187_mouse_2,
    dx(5), jmp(Mscurry2), // goto ":loop"

    LABEL(retreatstep) // move back with sword exactly 1 step back to avoid falling
    act(actions_1_run_jump), set_fall(-1, 0),
    dx(-2), frame_157_walk_with_sword,
    jmp(ready), // goto "ready"

    LABEL(retreatkid) // kid moves back with sword (PoP2 sprite requires a different logic)
    act(actions_1_run_jump), set_fall(-1, 0),
    dx(-8), frame_160_fighting,
    dx(3), frame_157_walk_with_sword,
    jmp(ready), // goto "ready"

    LABEL(advancekid) // kid moves forward with sword (PoP2 sprite requires a different logic)
    act(actions_1_run_jump), set_fall(1, 0),
    frame_163_fighting, dx(4), frame_164_fighting,
    frame_165_walk_with_sword, dx(2),
    jmp(ready), // goto "ready"

    LABEL(fistfightadv) // kid moves forward when fist fighting
    act(actions_1_run_jump), set_fall(1, 0),
    dx(4), frame_164_fighting,
    frame_165_walk_with_sword, dx(2),
    jmp(ready), // goto "ready"

    LABEL(fistfightkick) // kick in fist fight
    act(actions_5_bumped),
    dx(-9), frame_4_start_run,
    dx(12), frame_213_kick,
    dx(-2), frame_57_runturn,
    dx(-1), frame_158_stand_with_sword,
    frame_158_stand_with_sword,
    jmp(ready), // goto "ready"
};

// custom logic - inactive guard mini-sequence table
byte inactive_seqtbl[] = {
    LABEL(inactive_seq_strike) // strike
    frame_151_strike_1, frame_152_strike_2,
    frame_153_strike_3, frame_154_poking,
    inactive_seq_kid_hurt, frame_155_post_strike_1,
    frame_156_post_strike_2,
    inactive_seq_stand, // stand inactive

    LABEL(inactive_seq_block_retreat) // block, then retreat
    isnd(sound_10_sword_vs_sword),
    frame_150_parry, frame_161_parry,
    frame_158_stand_with_sword, // falls through

    LABEL(inactive_seq_retreat) // retreat
    idx(-3), frame_160_fighting,
    idx(-2), frame_157_walk_with_sword,
    frame_158_stand_with_sword,
    frame_170_stand_with_sword,
    inactive_seq_stand,

    LABEL(inactive_seq_counter) // counter strike
    isnd(sound_10_sword_vs_sword),
    frame_150_parry, frame_161_parry,
    frame_158_stand_with_sword, frame_152_strike_2,
    frame_153_strike_3, frame_154_poking,
    inactive_seq_kid_hurt,
    inactive_seq_stand,

    LABEL(inactive_seq_advance) // advance
    frame_170_stand_with_sword,
    idx(2), frame_163_fighting,
    idx(4), frame_164_fighting,
    frame_165_walk_with_sword,
    frame_158_stand_with_sword,
    inactive_seq_stand,

    LABEL(inactive_seq_turn) // turn
    inactive_seq_flip, idx(18),
    inactive_seq_stand,

    LABEL(inactive_seq_turn_engarde) // turn engarde
    inactive_seq_flip, idx(8),
    inactive_seq_stand,

    LABEL(inactive_seq_dropdead) // dying
    frame_179_collapse_1, frame_180_collapse_2,
    frame_181_collapse_3, frame_182_collapse_4,
    idx(1), frame_183_collapse_5,
    idx(-4), frame_185_dead,
    inactive_seq_dead, // dead
};
