/*
SDLPoP, a port/conversion of the DOS game Prince of Persia.
Copyright (C) 2013-2022  David Nagy

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

The authors of this program may be contacted at https://forum.princed.org

This mod was created by "dmitrys".
*/

#ifndef SEQ_TABLES_H
#define SEQ_TABLES_H

// This expands a two-byte number into two comma-separated bytes, used for the JMP destinations
#define DW(data_word) (data_word) & 0x00FF, (((data_word) & 0xFF00) >> 8)

// The sequence operation value should be converted to a negative number. Ex: op 1 = -1 = 0xFF.
#define SEQ_OP(op_byte) (byte)(-op_byte)

// Shorter notation for the sequence table instructions
#define act(action) SEQ_ACTION, (action)
#define jmp(dest) SEQ_JMP, DW(dest)
#define jmp_if_shad_feather(dest) SEQ_JMP_IF_SHAD_FEATHER, DW(dest)
#define dx(amount) SEQ_DX, (byte) (amount)
#define dy(amount) SEQ_DY, (byte) (amount)
#define snd(sound) SEQ_SOUND, (sound)
#define set_fall(x, y) SEQ_SET_FALL, (byte) (x), (byte) (y)

#define LABEL(label) // do nothing

// Labels
#define running             0                       // 0x000
#define startrun            5  + running            // 0x005
#define runstt1             2  + startrun           // 0x007
#define runstt4             3  + runstt1            // 0x00A
#define runcyc1             9  + runstt4            // 0x013
#define runcyc7             20 + runcyc1            // 0x027
#define stand               11 + runcyc7            // 0x032
#define goalertstand        6  + stand              // 0x038
#define alertstand          2  + goalertstand       // 0x03A
#define arise               4  + alertstand         // 0x03E
#define guardengarde        21 + arise              // 0x053
#define engarde             3  + guardengarde       // 0x056
#define ready               16 + engarde            // 0x066
#define ready_loop          6  + ready              // 0x06C
#define stabbed             4  + ready_loop         // 0x070
#define strikeadv           29 + stabbed            // 0x08D
#define strikeret           14 + strikeadv          // 0x09B
#define advance             12 + strikeret          // 0x0A7
#define fastadvance         15 + advance            // 0x0B6
#define retreat             12 + fastadvance        // 0x0C2
#define strike              14 + retreat            // 0x0D0
#define faststrike          6  + strike             // 0x0D6
#define guy4                3  + faststrike         // 0x0D9
#define guy7                5  + guy4               // 0x0DE
#define guy8                3  + guy7               // 0x0E1
#define blockedstrike       7  + guy8               // 0x0E8
#define blocktostrike       6  + blockedstrike      // 0x0EE
#define readyblock          4  + blocktostrike      // 0x0F2
#define blocking            1  + readyblock         // 0x0F3
#define striketoblock       4  + blocking           // 0x0F7
#define landengarde         5  + striketoblock      // 0x0FC
#define bumpengfwd          6  + landengarde        // 0x102
#define bumpengback         7  + bumpengfwd         // 0x109
#define flee                7  + bumpengback        // 0x110
#define turnengarde         7  + flee               // 0x117
#define alertturn           8  + turnengarde        // 0x11F
#define standjump           8  + alertturn          // 0x127
#define sjland              29 + standjump          // 0x144
#define runjump             29 + sjland             // 0x161
#define rjlandrun           46 + runjump            // 0x18F
#define rdiveroll           9  + rjlandrun          // 0x198
#define rdiveroll_crouch    18 + rdiveroll          // 0x1A0
#define sdiveroll           4  + rdiveroll_crouch   // 0x1AE
#define crawl               1  + sdiveroll          // 0x1AF
#define crawl_crouch        14 + crawl              // 0x1BD
#define turndraw            4  + crawl_crouch       // 0x1C1
#define turn                12 + turndraw           // 0x1CD
#define turnrun             26 + turn               // 0x1E7
#define runturn             7  + turnrun            // 0x1EE
#define fightfall           43 + runturn            // 0x219
#define efightfall          28 + fightfall          // 0x235
#define efightfallfwd       30 + efightfall         // 0x253
#define stepfall            28 + efightfallfwd      // 0x26F
#define fall1               9  + stepfall           // 0x278
#define patchfall           22 + fall1              // 0x28E
#define stepfall2           7  + patchfall          // 0x295
#define stepfloat           5  + stepfall2          // 0x29A
#define jumpfall            22 + stepfloat          // 0x2B0
#define rjumpfall           28 + jumpfall           // 0x2CC
#define jumphangMed         28 + rjumpfall          // 0x2E8
#define jumphangLong        21 + jumphangMed        // 0x2FE
#define jumpbackhang        27 + jumphangLong       // 0x318
#define hang                29 + jumpbackhang       // 0x335
#define hang1               3  + hang               // 0x338
#define hangstraight        45 + hang1              // 0x365
#define hangstraight_loop   7  + hangstraight       // 0x36C
#define climbfail           4  + hangstraight_loop  // 0x370
#define climbdown           16 + climbfail          // 0x380
#define climbup             24 + climbdown          // 0x398
#define hangdrop            33 + climbup            // 0x3B9
#define hangfall            17 + hangdrop           // 0x3CA
#define freefall            19 + hangfall           // 0x3DD
#define freefall_loop       2  + freefall           // 0x3DF
#define runstop             4  + freefall_loop      // 0x3E3
#define jumpup              25 + runstop            // 0x3FC
#define highjump            21 + jumpup             // 0x411
#define superhijump         30 + highjump           // 0x42F
#define fallhang            91 + superhijump        // 0x48A
#define bump                6  + fallhang           // 0x490
#define bumpfall            10 + bump               // 0x49A
#define bumpfloat           31 + bumpfall           // 0x4B9
#define hardbump            22 + bumpfloat          // 0x4CF
#define testfoot            30 + hardbump           // 0x4ED
#define stepback            31 + testfoot           // 0x50C
#define step14              5  + stepback           // 0x511
#define step13              31 + step14             // 0x530
#define step12              31 + step13             // 0x54F
#define step11              31 + step12             // 0x56E
#define step10              29 + step11             // 0x58B
#define step10a             5  + step10             // 0x590
#define step9               23 + step10a            // 0x5A7
#define step8               6  + step9              // 0x5AD
#define step7               26 + step8              // 0x5C7
#define step6               21 + step7              // 0x5DC
#define step5               21 + step6              // 0x5F1
#define step4               21 + step5              // 0x606
#define step3               16 + step4              // 0x616
#define step2               16 + step3              // 0x626
#define step1               12 + step2              // 0x632
#define stoop               9  + step1              // 0x63B
#define stoop_crouch        8  + stoop              // 0x643
#define standup             4  + stoop_crouch       // 0x647
#define pickupsword         23 + standup            // 0x65E
#define resheathe           16 + pickupsword        // 0x66E
#define fastsheathe         33 + resheathe          // 0x68F
#define drinkpotion         14 + fastsheathe        // 0x69D
#define softland            34 + drinkpotion        // 0x6BF
#define softland_crouch     11 + softland           // 0x6CA
#define landrun             4  + softland_crouch    // 0x6CE
#define medland             32 + landrun            // 0x6EE
#define hardland            66 + medland            // 0x730
#define hardland_dead       9  + hardland           // 0x739
#define stabkill            4  + hardland_dead      // 0x73D
#define dropdead            5  + stabkill           // 0x742
#define dropdead_dead       12 + dropdead           // 0x74E
#define impale              4  + dropdead_dead      // 0x752
#define impale_dead         7  + impale             // 0x759
#define halve               4  + impale_dead        // 0x75D
#define halve_dead          4  + halve              // 0x761
#define crush               4  + halve_dead         // 0x765
#define deadfall            3  + crush              // 0x768
#define deadfall_loop       5  + deadfall           // 0x76D
#define climbstairs         4  + deadfall_loop      // 0x771
#define climbstairs_loop    81 + climbstairs        // 0x7C2
#define Vstand              4  + climbstairs_loop   // 0x7C6
#define Vraise              4  + Vstand             // 0x7CA
#define Vraise_loop         21 + Vraise             // 0x7DF
#define Vwalk               4  + Vraise_loop        // 0x7E3
#define Vwalk1              2  + Vwalk              // 0x7E5
#define Vwalk2              3  + Vwalk1             // 0x7E8
#define Vstop               18 + Vwalk2             // 0x7FA
#define Vexit               7  + Vstop              // 0x801
#define Pstand              40 + Vexit              // 0x829
#define Palert              4  + Pstand             // 0x82D
#define Pstepback           15 + Palert             // 0x83B
#define Pstepback_loop      16 + Pstepback          // 0x84B
#define Plie                4  + Pstepback_loop     // 0x850
#define Pwaiting            4  + Plie               // 0x854
#define Pembrace            4  + Pwaiting           // 0x858
#define Pembrace_loop       30 + Pembrace           // 0x876
#define Pstroke             4  + Pembrace_loop      // 0x87A
#define Prise               4  + Pstroke            // 0x87E
#define Prise_loop          14 + Prise              // 0x88C
#define Pcrouch             4  + Prise_loop         // 0x890
#define Pcrouch_loop        64 + Pcrouch            // 0x8D0
#define Pslump              4  + Pcrouch_loop       // 0x8D4
#define Pslump_loop         1  + Pslump             // 0x8D5
#define Mscurry             4  + Pslump_loop        // 0x8D9
#define Mscurry1            2  + Mscurry            // 0x8DB
#define Mstop               12 + Mscurry1           // 0x8E7
#define Mraise              4  + Mstop              // 0x8EB
#define Mleave              4  + Mraise             // 0x8EF
#define Mclimb              19 + Mleave             // 0x902
#define Mclimb_loop         2  + Mclimb             // 0x904
// custom logic - additional sequences
#define balcony             4  + Mclimb_loop        // 0x908
#define balcony_loop        41 + balcony            // 0x931
#define crawlrepeat         4  + balcony_loop       // 0x935
#define putdownpotion       6  + crawlrepeat        // 0x93B
#define landjump            8  + putdownpotion      // 0x943
#define stabbedfrombehind   19 + landjump           // 0x956
#define drinkmousepotion    31 + stabbedfrombehind  // 0x975
#define hangstraightdrop    34 + drinkmousepotion   // 0x997
#define hangstraight1       75 + hangstraightdrop   // 0x9E2
#define pickupsheathe       7  + hangstraight1      // 0x9E9
#define chompersprint       20 + pickupsheathe      // 0x9FD
#define birdbow             27 + chompersprint      // 0xA18
#define birdbowing          14 + birdbow            // 0xA26
#define birdturnbow         3  + birdbowing         // 0xA29
#define cellstoop           9  + birdturnbow        // 0xA32
#define cellcrouch          6  + cellstoop          // 0xA38
#define bumpcloseddoor      11 + cellcrouch         // 0xA43
#define kidturnengarde      20 + bumpcloseddoor     // 0xA57
#define kidturning          5  + kidturnengarde     // 0xA5C
#define kidturnwithsword    11 + kidturnengarde     // 0xA62
#define Pdead               10 + kidturnwithsword   // 0xA6C
#define impalehanging       4  + Pdead              // 0xA70
#define impalehangdead      5  + impalehanging      // 0xA75
#define birdimpalehang      4  + impalehangdead     // 0xA79
#define stepcrouch11        9  + birdimpalehang     // 0xA82
#define stepcrouch10        30 + stepcrouch11       // 0xAA0
#define stepcrouch9         29 + stepcrouch10       // 0xABD
#define stepcrouch8         27 + stepcrouch9        // 0xAD8
#define stepcrouch7         27 + stepcrouch8        // 0xAF3
#define stepcrouch6         24 + stepcrouch7        // 0xB0B
#define runjumphang         24 + stepcrouch6        // 0xB23
#define collapse            6  + runjumphang        // 0xB29
#define shadowrise          18 + collapse           // 0xB3B
#define stepbackhang4       15 + shadowrise         // 0xB4A
#define stepbackhang5       11 + stepbackhang4      // 0xB55
#define stepbackhang6       11 + stepbackhang5      // 0xB60
#define stepbackhang7       10 + stepbackhang6      // 0xB6A
#define stepbackhang8       10 + stepbackhang7      // 0xB74
#define stepbackjumphang    11 + stepbackhang8      // 0xB7F
#define stepfwdhang9        11 + stepbackjumphang   // 0xB8A
#define stepfwdhang8        20 + stepfwdhang9       // 0xB9E
#define stepfwdhang7        20 + stepfwdhang8       // 0xBB2
#define stepfwdhang6        20 + stepfwdhang7       // 0xBC6
#define stepfwdhang5        20 + stepfwdhang6       // 0xBDA
#define stepfwdhang4        17 + stepfwdhang5       // 0xBEB
#define stopfistfight       17 + stepfwdhang4       // 0xBFC
#define startfistfight      10 + stopfistfight      // 0xC06
#define fistfighthit        10 + startfistfight     // 0xC10
#define stabfallfar         26 + fistfighthit       // 0xC2A
#define stabfalldead        12 + stabfallfar        // 0xC36
#define mscurrymanual       12 + stabfalldead       // 0xC42
#define Mscurry2            2 + mscurrymanual       // 0xC44
#define retreatstep         11 + Mscurry2           // 0xC4F
#define retreatkid          12 + retreatstep        // 0xC5B
#define advancekid          14 + retreatkid         // 0xC69
#define fistfightadv        15 + advancekid         // 0xC78
#define fistfightkick       14 + fistfightadv       // 0xC86 - 0xC96

// operations (special values of 241-255; 0-240 are used by frames)
#define SEQ_JMP                  SEQ_OP(1)  // 0xFF
#define SEQ_FLIP                 SEQ_OP(2)  // 0xFE
#define SEQ_UP                   SEQ_OP(3)  // 0xFD
#define SEQ_DOWN                 SEQ_OP(4)  // 0xFC
#define SEQ_DX                   SEQ_OP(5)  // 0xFB
#define SEQ_DY                   SEQ_OP(6)  // 0xFA
#define SEQ_ACTION               SEQ_OP(7)  // 0xF9
#define SEQ_SET_FALL             SEQ_OP(8)  // 0xF8
#define SEQ_JMP_IF_SHAD_FEATHER  SEQ_OP(9)  // 0xF7
#define SEQ_DIE                  SEQ_OP(10) // 0xF6
#define SEQ_KNOCK_UP             SEQ_OP(11) // 0xF5
#define SEQ_KNOCK_DOWN           SEQ_OP(12) // 0xF4
#define SEQ_GET_ITEM             SEQ_OP(13) // 0xF3
#define SEQ_SOUND                SEQ_OP(14) // 0xF2
#define SEQ_END_LEVEL            SEQ_OP(15) // 0xF1

// sound operations
#define SND_SILENT      0
#define SND_FOOTSTEP    1
#define SND_BUMP        2
#define SND_DRINK       3
#define SND_LEVEL       4
#define SND_DRAW_SWORD  5

// inactive guard table macros
#define idx(x) inactive_seq_dx_fwd, (byte) (x)
#define isnd(sound) inactive_seq_snd, (sound)

// inactive guard table sequences
#define inactive_seq_strike         0                                // 0x00
#define inactive_seq_block_retreat  8  + inactive_seq_strike         // 0x08
#define inactive_seq_retreat        5  + inactive_seq_block_retreat  // 0x0D
#define inactive_seq_counter        14 + inactive_seq_block_retreat  // 0x1B
#define inactive_seq_advance        10 + inactive_seq_counter        // 0x25
#define inactive_seq_stop           9  + inactive_seq_advance        // 0x2E
#define inactive_seq_turn           1  + inactive_seq_stop           // 0x2F
#define inactive_seq_turn_engarde   4  + inactive_seq_turn           // 0x33
#define inactive_seq_dropdead       4  + inactive_seq_turn_engarde   // 0x37 - 0x42

// inactive guard table operations
#define inactive_seq_stand       SEQ_OP(1) // 0xFF
#define inactive_seq_dx_fwd      SEQ_OP(2) // 0xFE
#define inactive_seq_snd         SEQ_OP(3) // 0xFD
#define inactive_seq_kid_hurt    SEQ_OP(4) // 0xFC
#define inactive_seq_flip        SEQ_OP(5) // 0xFB
#define inactive_seq_dead        SEQ_OP(6) // 0xFA

// sequence IDs (in seqtbl_offsets)
enum seqids {
    seq_1_start_run = 1,
    seq_2_stand = 2,
    seq_3_standing_jump = 3,
    seq_4_run_jump = 4,
    seq_5_turn = 5,
    seq_6_run_turn = 6,
    seq_7_fall = 7,
    seq_8_jump_up_and_grab_straight = 8,
    seq_9_grab_while_jumping = 9,
    seq_10_climb_up = 10,
    seq_11_release_ledge_and_land = 11,
    seq_13_stop_run = 13,
    seq_14_jump_up_into_ceiling = 14,
    seq_15_grab_ledge_midair = 15,
    seq_16_jump_up_and_grab = 16,
    seq_17_soft_land = 17,
    seq_18_fall_after_standing_jump = 18,
    seq_19_fall = 19,
    seq_20_medium_land = 20,
    seq_21_fall_after_running_jump = 21,
    seq_22_crushed = 22,
    seq_23_release_ledge_and_fall = 23,
    seq_24_jump_up_and_grab_forward = 24,
    seq_26_crouch_while_running = 26,
    seq_28_jump_up_with_nothing_above = 28,
    seq_29_safe_step_1 = 29,
    seq_30_safe_step_2 = 30,
    seq_31_safe_step_3 = 31,
    seq_32_safe_step_4 = 32,
    seq_33_safe_step_5 = 33,
    seq_34_safe_step_6 = 34,
    seq_35_safe_step_7 = 35,
    seq_36_safe_step_8 = 36,
    seq_37_safe_step_9 = 37,
    seq_38_safe_step_10 = 38,
    seq_39_safe_step_11 = 39,
    seq_40_safe_step_12 = 40,
    seq_41_safe_step_13 = 41,
    seq_42_safe_step_14 = 42,
    seq_43_start_run_after_turn = 43,
    seq_44_step_on_edge = 44,
    seq_45_bumpfall = 45,
    seq_46_hardbump = 46,
    seq_47_bump = 47,
    seq_48_super_high_jump = 48,
    seq_49_stand_up_from_crouch = 49,
    seq_50_crouch = 50,
    seq_51_spiked = 51,
    seq_52_loose_floor_fell_on_kid = 52,
    seq_54_chomped = 54,
    seq_55_draw_sword = 55,
    seq_56_guard_forward_with_sword = 56,
    seq_57_back_with_sword = 57,
    seq_58_guard_strike = 58,
    seq_60_turn_with_sword = 60,
    seq_61_parry_after_strike = 61,
    seq_62_parry = 62,
    seq_63_guard_active_after_fall = 63,
    seq_64_pushed_back_with_sword = 64,
    seq_65_bump_forward_with_sword = 65,
    seq_66_strike_after_parry = 66,
    seq_68_climb_down = 68,
    seq_69_attack_was_parried = 69,
    seq_70_go_up_on_level_door = 70,
    seq_71_dying = 71,
    seq_73_climb_up_to_closed_gate = 73,
    seq_74_hit_by_sword = 74,
    seq_75_strike = 75,
    seq_77_guard_stand_inactive = 77,
    seq_78_drink = 78,
    seq_79_crouch_hop = 79,
    seq_80_stand_flipped = 80,
    seq_81_kid_pushed_off_ledge = 81,
    seq_82_guard_pushed_off_ledge = 82,
    seq_83_guard_fall = 83,
    seq_84_run = 84,
    seq_85_stabbed_to_death = 85,
    seq_86_forward_with_sword_fast = 86,
    seq_87_guard_become_inactive = 87,
    seq_88_skel_wake_up = 88,
    seq_89_turn_draw_sword = 89,
    seq_90_guard_stand_active = 90,
    seq_91_get_sword = 91,
    seq_92_put_sword_away = 92,
    seq_93_put_sword_away_fast = 93,
    seq_94_princess_stand_PV1 = 94,
    seq_95_Jaffar_stand_PV1 = 95,
    seq_96_Jaffar_walk_PV1 = 96,
    seq_97_Jaffar_stop_PV1 = 97,
    seq_98_princess_turn_PV1 = 98,
    seq_99_princess_back_PV1 = 99,
    seq_100_Jaffar_stop_conjuring_PV1 = 100,
    seq_101_mouse_stands_up = 101,
    seq_102_Jaffar_conjuring_PV1 = 102,
    seq_103_princess_lying_PV2 = 103,
    seq_104_start_fall_in_front_of_wall = 104,
    seq_105_mouse_forward = 105,
    seq_106_mouse = 106,
    seq_107_mouse_stand_up_and_go = 107,
    seq_108_princess_turn_and_hug = 108,
    seq_109_princess_stand_PV2 = 109,
    seq_110_princess_crouching_PV2 = 110,
    seq_111_princess_stand_up_PV2 = 111,
    seq_112_princess_crouch_down_PV2 = 112,
    seq_113_princess_look_down_PV1 = 113,
    seq_114_mouse_stand = 114,
    // custom logic - additional sequences
    seq_115_teleport = 115,
    seq_116_crawl = 116,
    seq_119_hit_by_sword_from_behind = 119,
    seq_120_drink_mouse_potion = 120,
    seq_121_hang_straight_drop = 121,
    seq_122_chomper_sprint = 122,
    seq_123_bird_bow = 123,
    seq_124_bird_turn_bow = 124,
    seq_125_sit_in_cell = 125,
    seq_126_crouch_in_cell = 126,
    seq_127_dead = 127,
    seq_128_bump_into_closed_door = 128,
    seq_129_kid_turn_engarde = 129,
    seq_130_kid_turn_with_sword = 130,
    seq_131_princess_dead_PV2 = 131,
    seq_132_kid_spiked_hanging = 132,
    seq_133_bird_spiked_hanging = 133,
    seq_140_grab_while_run_jump = 140,
    seq_141_skeleton_collapse = 141,
    seq_142_rise = 142,
    seq_148_step_back_jump_grab = 148,
    seq_155_start_fist_fight = 155,
    seq_156_stop_fist_fight = 156,
    seq_157_fist_fight_hit = 157,
    seq_158_guard_pushed_far = 158,
    seq_159_guard_fell_dead = 159,
    seq_160_mouse_forward_manual = 160,
    seq_161_step_back_with_sword = 161,
    seq_162_kid_back_with_sword = 162,
    seq_163_kid_forward_with_sword = 163,
    seq_164_kid_forward_fist_fight = 164,
    seq_165_kid_kick_fist_fight = 165,
};

extern const word seqtbl_offsets[];
extern byte seqtbl[];
extern byte inactive_seqtbl[];

#endif
